/*
 * apps/ctamc_client.c - ctamc utility functions used by s_client, s_server,
 * req, genrsa and ca
 */
/* Copyright (C) 1995-1998 Eric Young (eay@cryptsoft.com)
 * All rights reserved.
 *
 * This package is an SSL implementation written
 * by Eric Young (eay@cryptsoft.com).
 * The implementation was written so as to conform with Netscapes SSL.
 *
 * This library is free for commercial and non-commercial use as long as
 * the following conditions are aheared to.  The following conditions
 * apply to all code found in this distribution, be it the RC4, RSA,
 * lhash, DES, etc., code; not just the SSL code.  The SSL documentation
 * included with this distribution is covered by the same copyright terms
 * except that the holder is Tim Hudson (tjh@cryptsoft.com).
 *
 * Copyright remains Eric Young's, and as such any Copyright notices in
 * the code are not to be removed.
 * If this package is used in a product, Eric Young should be given attribution
 * as the author of the parts of the library used.
 * This can be in the form of a textual message at program startup or
 * in documentation (online or textual) provided with the package.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    "This product includes cryptographic software written by
 *     Eric Young (eay@cryptsoft.com)"
 *    The word 'cryptographic' can be left out if the rouines from the library
 *    being used are not cryptographic related :-).
 * 4. If you include any Windows specific code (or a derivative thereof) from
 *    the apps directory (application code) you must include an acknowledgement:
 *    "This product includes software written by Tim Hudson (tjh@cryptsoft.com)"
 *
 * THIS SOFTWARE IS PROVIDED BY ERIC YOUNG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 */
/* ====================================================================
 * Copyright (c) 1998-2006 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */
#ifdef CTAME_SUPPORT
# include "ctamc_client.h"
# include <openssl/err.h>
# include <openssl/rand.h>
# include <openssl/x509.h>
# include <openssl/ssl.h>
# include "../engines/vendor_defns/ctamc.h"
# include "../engines/ctame/ctame.h"
# include "safe_str_lib.h"

# ifdef USE_CISCO_SSM
#  include <ssm.h>
#  define SSMDIR "ssmdir"
# else
#  include "tamc_api.h"
#  include "tam_linux_ipc.h"
#  include "tam_linux_mem.h"
# endif

static void *ctamc_ctx = NULL;
static void *ctamc_session_id = NULL;

void *my_mem_alloc(size_t size)
{
    return (calloc(1, size));
}

void my_free(void *ptr)
{
    free(ptr);
}

/*
 * Load engine
 */
int app_load_ctamc_engine(ENGINE **e)
{

    ENGINE *e_ctamc = NULL;
    int rc = 1;

    if (!e)
        goto end;

    /*
     * Load the statically linked CTAM engine
     */
    ENGINE_load_ctamc();

    if ((e_ctamc = ENGINE_by_id("ctamc")) == NULL) {
        goto end;
    }

    if (!ENGINE_init(e_ctamc)) {
        goto end;
    }

    *e = e_ctamc;
    rc = 0;
 end:

    return (rc);
}

// Login app to default device
void *ctamc_login(char *app_id, char *password, void **context,
                  void **session_id)
{
    tam_status_e rc;
    void *ctx = NULL;
    void *session = NULL;
    tam_credentials_t credentials;

    if (!app_id || !password || !context || !session_id)
        goto end;

# ifdef USE_CISCO_SSM
    ssm_cfg_t cfg;
    memzero_s(&cfg, sizeof(ssm_cfg_t));
    strncpy_s(cfg.storage_path, PATH_MAX, SSMDIR, PATH_MAX);
    rc = ssm_initialize(&cfg);
    if (rc != TAM_SUCCESS) {
        printf("ssm_initialize failed in ssmUtil_initSuite1Dev1App\n");
        goto end;
    }
# else
    /*
     * Register PD linux IPC and memory mgmt vectors
     */
    tam_register_ipc_cb(tam_linux_ipc_init,
                        tam_linux_ipc_cleanup,
                        tam_linux_ipc_open,
                        tam_linux_ipc_close,
                        tam_linux_ipc_send, tam_linux_ipc_receive);

    tam_register_mem_cb(my_mem_alloc, my_free);
# endif
    // setup credential
    credentials.credential_type = TAM_CREDENTIAL_PW;
    strncpy_s(&credentials.type.userpass.appid[0], TAM_APPID_LEN_MAX, app_id,
              TAM_APPID_LEN_MAX);
    strncpy_s(&credentials.type.userpass.pw[0], TAM_PW_LEN_MAX, password,
              TAM_PW_LEN_MAX);

    /* Initialize our context - must do before logging in */
    rc = tamc_init_context(&ctx, app_id, NULL);
    if (rc != TAM_SUCCESS)
        goto end;

    // login
    rc = tamc_app_login(ctx, NULL, &credentials, &session);
    if (rc != TAM_SUCCESS) {
        tamc_delete_context(&ctx);
        goto end;
    }
    // set results;
    *context = ctx;
    *session_id = session;
 end:
    return (session);
}

// Logout
void app_ctamc_logout(ENGINE *e)
{
    // logout
    if (ctamc_ctx) {
        tamc_app_logout(&ctamc_session_id);
        tamc_delete_context(&ctamc_ctx);
    }

    ctamc_ctx = NULL;
    ctamc_session_id = NULL;

    ENGINE_finish(e);
    return;
}

ENGINE *open_ctamc(char *app_id, char *password)
{
    int rc = 0;
    ENGINE *e_ctamc = NULL;
    void *session_id = NULL;
    ctamc_context_t *ctamc_session_ctx = NULL;

    /*
     * Login to TAM server
     */
    ctamc_login(app_id, password, &ctamc_ctx, &session_id);
    if (!ctamc_ctx) {
        goto bad;
    }

    /*
     * Load CTAMC Engine, Set Session_ID
     */
    app_load_ctamc_engine(&e_ctamc);
    if (!e_ctamc) {
        goto bad;
    }
# ifndef USE_CISCO_SSM
    tam_set_debug_level(TAMC_DBG_ERR_ALL);
# endif

    ctamc_session_ctx = calloc(1, sizeof(ctamc_context_t));
    ctamc_session_ctx->session_id = session_id;
    strncpy_s(&ctamc_session_ctx->app_id[0], TAM_APPID_LEN_MAX, app_id,
              TAM_APPID_LEN_MAX);
    ctamc_session_ctx->ctamc_ctx = ctamc_ctx;

    ctamc_item_t ctamc_item;
    ctamc_item.c = (unsigned char *)ctamc_session_ctx;
    ctamc_item.len = sizeof(ctamc_context_t);

    rc = ENGINE_ctrl_cmd(e_ctamc, "CTAMC_SET_CONTEXT", 0, &ctamc_item, 0, 0);

    if (rc == 1) {
    } else {
        goto bad;
    }
    ctamc_session_id = session_id;
 bad:
    free(ctamc_session_ctx);
    return (e_ctamc);
}

/*
 * convert openssl to tamd_ec_curve_e curve type
 */
ctamc_ec_curve_e swtam_nid_to_tam_curve_type(char *curve_name)
{
    int nid = 0;
    tam_ec_curve_e curve;
    if (curve_name != NULL) {
        /*
         * workaround for the SECG curve names secp192r1 and secp256r1 (which
         * are the same as the curves prime192v1 and prime256v1 defined in
         * X9.62)
         */
        if (!strcmp(curve_name, "secp192r1")) {
            nid = NID_X9_62_prime192v1;
        } else if (!strcmp(curve_name, "secp256r1")) {
            nid = NID_X9_62_prime256v1;
        } else {
            nid = OBJ_sn2nid(curve_name);
        }
    }

    switch (nid) {
    case NID_X9_62_prime192v1:
        curve = TAM_EC_CURVE_P192;
        break;
    case NID_X9_62_prime256v1:
        curve = TAM_EC_CURVE_P256;
        break;
    case NID_secp384r1:
        curve = TAM_EC_CURVE_P384;
        break;
    case NID_secp521r1:
        curve = TAM_EC_CURVE_P521;
        break;
    case NID_brainpoolP256r1:
        curve = TAM_EC_CURVE_BRAINPOOLP256;
        break;
    case NID_brainpoolP384r1:
        curve = TAM_EC_CURVE_BRAINPOOLP384;
        break;
    case NID_brainpoolP512r1:
        curve = TAM_EC_CURVE_BRAINPOOLP512;
        break;
    default:
        curve = CTAMC_EC_CURVE_UNKN;
        break;
    }
    return ((ctamc_ec_curve_e) curve);
}
#endif                          /* CTAME_SUPPORT */
