/* -----------------------------------------------------------------
 * crypto/bn/asm/modexp-octeon.c - Cavium Octeon specific hardware assist
 * functionality.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#include <openssl/bn.h>
#include "bn_lcl.h"
#include <openssl/rsa.h>
#include <openssl/dsa.h>
#ifdef OPENSSL_FIPSCANISTER
# include <openssl/fipssyms.h>
#endif
#ifndef _MIPS_ARCH_OCTEON3
# include "ciscossl-cvmx.h"
#else
# include "cvmx.h"
#endif

# define ROUNDUP8(val) (((val) + 7)&0xfffffff8)
# define OCT_SUCCESS          1
# define OCT_FAILURE          0
# define ROUNDUP2(v)	((((v) + 1) / 2) * 2)

#define FATAL_ERROR CAV_ASSERT(0)

extern int MontMul576(uint64_t * product, uint64_t * mpcand,
                      uint64_t * mplier, uint64_t * mod, uint64_t * recip);
extern int MontMul(uint64_t * product, uint64_t * mpcand, uint64_t * mplier,
                   uint64_t * mod, uint64_t * recip, int len);
extern int Vadd(uint64_t * accum, uint64_t * addend, int len);
extern int Vsub(uint64_t * accum, uint64_t * addend, int len);
extern int MontMul512(uint64_t * product, uint64_t * mpcand, uint64_t * mplier,
                      uint64_t * mod, uint64_t * recip);
extern int MontMul_O3(uint64_t * product, uint64_t * mpcand, uint64_t * mplier,
                      uint64_t * mod, uint64_t * recip, int len);

#if !defined(USER_SPACE_MODEXP)
# if defined(__linux__)
#  include "cryptolinux.h"
# endif
#endif

static
int find_msw(uint64_t * wptr)
{
    int i = 0;
    while (wptr[i++]) ;
    return i - 1;
}


void
VMul(uint64_t * product, const uint64_t * mpcand, const uint64_t * mplier,
     int len)
{
    int i, k;

    for (i = 0; i < len * 2; i++) {
        product[i] = 0;
    }

    CVMX_MTM2(0);
    CVMX_MTM1(0);

    for (k = 0; k < len; k++) {
        CVMX_MTM0(mplier[k]);

        for (i = 0; i < len; i++) {
            CVMX_V3MULU(product[i + k], mpcand[i], product[i + k]);
        }
        CVMX_V3MULU(product[i + k], 0, product[i + k]);
    }
}


int64_t VCmp(const uint64_t * a, const uint64_t * b, int len)
{
    int i = 0;

    for (i = len - 1; i >= 0; i--) {
        if (a[i] != b[i]) {
            if (a[i] > b[i]) {
                return 1;
            } else {
                return -1;
            }
        }
    }
    return 0;
}

#ifndef _MIPS_ARCH_OCTEON3
static
void MMLoop(uint64_t * product, const uint64_t * base,
            const uint64_t * exponent, const uint64_t * mod,
            const uint64_t * recip, int len, int elen)
{
    int i, j = 0;
    int k, size;
    uint64_t *temp;
    int lenx3x8 = len * 3 * 8;
    size = (len * 64) + MODEXP_GUARD;
    temp = OPENSSL_malloc(size);
    if (temp == NULL) {
        return;
    }
    /*
     * We can not assume exponent to be of same length as modulus We need to
     * have atleast space for len*3 192bits tat 576bits == 64B
     */
    int elen8 = elen * 8;
    j = (elen8 / 64) - 1;
    if (elen % 8) {
        j++;
    }
    /*
     * Now find i. i is the MSb which is set in our msw of exponent
     */
    for (k = (elen8 - 1); k >= 0; k--) {
        if ((exponent[j] >> (k & 63)) & 1) {
            break;
        }
    }
    /*
     * When exponent is zero k=-1. we can set the output directly to 1, can
     * save computation time.
     */
    if (k < 0) {
        k = 0;
    }
    i = k;

    memset(temp, 0, size);

    for (i--; i >= 0; i--) {
        if (len == 3) {
            MontMul576((uint64_t *) temp, (uint64_t *) product,
                       (uint64_t *) product, (uint64_t *) mod,
                       (uint64_t *) recip);
        } else {
            MontMul((uint64_t *) temp, (uint64_t *) product,
                    (uint64_t *) product, (uint64_t *) mod,
                    (uint64_t *) recip, len);
        }
        if ((exponent[i / 64] >> (i & 63)) & 1) // if(!flipflop)
        {
            if (len == 3) {
                MontMul576((uint64_t *) product, (uint64_t *) temp,
                           (uint64_t *) base, (uint64_t *) mod,
                           (uint64_t *) recip);
            } else {
                MontMul((uint64_t *) product, (uint64_t *) temp,
                        (uint64_t *) base, (uint64_t *) mod,
                        (uint64_t *) recip, len);
            }
        } else {                // if (flipflop) copy temp to result
            memcpy(product, temp, lenx3x8);
        }
    }

    if (temp) {
        OPENSSL_free(temp);
    }
}

// len is in multiple of 192 bits
static int
_cvm_ModExp(uint64_t * product, uint64_t * base, uint64_t * exponent,
            uint64_t * mod, int len, int elen, int mlen, int blen,
            int convert)
{
    int success = 0;
#if !defined(TARGET_LINUX) || defined(USER_SPACE_MODEXP)
    if (1) {
#else
    if (!cvm_crypto_vmul_hwbug()) {
#endif
        uint64_t *innertemp = NULL;
        uint64_t recip[3];
        // first compute montgomery reciprocal
        int i, size;
        int msw = 0;
        uint64_t junk = 0;
        uint64_t *residue = NULL, *temp = NULL, *negmod = NULL; // ,*base=NULL;
        int lenx6 = len * 6;
        int lenx3 = len * 3;
        int lenx3x8 = len * 3 * 8;
        size = mlen + MODEXP_GUARD;
        /* Moved the memory from stack to heap */
        residue = (uint64_t *) OPENSSL_malloc(size);
        temp = (uint64_t *) OPENSSL_malloc(size);
        negmod = (uint64_t *) OPENSSL_malloc(size);
        uint64_t *temp_base = (uint64_t *) OPENSSL_malloc(lenx3x8);
        if (temp == NULL || negmod == NULL ||
            residue == NULL || temp_base == NULL) {
            goto err;
        }

        memset(residue, 0, size);
        memset(temp, 0, size);
        memset(negmod, 0, size);
        memset(temp_base, 0, lenx3x8);
        memset(recip, 0, 24);

        for (i = 0; i < len * 3; i++) {
            negmod[i] = 0;
            if (mod[i] != 0) {
                break;
            }
        }
        negmod[i] = ~mod[i] + 1;
        for (i++; i < len * 3 + 1; i++) {
            negmod[i] = ~mod[i];
        }

        // I need to know where most significant word of the modulus is
        for (msw = len * 3 - 1; msw >= 0; msw--) {
            if (mod[msw]) {
                break;
            }
        }
        if (msw == len * 3 - 1 && (mod[msw] >> 36)) {
            FATAL_ERROR;        // len needs to be incremented by 1
        }

        unsigned char normalize = !(mod[msw] >> 32);
        if (normalize) {
            for (i = len * 3 - 1; i >= 1; i--) {
                negmod[i] = (negmod[i] << 32) | (negmod[i - 1] >> 32);
            }
            negmod[0] = negmod[0] << 32;
        }
        recip[0] = 0;
        recip[1] = 0;
        recip[2] = 0;
        CVMX_MTM2(0);
        junk = 0;
        for (i = 63; i >= 0; i--) {
            recip[1] |= (uint64_t) 1 << i;
            CVMX_MTM0(recip[0]);
            CVMX_MTM1(recip[1]);
            CVMX_V3MULU(junk, negmod[msw - 1], 0);
            CVMX_V3MULU(junk, negmod[msw], 0);
            CVMX_V3MULU(junk, (int64_t) - 1, 0);
            CVMX_V3MULU(junk, (int64_t) - 1, (uint64_t) 1 << 32);
            if ((junk >> 63)) {
                recip[1] ^= (uint64_t) 1 << i;
            }
        }
        junk = 0;
        for (i = 63; i >= 0; i--) {
            recip[0] |= (uint64_t) 1 << i;
            CVMX_MTM0(recip[0]);
            CVMX_MTM1(recip[1]);
            CVMX_V3MULU(junk, negmod[msw - 1], 0);
            CVMX_V3MULU(junk, negmod[msw], 0);
            CVMX_V3MULU(junk, (int64_t) - 1, 0);
            CVMX_V3MULU(junk, (int64_t) - 1, (uint64_t) 1 << 32);
            if ((junk >> 63)) {
                recip[0] ^= (uint64_t) 1 << i;
            }
        }

        memset(residue, 0, sizeof(uint64_t) * (len * 3 + 1));
        residue[0] = 1;
        // each iteration I'll left shift things 64 bits and do a modular
        // reduction
        uint64_t q[4];
        innertemp = OPENSSL_malloc(size);
        if (innertemp == NULL) {
            goto err;
        }
        memset(q, 0, 4 * 8);
        memset(innertemp, 0, size);
        for (i = 0; i < lenx6; i++) {
            int k;
            memmove(residue + 1, residue, lenx3x8);
            residue[0] = 0;
            CVMX_MTM0(recip[0]);
            CVMX_MTM1(recip[1]);
            CVMX_MTM2(0);
            CVMX_V3MULU(q[0], (residue[msw] >> 32) | (residue[msw + 1] << 32),
                        0);
            CVMX_V3MULU(q[1], (residue[msw + 1] >> 32), 0);
            CVMX_V3MULU(q[2], 0, 0);
            CVMX_V3MULU(q[3], 0, 0);
            CVMX_MTM0(q[2]);
            CVMX_MTM1(q[3]);

            for (k = 0; k < lenx3 + 1; k++) {
                CVMX_V3MULU(innertemp[k], negmod[k], 0);
            }
            Vadd(residue, innertemp, lenx3 + 1);
            if (residue[msw + 2] != 0) {
                FATAL_ERROR;
            }
        }
        if (len == 3) {
            recip[0] = 0;
#if defined(USER_SPACE_MODEXP)
            CVMX_MTM2(0);
            CVMX_MTM1(0);
#endif
            for (i = 0; i < 64; i++) {
                uint64_t *sub = &recip[0];
                int k;
                char undo = 0;
                *sub |= (uint64_t) 1 << (i & 63);

                CVMX_MTM0(recip[0]);
#if !defined(USER_SPACE_MODEXP)
                CVMX_VMULU(product[0], mod[0], 1);
#else
                CVMX_V3MULU(product[0], mod[0], 1);
#endif

                for (k = 0; k < (i / 64); k++) {
                    undo = undo || product[k];
                }

                undo = undo || (product[i / 64] << (63 - (i & 63)));

                if (undo) {
                    *sub ^= (uint64_t) 1 << (i & 63);
                }
            }
        } else {
            for (i = 0; i < 3; i++) {
                recip[i] = 0;
            }
            for (i = 0; i < 192; i++) {
                uint64_t *sub = &recip[i / 64];
                int k;
                char undo = 0;
                *sub |= (uint64_t) 1 << (i & 63);

                CVMX_MTM0(recip[0]);
                CVMX_MTM1(recip[1]);
                CVMX_MTM2(recip[2]);
                CVMX_V3MULU(product[0], mod[0], 1);
                CVMX_V3MULU(product[1], mod[1], 0);
                CVMX_V3MULU(product[2], mod[2], 0);

                for (k = 0; k < (i / 64); k++) {
                    undo = undo || product[k];
                }

                undo = undo || (product[i / 64] << (63 - (i & 63)));

                if (undo) {
                    *sub ^= (uint64_t) 1 << (i & 63);
                }
            }
        }

        memset(product, 0, sizeof(uint64_t) * (len * 3));

        if (len == 3) {
            MontMul576((uint64_t *) product, (uint64_t *) base,
                       (uint64_t *) residue, (uint64_t *) mod,
                       (uint64_t *) recip);
        } else {
            MontMul((uint64_t *) product, (uint64_t *) base,
                    (uint64_t *) residue, (uint64_t *) mod,
                    (uint64_t *) recip, len);
        }
        memcpy(temp_base, product, lenx3x8);

        MMLoop(product, temp_base, exponent, mod, recip, len, elen);

        memset(residue, 0, sizeof(uint64_t) * (lenx3 + 1));
        residue[0] = 1;
        if (len == 3) {
            MontMul576((uint64_t *) temp, (uint64_t *) product,
                       (uint64_t *) residue, (uint64_t *) mod,
                       (uint64_t *) recip);
        } else {
            MontMul((uint64_t *) temp, (uint64_t *) product,
                    (uint64_t *) residue, (uint64_t *) mod,
                    (uint64_t *) recip, len);
        }
        memcpy(product, temp, lenx3x8);

        Vsub(temp, (uint64_t *) mod, lenx3 + 1);

        if (temp[len * 3 - 1] >> 63) ;
        else {
            memcpy(product, temp, lenx3x8);
        }
        success = 1;
 err:
        if (residue) {
            OPENSSL_free(residue);
        }
        if (temp) {
            OPENSSL_free(temp);
        }
        if (negmod) {
            OPENSSL_free(negmod);
        }
        if (temp_base) {
            OPENSSL_free(temp_base);
        }
        if (innertemp) {
            OPENSSL_free(innertemp);
        }
    }
/* Not used by CiscoSSL */
#if defined(TARGET_LINUX)
    else {
        int retc = 0;
        if (cryptfd <= 0) {
            retc = crypto_init();
            if (retc <= 0) {
                return retc;
            }
        }
# if _MIPS_SIM == _MIPS_SIM_NABI32
        cvm_crypto_op_t tokernel;
        tokernel.sizeofptr = sizeof(void *);
        tokernel.arg1 = (uint64_t) (uint32_t) product;
        tokernel.arg2 = (uint64_t) (uint32_t) base;
        tokernel.arg3 = (uint64_t) (uint32_t) exponent;
        tokernel.arg4 = (uint64_t) (uint32_t) mod;
        tokernel.arg5 = (uint64_t) (uint32_t) 0; /* recip */
        tokernel.arg6 = (int64_t) len;
        tokernel.arg7 = (int64_t) elen;
        tokernel.arg8 = (int64_t) mlen;
        tokernel.arg9 = (int64_t) blen;
        success =
            !crypto_mult(CRYPT_MODEXP, (uint64_t) (uint32_t) & tokernel);
# else
        cvm_crypto_op_t tokernel;
        tokernel.sizeofptr = sizeof(void *);
        tokernel.arg1 = (uint64_t) product;
        tokernel.arg2 = (uint64_t) base;
        tokernel.arg3 = (uint64_t) exponent;
        tokernel.arg4 = (uint64_t) mod;
        tokernel.arg5 = (uint64_t) 0; /* recip */
        tokernel.arg6 = (int64_t) len;
        tokernel.arg7 = (int64_t) elen;
        tokernel.arg8 = (int64_t) mlen;
        tokernel.arg9 = (int64_t) blen;
        success = !crypto_mult(CRYPT_MODEXP, (uint64_t) & tokernel);
# endif
    }
#endif
    return success;
}

int
cav_mod_exp(BIGNUM *res, const BIGNUM *base, const BIGNUM *exp,
            const BIGNUM *mod, BN_CTX *ctx)
{
    uint8_t *copy_base = NULL;
    uint8_t *copy_exp = NULL;
    uint8_t *copy_mod = NULL;
    int ret = 0;                // BN_mod_exp returns 1 on success and 0 on
    // error
    int mlen, elen, blen;

    copy_base = (uint8_t *) OPENSSL_malloc((mod->top + MUL_PAD) * 8);
    copy_exp = (uint8_t *) OPENSSL_malloc(mod->top * 8);
    copy_mod = (uint8_t *) OPENSSL_malloc((mod->top + MUL_PAD) * 8);
    if (copy_base == NULL || copy_exp == NULL || copy_mod == NULL) {
        goto modexp_err;
    }
    memset(copy_base, 0, (mod->top + MUL_PAD) * 8);
    memset(copy_exp, 0, (exp->top) * 8);
    memset(copy_mod, 0, (mod->top + MUL_PAD) * 8);

    memcpy(copy_exp, exp->d, exp->top * sizeof(BN_ULONG));
    memcpy(copy_mod, mod->d, mod->top * sizeof(BN_ULONG));
    memcpy(copy_base, base->d, base->top * sizeof(BN_ULONG));

    if (bn_wexpand(res, mod->dmax + MUL_PAD + 1) == NULL) {
        goto modexp_err;
    }
    bn_correct_top(res);

    mlen = ROUNDUP8(mod->top * sizeof(BN_ULONG));
    elen = ROUNDUP8(exp->top * sizeof(BN_ULONG));
    blen = ROUNDUP8(base->top * sizeof(BN_ULONG));
    _cvm_ModExp((uint64_t *) res->d, (uint64_t *) copy_base,
                (uint64_t *) copy_exp, (uint64_t *) copy_mod,
                (mlen << 3) / 192 + 1, elen, mlen, blen, 1);
    res->top = mod->top;
    bn_correct_top(res);

    ret = 1;                    // Success
 modexp_err:
    if (copy_base) {
        OPENSSL_free(copy_base);
    }
    if (copy_exp) {
        OPENSSL_free(copy_exp);
    }
    if (copy_mod) {
        OPENSSL_free(copy_mod);
    }
    return ret;
}
#endif


#ifdef _MIPS_ARCH_OCTEON3

typedef struct prec_array_ {
	uint64_t data[16][128 + 8];
} prec_array;



void
MMLoop_O3(uint64_t *product, const uint64_t *base, const uint64_t *exponent,
    const uint64_t *mod, const uint64_t *recip, int len, int elen)
{
    uint64_t *temp;

    int i, size, bits, psize;
    prec_array *precompute;
    int lenx8 = len * sizeof(uint64_t);


    size = sizeof(uint64_t) * (len + 6);

    if((temp = OPENSSL_malloc(size)) == NULL) {
        printf("memory allocation failed\n");
        return;
    }
    memset(temp, 0, size);

    if((precompute = OPENSSL_malloc(sizeof(prec_array))) == NULL) {
        free(temp);
        printf("memory allocation failed\n");
        return;
    }
    psize = (int)sizeof(precompute->data[0]);
    memset(precompute->data[0], 0, psize);
    memset(precompute->data[1], 0, psize);

    memcpy(precompute->data[0], product, size);
    memcpy(precompute->data[1], base, size);

    for(i = 2; i < 16; i++) {
       memset(precompute->data[i], 0, psize);
       memcpy(temp, precompute->data[i - 1], lenx8);

        if(len <= 8) {
            MontMul512((uint64_t *)precompute->data[i], (uint64_t *)temp,
                (uint64_t *)base, (uint64_t *)mod, (uint64_t *)recip);
        } else {
            MontMul_O3((uint64_t *)precompute->data[i], (uint64_t *)temp,
                (uint64_t *)base, (uint64_t *)mod, (uint64_t *)recip, len);
        }
    }

    for (i = (elen * 2) - 1; i >= 0; i--) {
        bits = ((exponent[i / 16] >> ((i & 15) << 2)) & 0xf);

        if (bits) break;
    }

    for (; i >= 0; i--) {
        bits = (exponent[i / 16] >> ((i & 15) << 2)) & 0xf;

        if (len <= 8) {
            MontMul512((uint64_t *)temp, (uint64_t *)product,
                (uint64_t *)product, (uint64_t *)mod, (uint64_t *)recip);

            MontMul512((uint64_t *)product, (uint64_t *)temp,
                (uint64_t *)temp, (uint64_t *)mod, (uint64_t *)recip);

            MontMul512((uint64_t *)temp,(uint64_t *)product,
                (uint64_t *)product, (uint64_t *)mod, (uint64_t *)recip);

            MontMul512((uint64_t *)product, (uint64_t *)temp,
                (uint64_t *)temp, (uint64_t *)mod, (uint64_t *)recip);

        } else {
            memset(temp, 0, size);
            MontMul_O3((uint64_t *)temp, (uint64_t *)product,
                (uint64_t *)product, (uint64_t *)mod, (uint64_t *)recip, len);

            memset(product, 0, lenx8);
            MontMul_O3((uint64_t *)product, (uint64_t *)temp,
                (uint64_t *)temp, (uint64_t *)mod,(uint64_t *)recip, len);

            memset(temp, 0, size);
            MontMul_O3((uint64_t *)temp, (uint64_t *)product,
                (uint64_t *)product, (uint64_t *)mod, (uint64_t *)recip, len);

            memset(product, 0, lenx8); 
            MontMul_O3((uint64_t *)product, (uint64_t *)temp,
                (uint64_t *)temp, (uint64_t *)mod, (uint64_t *)recip, len);
        }

        if (len <= 8) {
            MontMul512((uint64_t *)temp, (uint64_t *)product,
                (uint64_t *)precompute->data[bits], (uint64_t *)mod,
                (uint64_t *)recip);
        } else {
            memset(temp, 0, size);
            MontMul_O3((uint64_t *)temp, (uint64_t *)product,
                (uint64_t *)precompute->data[bits], (uint64_t *)mod,
                (uint64_t *)recip, len);
        }

        memcpy(product, temp, lenx8);
    }

    if (temp) OPENSSL_free(temp);
    if (precompute) OPENSSL_free(precompute);

}

static int
crypto_modexp_octeon3(uint64_t *product, uint64_t *base, uint64_t *exponent,
    uint64_t *mod, int len, int elen, int mlen, int blen, int convert)
{
    int success = 0;
    
    if(!cvm_crypto_vmul_hwbug())
    {
        uint64_t recip[6];
        // first compute montgomery reciprocal
        int i, size;
	int r;
        int lenx8;
	int normalize;
        int msw = 0;
        uint64_t junk = 0;
        uint64_t *residue = NULL, *temp = NULL,
                  *negmod = NULL, *temp_base = NULL;

        lenx8 = sizeof(uint64_t) * len;
        len   = (len + 3) & 0xfffc;
        size  = sizeof(uint64_t) * (len + 6);

        /* Moved the memory from stack to heap */
        residue   = (uint64_t *)OPENSSL_malloc(size);
        temp      = (uint64_t *)OPENSSL_malloc(size);
        negmod    = (uint64_t *)OPENSSL_malloc(size);
        temp_base = (uint64_t *)OPENSSL_malloc(size);

        if(!temp || !negmod || !residue || !temp_base)
        {
            printf("memory allocation failed\n");
            goto err;
        }

        memset(residue,   0, size);
        memset(temp,      0, size);
        memset(negmod,    0, size);
        memset(temp_base, 0, size);
        memset(recip,     0, sizeof(recip));

        for(i = 0; i < len; i++)
        {
            negmod[i] = 0;
            if(mod[i] != 0) break;
        }

        negmod[i] = ~mod[i] + 1;

        for(i++; i < len; i++)
            negmod[i] = ~mod[i];

        negmod[len] = ~(uint64_t)0;

        /* I need to know where most significant word of the modulus is */
        for(msw = len - 1; msw >= 0; msw--)
            if(mod[msw])
                break;

        normalize = mod[msw] < ((uint64_t)1 << 32);

        if(normalize)
        {
            for(i = msw; i >= 1; i--)
                negmod[i] = (negmod[i] << 32) | (negmod[i - 1] >> 32);

            negmod[0] = (negmod[0] << 32);
        }

        recip[0] = 0;
        recip[1] = 0;

        junk = 0;
        CVMX_MTM2_V3(0, 0);

        for(i = 63; i >= 0; i--)
        {
            recip[1] |= (uint64_t)1 << i;

            CVMX_MTM0_V3(recip[0], 0);
            CVMX_MTM1_V3(recip[1], 0);

            CVMX_VMULU(junk, negmod[msw - 1], 0);
            CVMX_VMULU(junk, negmod[msw], 0);
            CVMX_VMULU(junk, (int64_t)-1, 0);
            CVMX_VMULU(junk, (int64_t)-1, (uint64_t)1 << 32);

            if((junk >> 63))
                recip[1] ^= (uint64_t)1 << i;
        }

        junk = 0;

        for(i = 63; i >= 0; i--)
        {
            recip[0] |= (uint64_t)1 << i;

            CVMX_MTM0_V3(recip[0], 0);
            CVMX_MTM1_V3(recip[1], 0);

            CVMX_VMULU(junk, negmod[msw - 1], 0);
            CVMX_VMULU(junk, negmod[msw], 0);
            CVMX_VMULU(junk, (int64_t)-1, 0);
            CVMX_VMULU(junk, (int64_t)-1, (uint64_t)1 << 32);

            if((junk >> 63))
                recip[0] ^= (uint64_t)1 << i;
        }

        r = (len <= 8) ? 8 : ((len + 5) / 6) * 6;
        r = r * 2;

        residue[msw] = (normalize ? ((uint64_t)1 << 32) : 1);
        r -= msw;


        /* Each iteration I'll left shift things 64 bits and do
         * a modular reduction.
         */
        for(i = 0; i < r; i++)
        {
            int k;
            uint64_t q[4] = {0, 0, 0, 0};
            memmove(residue + 1, residue, lenx8);
            residue[0] = 0;

            CVMX_MTM0_V3(recip[0], 0);
            CVMX_MTM1_V3(recip[1], 0);
            CVMX_MTM2_V3(0, 0);

            CVMX_VMULU(q[0], (residue[msw] >> 32) | (residue[msw + 1] << 32), 0);
            CVMX_VMULU(q[1], (residue[msw + 1] >> 32), 0);
            CVMX_VMULU(q[2], 0, 0);
            CVMX_VMULU(q[3], 0, 0);

            CVMX_MTM0_V3(q[2], 0);
            CVMX_MTM1_V3(q[3], 0);
            CVMX_MTM2_V3(0, 0);

           for(k = 0; k < len + 1; k++)
                CVMX_VMULU(residue[k], negmod[k], residue[k]);

            if(residue[len] > 1) {FATAL_ERROR;}
        }

        if(normalize)
        {
            if(residue[0] << 32) {FATAL_ERROR;}

            for(i = 0; i < len; i++)
                residue[i] = (residue[i] >> 32) | (residue[i + 1] << 32);

            residue[len] = 0;
        }

        if(len <= 8)
        {
            recip[0] = 0;
            recip[1] = 0;

            CVMX_MTM2_V3(0, 0);

#if defined(USER_SPACE_MODEXP)
            CVMX_MTM0_V3(0, 0);
            CVMX_MTM1_V3(0, 0);
            CVMX_MTM2_V3(0, 0);
#endif

            for(i = 0; i < 128; i++)
            {
                uint64_t *sub = &recip[i / 64];
                int k;
                char undo = 0;
                uint64_t p[2];
                *sub |= (uint64_t)1 << (i & 63);

                CVMX_MTM0_V3(recip[0], 0);
                CVMX_MTM1_V3(recip[1], 0);

#if !defined(USER_SPACE_MODEXP)
                CVMX_VMULU(p[0], mod[0], 1);
                CVMX_VMULU(p[1], mod[1], 0);
#else
                CVMX_MTM2_V3(0, 0);
                CVMX_V3MULU(p[0], mod[0], 1);
                CVMX_V3MULU(p[1], mod[1], 0);
#endif

                for(k = 0; k < (i / 64); k++)
                    undo = undo || p[k];

                undo = undo || (p[i / 64] << (63 - (i & 63)));

                if(undo)
                    *sub ^= (uint64_t)1 << (i & 63);
            }
        }
        else
        {
           for(i = 0; i < 6; i++)
                recip[i] = (uint64_t)0;

            for(i = 0; i < 384; i++)
            {
                uint64_t *sub = &recip[i / 64];
                int k;
                char undo = 0;
                uint64_t p[6];

                *sub |= ((uint64_t)1 << (i & 63));

                CVMX_MTM0_V3(recip[0], recip[3]);
                CVMX_MTM1_V3(recip[1], recip[4]);
                CVMX_MTM2_V3(recip[2], recip[5]);

                CVMX_V3MULU(p[0], mod[0], 1);
                CVMX_V3MULU(p[1], mod[1], 0);
                CVMX_V3MULU(p[2], mod[2], 0);
                CVMX_V3MULU(p[3], mod[3], 0);
                CVMX_V3MULU(p[4], mod[4], 0);
                CVMX_V3MULU(p[5], mod[5], 0);

                for(k = 0; k < (i / 64); k++)
                    undo = undo || p[k];

                undo = undo || (p[i / 64] << (63 - (i & 63)));

                if(undo)
                    *sub ^= ((uint64_t)1 << (i & 63));
            }
        }

        if(len <= 8)
        {
            MontMul512((uint64_t *)product, (uint64_t *)base,
                (uint64_t *)residue, (uint64_t *)mod, (uint64_t *)recip);
        }
        else
        {
            MontMul_O3((uint64_t *)product, (uint64_t *)base,
                (uint64_t *)residue, (uint64_t *)mod, (uint64_t *)recip, len);
        }

        memcpy(temp_base, product, lenx8);

#if 0
#ifdef SPEEDUP_CODE
        zero_cache_lines(temp, cache_lines);
#else
        memset(temp, 0, size);
#endif
#endif

        temp[0] = 1;

        if(len <= 8)
        {
            MontMul512((uint64_t *)product, (uint64_t *)temp,
                (uint64_t *)residue, (uint64_t *)mod, (uint64_t *)recip);
        }
        else
        {
            MontMul_O3((uint64_t *)product, (uint64_t *)temp,
                (uint64_t *)residue, (uint64_t *)mod, (uint64_t *)recip, len);
        }

        MMLoop_O3(product, temp_base, exponent, mod, recip, len, elen);

        memset(residue, 0, size);

        residue[0] = 1;
        memset(temp, 0, size);

        if(len <= 8)
        {
            MontMul512((uint64_t *)temp, (uint64_t *)product,
                (uint64_t *)residue, (uint64_t *)mod, (uint64_t *)recip);
        }
        else
        {
            MontMul_O3((uint64_t *)temp, (uint64_t *)product,
                (uint64_t *)residue, (uint64_t *)mod, (uint64_t *)recip, len);
        }

        memcpy(product, temp, lenx8);

        Vsub(temp, mod, len + 1);

        if(!(temp[len] >> 63))
            memcpy(product, temp, lenx8);

        success = 1;

err:
        if(residue) {
	    OPENSSL_free(residue);
        }
	residue = NULL;
        if(temp) {
	    OPENSSL_free(temp);
        }
	temp = NULL;
        if(negmod) {
	    OPENSSL_free(negmod);
        }
	negmod = NULL;
        if(temp_base) {
	    OPENSSL_free(temp_base);
        }
	temp_base = NULL;


    }
#if defined(TARGET_LINUX)
    else
    {
        int fd = oct_crypto_init();
        if (fd < 0) {
            printf("Octeon Crypto Library Kernel module[cavmodexp] not available\n");
            return OCT_FAILURE;
        }

#if _MIPS_SIM == _MIPS_SIM_NABI32
        cvm_crypto_op_t tokernel;
        tokernel.sizeofptr = sizeof (void *);
        tokernel.arg1 = (uint64_t) (uint32_t) product;
        tokernel.arg2 = (uint64_t) (uint32_t) base;
        tokernel.arg3 = (uint64_t) (uint32_t) exponent;
        tokernel.arg4 = (uint64_t) (uint32_t) mod;
        tokernel.arg5 = (uint64_t) (uint32_t) 0;      /* recip */
        tokernel.arg6 = (int64_t) len;
        tokernel.arg7 = (int64_t) elen;
        tokernel.arg8 = (int64_t) mlen;
        tokernel.arg9 = (int64_t) blen;
        success = !crypto_mult (fd, CRYPT_MODEXP, (uint64_t) (uint32_t) & tokernel);
#else
        cvm_crypto_op_t tokernel;
        tokernel.sizeofptr = sizeof (void *);
        tokernel.arg1 = (uint64_t) product;
        tokernel.arg2 = (uint64_t) base;
        tokernel.arg3 = (uint64_t) exponent;
        tokernel.arg4 = (uint64_t) mod;
        tokernel.arg5 = (uint64_t) 0; /* recip */
        tokernel.arg6 = (int64_t) len;
        tokernel.arg7 = (int64_t) elen;
        tokernel.arg8 = (int64_t) mlen;
        tokernel.arg9 = (int64_t) blen;
        success = !crypto_mult (fd, CRYPT_MODEXP, (uint64_t) & tokernel);
#endif
    }
#endif
    return success;
}
#endif


/**
 * cvm_ModExp function
 *
 * Function which performs modulus exponent operation
 *
 * Prerequisites to use this function:
 * Parameters(mod,exponent,base) should be padded with 64 bytes of data. Also
 * the parameters should be reversed taken 8 bytes at a time before
 * passing to this function.
 *
 * @param  product  pointer to result of modular exponent operation.
 * @param  base     pointer to base value used in modular exponent operation.
 * @param  exponent pointer to exponent value used for modular exponent operation.
 * @param  mod      pointer to modulus value used in modular exponent operation.
 * @param  len      length  of product buffer
 * @param  elen     length  of exponent buffer
 * @param  mlen     length  of modulus buffer
 * @param  blen     length  of base buffer
 *
 *
 * @return 0 on success, -1 on failure
 */

int
cvm_ModExp(uint64_t * product, uint64_t * base, uint64_t * exponent,
           uint64_t * mod, int len, int elen, int mlen, int blen)
{
#ifdef _MIPS_ARCH_OCTEON3
        return crypto_modexp_octeon3(product, base, exponent, mod,
                            len, elen, mlen, blen, 0);
#else
        return _cvm_ModExp(product, base, exponent, mod,
                           len, elen, mlen, blen, 0);
#endif
}

static
void ModReduce(uint64_t * result, uint64_t * base, int baselen, uint64_t
               * mod, int modlen)
{
    uint64_t *tmp = NULL;
    uint64_t recip[3];
    uint64_t *negmod = NULL;
    int i, size, negsize;
    int msw, msm;
    int len = baselen;
    size = (baselen * 8) + MODEXP_GUARD;
    negsize = (baselen * 8 * 3);
    tmp = (uint64_t *) OPENSSL_malloc(size);
    negmod = (uint64_t *) OPENSSL_malloc(negsize);
    if (tmp == NULL || negmod == NULL) {
        goto err;
    }
    memset(tmp, 0, size);
    memset(negmod, 0, negsize);
    /* Calculates negmod i.e. -1 * modulus */
    for (i = 0; i < modlen; i++) {
        *(negmod + i) = 0;
        if (mod[i] != 0) {
            break;
        }
    }
    *(negmod + i) = ~mod[i] + 1;
    for (i++; i < modlen + 1; i++) {
        *(negmod + i) = ~mod[i];
    }
    /* Finding most significant bit of modulus */
    // I need to know where most significant word of the modulus is
    for (msm = modlen - 1; msm >= 0; msm--) {
        if (mod[msm]) {
            break;
        }
    }

    if ((msm == len - 1) && (mod[msm] >> 36)) {
        return;                 /* len needs to be incremented by 1 */
    }
    // I also need to put the modulus msb into the high 32b
    unsigned char normalize = !(mod[msm] >> 32);
    if (normalize) {
        for (i = len * 3 - 1; i >= 1; i--) {
            *(negmod + i) =
                (*(negmod + i) << 32) | (*(negmod + (i - 1)) >> 32);
        }
        *negmod = (*negmod) << 32;
    }
    /* Calculating reciprocal of mod */
    recip[0] = 0;
    recip[1] = 0;
    recip[2] = 0;
    CVMX_MTM2(0);
    for (i = 63; i >= 0; i--) {
        uint64_t junk;
        recip[1] |= (uint64_t) 1 << i;
        CVMX_MTM0(recip[0]);
        CVMX_MTM1(recip[1]);
        CVMX_V3MULU(junk, negmod[msm - 1], 0);
        CVMX_V3MULU(junk, negmod[msm], 0);
        CVMX_V3MULU(junk, (int64_t) - 1, 0);
        CVMX_V3MULU(junk, (int64_t) - 1, (uint64_t) 1 << 32);
        if ((junk >> 63)) {
            recip[1] ^= (uint64_t) 1 << i;
        }
    }
    for (i = 63; i >= 0; i--) {
        uint64_t junk;
        recip[0] |= (uint64_t) 1 << i;
        CVMX_MTM0(recip[0]);
        CVMX_MTM1(recip[1]);
        CVMX_V3MULU(junk, negmod[msm - 1], 0);
        CVMX_V3MULU(junk, negmod[msm], 0);
        CVMX_V3MULU(junk, (int64_t) - 1, 0);
        CVMX_V3MULU(junk, (int64_t) - 1, (uint64_t) 1 << 32);
        if ((junk >> 63))
            recip[0] ^= (uint64_t) 1 << i;
    }
    /*
     * This is equivalent to the standard paper-pencil division method
     * Starting from the highest bits of base, calculate quotient,
     * subtract (quotient * mod ) from base. Now we have a base with
     * fewer bits than the original. Keep going till base < mod
     */
    {
        uint64_t q[4];
        uint64_t *temp = NULL;
        int k;
        temp = (uint64_t *) OPENSSL_malloc(size);
        if (temp == NULL) {
            goto err;
        }
        if (normalize) {
            for (i = len * 3 - 1; i >= 1; i--) {
                result[i] = (base[i] << 32) | (base[i - 1] >> 32);
            }
            result[0] = base[0] << 32;
        } else {
            memcpy(result, base, sizeof(uint64_t) * (len + 1));
        }
        result[len] = 0;

        CVMX_MTM0(recip[0]);
        CVMX_MTM1(recip[1]);
        CVMX_MTM2(0);
        for (k = 0; k < modlen; k++) {
            CVMX_V3MULU(temp[k], mod[k], 0);
        }
        CVMX_V3MULU(temp[k], 0, 0);
        k++;
        CVMX_V3MULU(temp[k], 0, 0);
        k++;

        /* Find out where msw of base is */
        msw = baselen - 1;
        do {
            memset(temp, 0, size);

            /*
             * Take reciprocal and multiply with high bits of "base", to get
             * Q ~= base/mod
             */
            CVMX_MTM0(recip[0]);
            CVMX_MTM1(recip[1]);
            CVMX_MTM2(0);
            CVMX_V3MULU(q[0], (result[msw] >> 32) | (result[msw + 1] << 32),
                        0);
            CVMX_V3MULU(q[1], (result[msw + 1] >> 32), 0);
            CVMX_V3MULU(q[2], 0, 0);
            CVMX_V3MULU(q[3], 0, 0);

            CVMX_MTM0(q[2]);
            CVMX_MTM1(q[3]);
            CVMX_MTM2(0);

            for (k = 0; k < baselen; k++) {
                CVMX_V3MULU(temp[k], mod[k], 0);
            }
            for (k = 0; k < baselen; k++) {
                CVMX_V3MULU(temp[k], negmod[k], 0);
            }
            for (k = 0; k < baselen; k++) { // could be smarter and stop
                // earlier
                CVMX_V3MULU(result[k + msw - msm], negmod[k],
                            result[k + msw - msm]);
            }
            result[msw + 4] = 0;
            result[msw + 3] = 0;
            // this gets rid of junk left behind because
            // negmod was signextended out to baselen bits
            result[msw + 2] = 0;
            msw--;
        } while (msw >= (modlen - 1));
        OPENSSL_free(temp);
    }
    if (normalize) {
        for (i = 0; i < len * 3; i++) {
            result[i] = (result[i + 1] << 32) | (result[i] >> 32);
        }
    }
    memcpy(tmp, result, len * sizeof(uint64_t));

    /* Now finish off by subtracting modulus */
    while (1) {
        Vsub(tmp, mod, modlen + 2);
        if (tmp[modlen] >> 63) {
            break;
        } else {
            memcpy(result, tmp, modlen * sizeof(uint64_t));
        }
    }
 err:
    if (tmp) {
        OPENSSL_free(tmp);
    }
    if (negmod) {
        OPENSSL_free(negmod);
    }
}

/*
 * CRT Algorithm: m1 = c^dP mod p m2 = c^dQ mod q h = qInv(m1 - m2) mod p m =
 * m2 + hq
 */
static
int _cvm_ModExpCrt(uint64_t * product, uint64_t * base,
                   cvm_rsa_key_t * rkey, int nonwalign, int convert)
{
    int retc = 1;
#if !defined(TARGET_LINUX) || defined(USER_SPACE_MODEXP)
    if (1) {
#else
    if (!cvm_crypto_vmul_hwbug()) {
#endif
        uint64_t *mod1 = NULL;
        uint64_t *mod2 = NULL;
        uint64_t *tmp = NULL;
        int len = rkey->len / 2, size = 0, tmpsize;
        int wlen = (rkey->len / 2) >> 3; // length in (8 byte) words
        int blen = rkey->len >> 3;
        len = len << 3;
        if (nonwalign) {
            wlen++;
            blen += 2;          // zero extended in case of non 8B
        } else if (rkey->len % 16) {
            wlen++;
            blen += 2;          // zero extended in case of non 8B
        }
        size = (blen * 8) + MODEXP_GUARD;
        tmpsize = (blen * 24);
        mod1 = (uint64_t *) OPENSSL_malloc(size);
        mod2 = (uint64_t *) OPENSSL_malloc(size);
        tmp = (uint64_t *) OPENSSL_malloc(tmpsize);
        if (mod1 == NULL || mod2 == NULL || tmp == NULL) {
            goto err;
        }
        /* First do modular reduction */
        ModReduce(tmp, base, blen, rkey->p, wlen);

        if (cvm_ModExp(mod1, tmp, rkey->expp, rkey->p, len / 192 + 1,
                       rkey->eplen * 8, rkey->len, wlen * 8) <= 0) {
            return 0;
        }

        /* Same procedure for second round */
        int wlen1;
        wlen1 = find_msw(rkey->q);
        ModReduce(tmp, base, blen, rkey->q, wlen1);
        if (cvm_ModExp(mod2, tmp, rkey->expq, rkey->q, len / 192 + 1,
                       rkey->eqlen * 8, rkey->len, wlen * 8) <= 0) {
            return 0;
        }
        if (VCmp(mod1, mod2, wlen) < 0) {
            memcpy(tmp, mod2, wlen * sizeof(uint64_t));
            Vsub(tmp, mod1, wlen);
            /* Since it is negative, add p to make it positive ?? */
            if (VCmp(rkey->p, tmp, wlen) < 0) {
                /*
                 * p is less than tmp, so a single correction will not make
                 * it positive
                 */
                memcpy(mod1, rkey->p, wlen * sizeof(uint64_t));
                Vadd(mod1, rkey->p, wlen);
                mod1[wlen + 1] = 0;
                tmp[wlen] = 0;
                Vsub(mod1, tmp, wlen + 1);
            } else {
                memcpy(mod1, rkey->p, wlen * sizeof(uint64_t));
                mod1[wlen] = 0;
                tmp[wlen] = 0;
                Vsub(mod1, tmp, wlen + 1);
            }

        } else {
            Vsub(mod1, mod2, wlen);
        }

        /*
         * zero off any higher bytes that might have been set during mod
         * operations
         */
        memset(&mod1[wlen], 0, wlen * sizeof(uint64_t));
        memset(&mod2[wlen], 0, wlen * sizeof(uint64_t));

        VMul(product, mod1, rkey->coeff, wlen);

        ModReduce(tmp, product, wlen * 2, rkey->p, wlen);

        /* Multiply tmp & rkey->q and put result in product */
        VMul(product, tmp, rkey->q, wlen);

        /* Assuming Vadd needs length in words */
        Vadd(product, mod2, wlen * 2);
 err:
        if (mod1) {
            OPENSSL_free(mod1);
        }
        if (mod2) {
            OPENSSL_free(mod2);
        }
        if (tmp) {
            OPENSSL_free(tmp);
        }
    }
/* Not used by CiscoSSL */
#if defined(TARGET_LINUX)
    else {
        if (cryptfd <= 0) {
            retc = crypto_init();
            if (retc <= 0) {
# ifdef PRINT_DEBUG
                printf("Create device /dev/octcrypto \n");
# endif
                return 0;
            }
        }
# if _MIPS_SIM == _MIPS_SIM_NABI32
        cvm_crypto_op_t tokernel;
        tokernel.sizeofptr = sizeof(void *);
        tokernel.arg1 = (uint64_t) (uint32_t) product;
        tokernel.arg2 = (uint64_t) (uint32_t) base;
        tokernel.arg3 = (uint64_t) (uint32_t) rkey;
        tokernel.arg6 = (int64_t) nonwalign;
        // ioctl on success returns 0
        crypto_mult(CRYPT_MODEXPCRT, (uint64_t) (uint32_t) & tokernel);
# else
        cvm_crypto_op_t tokernel;
        tokernel.sizeofptr = sizeof(void *);
        tokernel.arg1 = (uint64_t) product;
        tokernel.arg2 = (uint64_t) base;
        tokernel.arg3 = (uint64_t) rkey;
        tokernel.arg6 = (int64_t) nonwalign;
        // ioctl on success returns 0
        crypto_mult(CRYPT_MODEXPCRT, (uint64_t) & tokernel);

# endif
    }
#endif
    return retc;
}

int
cvm_ModExpCrt(uint64_t * product, uint64_t * base,
              cvm_rsa_key_t * rkey, int nonwalign)
{
    return _cvm_ModExpCrt(product, base, rkey, nonwalign, 0);
}

static int rsa_to_cry_private(RSA *rsa, cvm_rsa_key_t * rkey)
{
    BN_ULONG *buf;
    int buflen;

    buf = rsa->n->d;
    buflen = rsa->n->top;
    memset(rkey, 0, sizeof(cvm_rsa_key_t));
    rkey->len = BN_num_bytes(rsa->n);

    /* Used for CRT */
    /*
     * Will all these be the same size? Just use one for loop in that case
     */
    buf = rsa->dmp1->d;
    buflen = rsa->dmp1->top;

    if (buflen > 256) {
        return -1;
    }

    rkey->eplen = buflen;
    rkey->expp = (uint64_t *) OPENSSL_malloc((buflen * 8) + MODEXP_GUARD);
    memset(rkey->expp, 0, (buflen * 8) + MODEXP_GUARD);
    memcpy(rkey->expp, buf, buflen * sizeof(BN_ULONG));

    buf = rsa->dmq1->d;
    buflen = rsa->dmq1->top;

    if (buflen > 256) {
        return -1;
    }

    /* ModExp (Octeon) Expects length in 8-byte words */
    rkey->eqlen = buflen;
    rkey->expq = (uint64_t *) OPENSSL_malloc((buflen * 8) + MODEXP_GUARD);
    memset(rkey->expq, 0, (buflen * 8) + MODEXP_GUARD);
    memcpy(rkey->expq, buf, buflen * sizeof(BN_ULONG));

    buf = rsa->p->d;
    buflen = rsa->p->top;

    if (buflen > 256) {
        return -1;
    }
    rkey->p = (uint64_t *) OPENSSL_malloc(buflen * 8 + MODEXP_GUARD);
    memset(rkey->p, 0, buflen * 8 + MODEXP_GUARD);
    memcpy(rkey->p, buf, buflen * sizeof(BN_ULONG));
    buf = rsa->q->d;
    buflen = rsa->q->top;

    if (buflen > 256) {
        return -1;
    }

    rkey->q = (uint64_t *) OPENSSL_malloc(buflen * 8 + MODEXP_GUARD);
    memset(rkey->q, 0, buflen * 8 + MODEXP_GUARD);
    memcpy(rkey->q, buf, buflen * sizeof(BN_ULONG));
    buf = rsa->iqmp->d;
    buflen = rsa->iqmp->top;
    if (buflen > 256) {
        return -1;
    }
    rkey->coeff = (uint64_t *) OPENSSL_malloc(buflen * 8 + MODEXP_GUARD);
    memset(rkey->coeff, 0, buflen * 8 + MODEXP_GUARD);
    memcpy(rkey->coeff, buf, buflen * sizeof(BN_ULONG));
    return 1;
}

#ifdef _MIPS_ARCH_OCTEON3
/*
   CRT Algorithm:
   m1 = c^dP mod p
   m2 = c^dQ mod q
   h = qInv(m1 - m2) mod p
   m = m2 + hq
 */
int
_crypto_modexp_crt (uint64_t * product, uint64_t * base,
  cvm_rsa_key_t * rkey, int nonwalign, int convert)
{
    int retc=1; 
#if !defined(TARGET_LINUX) || defined(USER_SPACE_MODEXP)
    if(1)
#else
    if(!cvm_crypto_vmul_hwbug())
#endif
    {
        uint64_t *mod1 = NULL;
        uint64_t *mod2 = NULL;
        uint64_t *tmp = NULL;
        int len = (rkey->len + 1) / 2, size = 0, tmpsize;
        int wlen = (rkey->len / 2) >> 3;      // length in (8 byte) words
        int blen = rkey->len >> 3;
        int wlen1;

        len = len << 3;
        if(nonwalign)
        {
            wlen++;
            blen += 2;                  //zero extended in case of non 8B 
        }
        else if(rkey->len % 16)
        {
            wlen++;
            blen += 2;                  //zero extended in case of non 8B 
        }

        size    = (blen * 8) + MODEXP_GUARD;
        tmpsize = (blen * 24);

        mod1 = (uint64_t *)OPENSSL_malloc(size);
        mod2 = (uint64_t *)OPENSSL_malloc(size);
        tmp  = (uint64_t *)OPENSSL_malloc(tmpsize);

        if(mod1 == NULL || mod2 == NULL || tmp == NULL)
            goto err;

        /* First do modular reduction */
        ModReduce(tmp, base, blen, rkey->p, wlen);

        if(cvm_ModExp(mod1, tmp, rkey->expp, rkey->p, (len + 63) / 64,
            rkey->eplen * 8, rkey->len, wlen * 8) <= 0)
        {
            return 0;
        }

        /* Same procedure for second round */
        wlen1 = find_msw (rkey->q);
        ModReduce(tmp, base, blen, rkey->q, wlen1);
        if(cvm_ModExp(mod2, tmp, rkey->expq, rkey->q, (len + 63) / 64,
            rkey->eqlen * 8, rkey->len, wlen * 8) <= 0)
        {
            return 0;
        }

        if(VCmp (mod1, mod2, wlen) < 0)
        {
            memcpy(tmp, mod2, wlen * sizeof (uint64_t));
            Vsub(tmp, mod1, wlen);
            /* Since it is negative, add p to make it positive ?? */
            if(VCmp(rkey->p, tmp, wlen) < 0)
            {
                /* p is less than tmp, so a single correction will not make it
                   positive */
                memcpy (mod1, rkey->p, wlen * sizeof (uint64_t));
                Vadd(mod1, rkey->p, wlen);

                mod1[wlen + 1] = 0;
                tmp[wlen] = 0;

                Vsub(mod1, tmp, wlen + 1);
            }
            else
            {
                memcpy (mod1, rkey->p, wlen * sizeof (uint64_t));
                mod1[wlen] = 0;
                tmp[wlen] = 0;

                Vsub(mod1, tmp, wlen + 1);
            }
        }
        else
        {
            Vsub(mod1, mod2, wlen);
        }

        /* zero off any higher bytes that might have been set during
         * mod operations
         */
        memset(&mod1[wlen], 0, wlen * sizeof (uint64_t));
        memset(&mod2[wlen], 0, wlen * sizeof (uint64_t));

        VMul(product, mod1, rkey->coeff, wlen);

        ModReduce(tmp, product, wlen * 2, rkey->p, wlen);

        /* Multiply tmp & rkey->q and put result in product */
        VMul(product, tmp, rkey->q, wlen);

        /* Assuming Vadd needs length in words */
        Vadd(product, mod2, wlen * 2);

err:
        if(mod1)
          OPENSSL_free(mod1);
        if(mod2)
          OPENSSL_free(mod2);
        if(tmp)
          OPENSSL_free(tmp);
    }
#if defined(TARGET_LINUX)
  else {
     if (cryptfd <= 0) {
       retc = crypto_init ();
       if (retc <= 0) {
       #ifdef PRINT_DEBUG
         printf ("Create device /dev/octcrypto \n");
       #endif
        return 0;
       }
     }
#if _MIPS_SIM == _MIPS_SIM_NABI32
     cvm_crypto_op_t tokernel;
     tokernel.sizeofptr = sizeof (void *);
     tokernel.arg1 = (uint64_t) (uint32_t) product;
     tokernel.arg2 = (uint64_t) (uint32_t) base;
     tokernel.arg3 = (uint64_t) (uint32_t) rkey;
     tokernel.arg6 = (int64_t) nonwalign;
     //ioctl on success returns 0
     crypto_mult (CRYPT_MODEXPCRT, (uint64_t) (uint32_t) & tokernel);
#else
     cvm_crypto_op_t tokernel;
     tokernel.sizeofptr = sizeof (void *);
     tokernel.arg1 = (uint64_t) product;
     tokernel.arg2 = (uint64_t) base;
     tokernel.arg3 = (uint64_t) rkey;
     tokernel.arg6 = (int64_t) nonwalign;
     //ioctl on success returns 0
     crypto_mult (CRYPT_MODEXPCRT, (uint64_t) & tokernel);

#endif
  }
#endif
  return retc;
}

static int
crypto_modexp_crt (uint64_t * product, uint64_t * base,
  cvm_rsa_key_t * rkey, int nonwalign)
{
  return _crypto_modexp_crt (product, base, rkey, nonwalign, 0);
}
static 
void rsa_cry_private_free(cvm_rsa_key_t *key)
{
    if (key == NULL)  return;
    if (key->mod)     OPENSSL_free(key->mod);
    if (key->privexp) OPENSSL_free(key->privexp);
    if (key->pubexp)  OPENSSL_free(key->pubexp);
    if (key->expp)    OPENSSL_free(key->expp);
    if (key->expq)    OPENSSL_free(key->expq);
    if (key->p)       OPENSSL_free(key->p);
    if (key->q)       OPENSSL_free(key->q);
    if (key->coeff)   OPENSSL_free(key->coeff);
}

static 
int oct_mod_exp_crt(BIGNUM *res, BIGNUM *base, RSA *rsa)
{
   cvm_rsa_key_t rkey;
   uint8_t *tmpbase = NULL;
   int num, ret = OCT_FAILURE;

   rsa_to_cry_private(rsa, &rkey);
   num = BN_num_bytes(rsa->n);
   if (bn_wexpand(res, rsa->n->dmax + MUL_PAD) == NULL) {
      goto modexp_crt_err;
   }
   tmpbase = (uint8_t *) OPENSSL_malloc((base->top + MUL_PAD) * 8);
   if (tmpbase == NULL)
      goto modexp_crt_err;
   memset(tmpbase, 0, (base->top + MUL_PAD) * 8);
   memcpy(tmpbase, base->d, base->top * sizeof (BN_ULONG));
#if _MIPS_SIM == _MIPS_SIM_NABI32
     convert_bn32_to_hw ((uint8_t *) tmpbase, ROUNDUP8 (rkey.len));
     /* Converting rkey to H/W specific format */
     convert_bn32_to_hw ((uint8_t *) (rkey.expp),
                          rkey.eplen * sizeof (BN_ULONG));
     convert_bn32_to_hw ((uint8_t *) (rkey.expq),
                         rkey.eqlen * sizeof (BN_ULONG));
    /* eplen, eqlen should be 8-byte words for Octeon ModExp in Kernel */
     rkey.eplen = (rkey.eplen + 1) * sizeof (BN_ULONG) / 8;
     rkey.eqlen = (rkey.eqlen + 1) * sizeof (BN_ULONG) / 8;
     convert_bn32_to_hw ((uint8_t *) (rkey.p), rkey.len);
     convert_bn32_to_hw ((uint8_t *) (rkey.q), rkey.len);
     convert_bn32_to_hw ((uint8_t *) (rkey.coeff), rkey.len);
#endif
     ret = crypto_modexp_crt((uint64_t *)res->d, (uint64_t *)tmpbase, &rkey, num % 8);
     if (ret != OCT_SUCCESS) {
        goto modexp_crt_err;
     }
#if _MIPS_SIM == _MIPS_SIM_NABI32
     convert_hw_to_bn32 ((uint8_t *)res->d, rkey.len);
#endif
     res->top = rsa->n->top;
    ret = OCT_SUCCESS;
   
modexp_crt_err:
   if (tmpbase)
      OPENSSL_free(tmpbase);
    rsa_cry_private_free(&rkey);

   return ret;
}

#else
static int oct_mod_exp_crt(BIGNUM *res, BIGNUM *base, RSA *rsa)
{
    cvm_rsa_key_t rkey;
    uint8_t *tmpbase = NULL;
    int num, ret = OCT_FAILURE;

    rsa_to_cry_private(rsa, &rkey);
    num = BN_num_bytes(rsa->n);
    if (bn_wexpand(res, rsa->n->dmax + MUL_PAD) == NULL) {
        goto modexp_crt_err;
    }
    tmpbase = (uint8_t *) OPENSSL_malloc((base->top + MUL_PAD) * 8);
    if (tmpbase == NULL) {
        goto modexp_crt_err;
    }
    memset(tmpbase, 0, (base->top + MUL_PAD) * 8);
    memcpy(tmpbase, base->d, base->top * sizeof(BN_ULONG));
    ret = cvm_ModExpCrt(res->d, (uint64_t *) tmpbase, &rkey, num % 8);
    if (ret != OCT_SUCCESS) {
        goto modexp_crt_err;
    }
    res->top = rsa->n->top;
    ret = OCT_SUCCESS;

 modexp_crt_err:
    if (tmpbase) {
        OPENSSL_free(tmpbase);
    }
    if (rkey.mod) {
        OPENSSL_free(rkey.mod);
    }
    if (rkey.privexp) {
        OPENSSL_free(rkey.privexp);
    }
    if (rkey.pubexp) {
        OPENSSL_free(rkey.pubexp);
    }
    if (rkey.expp) {
        OPENSSL_free(rkey.expp);
    }
    if (rkey.expq) {
        OPENSSL_free(rkey.expq);
    }
    if (rkey.p) {
        OPENSSL_free(rkey.p);
    }
    if (rkey.q) {
        OPENSSL_free(rkey.q);
    }
    if (rkey.coeff) {
        OPENSSL_free(rkey.coeff);
    }
    return ret;
}
#endif
/**
 * oct_rsa_mod_exp
 *
 * RSA computation of r0 = r0 ^ I mod rsa->n
 *
 * \retval SUCCESS 1
 * \retval FAILURE 0
 */
int oct_rsa_mod_exp(BIGNUM *r0, const BIGNUM *I, RSA *rsa, BN_CTX *ctx)
{
    int ret;
    ret = oct_mod_exp_crt(r0, (BIGNUM *)I, rsa);
    return ret;
}

void oct_bn_sqr_words(BN_ULONG *r, const BN_ULONG *a, int n)
{
    CAV_ASSERT(n >= 0);
    if (n <= 0)
        return;

    while (n) {
        cav_sqr64(r[0], r[1], a[0]);
        a++;
        r += 2;
        n--;
    }
    return;
}

BN_ULONG
oct_bn_mul_add_words(BN_ULONG *rp, const BN_ULONG *ap, int num, BN_ULONG w)
{
    BN_ULONG c = 0;
    BN_ULONG bl, bh;

    CAV_ASSERT(num >= 0);
    if (num <= 0)
        return ((BN_ULONG)0);

    bl = LBITS(w);
    bh = HBITS(w);

    while (num) {
        cav_mul_add(rp[0], ap[0], bl, bh, c);
        ap++;
        rp++;
        num--;
    }
    return (c);
}

/**
 *   oct_bn_mod_exp
 *
 *   RSA computation of r = a ^ p mod m
 *   \retval SUCCESS 1
 *   \retval FAILURE 0
 */

int
oct_bn_mod_exp(BIGNUM *r, const BIGNUM *a, const BIGNUM *p, const BIGNUM *m,
               BN_CTX *ctx, BN_MONT_CTX *m_ctx)
{
    int ret;
    ret = BN_mod_exp(r, (BIGNUM *)a, (BIGNUM *)p, (BIGNUM *)m, ctx);
    return ret;
}

/**
 *   oct_dsa_mod_exp
 *
 *   DSA computation of rr = a1^p1 * a2^p2 mod m
 *   \retval SUCCESS 1
 *   \retval FAILURE 0
 */
int oct_dsa_mod_exp(DSA *dsa, BIGNUM *rr, BIGNUM *a1, BIGNUM *p1,
                    BIGNUM *a2, BIGNUM *p2, BIGNUM *m, BN_CTX *ctx,
                    BN_MONT_CTX *in_mont)
{
    BIGNUM t2;
    int ret = OCT_FAILURE;

    if (OCTEON_HWA_CAPABLE) {
        /*
         * rr = t1, a1 = dsa->g , p1 =u1 , a2 = dsa->pub_key, p2 = u2, m =
         * dsa->p
         */
        BN_init(&t2);
        /* let rr = g ^ p1 mod m */
        if (!BN_mod_exp(rr, dsa->g, p1, dsa->p, ctx))
            goto err;
        /* let t2 = a2 ^ p2 mod m */
        if (!BN_mod_exp(&t2, dsa->pub_key, p2, dsa->p, ctx))
            goto err;
        /* let rr = rr * t2 mod m */
        if (!BN_mod_mul(rr, rr, &t2, dsa->p, ctx))
            goto err;
        ret = OCT_SUCCESS;
 err:
        BN_free(&t2);
        return ret;
    } else {
        return (BN_mod_exp2_mont(rr, a1, p1, a2, p2, m, ctx, in_mont));
    }

}

static int
oct_mod_exp_internal(BIGNUM *res, BIGNUM *base, BIGNUM *exp, BIGNUM *mod, BN_CTX * ctx)
{
    uint8_t* copy_base = NULL;
    uint8_t* copy_exp = NULL;
    uint8_t* copy_mod = NULL;
    int ret = OCT_FAILURE; 
    int len, mlen, elen, blen;
    int i = 0;
 
    copy_base = (uint8_t *)OPENSSL_malloc((ROUNDUP2(mod->top) + MUL_PAD) * 8);
    copy_exp  = (uint8_t *)OPENSSL_malloc((ROUNDUP2(mod->top) + MUL_PAD) * 8);
    copy_mod  = (uint8_t *)OPENSSL_malloc((ROUNDUP2(mod->top) + MUL_PAD) * 8);
     
    if(copy_base == NULL || copy_exp == NULL || copy_mod == NULL) 
         goto  modexp_err;
 
    memset(copy_base, 0, (ROUNDUP2(mod->top) + MUL_PAD) * 8);
    memset(copy_exp,  0, (ROUNDUP2(mod->top) + MUL_PAD) * 8);
    memset(copy_mod,  0, (ROUNDUP2(mod->top) + MUL_PAD) * 8);

    memcpy(copy_exp,  exp->d,  exp->top  * sizeof (BN_ULONG));
    memcpy(copy_mod,  mod->d,  mod->top  * sizeof (BN_ULONG));
    memcpy(copy_base, base->d, base->top * sizeof (BN_ULONG));

    if (bn_wexpand(res, ROUNDUP2(mod->dmax) + MUL_PAD) == NULL)
        goto modexp_err;

    memset(res->d, 0, (mod->dmax + MUL_PAD) * sizeof(BN_ULONG));

    mlen = BN_num_bytes(mod);
    elen = BN_num_bytes(exp);
    blen = BN_num_bytes(base);
    len  = ((((mod->top * sizeof(BN_ULONG)) * 8) + 63) / 64);


    #if _MIPS_SIM == _MIPS_SIM_NABI32
    /* Convert all input data to format specific to crypto kernel module */
    convert_bn32_to_hw(copy_base, blen);
    convert_bn32_to_hw(copy_exp, elen);
    convert_bn32_to_hw(copy_mod, mlen);
    #endif
 
    ret = cvm_ModExp((uint64_t*)res->d, (uint64_t*)copy_base,
                         (uint64_t*)copy_exp, (uint64_t*)copy_mod,
                         len, elen, mlen, blen);
     
    if (ret != OCT_SUCCESS)
        goto modexp_err;
 
    #if _MIPS_SIM == _MIPS_SIM_NABI32
     convert_hw_to_bn32((uint8_t *)res->d, mlen);
    #endif
 
    res->top = mod->top;
    for(i = mod->top-1; i >=0; i--) {
        if(res->d[i])
            break;
        res->top--;
    }
    ret = OCT_SUCCESS; // Success

modexp_err:
    if (ret  == OCT_FAILURE) 
        printf("cav_mod_exp error !\n");
    if (copy_base)        OPENSSL_free(copy_base);
    if (copy_exp)         OPENSSL_free(copy_exp);
    if (copy_mod)         OPENSSL_free(copy_mod);
    return ret;
}


/* This code is obtained from crypto component's ModExp */
int oct_mod_exp(BIGNUM *r, BIGNUM *a, BIGNUM *p, BIGNUM *m, BN_CTX *ctx)
{
    
    BIGNUM *pdash,*mdash;
    int ret;
    
    bn_check_top (a);
    bn_check_top (p);
    bn_check_top (m);

    pdash = BN_new();
    mdash = BN_new();

    if (!pdash || !mdash) {
        if(pdash) BN_free(pdash);
        if(mdash) BN_free(mdash);
        /* return error */
        return OCT_FAILURE;
    }

    if(!BN_sub(mdash, m, BN_value_one())) {
        BN_free(pdash);
        BN_free(mdash);
        return OCT_FAILURE;       
    }

    if(!BN_div(NULL, pdash, p, mdash, ctx)) {
        BN_free(pdash);
        BN_free(mdash);
        return OCT_FAILURE;
    }

    ret = oct_mod_exp_internal(r, a, pdash, m, ctx);

    BN_free(pdash);
    BN_free(mdash);
    return ret;  
}

