/* crypto/bn/bn_prime.c */
/* Copyright (C) 1995-1998 Eric Young (eay@cryptsoft.com)
 * All rights reserved.
 *
 * This package is an SSL implementation written
 * by Eric Young (eay@cryptsoft.com).
 * The implementation was written so as to conform with Netscapes SSL.
 *
 * This library is free for commercial and non-commercial use as long as
 * the following conditions are aheared to.  The following conditions
 * apply to all code found in this distribution, be it the RC4, RSA,
 * lhash, DES, etc., code; not just the SSL code.  The SSL documentation
 * included with this distribution is covered by the same copyright terms
 * except that the holder is Tim Hudson (tjh@cryptsoft.com).
 *
 * Copyright remains Eric Young's, and as such any Copyright notices in
 * the code are not to be removed.
 * If this package is used in a product, Eric Young should be given attribution
 * as the author of the parts of the library used.
 * This can be in the form of a textual message at program startup or
 * in documentation (online or textual) provided with the package.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    "This product includes cryptographic software written by
 *     Eric Young (eay@cryptsoft.com)"
 *    The word 'cryptographic' can be left out if the rouines from the library
 *    being used are not cryptographic related :-).
 * 4. If you include any Windows specific code (or a derivative thereof) from
 *    the apps directory (application code) you must include an acknowledgement:
 *    "This product includes software written by Tim Hudson (tjh@cryptsoft.com)"
 *
 * THIS SOFTWARE IS PROVIDED BY ERIC YOUNG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 */
/* ====================================================================
 * Copyright (c) 1998-2001 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include <stdio.h>
#include <time.h>
#include "cryptlib.h"
#include "bn_lcl.h"
#include <openssl/rand.h>

/*
 * NB: these functions have been "upgraded", the deprecated versions (which
 * are compatibility wrappers using these functions) are in bn_depr.c. -
 * Geoff
 */

/*
 * The quick sieve algorithm approach to weeding out primes is Philip
 * Zimmermann's, as implemented in PGP.  I have had a read of his comments
 * and implemented my own version.
 */
#include <openssl/sha.h>
#include <openssl/evp.h>
#include "modes_lcl.h"
#include "bn_prime.h"

/* The delta between adjacent primes between 3 and 500 */
static const unsigned char delta_prime[] = {
    2, 2, 4, 2, 4, 2, 4, 6, 2, 6, 4, 2, 4, 6, 6, 2, 6, 4, 2, 6, 4, 6,
    8, 4, 2, 4, 2, 4, 14, 4, 6, 2, 10, 2, 6, 6, 4, 6, 6, 2, 10, 2, 4,
    2, 12, 12, 4, 2, 4, 6, 2, 10, 6, 6, 6, 2, 6, 4, 2, 10, 14, 4, 2,
    4, 14, 6, 10, 2, 4, 6, 8, 6, 6, 4, 6, 8, 4, 8, 10, 2, 10, 2, 6, 4,
    6, 8, 4, 2, 4, 12, 8, 4, 8,
};

static int witness(BIGNUM *w, const BIGNUM *a, const BIGNUM *a1,
                   const BIGNUM *a1_odd, int k, BN_CTX *ctx,
                   BN_MONT_CTX *mont);
static int probable_prime(BIGNUM *rnd, int bits);
static int probable_prime_dh(BIGNUM *rnd, int bits,
                             const BIGNUM *add, const BIGNUM *rem,
                             BN_CTX *ctx);
static int probable_prime_dh_safe(BIGNUM *rnd, int bits, const BIGNUM *add,
                                  const BIGNUM *rem, BN_CTX *ctx);

int BN_GENCB_call(BN_GENCB *cb, int a, int b)
{
    /* No callback means continue */
    if (!cb)
        return 1;
    switch (cb->ver) {
    case 1:
        /* Deprecated-style callbacks */
        if (!cb->cb.cb_1)
            return 1;
        cb->cb.cb_1(a, b, cb->arg);
        return 1;
    case 2:
        /* New-style callbacks */
        return cb->cb.cb_2(a, b, cb);
    default:
        break;
    }
    /* Unrecognised callback type */
    return 0;
}

int BN_generate_prime_ex(BIGNUM *ret, int bits, int safe,
                         const BIGNUM *add, const BIGNUM *rem, BN_GENCB *cb)
{
    BIGNUM *t;
    int found = 0;
    int i, j, c1 = 0;
    BN_CTX *ctx;
    int checks = BN_prime_checks_for_size(bits);

    ctx = BN_CTX_new();
    if (ctx == NULL)
        goto err;
    BN_CTX_start(ctx);
    t = BN_CTX_get(ctx);
    if (!t)
        goto err;
 loop:
    /* make a random number and set the top and bottom bits */
    if (add == NULL) {
        if (!probable_prime(ret, bits))
            goto err;
    } else {
        if (safe) {
            if (!probable_prime_dh_safe(ret, bits, add, rem, ctx))
                goto err;
        } else {
            if (!probable_prime_dh(ret, bits, add, rem, ctx))
                goto err;
        }
    }
    /* if (BN_mod_word(ret,(BN_ULONG)3) == 1) goto loop; */
    if (!BN_GENCB_call(cb, 0, c1++))
        /* aborted */
        goto err;

    if (!safe) {
        i = BN_is_prime_fasttest_ex(ret, checks, ctx, 0, cb);
        if (i == -1)
            goto err;
        if (i == 0)
            goto loop;
    } else {
        /*
         * for "safe prime" generation, check that (p-1)/2 is prime. Since a
         * prime is odd, We just need to divide by 2
         */
        if (!BN_rshift1(t, ret))
            goto err;

        for (i = 0; i < checks; i++) {
            j = BN_is_prime_fasttest_ex(ret, 1, ctx, 0, cb);
            if (j == -1)
                goto err;
            if (j == 0)
                goto loop;

            j = BN_is_prime_fasttest_ex(t, 1, ctx, 0, cb);
            if (j == -1)
                goto err;
            if (j == 0)
                goto loop;

            if (!BN_GENCB_call(cb, 2, c1 - 1))
                goto err;
            /* We have a safe prime test pass */
        }
    }
    /* we have a prime :-) */
    found = 1;
 err:
    if (ctx != NULL) {
        BN_CTX_end(ctx);
        BN_CTX_free(ctx);
    }
    bn_check_top(ret);
    return found;
}

int BN_is_prime_ex(const BIGNUM *a, int checks, BN_CTX *ctx_passed,
                   BN_GENCB *cb)
{
    return BN_is_prime_fasttest_ex(a, checks, ctx_passed, 0, cb);
}

int BN_is_prime_fasttest_ex(const BIGNUM *a, int checks, BN_CTX *ctx_passed,
                            int do_trial_division, BN_GENCB *cb)
{
    int i, j, ret = -1;
    int k;
    BN_CTX *ctx = NULL;
    BIGNUM *A1, *A1_odd, *check; /* taken from ctx */
    BN_MONT_CTX *mont = NULL;
    const BIGNUM *A = NULL;

    if (BN_cmp(a, BN_value_one()) <= 0)
        return 0;

    if (checks == BN_prime_checks)
        checks = BN_prime_checks_for_size(BN_num_bits(a));

    /* first look for small factors */
    if (!BN_is_odd(a))
        /* a is even => a is prime if and only if a == 2 */
        return BN_is_word(a, 2);
    if (do_trial_division) {
        for (i = 1; i < NUMPRIMES; i++)
            if (BN_mod_word(a, primes[i]) == 0)
                return 0;
        if (!BN_GENCB_call(cb, 1, -1))
            goto err;
    }

    if (ctx_passed != NULL)
        ctx = ctx_passed;
    else if ((ctx = BN_CTX_new()) == NULL)
        goto err;
    BN_CTX_start(ctx);

    /* A := abs(a) */
    if (a->neg) {
        BIGNUM *t;
        if ((t = BN_CTX_get(ctx)) == NULL)
            goto err;
        if (BN_copy(t, a) == NULL)
            goto err;
        t->neg = 0;
        A = t;
    } else
        A = a;
    A1 = BN_CTX_get(ctx);
    A1_odd = BN_CTX_get(ctx);
    check = BN_CTX_get(ctx);
    if (check == NULL)
        goto err;

    /* compute A1 := A - 1 */
    if (!BN_copy(A1, A))
        goto err;
    if (!BN_sub_word(A1, 1))
        goto err;
    if (BN_is_zero(A1)) {
        ret = 0;
        goto err;
    }

    /* write  A1  as  A1_odd * 2^k */
    k = 1;
    while (!BN_is_bit_set(A1, k))
        k++;
    if (!BN_rshift(A1_odd, A1, k))
        goto err;

    /* Montgomery setup for computations mod A */
    mont = BN_MONT_CTX_new();
    if (mont == NULL)
        goto err;
    if (!BN_MONT_CTX_set(mont, A, ctx))
        goto err;

    for (i = 0; i < checks; i++) {
        if (!BN_pseudo_rand_range(check, A1))
            goto err;
        if (!BN_add_word(check, 1))
            goto err;
        /* now 1 <= check < A */

        j = witness(check, A, A1, A1_odd, k, ctx, mont);
        if (j == -1)
            goto err;
        if (j) {
            ret = 0;
            goto err;
        }
        if (!BN_GENCB_call(cb, 1, i))
            goto err;
    }
    ret = 1;
 err:
    if (ctx != NULL) {
        BN_CTX_end(ctx);
        if (ctx_passed == NULL)
            BN_CTX_free(ctx);
    }
    if (mont != NULL)
        BN_MONT_CTX_free(mont);

    return (ret);
}

static int witness(BIGNUM *w, const BIGNUM *a, const BIGNUM *a1,
                   const BIGNUM *a1_odd, int k, BN_CTX *ctx,
                   BN_MONT_CTX *mont)
{
    if (!BN_mod_exp_mont(w, w, a1_odd, a, ctx, mont)) /* w := w^a1_odd mod a */
        return -1;
    if (BN_is_one(w))
        return 0;               /* probably prime */
    if (BN_cmp(w, a1) == 0)
        return 0;               /* w == -1 (mod a), 'a' is probably prime */
    while (--k) {
        if (!BN_mod_mul(w, w, w, a, ctx)) /* w := w^2 mod a */
            return -1;
        if (BN_is_one(w))
            return 1;           /* 'a' is composite, otherwise a previous 'w'
                                 * would have been == -1 (mod 'a') */
        if (BN_cmp(w, a1) == 0)
            return 0;           /* w == -1 (mod a), 'a' is probably prime */
    }
    /*
     * If we get here, 'w' is the (a-1)/2-th power of the original 'w', and
     * it is neither -1 nor +1 -- so 'a' cannot be prime
     */
    bn_check_top(w);
    return 1;
}

static int probable_prime(BIGNUM *rnd, int bits)
{
    int i;
    prime_t mods[NUMPRIMES];
    BN_ULONG delta, maxdelta;

 again:
    if (!BN_rand(rnd, bits, 1, 1))
        return (0);
    /* we now have a random number 'rand' to test. */
    for (i = 1; i < NUMPRIMES; i++)
        mods[i] = (prime_t) BN_mod_word(rnd, (BN_ULONG)primes[i]);
    maxdelta = BN_MASK2 - primes[NUMPRIMES - 1];
    delta = 0;
 loop:for (i = 1; i < NUMPRIMES; i++) {
        /*
         * check that rnd is not a prime and also that gcd(rnd-1,primes) == 1
         * (except for 2)
         */
        if (((mods[i] + delta) % primes[i]) <= 1) {
            delta += 2;
            if (delta > maxdelta)
                goto again;
            goto loop;
        }
    }
    if (!BN_add_word(rnd, delta))
        return (0);
    bn_check_top(rnd);
    return (1);
}

static int probable_prime_dh(BIGNUM *rnd, int bits,
                             const BIGNUM *add, const BIGNUM *rem,
                             BN_CTX *ctx)
{
    int i, ret = 0;
    BIGNUM *t1;

    BN_CTX_start(ctx);
    if ((t1 = BN_CTX_get(ctx)) == NULL)
        goto err;

    if (!BN_rand(rnd, bits, 0, 1))
        goto err;

    /* we need ((rnd-rem) % add) == 0 */

    if (!BN_mod(t1, rnd, add, ctx))
        goto err;
    if (!BN_sub(rnd, rnd, t1))
        goto err;
    if (rem == NULL) {
        if (!BN_add_word(rnd, 1))
            goto err;
    } else {
        if (!BN_add(rnd, rnd, rem))
            goto err;
    }

    /* we now have a random number 'rand' to test. */

 loop:for (i = 1; i < NUMPRIMES; i++) {
        /* check that rnd is a prime */
        if (BN_mod_word(rnd, (BN_ULONG)primes[i]) <= 1) {
            if (!BN_add(rnd, rnd, add))
                goto err;
            goto loop;
        }
    }
    ret = 1;
 err:
    BN_CTX_end(ctx);
    bn_check_top(rnd);
    return (ret);
}

static int probable_prime_dh_safe(BIGNUM *p, int bits, const BIGNUM *padd,
                                  const BIGNUM *rem, BN_CTX *ctx)
{
    int i, ret = 0;
    BIGNUM *t1, *qadd, *q;

    bits--;
    BN_CTX_start(ctx);
    t1 = BN_CTX_get(ctx);
    q = BN_CTX_get(ctx);
    qadd = BN_CTX_get(ctx);
    if (qadd == NULL)
        goto err;

    if (!BN_rshift1(qadd, padd))
        goto err;

    if (!BN_rand(q, bits, 0, 1))
        goto err;

    /* we need ((rnd-rem) % add) == 0 */
    if (!BN_mod(t1, q, qadd, ctx))
        goto err;
    if (!BN_sub(q, q, t1))
        goto err;
    if (rem == NULL) {
        if (!BN_add_word(q, 1))
            goto err;
    } else {
        if (!BN_rshift1(t1, rem))
            goto err;
        if (!BN_add(q, q, t1))
            goto err;
    }

    /* we now have a random number 'rand' to test. */
    if (!BN_lshift1(p, q))
        goto err;
    if (!BN_add_word(p, 1))
        goto err;

 loop:for (i = 1; i < NUMPRIMES; i++) {
        /* check that p and q are prime */
        /*
         * check that for p and q gcd(p-1,primes) == 1 (except for 2)
         */
        if ((BN_mod_word(p, (BN_ULONG)primes[i]) == 0) ||
            (BN_mod_word(q, (BN_ULONG)primes[i]) == 0)) {
            if (!BN_add(p, p, padd))
                goto err;
            if (!BN_add(q, q, qadd))
                goto err;
            goto loop;
        }
    }
    ret = 1;
 err:
    BN_CTX_end(ctx);
    bn_check_top(p);
    return (ret);
}

/* Start of 186-3 prime generation code */

/*
 * This computes a value c such that a*c = 1 (mod b)
 * This uses Knuth's algorithm 4.5.2X
 */
static int modinv(int a, int b)
{
    int u1 = 1;
    int u3 = a;
    int v1 = 0;
    int v3 = b;

    for (;;) {
        int q;
        int t1, t3;

        if (v3 == 0) {
            break;
        }
        q = u3 / v3;
        t1 = u1 - q * v1;
        t3 = u3 - q * v3;
        u1 = v1;
        u3 = v3;
        v1 = t1;
        v3 = t3;
    }

    if (u3 != 1) {
        return -1;              /* a, b not relatively prime */
    }
    if (u1 < 0) {
        u1 += b;
    }
    return u1;
}

/*
 * This marks all multiples of p within the sieve, assuming the bit 0 of the
 * sieve corresponds to the value x (with x % p == s), and bit 1 of the sieve
 * corresponds to the value y (with (y-x) % p == t)
 */
static void mark_sieve(unsigned char *sieve, unsigned int maximum,
                       unsigned int p, unsigned int s, unsigned int t)
{
    unsigned int r;

    if (!t) {
        return;
    }
    if (!s) {
        r = 0;
    } else {
        int tinv = modinv(t, p);
        if (tinv < 0) {
            return;             /* This can happen if p is not prime */
        }
        r = p - ((unsigned long long)s * tinv) % p;
    }

    /* Mark all the multiples of p within the sieve */
    for (; r < maximum; r += p) {
        sieve[r / 8] |= 1 << (r % 8);
    }
}

/*
 * This initializes the sieve, so that the first index corresponds to the
 * number start, with a constant increment after that
 *
 * b is the number of bits of start -- we could compute it, but the caller
 * knows it already
 */
static void setup_sieve(prime_sieve * s, const BIGNUM *start,
                        const BIGNUM *increment, unsigned int b)
{
    unsigned char *sieve = s->sieve;
    unsigned char *subsieve = s->subsieve;
    unsigned int p, j;
    unsigned int max_sieve_this_iter;

    memset(sieve, 0, SEGMENT_SIZE / 8); /* Everything is a potential */
    /* prime until proved otherwise */
    /* Mark the primes between 3 and 499 */
    for (p = 3, j = 0; j < sizeof delta_prime / sizeof *delta_prime; j++) {

        BN_ULONG s1 = BN_mod_word(start, p);
        BN_ULONG t = BN_mod_word(increment, p);

        mark_sieve(sieve, SEGMENT_SIZE, p, s1, t);

        p += delta_prime[j];
    }

    /*
     * Note: based the number of primes we sieve out based on the size of
     * the prime we're looking for.  For smaller primes, it isn't cost
     * effective to sieve out as many primes (as the time taken to sieve
     * them won't eliminate enough composites to justify the expense)
     */
    max_sieve_this_iter = 200 * b;
    if (max_sieve_this_iter > s->max_sieve) {
        max_sieve_this_iter = s->max_sieve; /* Shouldn't happen */
    }

    /* Now, eliminate the primes from 500 to the maximum we selected */
    for (j = (501 - 1) / 2; 2 * j < max_sieve_this_iter; j++) {
        unsigned int s1, t, p;

        if (subsieve[j / 8] & (1 << (j % 8))) {
            continue;
        }
        p = 2 * j + 1;

        s1 = BN_mod_word(start, p);
        t = BN_mod_word(increment, p);

        mark_sieve(sieve, SEGMENT_SIZE, p, s1, t);
    }
}

void BN_free_prime_sieve(prime_sieve * sieve)
{
    if (sieve) {
        OPENSSL_free(sieve->subsieve);
        OPENSSL_free(sieve);
    }
}

/*
 * From FIPS PUBS 186-3 C.8
 * This will allocate a sieve.  The max_prime parameter is the size of the
 * largest prime that we'll use this sieve to find
 *
 * This is a bitmask listing all the small primes
 * (where small is up to at least 65535; possibly a bit larger)
 *
 * First, estimate the size of the small primes we want to eliminate.
 * Letting another composite through is approximately 10000 times as
 * costly as sieving out another prime, so we attempt to set the maximum to
 * the point where sieving out another prime reduces the expected number of
 * composites visited by 0.0001 .  This formula should approximate that
 * point
 */
prime_sieve *BN_allocate_prime_sieve(unsigned int max_prime)
{
    unsigned int max_sieve;
    unsigned int p, j;
    unsigned char *subsieve;

    prime_sieve *s = OPENSSL_malloc(sizeof *s);
    if (!s) {
        return NULL;
    }
    max_sieve = 200 * max_prime;

    subsieve = OPENSSL_malloc(max_sieve / 16 + 1);
    if (!subsieve) {
        OPENSSL_free(s);
        return NULL;
    }

    for (p = 3, j = 0;; j++) {

        mark_sieve(subsieve, max_sieve / 2, p, 1, 2);

        if (j == (sizeof delta_prime / sizeof *delta_prime)) {
            break;
        }
        p += delta_prime[j];
    }
    /* Now, if bit subsieve[ j/8 ] & (1 << (j%8)) is set, then 2*j + 1 */
    /* is known to be composite (for j > 250) */

    s->max_sieve = max_sieve;
    s->subsieve = subsieve;

    return s;
}

/*
 * This increments the given seed by the passed increment.  We use this to
 * 'step' our random number generator
 */
static void increment_seed(unsigned char *seed, unsigned int seed_len,
                           unsigned int increment)
{
    unsigned int t;

    while ((seed_len-- > 0) && (increment > 0)) {
        t = seed[seed_len] + increment;
        seed[seed_len] = t & 0xff;
        increment = (t >> 8);
    }
}

BIGNUM *BN_2PowerN(unsigned int bits)
{
    unsigned int digits = bits / (sizeof(BN_ULONG) * 8) + 1;
    unsigned int i;
    BIGNUM *t = BN_new();

    if (!t) {
        return t;
    }
    bn_wexpand(t, (int)digits);
    for (i = 0; i < digits - 1; i++) {
        t->d[i] = 0;
    }
    t->d[i] = ((BN_ULONG)1 << (bits % (sizeof(BN_ULONG) * 8)));
    t->top = digits;
    return t;
}

/*
 * This computes Hash(input_seed) and the increments input_seed
 * This is what this file uses in lieu of a random number generator
 *
 * FIPS 186-3 asks us to use an approved hash -- we pick SHA256
 */
#define OUTLEN SHA256_DIGEST_LENGTH /* The size of a hash output */
static char hash_seed(void *hash_output, void *input_seed,
                      unsigned int seed_len)
{
    const EVP_MD *evp_md;
    char ret = 0;

    evp_md = EVP_sha256();
    if (evp_md == NULL) {
        return ret;
    }
    ret = EVP_Digest(input_seed, seed_len, hash_output, NULL, evp_md, NULL);

    increment_seed(input_seed, seed_len, 1);
    return ret;
}

/*
 * This generates a random bignum out of iterations+1 calls to the underlying
 * 'random number generator'
 */
static BIGNUM *hash_seed_bignum(void *input_seed, unsigned int seed_len,
                                unsigned int iterations, BN_CTX *ctx)
{
    BIGNUM *x;
    unsigned int i;
    unsigned int buffer_len = (iterations + 1) * OUTLEN;
    unsigned char *x_buffer = OPENSSL_malloc(buffer_len);
    char ret;

    if (!x_buffer) {
        return NULL;
    }
    x = BN_CTX_get(ctx);
    if (!x) {
        OPENSSL_free(x_buffer);
        return x;
    }

    for (i = 0; i <= iterations; i++) {
        ret =
            hash_seed(&x_buffer[(iterations - i) * OUTLEN], input_seed,
                      seed_len);
        if (!ret) {
            return x;
        }
    }

    BN_bin2bn(x_buffer, buffer_len, x);
    OPENSSL_free(x_buffer);

    return x;
}

static unsigned int st_reduce(unsigned int target_size)
{
    return ((target_size + 3) / 2);
}

static unsigned int do_sqrt(unsigned int n)
{
    unsigned int m = n / 4 + 1;
    unsigned int first_iter = 1;

    for (;;) {
        unsigned int next_m = (n / m + m) / 2;
        if (!first_iter && next_m >= m) {
            return next_m;
        }
        m = next_m;
        first_iter = 0;
    }
}

/*
 * This returns TRUE if a and b are relatively prime
 */
static int BN_IsRelativelyPrime(const BIGNUM *a, const BIGNUM *b, BN_CTX *ctx)
{
    BIGNUM *gcd;
    int result = 0;

    gcd = BN_CTX_get(ctx);
    if (!gcd) {
        return 0;
    }

    BN_gcd(gcd, a, b, ctx);
    result = gcd && BN_is_one(gcd);

    return result;
}

/*
 * This returns the log (base 2) of the given BigNum
 */
static unsigned int BN_Log2(const BIGNUM *a)
{
    int n;
    int result = 0;
    BN_ULONG msdigit;

    if (!a)
        return -1;
    n = a->top;
    while (n) {
        msdigit = a->d[n - 1];
        if (!msdigit) {
            n--;                /* The number is supposed to be normalized */
            continue;
        }
        result = BN_BITS2 * n - 1;
        while ((msdigit & BN_TBIT) == 0) {
            result--;
            msdigit <<= 1;
        }
        return result;
    }
    return -1;
}

/*
 * Compute the square root of n (rounded down, naturally)
 * This uses Newton-Raphson, modified to work with integers
 */
static BIGNUM *BN_SquareRoot(BIGNUM *n)
{
    BIGNUM *m = NULL;
    BIGNUM *div;
    int first_iter;
    BN_CTX *ctx;
    int b;

    if (!n) {
        return NULL;
    }
    ctx = BN_CTX_new();
    if (!ctx) {
        return NULL;
    }
    BN_CTX_start(ctx);
    div = BN_CTX_get(ctx);

    if (!div) {
        goto error;
    }

    /*
     * Potential optimization: if m was initialized to a closer approximation of
     * sqrt(n), we'd use far fewer iterations
     */
    b = BN_Log2(n) / 2;
    m = BN_2PowerN(b);
    if (!m) {
        goto error;
    }
    for (first_iter = 1;; first_iter = 0) {
        BIGNUM *next_m;

        BN_div(div, NULL, n, m, ctx);
        next_m = BN_dup(div);
        if (!next_m) {
            goto error;
        }
        BN_add(next_m, next_m, m);
        BN_div_word(next_m, 2);

        if (!next_m) {
            goto error;
        }
        /*
         * if next_m >= m, them m is the square root (unless this is the first
         * iteration)
         */
        if (!first_iter && (BN_cmp(next_m, m) >= 0)) {
            BN_free(next_m);
            BN_free(n);
            BN_CTX_end(ctx);
            BN_CTX_free(ctx);
            return m;
        }

        /* It's not -- set m = next_m, and repeat the iteration */
        BN_free(m);
        m = next_m;
    }

 error:
    BN_CTX_end(ctx);
    BN_CTX_free(ctx);
    BN_free(n);
    BN_free(m);
    return NULL;
}

/*
 * This will return a prime using algorithm C.6 from FIPS 186-3.  This is
 * the Shawe-Taylor algorithm, and returns a deterministic prime
 *
 * Parameters:
 * length - the length of the target prime, in bits
 * input_seed, seed_len - the seed that is used to generate the prime.
 *     It's effectively the 'random number generator state'
 * s = the precomputed prime sieve to use
 */
static
BIGNUM *generate_prime_sane(unsigned int length, void *input_seed,
                            unsigned int seed_len, prime_sieve * s,
                            BN_GENCB *cb)
{
    unsigned int target_size;
    struct prime_stack {
        struct prime_stack *next;
        unsigned int target_size; /* Size of the prime we want, in bits */
    } *stack = NULL;
#define MANUAL_SIZE 33          /* If we get to smaller than 33 bits, switch
                                 * to a */
    /* brute force method */
    BIGNUM *c = NULL;
    BIGNUM *c0 = NULL;
    BIGNUM *top = NULL, *bottom = NULL;
    BIGNUM *x = NULL;
    BIGNUM *t = NULL;
    BIGNUM *two_t = NULL;
    BIGNUM *two_c0 = NULL;
    BIGNUM *a = NULL;
    BIGNUM *z = NULL;
    BIGNUM *zr = NULL;
    BIGNUM *c3, *z_minus_1;
    unsigned char *sieve = s->sieve;
    BN_CTX *ctx = NULL;
    unsigned int prime_gen_counter, relatively_prime;
    char ret = 0;

    /* Step 1: if we're asked to look for a prime < 2, error */
    if (length < 2) {
        goto error;
    }
    /* Push the first item onto the stack */
    stack = OPENSSL_malloc(sizeof *stack);
    if (!stack) {
        return NULL;
    }
    stack->next = NULL;         /* Bottom element */
    stack->target_size = length;
    target_size = st_reduce(length);

    ctx = BN_CTX_new();
    if (!ctx) {
        OPENSSL_free(stack);
        return NULL;
    }
    BN_CTX_start(ctx);
    /* Recurse down until we get small enough */
    while (target_size >= MANUAL_SIZE) {
        struct prime_stack *elem = OPENSSL_malloc(sizeof *stack);
        if (!elem)
            goto error;
        elem->next = stack;
        elem->target_size = target_size;
        stack = elem;
        target_size = st_reduce(target_size);
    }

    /*
     * Ok, step 2; target_size < 33, and so we can use a brute force algorithm
     * to find a prime of the target size; brute force == try random numbers of
     * the right size, and check for primality by trial division
     */
    {
        unsigned int target_mask = (1 << (target_size - 1)) - 1; /* We want
                                                                  * to */
        /*
         * mask off the bits that would give us a larger prime than
         */
        /* we're looking for */
        unsigned int bit_set = (target_mask + 1) | 1; /* We know we need to
                                                       * set */
        /* these bits; bit 1<<(target_size-1) because that's the */
        /* MSBit of numbers of target_size bits in length, and the */
        /* LSBit, because the prime we're looking for is odd */

        /*
         * Steps 3, 4: adjust counters
         */
        prime_gen_counter = 0;

        for (;;) {
            unsigned int c1, s1, p;

            /* Step 5, 9: generate the initial c value */
            {
                unsigned char *h1 = OPENSSL_malloc(OUTLEN);
                unsigned char *h2 = OPENSSL_malloc(OUTLEN);

                ret = hash_seed(h1, input_seed, seed_len);
                if (!ret) {
                    goto error;
                }
                ret = hash_seed(h2, input_seed, seed_len);
                if (!ret) {
                    goto error;
                }
                c1 = (h1[OUTLEN - 1] ^ h2[OUTLEN - 1]);
                c1 |= (h1[OUTLEN - 2] ^ h2[OUTLEN - 2]) << 8;
                c1 |= (h1[OUTLEN - 3] ^ h2[OUTLEN - 3]) << 16;
                c1 |= (h1[OUTLEN - 4] ^ h2[OUTLEN - 4]) << 24;
                OPENSSL_cleanse(h1, OUTLEN);
                OPENSSL_cleanse(h2, OUTLEN);
                OPENSSL_free(h1);
                OPENSSL_free(h2);
            }

            /*
             * Step 6, 7: set bit length-1, bit 0; clear bits higher than
             * length-1
             */
            c1 &= target_mask;  /* Clear the bits that put this random number
                                 */
            /* above the requested range */
            c1 |= bit_set;      /* Set the bits we know need to be set */

            /*
             * Step 8: adjust counters
             */
            prime_gen_counter += 1;

            /*
             * Step 10: perform a determanistic primality test of c
             * Check to see if z has any prime factors <= sqrt(z).
             */
            s1 = do_sqrt(c1);
            for (p = 3; p <= s1; p += 2) {

                if ((c1 % p) == 0)
                    break;      /* Oops, not prime */
            }

            /*
             * Step 11: if c is a prime number, then that's the prime we're
             * looking for (at least, at this recursive call)
             */
            if (p > s1) {
                c0 = BN_new();
                if (!c0) {
                    goto error;
                }
                BN_set_word(c0, c1);
                break;
            }
            if (!BN_GENCB_call(cb, 0, 0)) {
                /* aborted */
                goto error;
            }

            /*
             * Step 12: check if we've iterated too many times, and fail if we
             * have.  I really don't see the point of that
             */
            if (prime_gen_counter > (PRIME_SANITY_LIMIT * target_size)) {
                goto error;
            }
            /* Step 13: go back to step 5 */
        }
    }

    while (stack) {
        unsigned int tar_len = stack->target_size; /* Size of the prime we're
                                                    */
        /* looking for, in bits */
        int i;
        unsigned int iterations;

        /*
         * Step 14: call the ST algorithm recurively.  We've effectively done
         * this already by creating the stack.
         * Here, c0 = ST_Random_Prime((b+3)/2)
         */

        /* Step 15: on failure, error out */
        if (!c0) {
            goto error;
        }
        BN_free(two_c0);
        two_c0 = BN_dup(c0);
        if (!two_c0) {
            goto error;
        }
        BN_mul_word(two_c0, 2);

        /* Step 16: compute iterations */
        iterations = (tar_len + 8 * OUTLEN - 1) / (8 * OUTLEN) - 1;
        /*
         * Step 17: we're supposed to track a counter that's used to fail if
         * we go longer than expected.
         */
        prime_gen_counter = 0;

        /*
         * Step 18, 19, 20: generate a random value using iterations+1 calls
         * to select a random starting point
         */
        x = hash_seed_bignum(input_seed, seed_len, iterations, ctx);

        /* Step 21: normalize that number into the range [2^(b-1), 2^b-1] */
        BN_free(bottom);
        bottom = BN_2PowerN(tar_len - 1);
        BN_mod(x, x, bottom, ctx);
        BN_add(x, x, bottom);
        BN_free(top);
        top = BN_2PowerN(tar_len);
        if (!top) {
            goto error;
        }
        /*
         * Step 22: let t be the least integer greater than x/(2 * c0)
         */

 step_22:
        BN_add(x, x, two_c0);
        BN_sub_word(x, 1);

        BN_free(t);
        t = BN_new();
        if (!t) {
            goto error;
        }

        BN_div(t, NULL, x, two_c0, ctx);

        /*
         * Note: the below algorithm always references 2*t.  Hence, we compute
         * 2*t up front, and discard the value 't' as unneeded
         */
        BN_free(two_t);
        two_t = BN_dup(t);
        if (!two_t) {
            goto error;
        }
        BN_mul_word(two_t, 2);
        BN_free(t);             /* Note: the below algorithm modifies the
                                 * variable */
        t = NULL;               /* it calls 't'.  As we won't update our copy
                                 * of t */
        /* (we'll update two_t instead), discard the */
        /* variable, lest some later maintainer attempt to */
        /* use it */
 step_23:
        /*
         * Step 24: let c = 2*t*c0 + 1
         * Note: we take steps 23 and 24 out of order, so that we can reuse the
         * computation of c in step 5
         */
        BN_free(c);
        c = BN_dup(two_t);
        if (!c) {
            goto error;
        }
        BN_mul(c, c, c0, ctx);
        BN_add_word(c, 1);
        if (!c) {
            goto error;
        }
        /*
         * Ok, here the S.T. algorithm expects us to step through each value
         * 2*t*c0+i*2*c0+1 (for i=0,1,2,3,...), testing each one until we find
         * a prime. However, a straight-forward implementation is going to
         * waste a lot of time testing numbers that are a multiple of 3, 5 or
         * other small primes.  To avoid that, we set up a sieve that tells us
         * which numbers have small factors (and so we needn't waste the time
         * doing tests involving modexp's on them)
         */
        setup_sieve(s, c, two_c0, tar_len);

        /*
         * Now, scan for the next integer not marked as composite
         */
        for (i = 0; i < SEGMENT_SIZE; i++,
             BN_add(c, c, two_c0), BN_add_word(two_t, 2)) {

            /* Step 23: if c is too big, wrap around to the bottom */
            if (BN_cmp(c, top) > 0) { /* top == 2**b */
                x = BN_CTX_get(ctx); /* Freed in BN_CTX_end(ctx) */
                if (!x) {
                    goto error;
                }
                BN_copy(x, bottom); /* This will start back at the */
                /*
                 * begining of the interval
                 */
                goto step_22;
            }

            /* Step 25: increment test counter */
            prime_gen_counter += 1;

            /* Check if c has any small factors */
            if (sieve[i / 8] & (1 << (i % 8))) {
                /* Composite!  Skip this one */

                /* bump the seed */
                increment_seed(input_seed, seed_len, iterations + 1);
                continue;
            }

            /*
             * Now we have a 'c' that might be prime, back to the standard
             * algorithm
             */

            /*
             * Step 26, 27, 28: Select a random integer a in [0, 2**n)
             */
            a = hash_seed_bignum(input_seed, seed_len, iterations, ctx);

            /* Step 29: normalize it into the range [2, c-2] */
            {
                c3 = BN_dup(c);
                if (!c3) {
                    goto error;
                }
                BN_sub_word(c3, 3);

                BN_mod(a, a, c3, ctx);
                BN_add_word(a, 2);
                BN_free(c3);
            }

            /*
             * Step 30: let x = a^(2t) mod c
             */
            BN_free(z);
            z = BN_new();
            if (!z) {
                goto error;
            }
            BN_mod_exp(z, a, two_t, c, ctx);

            if (!z) {
                goto error;
            }
            /*
             * Step 31: if z!=1 && gcd(z-1, c)==1, and z**c0==1 mod c, then c is prime
             */
            if (!BN_is_one(z)) {
                z_minus_1 = BN_dup(z);
                if (!z_minus_1) {
                    goto error;
                }
                BN_sub_word(z_minus_1, 1);

                relatively_prime = BN_IsRelativelyPrime(z_minus_1, c, ctx);
                BN_free(z_minus_1);
                if (relatively_prime) {
                    zr = BN_new();
                    if (!zr) {
                        goto error;
                    }
                    BN_mod_exp(zr, z, c0, c, ctx);

                    if (!zr) {
                        goto error;
                    }
                    if (BN_is_one(zr)) {
                        BN_free(zr);
                        zr = NULL;
                        goto found_prime;
                    }
                    BN_free(zr);
                    zr = NULL;
                }
            }

            /*
             * Step 32: adjust counters, and trigger possible failure if we
             * haven't seen a prime as fast as we expect.
             */
            if (prime_gen_counter >= PRIME_SANITY_LIMIT * tar_len) {
                goto error;
            }

            /* Step 33: if not prime, then t = t+1, goto step 5 */
            continue;           /* That's handled in the loop iteration */
            /* And, because we actually track 2*t, we actually */
            /* do: two_t = two_t + 2 */
        }

        /*
         * Went through the entire sieve segment, and didn't find a
         * single prime.  Start again on the next segment
         */
        if (!BN_GENCB_call(cb, 1, 0)) {
            /* aborted */
            goto error;
        }
        goto step_23;

 found_prime:
        /* found the prime */
        BN_free(c0);
        c0 = c;
        c = NULL;

        /* Pop the item we just processed off the stack */
        {
            struct prime_stack *old_stack = stack;
            stack = old_stack->next;
            OPENSSL_free(old_stack);
        }
        /*
         * Start on the next item on the stack (or just return if we're all
         * done)
         */
    }

    /*
     * Sweet success! Clean up everything we've allocated and return the
     * prime
     */
    BN_free(top);
    BN_free(bottom);
    BN_free(t);
    BN_free(two_c0);
    BN_free(two_t);
    BN_free(z);
    BN_free(zr);
    BN_free(c);
    BN_CTX_end(ctx);
    BN_CTX_free(ctx);
    return c0;

 error:
    /* Come here on any error */
    while (stack) {
        struct prime_stack *old_stack = stack;
        stack = old_stack->next;
        OPENSSL_free(old_stack);
    }

    BN_free(c0);
    BN_free(top);
    BN_free(bottom);
    BN_free(t);
    BN_free(two_c0);
    BN_free(two_t);
    BN_free(z);
    BN_free(zr);
    BN_free(c);
    BN_CTX_end(ctx);
    BN_CTX_free(ctx);

    return 0;
}

/*
 * This will return a prime using algorithm C.10 from FIPS 186-3.  This is
 * rather more complex than necessary; apparently, NIST believes that
 * unnecessary complexity adds to security
 *
 * Parameters (taken from the algorithm description)
 * L - the length of the target prime, in bits
 * N1 - the size, in bits, of the requested auxiliary prime p1; the resulting
 *       prime p = 1 (mod p1)
 * N2 - the size, in bits, of the requested auxiliary prime p2; the resulting
 *       prime p = -1 (mod p2)
 *      Note that neither p1 nor p2 will be returned; it's not important that
 *      we know those values; we just care whether those values exist
 * firstseed, seed_len - the seed that is used to generate the primes.  Think
 *       of it as a rando number generator stand-in
 * e - The public exponent this prime will be used with
 * s - the precomputed prime sieve to use
 */
BIGNUM *BN_generate_1863_prime(unsigned int L, unsigned int N1,
                               unsigned int N2, void *firstseed,
                               unsigned int seed_len, BIGNUM *e_value,
                               prime_sieve * s, BN_GENCB *cb)
{
    BN_CTX *ctx;
    BIGNUM *p = NULL, *diff = NULL;
    BIGNUM *p0 = NULL, *p1 = NULL, *p2 = NULL, *p0p1 = NULL;
    BIGNUM *x = NULL, *y = NULL, *t = NULL, *z = NULL, *zr = NULL;
    BIGNUM *increment = NULL;
    BIGNUM *low = NULL, *high = NULL;
    BIGNUM *a = NULL, *test = NULL;
    BIGNUM *p_minus_one = NULL;
    BIGNUM *gcd, *div, *z_minus_1, *p3, *divisor;
    BIGNUM *initial_exponent = NULL, *delta_exponent = NULL;
    unsigned int iterations;
    unsigned int pgen_counter;
    unsigned int i, relatively_prime;
    unsigned char *sieve = s->sieve;

    ctx = BN_CTX_new();
    if (!ctx) {
        return NULL;
    }
    BN_CTX_start(ctx);

    /* Step 1: sanity check the parameters.  For now, we don't bother */

    /* Step 2, 3: generate p1 */
    if (N1 == 1) {
        p1 = BN_new();
        BN_set_word(p1, 1);
    } else {
        p1 = generate_prime_sane(N1, firstseed, seed_len, s, cb);
    }
    if (!p1) {
        goto error;
    }
    /* Step 4, 5: generate p2 */
    if (N2 == 1) {
        p2 = BN_new();
        BN_set_word(p2, 1);
    } else {
        p2 = generate_prime_sane(N2, firstseed, seed_len, s, cb);
    }
    if (!p2) {
        goto error;
    }
    /* Step 6: generate the seed prime */
    p0 = generate_prime_sane((L + 1) / 2 + 1, firstseed, seed_len, s, cb);
    if (!p0) {
        goto error;
    }
    /* Step 7: compute iterations */
    iterations = (L + 8 * OUTLEN - 1) / (8 * OUTLEN) - 1;

    /* Step 8: initialize pgen_counter */
    pgen_counter = 0;

    /* Step 9, 10, 11: initialize x randomly based on the seed */
    BN_free(x);
    x = hash_seed_bignum(firstseed, seed_len, iterations, ctx);

    /* Step 12: adjust x into the range sqrt(2)2**(L-1) and 2**L */
    low = BN_SquareRoot(BN_2PowerN(2 * L - 1));
    high = BN_2PowerN(L);
    diff = BN_CTX_get(ctx);
    if (!diff) {
        goto error;
    }
    BN_sub(diff, high, low);
    BN_mod(x, x, diff, ctx);
    BN_add(x, x, low);

    /* Step 13: check that p0p1 and p2 are relatively prime */
    /*
     * Given that p0, p1, p2 are primes of different sizes, this should never
     * fail
     */
    test = BN_CTX_get(ctx);

    p0p1 = BN_CTX_get(ctx);
    if (!p0p1) {
        goto error;
    }
    BN_copy(p0p1, p0);
    BN_mul(p0p1, p0p1, p1, ctx);
    gcd = BN_CTX_get(ctx);

    if (!gcd) {
        goto error;
    }
    BN_gcd(gcd, p0p1, p2, ctx);
    if (!gcd || !BN_is_one(gcd)) {
        goto error;
    }
    /* Step 14: compute (p0p1)**-1 mod p2 */
    y = BN_CTX_get(ctx);
    if (!y) {
        goto error;
    }
    BN_mod_inverse(y, p0p1, p2, ctx);

    BN_copy(test, p0p1);
    BN_mul(test, test, y, ctx);
    BN_sub_word(test, 1);
    BN_mod(test, test, p2, ctx);

    /* Step 15: compute t */
    div = BN_CTX_get(ctx);
    if (!div) {
        goto error;
    }
    BN_copy(div, p0p1);
    BN_mul(div, div, p2, ctx);
    BN_mul_word(div, 2);

    t = BN_CTX_get(ctx);
    if (!t) {
        goto error;
    }
    BN_copy(t, y);
    BN_mul(t, t, p0p1, ctx);
    BN_mul_word(t, 2);

    BN_add(t, t, x);
    BN_add(t, t, div);
    BN_sub_word(t, 1);
    BN_div(t, NULL, t, div, ctx);

    /* increment is the delta between successive prime candidates */
    increment = BN_CTX_get(ctx);
    if (!increment) {
        goto error;
    }
    BN_copy(increment, p0p1);
    BN_mul(increment, increment, p2, ctx);
    BN_mul_word(increment, 2);

 step_17:
    /*
     * initial_exponent is power we raise the witness to when testing
     * primality
     */
    initial_exponent = BN_CTX_get(ctx);
    if (!initial_exponent) {
        goto error;
    }
    BN_copy(initial_exponent, t);
    BN_mul(initial_exponent, initial_exponent, p2, ctx);
    BN_sub(initial_exponent, initial_exponent, y);
    BN_mul(initial_exponent, initial_exponent, p1, ctx);
    BN_mul_word(initial_exponent, 2);

    /*
     * delta_exponent is the amount initial_exponent changes when we move to
     * successive prime candiates
     */
    delta_exponent = BN_CTX_get(ctx);
    if (!delta_exponent) {
        goto error;
    }
    BN_copy(delta_exponent, p2);
    BN_mul(delta_exponent, delta_exponent, p1, ctx);
    BN_mul_word(delta_exponent, 2);

    /*
     * Step 17: let p = 2(t p2 - y) p0p1 + 1
     * Note: we take steps 16 and 17 out of order, so that we can reuse the
     * computation of y
     */
    BN_free(p);

    p = BN_dup(t);
    if (!p) {
        goto error;
    }
    BN_mul(p, p, p2, ctx);
    if (!BN_sub(p, p, y)) {
        goto error;
    }

    BN_mul(p, p, p0p1, ctx);
    BN_mul_word(p, 2);
    BN_add_word(p, 1);

    if (!p) {
        goto error;
    }
    /*
     * Ok, here the algorithm expects us to step through each value
     * p+i*2*p0p1p2+1 (for i=0,1,2,3,...), testing each one until we find
     * a prime. However, a straight-forward implementation is going to
     * waste a lot of time testing numbers that are a multiple of 3, 5 or
     * other small primes.  To avoid that, we set up a sieve that tells us
     * which numbers have small factors (and so we needn't waste the time
     * doing tests involving modexp's on them)
     */
    setup_sieve(s, p, increment, L);

    /*
     * Now, scan for the next integer not marked as composite
     */
    for (i = 0; i < SEGMENT_SIZE; i++,
         BN_add(p, p, increment),
         BN_add(initial_exponent, initial_exponent, delta_exponent),
         BN_add_word(t, 1)) {

        /* Step 16: if p is too big, wrap around to the bottom */
        if (BN_cmp(p, high) > 0) { /* high == 2**b */

            divisor = BN_dup(p0p1);
            if (!divisor) {
                goto error;
            }
            BN_mul(divisor, divisor, p2, ctx);
            BN_mul_word(divisor, 2);

            /* This will start back at the begining of the interval */
            t = BN_CTX_get(ctx);
            if (!t) {
                goto error;
            }
            BN_copy(t, y);
            BN_mul(t, t, p0p1, ctx);
            BN_mul_word(t, 2);
            BN_add(t, t, low);
            BN_add(t, t, divisor);
            BN_sub_word(t, 1);
            BN_div(t, NULL, t, divisor, ctx);
            BN_free(divisor);
            goto step_17;
        }

        /* Step 18: adjust the failure counter */
        pgen_counter += 1;

        /* Step 19: check if p != 1 mod e */
        if (!p) {
            goto error;
        }

        /* e might be composite; we need to full relatively prime check */
        p_minus_one = BN_CTX_get(ctx);
        if (!p_minus_one) {
            goto error;
        }
        BN_copy(p_minus_one, p);
        BN_sub_word(p_minus_one, 1);
        BN_gcd(p_minus_one, p_minus_one, e_value, ctx);
        if (!BN_is_one(p_minus_one)) {
            continue;
        }

        /*
         * Check if p has any small factors.  We do this after we check whether
         * p != 1 mod e because the published algorithm doesn't expect us to
         * use any rng's if step 19 fails
         */
        if (sieve[i / 8] & (1 << (i % 8))) {
            /* Composite!  Skip this one */

            /*
             * The algorithm expects us to have used a random 'a' value;
             * adjust the seed just as if we did
             */
            increment_seed(firstseed, seed_len, iterations + 1);

            /* Wasn't that a lot easier than doing a modexp? */
            continue;
        }

        /* Now we have a 'p' that might be prime, back to the ST algorithm */
        /* Step 19.1 - 19.3: generate a random a */
        a = hash_seed_bignum(firstseed, seed_len, iterations, ctx);

        /* Step 19.4: normalize it into the range [2, p-2] */
        {
            p3 = BN_dup(p);
            if (!p3) {
                goto error;
            }
            BN_sub_word(p3, 3);
            BN_mod(a, a, p3, ctx);
            BN_add_word(a, 2);
            BN_free(p3);
            p3 = NULL;
        }

        /*
         * Step 19.5: let z = a^(2(tp2 - y)p1) mod p
         */
        z = BN_CTX_get(ctx);
        if (!z) {
            goto error;
        }
        BN_mod_exp(z, a, initial_exponent, p, ctx);
        if (!z) {
            goto error;
        }
        /*
         * Step 19.6: if z!=1 && gcd(z-1, p)==1, and z**p0==1 mod p, then p is prime
         */
        if (!BN_is_one(z)) {
            z_minus_1 = BN_dup(z);
            if (!z_minus_1) {
                goto error;
            }
            BN_sub_word(z_minus_1, 1);

            relatively_prime = BN_IsRelativelyPrime(z_minus_1, p, ctx);
            BN_free(z_minus_1);
            if (relatively_prime) {
                zr = BN_new();
                BN_mod_exp(zr, z, p0, p, ctx);

                if (!zr) {
                    goto error;
                }
                if (BN_is_one(zr)) {
                    if (!BN_GENCB_call(cb, 3, 0)) {
                        goto error;
                    }
                    goto found_prime;
                }
                BN_free(zr);
                zr = NULL;
            }
        }

        /* Step 20, 21: increment counters, try next candidate */
        if (pgen_counter >= (PRIME_SANITY_LIMIT * L)) {
            goto error;
        }
    }
    /*
     * Oops, we hit the end of the sieve without hitting any primes
     * Generate a fresh sieve starting at where we are now
     */
    goto step_17;

 error:
    BN_free(p);
    p = NULL;
 found_prime:
    BN_CTX_end(ctx);
    BN_CTX_free(ctx);
    BN_free(p0);
    BN_free(p1);
    BN_free(p2);
    BN_free(zr);
    BN_free(low);
    BN_free(high);
    return p;
}
