/* -----------------------------------------------------------------
 * crypto/des/asm/ecb_enc-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include "des_locl.h"
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif

/*
 * DES_ecb_encrypt - the basic DES encryption routine that encrypts or decrypts
 * a single 8-byte DES_cblock in electronic code book ( ECB ) mode.
 */
void DESOCT_ecb_encrypt(const_DES_cblock *input, DES_cblock *output,
                        DES_key_schedule *ks, int enc)
{
    register uint64_t inp;

    inp = *(uint64_t *) input;

    CVMX_MT_3DES_KEY (cvmx_htobe64(ks->cvmkey), 0);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks->cvmkey), 1);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks->cvmkey), 2);

    if (DES_ENCRYPT == enc) {
	CVMX_MT_3DES_ENC (cvmx_htobe64(inp));
        CVMX_MF_3DES_RESULT(inp);
	*(uint64_t *)output = cvmx_betoh64(inp);
    } else {
       CVMX_MT_3DES_DEC (cvmx_htobe64(inp));
       CVMX_MF_3DES_RESULT (inp);
       *(uint64_t *)output = cvmx_betoh64(inp);
    }
}
