/* -----------------------------------------------------------------
 * crypto/des/asm/ncbc_enc-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <openssl/crypto.h>
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif

/*
 * DES_ncbc_encrypt()/DES_cbc_encrypt - encrypts/decrypts using the
 * cipher-block-chaining (CBC) mode of DES
 */
#ifdef CBC_ENC_C__DONT_UPDATE_IV
void DESOCT_cbc_encrypt(const unsigned char *in, unsigned char *out,
                        long length, DES_key_schedule *key, DES_cblock *ivec,
                        int enc)
#else
void DESOCT_ncbc_encrypt(const unsigned char *in, unsigned char *out,
                         long length, DES_key_schedule *key, DES_cblock *ivec,
                         int enc)
#endif
{
    uint64_t *iv;
    uint64_t *inp, *outp;
    uint64_t in64 = 0, i0, r0;
    uint64_t out64 = 0;

    iv = (uint64_t *) & (*ivec)[0];
    inp = (uint64_t *) in;
    outp = (uint64_t *) out;
    i0 = inp[0];
    /* Initialise the keys */
    CVMX_MT_3DES_KEY(cvmx_htobe64(key->cvmkey), 0);
    CVMX_MT_3DES_KEY(cvmx_htobe64(key->cvmkey), 1);
    CVMX_MT_3DES_KEY(cvmx_htobe64(key->cvmkey), 2);
    if (DES_ENCRYPT == enc) {
        /* Encrypt */
        /* Initialise IV */
        CVMX_MT_3DES_IV (cvmx_htobe64(*iv));
        if (length >= 16) {
            CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(i0));
            length -= 8;
            inp++;
            outp++;
            if (length >= 8) {
                i0 = inp[0];
                CVMX_MF_3DES_RESULT(r0);
                CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(i0));
                while (1) {
                    outp[-1] = cvmx_betoh64(r0);
                    length -= 8;
                    inp++;
                    outp++;
                    if (length < 8) {
                        break;
                    }
                    i0 = inp[0];
                    CVMX_MF_3DES_RESULT(r0);
                    CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(i0));
                }
            }
            CVMX_MF_3DES_RESULT(r0);
            outp[-1] = cvmx_betoh64(r0);
        }
        if (length) {
            in64 = 0;
            if (length > 8) {
                CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(i0));
                CVMX_MF_3DES_RESULT(out64);
		*((uint64_t *) outp) = cvmx_betoh64(out64);
                outp++;
                inp++;
                length -= 8;
                memcpy((void *)&in64, (void *)inp, length);
            } else {
                memcpy((void *)&in64, (void *)inp, length);
            }
            CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(in64));
            CVMX_MF_3DES_RESULT(out64);
#ifndef CBC_ENC_C__DONT_UPDATE_IV
	    *((uint64_t *) outp) = cvmx_betoh64(out64);
	    *iv = cvmx_htobe64(*(uint64_t *) outp);
        } else {
            *iv = cvmx_htobe64(*(uint64_t *) (outp - 1));
#endif
        }
    } else {
        /* Decrypt */
        /* Initialise IV */
        CVMX_MT_3DES_IV (cvmx_htobe64(*iv));
        if (length >= 16) {
            CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(i0));
            length -= 8;
            inp++;
            outp++;
            if (length >= 8) {
                i0 = inp[0];
                CVMX_MF_3DES_RESULT(r0);
                CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(i0));
                while (1) {
                    outp[-1] = cvmx_betoh64(r0);
                    length -= 8;
                    inp++;
                    outp++;
                    if (length < 8) {
                        break;
                    }
                    i0 = inp[0];
                    CVMX_MF_3DES_RESULT(r0);
                    CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(i0));
                }
                CVMX_MF_3DES_RESULT(r0);
#ifndef CBC_ENC_C__DONT_UPDATE_IV
		*iv = cvmx_htobe64(*(uint64_t *) (inp - 1));
#endif
		outp[-1] = cvmx_betoh64(r0);
            }
        }
        if (length) {
            in64 = 0;
#ifndef CBC_ENC_C__DONT_UPDATE_IV
            *iv = cvmx_htobe64(*(uint64_t *) inp);
#endif
            memcpy((void *)&in64, (void *)inp, 8);
            CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(in64));
            CVMX_MF_3DES_RESULT(out64);
	    *((uint64_t *) outp) = cvmx_betoh64(out64);
        }
    }
}
