/***********************license start***************
 * Copyright (c) 2003-2016  Cavium Inc. (support@cavium.com). All rights
 * reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.

 *   * Neither the name of Cavium Inc. nor the names of
 *     its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.

 * This Software, including technical data, may be subject to U.S. export  control
 * laws, including the U.S. Export Administration Act and its  associated
 * regulations, and may be subject to export or import  regulations in other
 * countries.

 * TO THE MAXIMUM EXTENT PERMITTED BY LAW, THE SOFTWARE IS PROVIDED "AS IS"
 * AND WITH ALL FAULTS AND CAVIUM INC. MAKES NO PROMISES, REPRESENTATIONS OR
 * WARRANTIES, EITHER EXPRESS, IMPLIED, STATUTORY, OR OTHERWISE, WITH RESPECT TO
 * THE SOFTWARE, INCLUDING ITS CONDITION, ITS CONFORMITY TO ANY REPRESENTATION OR
 * DESCRIPTION, OR THE EXISTENCE OF ANY LATENT OR PATENT DEFECTS, AND CAVIUM
 * SPECIFICALLY DISCLAIMS ALL IMPLIED (IF ANY) WARRANTIES OF TITLE,
 * MERCHANTABILITY, NONINFRINGEMENT, FITNESS FOR A PARTICULAR PURPOSE, LACK OF
 * VIRUSES, ACCURACY OR COMPLETENESS, QUIET ENJOYMENT, QUIET POSSESSION OR
 * CORRESPONDENCE TO DESCRIPTION. THE ENTIRE  RISK ARISING OUT OF USE OR
 * PERFORMANCE OF THE SOFTWARE LIES WITH YOU.
 ***********************license end**************************************/
#include <openssl/opensslconf.h>

#include <stdio.h>
#include <string.h>
#ifdef OCTEON_SINGLE_PASS
#if !defined(OPENSSL_NO_DES) && !defined(OPENSSL_NO_SHA1)

# include <openssl/evp.h>
# include <openssl/objects.h>
# include <openssl/des.h>
# include <openssl/sha.h>
# include <openssl/rand.h>
# include "modes_lcl.h"
#include "octeon_papi.h"

# define NO_PAYLOAD_LENGTH       ((size_t)-1)
# ifndef EVP_CIPH_FLAG_AEAD_CIPHER
#  define EVP_CIPH_FLAG_AEAD_CIPHER       0x200000
#  define EVP_CTRL_AEAD_TLS1_AAD          0x16
#  define EVP_CTRL_AEAD_SET_MAC_KEY       0x17
# endif

# if !defined(EVP_CIPH_FLAG_DEFAULT_ASN1)
#  define EVP_CIPH_FLAG_DEFAULT_ASN1 0
# endif


# define SSL3_VERSION   0x0300
# define TLS1_VERSION   0x0301
# define TLS1_1_VERSION 0x0302
# define TLS1_2_VERSION 0x0303

#  define ddata(ctx) ((EVP_DESOCT_HMAC_SHA1 *)(ctx)->cipher_data)

static
int desoct_cbc_hmac_sha1_init_key (EVP_CIPHER_CTX *ctx, const unsigned char * inkey,
                                    const unsigned char *iv, int enc)
{
    EVP_DESOCT_HMAC_SHA1 *key = ddata(ctx);
    
    switch(EVP_CIPHER_key_length(ctx->cipher)) {
        case 24:
            key->cipher_type = OCT_3DES;
            break;
        case 8:
            key->cipher_type = OCT_DES_64;
            break;
        default:
            return 0;
    }
    
    if(inkey)
        memcpy((unsigned char *) key->cipher_key,inkey,EVP_CIPHER_key_length(ctx->cipher));
    if(iv)
        memcpy((unsigned char *)key->cipher_iv,iv,EVP_CIPHER_iv_length(ctx->cipher));
    key->payload_length = NO_PAYLOAD_LENGTH;
    key->ex_mode = EX_MODE_DEFAULT;
    return  1;
}

static
int desoct_cbc_hmac_sha1_ctrl(EVP_CIPHER_CTX *ctx, int type, int arg,
                                    void *ptr)
{
    EVP_DESOCT_HMAC_SHA1 *key = ddata(ctx);

    switch (type) {
    case EVP_CTRL_AEAD_SET_MAC_KEY:
        {
            unsigned char hmac_key[64];

            memset(hmac_key, 0, sizeof(hmac_key));

            if (arg > (int)sizeof(hmac_key)) {
                SHA_CTX key_ex;
                SHA1OCT_Init(&key_ex);
                SHA1OCT_Update(&key_ex, ptr, arg);
                SHA1OCT_Final(hmac_key, &key_ex);
            } else {
                memcpy(hmac_key, ptr, arg);
            }
            memcpy(key->mac_secret, hmac_key,64);
            OPENSSL_cleanse(hmac_key, sizeof(hmac_key));

            return 1;
        }
    case EVP_CTRL_AEAD_TLS1_AAD:
        {
            unsigned char *p = ptr;
            unsigned int len;

            if (arg != EVP_AEAD_TLS1_AAD_LEN)
                return -1;
 
            if(key->ex_mode == EX_MODE_DEFAULT) {
                len = p[arg - 2] << 8 | p[arg - 1];

                 if (ctx->encrypt) {
                    key->payload_length = len;
                    if ((p[arg - 4] << 8 | p[arg - 3]) >= TLS1_1_VERSION) {
                        len -= DES_BLOCK_SIZE;
                        p[arg - 2] = len >> 8;
                        p[arg - 1] = len;
                    }
                    memcpy(key->aux.tls_aad_8, ptr, arg);
                    return (int)(((len + SHA_DIGEST_LENGTH +
                                   DES_BLOCK_SIZE) & -DES_BLOCK_SIZE)
                                 - len);
                }
                 else {
                    memcpy(key->aux.tls_aad_8, ptr, arg);
                    key->payload_length = arg;
                    return SHA_DIGEST_LENGTH;
                }
            } else {
                memcpy(key->aux.tls_aad_8, ptr, arg);
                return 1;
            }
        }
    case EVP_CTRL_CIPH_SET_MODE :
        {
            switch(arg) {
                case EX_MODE_ENCRYPT:
                    key->ex_mode = EX_MODE_ENCRYPT;
                    return 1;
                case EX_MODE_DIGEST:
                    key->ex_mode = EX_MODE_DIGEST;
                    return 1;
                case EX_MODE_HMAC:
                    key->ex_mode = EX_MODE_HMAC;
                    return 1;
                case EX_MODE_DEFAULT:
                    key->ex_mode = EX_MODE_DEFAULT;
                    return 1;
                default:
                    return 0;
            }
        }
    default:
        return -1;
    }
}

static
int desoct_cbc_hmac_sha_cipher(EVP_CIPHER_CTX *ctx, unsigned char *out, const unsigned char *in, size_t in_len)
{
    EVP_DESOCT_HMAC_SHA1 *key = ddata(ctx);
    struct iovec_cav inp_vec[1],outp_vec[1];
    uint16_t inp_cnt = 1, outp_cnt = 1;
    uint8_t *ext_iv_ptr = NULL;
    int ret = 0;
    struct tls_rec_hdr *rec_header = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];
    if(key->ex_mode == EX_MODE_DEFAULT) {
        if(cvmx_htobe16(rec_header->u.version)  >= TLS1_1_VERSION && ctx->encrypt){
            ext_iv_ptr = (uint8_t *)in;
            inp_vec[0].ptr = (uint8_t *)in+EVP_CIPHER_iv_length(ctx->cipher);
            inp_vec[0].size = key->payload_length - EVP_CIPHER_iv_length(ctx->cipher);
            outp_vec[0].ptr = (uint8_t *)out;
        } else {
            inp_vec[0].ptr = (uint8_t *)in;
            inp_vec[0].size = (ctx->encrypt)?key->payload_length:in_len;
            outp_vec[0].ptr = (uint8_t *)out;
            if(cvmx_htobe16(rec_header->u.version) >= TLS1_1_VERSION)
                outp_vec[0].ptr += EVP_CIPHER_iv_length(ctx->cipher);
        }
    } else {
        inp_vec[0].ptr = (uint8_t *)in;
        inp_vec[0].size = in_len;
        outp_vec[0].ptr = out;
    }
    outp_vec[0].size = 16384;
    switch(cvmx_htobe16(rec_header->u.version)) {
        case SSL3_VERSION:
            if(ctx->encrypt)
                ret = oct_des_sha1_enc_ssl30 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,(uint16_t *)&in_len);
            else
                ret = oct_des_sha1_dec_ssl30 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,(uint16_t *)&in_len);
            break;
        case TLS1_VERSION:
            if(ctx->encrypt)
                ret = oct_des_sha1_enc_tls10 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,(uint16_t *)&in_len);
            else
                ret = oct_des_sha1_dec_tls10 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,(uint16_t *)&in_len);
            break;
        case TLS1_1_VERSION:
        case TLS1_2_VERSION:
            if(ctx->encrypt)
                ret = oct_des_sha1_enc_tls11 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,ext_iv_ptr, (uint16_t *)&in_len);
            else
                ret = oct_des_sha1_dec_tls11 (key, inp_vec, inp_cnt, outp_vec, outp_cnt,(uint16_t *)&in_len);
            break;
        default:
            return ret;
    }
    return ret;
}

static
EVP_CIPHER desoct_cbc_hmac_sha1_cipher = {
#  ifdef NID_aes_128_cbc_hmac_sha1
    NID_des_cbc_hmac_sha1,
#  else
    NID_undef,
#  endif
    8, 8, 8,
    EVP_CIPH_CBC_MODE | EVP_CIPH_FLAG_DEFAULT_ASN1 |
        EVP_CIPH_FLAG_AEAD_CIPHER,
    desoct_cbc_hmac_sha1_init_key,
    desoct_cbc_hmac_sha_cipher,
    NULL,
    sizeof(EVP_DESOCT_HMAC_SHA1),
    EVP_CIPH_FLAG_DEFAULT_ASN1 ? NULL : EVP_CIPHER_set_asn1_iv,
    EVP_CIPH_FLAG_DEFAULT_ASN1 ? NULL : EVP_CIPHER_get_asn1_iv,
    desoct_cbc_hmac_sha1_ctrl,
    NULL
};

static
EVP_CIPHER desoct_cbc3_hmac_sha1_cipher = {
#  ifdef NID_aes_128_cbc_hmac_sha1
    NID_des_cbc3_hmac_sha1,
#  else
    NID_undef,
#  endif
    8, 24, 8,
    EVP_CIPH_CBC_MODE | EVP_CIPH_FLAG_DEFAULT_ASN1 |
        EVP_CIPH_FLAG_AEAD_CIPHER,
    desoct_cbc_hmac_sha1_init_key,
    desoct_cbc_hmac_sha_cipher,
    NULL,
    sizeof(EVP_DESOCT_HMAC_SHA1),
    EVP_CIPH_FLAG_DEFAULT_ASN1 ? NULL : EVP_CIPHER_set_asn1_iv,
    EVP_CIPH_FLAG_DEFAULT_ASN1 ? NULL : EVP_CIPHER_get_asn1_iv,
    desoct_cbc_hmac_sha1_ctrl,
    NULL
};
const EVP_CIPHER *EVP_octdes_cbc_hmac_sha1(void)
{
    return &desoct_cbc_hmac_sha1_cipher;
}

const EVP_CIPHER *EVP_octdes3_cbc_hmac_sha1(void)
{
    return &desoct_cbc3_hmac_sha1_cipher;
}
#endif
#endif
