/** @file */
/*------------------------------------------------------------------
 * kdf/kdf.h - Key Derivation Function header file
 *
 * February 2015
 *
 * Copyright (c) 2015 by cisco Systems, Inc.
 * All rights reserved.
 **------------------------------------------------------------------
 */

#ifndef HEADER_KDF_H
#define HEADER_KDF_H

#include <openssl/opensslconf.h>

#ifdef __cplusplus
extern "C" {
#endif

#define PASSWORD_HASH_AMOUNT (1024 * 1024)

#define KDF_SRTP_AUTH_KEY_LEN 20
#define KDF_SRTP_SALT_KEY_LEN  14
#define KDF_SRTCP_AUTH_KEY_LEN KDF_SRTP_AUTH_KEY_LEN
#define KDF_SRTCP_SALT_KEY_LEN  KDF_SRTP_SALT_KEY_LEN
#define KDF_SRTP_SALT_LEN 14
#define KDF_SRTP_KDR_LEN 6
#define KDF_SRTP_IDX_LEN 6
#define KDF_SRTCP_IDX_LEN 4
#define KDF_SRTP_IV_LEN 16

int kdf_snmp(unsigned char *e_id, int e_len, const char *password, 
             int pw_len, unsigned char *digest);

int kdf_srtp(const EVP_CIPHER *cipher, char *km, char *ms, char *idx, 
             char *kdr, int label, char *buffer);

int kdf_ssh(const EVP_MD *evp_md, int id, unsigned int need, char *shared_secret,
            int ss_len, char *session_id, int session_id_len, char *hash, 
	    int hash_len, unsigned char *digest);

int kdf_802_11i(unsigned char *key, int key_len,
                unsigned char *prefix, int prefix_len,
		unsigned char *data, int data_len,
		unsigned char *out, unsigned int len, const EVP_MD *evp_md);

int kdf_tls12_P_hash(const EVP_MD *evp_md, const unsigned char *sec,
                     int sec_len,
		     const void *seed1, int seed1_len,
		     const void *seed2, int seed2_len,
		     const void *seed3, int seed3_len,
		     const void *seed4, int seed4_len,
		     const void *seed5, int seed5_len,
		     unsigned char *out, int olen);

int kdf_ikev2_gen(unsigned char *seedkey, const EVP_MD *evp_md, const void *nonce, 
                  unsigned int nonce_len, const void *shared_secret, 
		  unsigned int shared_secret_len);

int kdf_ikev2_rekey(unsigned char *seedkey, const EVP_MD *evp_md, const void *nonce, 
                    unsigned int nonce_len, const void *shared_secret, 
		    unsigned int shared_secret_len, int dh,
		    const void *sk_d, unsigned int skd_len);

int kdf_ikev2_dkm(unsigned char *dkm, unsigned int req_len, const EVP_MD *evp_md, 
                  const void *seedkey, unsigned int seedkey_len,
		  unsigned char *nonce, unsigned int nonce_len,
		  unsigned char *shared_secret, unsigned int shared_secret_len);


/* BEGIN ERROR CODES */
/*
 * The following lines are auto generated by the script mkerr.pl. Any changes
 * made after this point may be overwritten when the script is next run.
 */
void ERR_load_KDF_strings(void);

/* Error codes for the KDF functions. */

/* Function codes. */
# define KDF_F_KDF_802_11I                                103
# define KDF_F_KDF_IKEV2_DKM                              105
# define KDF_F_KDF_IKEV2_GEN                              106
# define KDF_F_KDF_IKEV2_REKEY                            107
# define KDF_F_KDF_SNMP                                   100
# define KDF_F_KDF_SRTP                                   101
# define KDF_F_KDF_SSH                                    102
# define KDF_F_KDF_TLS12_P_HASH                           104

/* Reason codes. */
# define KDF_R_BAD_CIPHER                                 104
# define KDF_R_BAD_DIGEST                                 103
# define KDF_R_BN_CTX_ERR                                 101
# define KDF_R_BN_GET_ERR                                 102
# define KDF_R_INPUT_PARAMETER_ERROR                      100
# define KDF_R_MALLOC                                     106

#ifdef  __cplusplus
}
#endif
#endif
