/*
 * Written by Stephen henson (steve@openssl.org) for the OpenSSL project
 * 2011.
 */
/* ====================================================================
 * Copyright (c) 2011 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include "cryptlib.h"
#ifdef OPENSSL_FIPS
# include <openssl/fips.h>
# include <openssl/fips_rand.h>
# include <openssl/rand.h>
#endif

int FIPS_mode(void)
{
    OPENSSL_init();
#ifdef OPENSSL_FIPS
    return FIPS_module_mode();
#else
    return 0;
#endif
}

/*
 * This value needs to match the value in the FOM. In the FOM
 * it's defined in fips/fips_local.h as FIPS_MODULE_VERSION_NUMBER.
 * This is used to prevent a user from using this version of 
 * CiscoSSL with an outdated FOM, which will result in seg faults
 * and other nasty behavior.
 */
#define FIPS_MIN_FOM_VERSION 0x60000000L
#define FIPS_MAX_FOM_VERSION 0x70000000L

int FIPS_mode_set(int r)
{
    OPENSSL_init();
#ifdef OPENSSL_FIPS
    if (FIPS_module_version() < FIPS_MIN_FOM_VERSION ||
	FIPS_module_version() > FIPS_MAX_FOM_VERSION) {
	CRYPTOerr(CRYPTO_F_FIPS_MODE_SET, CRYPTO_R_FIPS_FOM_VERSION_MISMATCH);
	return 0;
    }
# ifndef FIPS_AUTH_USER_PASS
#  define FIPS_AUTH_USER_PASS     "Default FIPS Crypto User Password"
# endif
    if (!FIPS_module_mode_set(r, FIPS_AUTH_USER_PASS))
        return 0;
    if (r)
        RAND_set_rand_method(FIPS_rand_get_method());
    else
        RAND_set_rand_method(RAND_get_drbg_method());
    return 1;
#else
    if (r == 0)
        return 1;
    CRYPTOerr(CRYPTO_F_FIPS_MODE_SET, CRYPTO_R_FIPS_MODE_NOT_SUPPORTED);
    return 0;
#endif
}

const char * FIPS_module_vertxt(void)
{
#ifdef OPENSSL_FIPS
    return FIPS_module_version_text();
#else
    return "NA";
#endif
}

int FIPS_cc_mode(void)
{
    OPENSSL_init();
#ifdef OPENSSL_FIPS
    return FIPS_module_cc_mode();
#else
    return 0;
#endif
}


int FIPS_cc_set(int r)
{
    OPENSSL_init();
#ifdef OPENSSL_FIPS
    if (!FIPS_module_cc_mode_set(r)) {
        return 0;
    }
    return 1;
#else
    if (r == 0) {
        return 1;
    }
    CRYPTOerr(CRYPTO_F_FIPS_CC_SET, CRYPTO_R_CC_MODE_NOT_SUPPORTED);
    return 0;
#endif
}
