/***********************license start***************
 * Copyright (c) 2003-2016  Cavium Inc. (support@cavium.com). All rights
 * reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.

 *   * Neither the name of Cavium Inc. nor the names of
 *     its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.

 * This Software, including technical data, may be subject to U.S. export  control
 * laws, including the U.S. Export Administration Act and its  associated
 * regulations, and may be subject to export or import  regulations in other
 * countries.

 * TO THE MAXIMUM EXTENT PERMITTED BY LAW, THE SOFTWARE IS PROVIDED "AS IS"
 * AND WITH ALL FAULTS AND CAVIUM INC. MAKES NO PROMISES, REPRESENTATIONS OR
 * WARRANTIES, EITHER EXPRESS, IMPLIED, STATUTORY, OR OTHERWISE, WITH RESPECT TO
 * THE SOFTWARE, INCLUDING ITS CONDITION, ITS CONFORMITY TO ANY REPRESENTATION OR
 * DESCRIPTION, OR THE EXISTENCE OF ANY LATENT OR PATENT DEFECTS, AND CAVIUM
 * SPECIFICALLY DISCLAIMS ALL IMPLIED (IF ANY) WARRANTIES OF TITLE,
 * MERCHANTABILITY, NONINFRINGEMENT, FITNESS FOR A PARTICULAR PURPOSE, LACK OF
 * VIRUSES, ACCURACY OR COMPLETENESS, QUIET ENJOYMENT, QUIET POSSESSION OR
 * CORRESPONDENCE TO DESCRIPTION. THE ENTIRE  RISK ARISING OUT OF USE OR
 * PERFORMANCE OF THE SOFTWARE LIES WITH YOU.
 ***********************license end**************************************/

#ifndef __OCTEON_PAPI_H
#define __OCTEON_PAPI_H

//#define DEBUG

//#define NO_SEQ_UPDATE
//#define DEBUG
#ifdef DEBUG
#define cavium_dbgprint printf
#else
#define cavium_dbgprint(format,...) do {} while(0);
#endif

#define bit_start7 (~((uint64_t)0xFF))
#define upper_32_bits(n)	((uint32_t)(((n) >> 16) >> 16))
#define upper_8_bits(n)		((uint8_t)((upper_32_bits(n) >> 16) >> 8))	
#define lower_32_bits(n) ((uint32_t)(n))

#define HMAC_SHA1_KEY_LEN	SHA_DIGEST_LENGTH

#ifndef DES_BLOCK_SIZE
#define DES_BLOCK_SIZE	8
#endif

#ifndef ROUNDUP16
#define ROUNDUP16(val) (((val) + 15)&0xfffffff0)
#endif

#ifndef ROUNDUP8
#define ROUNDUP8(val) (((val) + 7)&0xfffffff8)
#endif

#ifndef TARGET_LINUX
#define CAST_TO_PTRLONG(x)    ((uint64_t)x)      
#else
#if _MIPS_SIM == _ABIN32
#define CAST_TO_PTRLONG(x)    ((uint32_t)x)      
#else
#define CAST_TO_PTRLONG(x)    ((uint64_t)x)      
#endif
#endif

#define CRY_SHA1_INIT0	0x67452301EFCDAB89ULL
#define CRY_SHA1_INIT1	0x98BADCFE10325476ULL
#define CRY_SHA1_INIT2	0xC3D2E1F000000000ULL
#define CRY_HSH_BLK1	0x8000000000000000ULL
#define CRY_HSH_BLK2	0x0ull


#define CRY_SHA256_INIT0 0x6a09e667bb67ae85ULL
#define CRY_SHA256_INIT1 0x3c6ef372a54ff53aULL
#define CRY_SHA256_INIT2 0x510e527f9b05688cULL
#define CRY_SHA256_INIT3 0x1f83d9ab5be0cd19ULL

#define CRY_SHA384_INIT0 0xcbbb9d5dc1059ed8ULL
#define CRY_SHA384_INIT1 0x629a292a367cd507ULL
#define CRY_SHA384_INIT2 0x9159015a3070dd17ULL
#define CRY_SHA384_INIT3 0x152fecd8f70e5939ULL
#define CRY_SHA384_INIT4 0x67332667ffc00b31ULL
#define CRY_SHA384_INIT5 0x8eb44a8768581511ULL
#define CRY_SHA384_INIT6 0xdb0c2e0d64f98fa7ULL
#define CRY_SHA384_INIT7 0x47b5481dbefa4fa4ULL


#define SSL_CRY_SHA1_SSL3_LEN \
	( SHA_DIGEST_LENGTH + SHA_CBLOCK - 4 )



#define SHA1_FINISH(ptr, num, _bits) ({					\
	uint64_t *_my_ptr = (uint64_t *)ptr;				\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[0]), 0);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[1]), 1);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[2]), 2);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[3]), 3);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[4]), 4);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[5]), 5);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[6]), 6);					\
	if (num < 56) {							\
        	CVMX_MT_HSH_STARTSHA(_bits);				\
    	} else {							\
        	CVMX_MT_HSH_STARTSHA(cvmx_htobe64(_my_ptr[7]) );			\
		if (num == 64)						\
			CVMX_MT_HSH_DAT(0x8000000000000000ULL, 0);	\
		else							\
			CVMX_MT_HSH_DATZ(0);				\
		CVMX_MT_HSH_DATZ(1);					\
		CVMX_MT_HSH_DATZ(2);					\
		CVMX_MT_HSH_DATZ(3);					\
		CVMX_MT_HSH_DATZ(4);					\
		CVMX_MT_HSH_DATZ(5);					\
		CVMX_MT_HSH_DATZ(6);					\
		CVMX_MT_HSH_STARTSHA(_bits);				\
	}								\
})

#define SHA256_FINISH(ptr, num, _bits) ({				\
	uint64_t *_my_ptr = (uint64_t *)ptr;				\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[0]), 0);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[1]), 1);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[2]), 2);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[3]), 3);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[4]), 4);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[5]), 5);					\
	CVMX_MT_HSH_DAT(cvmx_htobe64(_my_ptr[6]), 6);					\
	if (num < 56) {							\
        	CVMX_MT_HSH_STARTSHA256(_bits);				\
    	} else {							\
        	CVMX_MT_HSH_STARTSHA256(cvmx_htobe64(_my_ptr[7]) );			\
		if (num == 64)						\
			CVMX_MT_HSH_DAT(0x8000000000000000ULL, 0);	\
		else							\
			CVMX_MT_HSH_DATZ(0);				\
		CVMX_MT_HSH_DATZ(1);					\
		CVMX_MT_HSH_DATZ(2);					\
		CVMX_MT_HSH_DATZ(3);					\
		CVMX_MT_HSH_DATZ(4);					\
		CVMX_MT_HSH_DATZ(5);					\
		CVMX_MT_HSH_DATZ(6);					\
		CVMX_MT_HSH_STARTSHA256(_bits);				\
	}								\
})

#define SHA384_FINISH(ptr, num, _bytes) ({				\
	uint64_t *_my_ptr = (uint64_t *)ptr;				\
	uint64_t _ab[2] = {0};						\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[0]), 0);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[1]), 1);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[2]), 2);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[3]), 3);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[4]), 4);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[5]), 5);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[6]), 6);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[7]), 7);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[8]), 8);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[9]), 9);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[10]), 10);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[11]), 11);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[12]), 12);				\
	CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[13]), 13);				\
	if (num < 112) {						\
		uint64_t_mul(_ab[0], _ab[1], _bytes, 8);		\
		CVMX_MT_HSH_DATW(_ab[0], 14);				\
        	CVMX_MT_HSH_STARTSHA512(_ab[1]);			\
    	} else {							\
		CVMX_MT_HSH_DATW(cvmx_htobe64(_my_ptr[14]), 14);			\
        	CVMX_MT_HSH_STARTSHA512(cvmx_htobe64(_my_ptr[15]));			\
		if (num == 128)						\
			CVMX_MT_HSH_DATW(0x8000000000000000ULL, 0);	\
		else							\
			CVMX_MT_HSH_DATWZ(0);				\
		CVMX_MT_HSH_DATWZ(1);					\
		CVMX_MT_HSH_DATWZ(2);					\
		CVMX_MT_HSH_DATWZ(3);					\
		CVMX_MT_HSH_DATWZ(4);					\
		CVMX_MT_HSH_DATWZ(5);					\
		CVMX_MT_HSH_DATWZ(6);					\
		CVMX_MT_HSH_DATWZ(7);					\
		CVMX_MT_HSH_DATWZ(8);					\
		CVMX_MT_HSH_DATWZ(9);					\
		CVMX_MT_HSH_DATWZ(10);					\
		CVMX_MT_HSH_DATWZ(11);					\
		CVMX_MT_HSH_DATWZ(12);					\
		CVMX_MT_HSH_DATWZ(13);					\
		uint64_t_mul(_ab[0], _ab[1], _bytes, 8);		\
		CVMX_MT_HSH_DATW(_ab[0], 14);				\
		CVMX_MT_HSH_STARTSHA512(_ab[1]);			\
	}								\
})


#define CRY_MD5_INIT0   0x0123456789abcdefULL
#define CRY_MD5_INIT1   0xfedcba9876543210ULL

#define MD5_PAD_LENGTH 48
#define SSL_CRY_MD5_SSL3_LEN \
	( MD5_DIGEST_LENGTH + MD5_PAD_LENGTH + MD5_DIGEST_LENGTH )

#define MD5_FINISH(ptr, num, _bits) ({					\
	uint64_t *_my_ptr = (uint64_t *)ptr;				\
	CVMX_MT_HSH_DAT(_my_ptr[0], 0);					\
	CVMX_MT_HSH_DAT(_my_ptr[1], 1);					\
	CVMX_MT_HSH_DAT(_my_ptr[2], 2);					\
	CVMX_MT_HSH_DAT(_my_ptr[3], 3);					\
	CVMX_MT_HSH_DAT(_my_ptr[4], 4);					\
	CVMX_MT_HSH_DAT(_my_ptr[5], 5);					\
	CVMX_MT_HSH_DAT(_my_ptr[6], 6);					\
	if (num < 56) {							\
        	CVMX_MT_HSH_STARTMD5(_bits);				\
    	} else {							\
        	CVMX_MT_HSH_STARTMD5(_my_ptr[7] );			\
		if (num == 64)						\
			CVMX_MT_HSH_DAT(0x8000000000000000ULL, 0);	\
		else							\
			CVMX_MT_HSH_DATZ(0);				\
		CVMX_MT_HSH_DATZ(1);					\
		CVMX_MT_HSH_DATZ(2);					\
		CVMX_MT_HSH_DATZ(3);					\
		CVMX_MT_HSH_DATZ(4);					\
		CVMX_MT_HSH_DATZ(5);					\
		CVMX_MT_HSH_DATZ(6);					\
		CVMX_MT_HSH_STARTMD5(_bits);				\
	}								\
})


#define MAX_MAC_LENGTH	SHA512_DIGEST_LENGTH
#define MAX_PAD_LENGTH	AES_BLOCK_SIZE

#define SHA384_DIGEST_LENGTH 48
#define SHA512_DIGEST_LENGTH 64
#define SHA384_CBLOCK 128
#define HMAC_SHA1_OUTER_BITS 640

typedef enum {
	OCT_RC4_128	= 0x0,   
	OCT_RC4_56	= 0x1,   
	OCT_RC4_40	= 0x2,   
	OCT_DES_64	= 0x3,
	OCT_DES_40	= 0x4,
	OCT_3DES	= 0x5,
	OCT_AES_128	= 0x6,
	OCT_AES_256	= 0x7,
	OCT_EXP_DES_64	= 0x8,
	OCT_AES128_GCM	= 0x9,
	OCT_AES256_GCM	= 0xa,
} oct_cipher_type_t;

typedef enum {
   OCT_MD5  = 0x0,
   OCT_SHA1 = 0x1, 
   OCT_SHA224 = 0x2,
   OCT_SHA256 = 0x3, 
   OCT_SHA384 = 0x4, 
   OCT_SHA512 = 0x5
}oct_hash_type_t;

struct ssl3_rec_hdr {
	uint8_t type;
	union {
		uint16_t version;
		struct {
			uint8_t major;
			uint8_t minor;
		} s;
	} u;
	uint16_t length;
} __attribute__((packed));

struct tls_rec_hdr {
	uint8_t type;
	union {
		uint16_t version;
		struct {
			uint8_t major;
			uint8_t minor;
		} s;
	} u;
	uint16_t length;
} __attribute__((packed));

struct iovec_cav {
   uint8_t *ptr;
   uint32_t size;
   uint8_t free_buf;
};

typedef struct {
    uint8_t cipher_key[32];
    uint8_t mac_secret[64];
    uint8_t cipher_iv[16];
    uint8_t cipher_type;
    uint8_t hash_type;
    uint8_t ex_mode;
    size_t payload_length;
    union {
        uint64_t tls_aad_64[2];
        uint32_t tls_aad_32[4]; 
        uint8_t tls_aad_8[16]; /* 13 used */
    } aux;
} EVP_AESOCT_HMAC_SHA1;

typedef struct {
    uint8_t cipher_key[32];
    uint8_t mac_secret[64];
    uint8_t cipher_iv[32];
    uint8_t cipher_type;
    uint8_t hash_type;
    size_t payload_length;
    uint8_t ex_mode;
    union {
        uint64_t tls_aad_64[2];
        uint32_t tls_aad_32[4]; 
        uint8_t tls_aad_8[16]; /* 13 used */
    } aux;
} EVP_DESOCT_HMAC_SHA1;

typedef struct iovec_cav cav_iovec_t;

int32_t
oct_aes_sha1_enc_ssl30(EVP_AESOCT_HMAC_SHA1 *key,
                       struct iovec_cav *inp_vec, uint16_t inp_cnt,
                       struct iovec_cav *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);

int32_t
oct_des_sha1_enc_ssl30(EVP_DESOCT_HMAC_SHA1 *key,
                       struct iovec_cav *inp_vec, uint16_t inp_cnt,
                       struct iovec_cav *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);

int32_t
oct_aes_sha1_dec_ssl30(EVP_AESOCT_HMAC_SHA1 *key,
                       struct iovec_cav *inp_vec, uint16_t inp_cnt,
                       struct iovec_cav *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);

int32_t
oct_des_sha1_dec_ssl30(EVP_DESOCT_HMAC_SHA1 *key,
                       struct iovec_cav *inp_vec, uint16_t inp_cnt,
                       struct iovec_cav *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);

int32_t
oct_aes_sha1_enc_tls10(EVP_AESOCT_HMAC_SHA1 *key,
                       struct iovec_cav *inp_vec, uint16_t inp_cnt,
                       struct iovec_cav *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);

int32_t
oct_des_sha1_enc_tls10(EVP_DESOCT_HMAC_SHA1 *key,
                       struct iovec_cav *inp_vec, uint16_t inp_cnt,
                       struct iovec_cav *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);

int32_t
oct_aes_sha1_dec_tls10(EVP_AESOCT_HMAC_SHA1 *key,
                       struct iovec_cav *inp_vec, uint16_t inp_cnt,
                       struct iovec_cav *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);

int32_t
oct_des_sha1_dec_tls10(EVP_DESOCT_HMAC_SHA1 *key,
                       struct iovec_cav *inp_vec, uint16_t inp_cnt,
                       struct iovec_cav *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);

int32_t
oct_aes_sha1_enc_tls11(EVP_AESOCT_HMAC_SHA1 *key,
                       cav_iovec_t *inp_vec, uint16_t inp_cnt,
                       cav_iovec_t *outp_vec, uint16_t outp_cnt,
                       uint8_t *ext_iv_data, uint16_t *len);

int32_t
oct_des_sha1_enc_tls11(EVP_DESOCT_HMAC_SHA1 *key,
                       cav_iovec_t *inp_vec, uint16_t inp_cnt,
                       cav_iovec_t *outp_vec, uint16_t outp_cnt,
                       uint8_t *ext_iv_data, uint16_t *len);

int32_t
oct_des_sha1_dec_tls11(EVP_DESOCT_HMAC_SHA1 *key,
                       cav_iovec_t *inp_vec, uint16_t inp_cnt,
                       cav_iovec_t *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);

int32_t
oct_aes_sha1_dec_tls11(EVP_AESOCT_HMAC_SHA1 *key,
                       cav_iovec_t *inp_vec, uint16_t inp_cnt,
                       cav_iovec_t *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);


int32_t
oct_aes_sha256_dec_tls12(EVP_AESOCT_HMAC_SHA1 *key,
                       cav_iovec_t *inp_vec, uint16_t inp_cnt,
                       cav_iovec_t *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);
int32_t
oct_aes_sha384_dec_tls12(EVP_AESOCT_HMAC_SHA1 *key,
                       cav_iovec_t *inp_vec, uint16_t inp_cnt,
                       cav_iovec_t *outp_vec, uint16_t outp_cnt,
                       uint16_t *len);
int32_t
oct_aes_sha256_enc_tls12(EVP_AESOCT_HMAC_SHA1 *key,
                       cav_iovec_t *inp_vec, uint16_t inp_cnt,
                       cav_iovec_t *outp_vec, uint16_t outp_cnt,
                       uint8_t *ext_iv_data, uint16_t *len);

int32_t
oct_aes_sha384_enc_tls12(EVP_AESOCT_HMAC_SHA1 *key,
                       cav_iovec_t *inp_vec, uint16_t inp_cnt,
                       cav_iovec_t *outp_vec, uint16_t outp_cnt,
                       uint8_t *ext_iv_data, uint16_t *len);

#endif /* __SSL_PAPI_H */
