/* crypto/rand/rdrand.c */
/* ====================================================================
 * Copyright (c) 2015 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include <openssl/rand.h>
#include <openssl/err.h>

/*
 * If we have an Intel processor with assembly code support, then
 * enable rdrand support in the entropy collector
 */
#if (defined(__i386)   || defined(__i386__)   || defined(_M_IX86) || \
     defined(__x86_64) || defined(__x86_64__) || \
     defined(_M_AMD64) || defined(_M_X64)) && \
    defined(OPENSSL_CPUID_OBJ) && \
    !defined(OPENSSL_NO_ASM)
extern unsigned int OPENSSL_ia32cap_P[4];
extern size_t OPENSSL_ia32_rdrand(void);

# define RDRAND_STIR_BYTES   16 * 4  /* should be multiple of 4 */

int RDRAND_enabled (void)
{
    /* bit 30 indicates if rdrand is supported */
    return ((OPENSSL_ia32cap_P[1] & (1 << 30)) ? 1 : 0);
}

void rdrand_stir(void)
{
    unsigned int rdrand_val;
    int i;

    if (RDRAND_enabled()) {
	for (i=0; i < (RDRAND_STIR_BYTES/4); i++) {
	    rdrand_val = OPENSSL_ia32_rdrand();
	    if (rdrand_val == 0) {
		RANDerr(RAND_F_RDRAND_STIR, RAND_R_RDRAND_FAILED);
	    } else {
		RAND_add(&rdrand_val, sizeof(rdrand_val), 4);
	    }
	}
    }
    
    /* zeroize the last random value */
    rdrand_val = 0;
}
#else
int RDRAND_enabled (void)
{
    return (0);
}

void rdrand_stir(void) {}
#endif


