/*
 *------------------------------------------------------------------
 * ctame.h
 *
 * Jan 2014
 *
 * Copyright (c) 2014 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#ifndef _CTAME_H_
# define _CTAME_H_
# include <stdint.h>
# ifndef OPENSSL_SYS_MSDOS
#  ifdef OPENSSL_UNISTD
#   include OPENSSL_UNISTD
#  else
#   include <unistd.h>
#  endif
# endif
# include <openssl/opensslconf.h>
# ifndef OPENSSL_NO_ECDSA
#  include <openssl/ecdsa.h>
# endif
# define CTAME_OBJECT_NAME_LEN_MAX  32
int ctame_init(void);

CTAMC_STATUS ctame_rsa_private_key_encrypt(const char *ctam_context,
                                           int length,
                                           const unsigned char *input,
                                           unsigned char *output,
                                           unsigned short *output_length,
                                           const char *key_name, int padding);

CTAMC_STATUS ctame_rsa_public_key_encrypt(const char *ctam_context,
                                          int length,
                                          const unsigned char *input,
                                          unsigned char *output,
                                          unsigned short *output_length,
                                          const char *key_name, int padding);

CTAMC_STATUS ctame_rsa_private_key_decrypt(const char *ctam_context,
                                           int length,
                                           const unsigned char *input,
                                           unsigned char *output,
                                           unsigned short *output_length,
                                           const char *key_name, int padding);

CTAMC_STATUS ctame_rsa_public_key_decrypt(const char *ctam_context,
                                          int length,
                                          const unsigned char *input,
                                          unsigned char *output,
                                          unsigned short *output_length,
                                          const char *key_name, int padding);

CTAMC_STATUS ctame_load_rsa_private_key(const char *ctam_context,
                                        const char *key_name,
                                        BIGNUM **bn_mod, BIGNUM **bn_exp);

CTAMC_STATUS ctame_read_object(const char *ctam_context,
                               const char *object_name,
                               unsigned char **data, unsigned short *length);

CTAMC_STATUS ctame_rsa_gen_keypair(const char *ctam_context,
                                   unsigned short key_length,
                                   unsigned int e_value,
                                   unsigned char zeroize,
                                   unsigned char persistent,
                                   unsigned char exportable,
                                   const char *key_name);

CTAMC_STATUS ctame_ec_gen_keypair(const char *ctam_context,
                                  ctamc_ecc_key_attr_t * ec_keyparms);

CTAMC_STATUS ctame_ecdsa_do_sign(const char *ctamc_context,
                                 const unsigned char *dgst,
                                 int dlen, const char *key_id,
                                 ECDSA_SIG **sig);

CTAMC_STATUS ctame_ecdsa_do_verify(const char *ctamc_context,
                                   const unsigned char *dgst,
                                   int dgst_len,
                                   const char *key_id, const ECDSA_SIG *sig);

CTAMC_STATUS ctame_load_ecc_private_key(const char *ctam_context,
                                        const char *key_name, EC_KEY **eckey);

CTAMC_STATUS ctame_get_true_random_bytes(const char *ctam_context,
                                         unsigned char *buf, int rand_size);
CTAMC_STATUS ctame_get_pseudo_random_bytes(const char *ctam_context,
                                           unsigned char *buf, int rand_size);

CTAMC_STATUS ctame_check_object_type(const char *ctam_context,
                                     const char *object_name,
                                     unsigned int obj_type);

unsigned int ctame_get_object_size(const char *ctam_context,
                                   const char *object_name);

/*
 * Used to store extra info pertaining to a RSA/ECC or symmetric key.
 * When the key id first loaded (ENGINE_load ...) the key name stored in TAM is set.
 * If the key is used in a sign/verify op the context is set during the init.
 * Similarly, if a symmetric key is used for a HMAC, the context is set on init.
 */
typedef struct ctame_extra_data_ {
    uint64_t sign_context;      // For sign operations. Set on init
    uint64_t verify_context;    // For verify operations. Set on init
    uint64_t hmac_context;      // For HAMC operations. Set on init
    unsigned char key_id_str[CTAME_OBJECT_NAME_LEN_MAX]; // Key name. Set
    // when key is
    // loaded
} ctame_extra_data_t;

#endif                          /* _CTAME_H_ */
