/*
 *  ctame_rand.c
 *  Cisco TAM Engine for use with Cisco Trusted Anchor Module Client
 *  OpenSSL engine
 *
 *  Feb 2014
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */
#ifdef CTAME_SUPPORT
# include <stdlib.h>
# include <string.h>
# include "../vendor_defns/ctamc.h"
# include "e_ctame_err.h"
# include "tam_types.h"
# include "safe_mem_lib.h"
# include "safe_lib_errno.h"
# include "safe_str_lib.h"

# ifdef USE_CISCO_SSM
#  include "ssmc/ssmc_rnd.c"
# else
#  include "tamc_api.h"
# endif

CTAMC_STATUS ctame_get_true_random_bytes(const char *ctam_context,
                                         unsigned char *buf, int rand_size)
{
    tam_status_e status;
    uint8_t *trand_data = NULL;
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;
    int retry = 1;
# define TRAND_RETRY_MAX 5

    // validate input
    if (!context || !buf || (rand_size == 0)) {
        CTAMEerr(CTAME_F_CTAME_GET_TRUE_RANDOM_BYTES, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }
    // call trand
    status = tamc_trand_read(context->session_id,
                             (uint32_t) rand_size, &trand_data);
    // retry if needed
    while (retry < TRAND_RETRY_MAX && status == TAM_ERROR_NO_RESOURCES) {
        sleep(1);
        status = tamc_trand_read(context->session_id,
                                 (uint32_t) rand_size, &trand_data);
        retry++;
    }

    if (status == TAM_SUCCESS) {
        memcpy_s(buf, rand_size, trand_data, rand_size);
    }

    free(trand_data);
    trand_data = NULL;
    return ((status == TAM_SUCCESS) ? CTAMC_SUCCESS : CTAMC_FAILURE);
}

CTAMC_STATUS ctame_get_pseudo_random_bytes(const char *ctam_context,
                                           unsigned char *buf, int rand_size)
{
    tam_status_e status;
    uint8_t *drand_data = NULL;
    ctamc_context_t *context = (ctamc_context_t *) ctam_context;

    // validate input
    if (!context || !buf || (rand_size == 0)) {
        CTAMEerr(CTAME_F_CTAME_GET_PSEUDO_RANDOM_BYTES, CTAME_R_INVALID_PARM);
        return (CTAMC_INVALID_ARG);
    }
    // call drand
    status = tamc_drand_read(context->session_id,
                             (uint32_t) rand_size, &drand_data);
    if (status == TAM_SUCCESS) {
        memcpy_s(buf, rand_size, drand_data, rand_size);
    }

    free(drand_data);
    drand_data = NULL;
    return ((status == TAM_SUCCESS) ? CTAMC_SUCCESS : CTAMC_FAILURE);
}

#endif                          /* CTAME_SUPPORT */
