/*
 *------------------------------------------------------------------
 * e_ctamc.h
 *
 * Jan 2014
 *
 * Copyright (c) 2014 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#ifndef __E_CTAMC_H_
# define __E_CTAMC_H_
# include "../vendor_defns/ctamc.h"
# include <openssl/ecdsa.h>

# define CTAMC_SUDI_KEY_NAME        "CTAMC_SUDI_KEY"
# define CTAMC_SUDI_CERT_NAME       "CTAMC_SUDI_CERT"
# define CTAMC_SUDI_CERT_CHAIN_NAME "CTAMC_SUDI_CERT_CHAIN"

/*
 * Predeclare the function pointer types that we dynamically load from the DSO.
 */

typedef CTAMC_STATUS t_ctamc_rsa_private_key_encrypt(const char
                                                     *ctamc_context,
                                                     int length,
                                                     const unsigned char
                                                     *input,
                                                     unsigned char *output,
                                                     unsigned short
                                                     *output_length,
                                                     const char *key_name,
                                                     int padding);

typedef CTAMC_STATUS t_ctamc_rsa_public_key_encrypt(const char *ctamc_context,
                                                    int length,
                                                    const unsigned char
                                                    *input,
                                                    unsigned char *output,
                                                    unsigned short
                                                    *output_length,
                                                    const char *key_name,
                                                    int padding);

typedef CTAMC_STATUS t_ctamc_rsa_private_key_decrypt(const char
                                                     *ctamc_context,
                                                     int length,
                                                     const unsigned char
                                                     *input,
                                                     unsigned char *output,
                                                     unsigned short
                                                     *output_length,
                                                     const char *key_name,
                                                     int padding);

typedef CTAMC_STATUS t_ctamc_rsa_public_key_decrypt(const char *ctamc_context,
                                                    int length,
                                                    const unsigned char
                                                    *input,
                                                    unsigned char *output,
                                                    unsigned short
                                                    *output_length,
                                                    const char *key_name,
                                                    int padding);

typedef CTAMC_STATUS t_ctamc_rsa_gen_keypair(const char *ctamc_context,
                                             unsigned short key_length,
                                             unsigned int e_value,
                                             unsigned char zeroize,
                                             unsigned char persistent,
                                             unsigned char exportable,
                                             const char *key_name);

typedef CTAMC_STATUS t_ctamc_read_object(const char *ctamc_context,
                                         const char *object_name,
                                         unsigned char **data,
                                         unsigned short *length);

typedef CTAMC_STATUS t_ctamc_load_rsa_private_key(const char *ctamc_context,
                                                  const char *key_name,
                                                  BIGNUM **mod, BIGNUM **exp);

typedef CTAMC_STATUS t_ctamc_ec_gen_keypair(const char *ctamc_context,
                                            ctamc_ecc_key_attr_t * params);

typedef CTAMC_STATUS t_ctamc_ecdsa_do_sign(const char *ctamc_context,
                                           const unsigned char *dgst,
                                           int dlen, const char *key_id,
                                           ECDSA_SIG **sig);

typedef CTAMC_STATUS t_ctamc_ecdsa_do_verify(const char *ctamc_context,
                                             const unsigned char *dgst,
                                             int dgst_len,
                                             const char *key_id,
                                             const ECDSA_SIG *sig);

typedef CTAMC_STATUS t_ctamc_load_ecc_private_key(const char *ctamc_context,
                                                  const char *key_name,
                                                  EC_KEY **eckey);

# ifndef OPENSSL_NO_RSA
extern t_ctamc_rsa_private_key_encrypt *p_ctamc_rsa_private_key_encrypt;
extern t_ctamc_rsa_public_key_encrypt *p_ctamc_rsa_public_key_encrypt;
extern t_ctamc_rsa_private_key_decrypt *p_ctamc_rsa_private_key_decrypt;
extern t_ctamc_rsa_public_key_decrypt *p_ctamc_rsa_public_key_decrypt;
extern t_ctamc_rsa_gen_keypair *p_ctamc_rsa_gen_keypair;
# endif                         // OPENSSL_NO_RSA
extern t_ctamc_read_object *p_ctamc_read_object;
extern t_ctamc_load_rsa_private_key *p_ctamc_load_rsa_private_key;

# ifndef OPENSSL_NO_EC
extern t_ctamc_ec_gen_keypair *p_ctamc_ec_gen_keypair;
extern t_ctamc_ecdsa_do_sign *p_ctamc_ecdsa_do_sign;
extern t_ctamc_ecdsa_do_verify *p_ctamc_ecdsa_do_verify;
extern t_ctamc_load_ecc_private_key *p_ctamc_load_ecc_private_key;
# endif

#endif                          // __E_CTAMC_H_
