/** @file */
/* -----------------------------------------------------------------
 * ssmc.h
 *
 * SSMc cryptographic functions.
 *
 * Trusted Services interest-ctam@cisco.com
 *
 * Copyright (c) 2014 by Cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#ifndef SSMC_TAMC_API_H
# define SSMC_TAMC_API_H

# include "tamc_ssm_api.h"
# include "tam_types.h"
# include <openssl/evp.h>
# include "ssm.h"

/*
 * SSM Key Store Structure
 */
typedef struct ssmc_key_buf_ {
    uint8_t buffer[0];
} ssmc_key_buf_t;

uint32_t ssmc_tam_curve_type_to_evp(tam_ec_curve_e curve);

tam_status_e ssmc_build_evp_key(tam_object_attr_t * attrs,
                                uint8_t * key_buffer, EVP_PKEY **key);
EVP_PKEY *ssmc_build_ec_key(ssmc_key_buf_t * ecc_key_buff,
                            tam_ec_curve_e ec_curve, uint32_t key_buff_len);
tam_status_e ssmc_load_key(void *session, char *key_name, char *key_owner,
                           EVP_PKEY **evp_key);
const EVP_MD *ssmc_tam_hash_type_to_evp(tam_hash_algo_e algo);
const RSA_METHOD *ssmc_get_rsa_method(void);
const ECDSA_METHOD *ssmc_get_ecdsa_method(void);

#endif                          // SSMC_ssmc_API_H
