/*
 *  ctamc_test_device.c
 *  Openssl Engine Cisco Trusted Anchor Module client Device Tests
 *
 *  January, 2014
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */

#include <openssl/engine.h>
#include "../../vendor_defns/ctamc.h"
#include "../ctame.h"

extern void ENGINE_load_ctamc(void);
static int debug = 0;

int main(int argc, char *argv[])
{
    static BIO *bio_out = NULL;
    ENGINE *e_ctamc = NULL;

    bio_out = BIO_new_fp(stdout, BIO_NOCLOSE);
    BIO_printf(bio_out, "Test entered\n");

#ifndef OPENSSL_NO_STATIC_ENGINE
    /*
     * Load the statically linked CTAM engine
     */
    if (debug)
        BIO_printf(bio_out, "In OPENSSL_NO_STATIC_ENGINE not defined.\n");
    ENGINE_load_ctamc();

    if ((e_ctamc = ENGINE_by_id("ctamc")) == NULL) {
        BIO_printf(bio_out, "invalid engine \"%s\"\n", "ctamc");
        ERR_print_errors(bio_out);
        goto end;
    }
#else
    /*
     * Dynamically load the CTAMC engine
     */
    if (debug)
        BIO_printf(bio_out, "In OPENSSL_NO_STATIC_ENGINE defined.\n");

    ENGINE_load_dynamic();
    e_ctamc = ENGINE_by_id("dynamic");
    if (!e_ctamc) {
        BIO_printf(bio_out, "ERROR Failed to load dynamic engine\n");
        ERR_print_errors(bio_out);
        goto end;
    }

    if (!ENGINE_ctrl_cmd(e_ctamc, "SO_PATH", 0, "../libctame.so", 0, 0)) {
        BIO_printf(bio_out, "ERROR Failed to set SO_PATH\n");
        goto end;
    }

    if (!ENGINE_ctrl_cmd(e_ctamc, "ID", 0, "ctamc", 0, 0)) {
        BIO_printf(bio_out, "ID ctamc\n");
        goto end;
    }

    if (!ENGINE_ctrl_cmd(e_ctamc, "LOAD", 0, NULL, 0, 0)) {
        BIO_printf(bio_out, "LOAD error\n");
        goto end;
    }
#endif                          /* OPENSSL_NO_STATIC_ENGINE */

    if (!ENGINE_init(e_ctamc)) {
        BIO_printf(bio_out, "ERROR Failed to initialize ctamc engine\n");
        ERR_print_errors(bio_out);
        goto end;
    }
    BIO_printf(bio_out, "PASS. Initialized CTAMC engine.\n");

 end:
    ERR_print_errors_fp(stderr);
    BIO_printf(bio_out, "Test complete!\n");

    return (0);
}
