/*
 *  ctamc_utility.c
 *  Openssl Engine Cisco Trusted Anchor Module client Device Tests
 *
 *  January, 2014
 *  Copyright 2014 Cisco Systems, Inc. All rights reserved.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// #include "ctamc_utility.h"
#include <openssl/engine.h>

#include "../../vendor_defns/ctamc.h"
#include "../ctame.h"
#include "safe_str_lib.h"

/* Client Interface APIs */

#ifdef USE_CISCO_SSM
# include "ssm.h"
# define SSMDIR "ssmdir"
#else
# include "tamc_api.h"
# include "tam_linux_ipc.h"
# include "tam_linux_mem.h"
#endif

extern void ENGINE_load_ctamc(void);

void *my_mem_alloc(size_t size)
{
    return (calloc(1, size));
}

void my_free(void *ptr)
{
    free(ptr);
}

/*
 * Load engine
 */
int load_ctamc_engine(ENGINE **e)
{

    ENGINE *e_ctamc = NULL;

    /*
     * Load the statically linked CTAM engine
     */
    ENGINE_load_ctamc();

    if ((e_ctamc = ENGINE_by_id("ctamc")) == NULL) {
        goto end;
    }

    if (!ENGINE_init(e_ctamc)) {
        goto end;
    }

    *e = e_ctamc;
 end:

    return (0);
}

// Login app to default device
void *ctamc_login(char *app_id, char *password, void **context,
                  void **session_id)
{
    tam_status_e rc;
    void *ctx = NULL;
    void *session = NULL;
    tam_credentials_t credentials;

    if (!app_id || !password || !context || !session_id)
        goto end;

#ifdef USE_CISCO_SSM
    ssm_cfg_t cfg;
    memzero_s(&cfg, sizeof(ssm_cfg_t));
    strncpy_s(cfg.storage_path, PATH_MAX, SSMDIR, PATH_MAX);
    rc = ssm_initialize(&cfg);
    if (rc != TAM_SUCCESS) {
        printf("ssm_initialize failed in ssmUtil_initSuite1Dev1App\n");
        goto end;
    }
#else
    /*
     * Register PD linux IPC and memory mgmt vectors
     */
    tam_register_ipc_cb(tam_linux_ipc_init,
                        tam_linux_ipc_cleanup,
                        tam_linux_ipc_open,
                        tam_linux_ipc_close,
                        tam_linux_ipc_send, tam_linux_ipc_receive);

    tam_register_mem_cb(my_mem_alloc, my_free);
#endif

    // setup credential
    credentials.credential_type = TAM_CREDENTIAL_PW;
    strncpy_s(&credentials.type.userpass.appid[0], TAM_APPID_LEN_MAX, app_id,
              TAM_APPID_LEN_MAX);
    strncpy_s(&credentials.type.userpass.pw[0], TAM_PW_LEN_MAX, password,
              TAM_PW_LEN_MAX);

    /* Initialize our context - must do before logging in */
    rc = tamc_init_context(&ctx, app_id, NULL);
    if (rc != TAM_SUCCESS)
        goto end;

    // login
    rc = tamc_app_login(ctx, NULL, &credentials, &session);
    if (rc != TAM_SUCCESS)
        goto end;

    // set results;
    *context = ctx;
    *session_id = session;
 end:
    return (session);
}

// Logout
void ctamc_logout(void **session_id)
{
    if (!session_id || !*session_id)
        return;

    // logout
    tamc_app_logout(session_id);
    return;
}
