/*
 *  ctamc.h
 *  Openssl Engine Cisco Trusted Anchor Module
 *
 *  Copyright 2014 Cisco Systems, Inc.. All rights reserved.
 */
#ifndef _CTAMC_H_
# define _CTAMC_H_
# define CTAMC_OBJECT_NAME_LEN_MAX  32
# define APP_ID_LEN 64

typedef enum {
    CTAMC_SUCCESS = 0,
    CTAMC_FAILURE = 1,
    CTAMC_NOT_SUPPORTED = 2,
    CTAMC_INVALID_ARG = 3,
    CTAMC_NO_RESOURCES = 4,
} CTAMC_STATUS;

typedef struct ctamc_context_ {
    unsigned char *ctame_device_handle;
    unsigned char *ctame_tam_handle;
    char app_id[APP_ID_LEN];
    void *session_id;
    void *ctamc_ctx;
} ctamc_context_t;

typedef struct ctamc_item_ {
    unsigned char *c;
    unsigned short len;
} ctamc_item_t;

typedef struct ctamc_rsa_keygen_parms_ {
    unsigned short key_len;
    unsigned int e_value;
    unsigned char zeroize;
    unsigned char persistent;
    unsigned char exportable;
    char *key_name;
} ctamc_rsa_keygen_parms_t;

typedef struct ctamc_rsa_key_attr_ {
    unsigned int key_len;       /* 1024, 2048, etc. */
    unsigned int e_value;       /* 0x03 or 0x10001 */
    unsigned char zeroize;
    unsigned char persistent;
    unsigned char exportable;   /* Exportable key? */
    char *key_name;
} ctamc_rsa_key_attr_t;

typedef enum ctamc_ec_curve_ {
    CTAMC_EC_CURVE_UNKN = 0,
    CTAMC_EC_CURVE_P192 = 1,
    CTAMC_EC_CURVE_P256 = 2,
    CTAMC_EC_CURVE_P384 = 3,
    CTAMC_EC_CURVE_P521 = 4,
    CTAMC_EC_CURVE_MAX_P = 5,
    CTAMC_EC_CURVE_BRAINPOOLP256 = 6,
    CTAMC_EC_CURVE_BRAINPOOLP384 = 7,
    CTAMC_EC_CURVE_BRAINPOOLP512 = 8,
    CTAMC_EC_CURVE_MAX
} ctamc_ec_curve_e;

typedef struct ctamc_ecc_key_attr_ {
    ctamc_ec_curve_e ec_curve;
    unsigned char zeroize;
    unsigned char persistent;
    unsigned char exportable;   /* Exportable key? */
    char key_name[CTAMC_OBJECT_NAME_LEN_MAX];
} ctamc_ecc_key_attr_t;

typedef struct ctamc_sym_key_attr_ {
    unsigned int key_size;      /* in bits */
    unsigned int iv_size;       /* in bits */
    unsigned char zeroize;
    unsigned char persistent;
    unsigned char exportable;   /* Exportable key? */
    char *key_name;
} ctamc_sym_key_attr_t;

typedef union ctamc_key_attr_ {
    ctamc_rsa_key_attr_t rsa;
    ctamc_ecc_key_attr_t ecc;
    ctamc_sym_key_attr_t sym;
} ctamc_key_attr_t;

typedef struct ctamc_file_attr_ {
    unsigned short file_size;
    unsigned char filename[CTAMC_OBJECT_NAME_LEN_MAX];
    unsigned char *file_buffer;
} ctamc_file_attr_t;

int ctamc_set_session_context();
#endif                          // _CTAMC_H_
