/* -----------------------------------------------------------------
 * crypto/aes/asm/aes_ctr-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */
#include <openssl/crypto.h>
#include <openssl/aes.h>
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif
#include "modes_lcl.h"
#include "cvmx-aes.h"

/*
 * octeon_ctr_initialize() - initialize the key for counter mode.
 */
void octeon_ctr_initialize(uint64_t * key, uint32_t key_len)
{
    CVMX_MT_AES_KEY (cvmx_htobe64(key[0]), 0);
    CVMX_MT_AES_KEY (cvmx_htobe64(key[1]), 1);
    CVMX_MT_AES_KEY (cvmx_htobe64(key[2]), 2);
    CVMX_MT_AES_KEY (cvmx_htobe64(key[3]), 3);
    CVMX_MT_AES_KEYLENGTH(key_len / 64 - 1);
}

/*
 * octeon_ctr_encrypt() - perform the encrypt/decrypt
 */
void octeon_ctr_encrypt(EVP_AES_OCT_CTR_CTX * cctx, uint64_t * out,
                        uint64_t * in, size_t len)
{
    int i;
    // step2
    int32_t nblocks = len / 16;
    int32_t pending = len % 16;
    int32_t lastblk = nblocks;
    lastblk -= (!pending);

    for (i = 0; i < lastblk; i++) {
        uint64_t tmpctrblk[2];
        CVMX_MT_AES_ENC0(cctx->ctrblk[0]);
        CVMX_MT_AES_ENC1(cctx->ctrblk[1]);

        // CTRBLK := CTRBLK + 1;
        if (cctx->ctrblk[1] != 0xFFffFFffFFffFFffULL) {
            cctx->ctrblk[1]++;
        } else {
            cctx->ctrblk[1] = 0;
            cctx->ctrblk[0]++;
        }

        CVMX_MF_AES_RESULT(tmpctrblk[0], 0);
        CVMX_MF_AES_RESULT(tmpctrblk[1], 1);

        // CT[i] := PT[i] xor AES(CTRBLK)
	out[2 * i] = in[2 * i] ^ cvmx_betoh64(tmpctrblk[0]);
	out[2 * i + 1] = in[2 * i + 1] ^ cvmx_betoh64(tmpctrblk[1]);
    }
    // step3
    // CT[n] := PT[n] XOR TRUNC(AES(CTRBLK))
    uint64_t tmpctrblk[2];

    CVMX_MT_AES_ENC0(cctx->ctrblk[0]);
    CVMX_MT_AES_ENC1(cctx->ctrblk[1]);
    CVMX_MF_AES_RESULT(tmpctrblk[0], 0);
    CVMX_MF_AES_RESULT(tmpctrblk[1], 1);

    // CTRBLK := CTRBLK + 1;
    if (cctx->ctrblk[1] != 0xFFffFFffFFffFFffULL) {
        cctx->ctrblk[1]++;
    } else {
        cctx->ctrblk[1] = 0;
        cctx->ctrblk[0]++;
    }

    cctx->ctrblk[0]     =  cvmx_betoh64(cctx->ctrblk[0]);
    cctx->ctrblk[1]     =  cvmx_betoh64(cctx->ctrblk[1]);
    tmpctrblk[0]  =  cvmx_betoh64(tmpctrblk[0]);
    tmpctrblk[1]  =  cvmx_betoh64(tmpctrblk[1]);
 

    uint8_t *ct = (uint8_t *) & out[2 * i];
    uint8_t *pt = (uint8_t *) & in[2 * i];
    uint8_t *ctr = (uint8_t *) & tmpctrblk[0];
    uint8_t k;
    uint32_t last_block_len = len % 16;
    if (!last_block_len) {
        last_block_len = 16;
    }
    for (k = 0; k < last_block_len; k++) {
        ct[k] = pt[k] ^ ctr[k];
    }
}
