
#ifdef OPENSSL_FIPSCANISTER
#include <openssl/fipssyms.h>
#endif
#include <machine/asm.h>
#include <machine/regdef.h>


#define accum	a0
#define addend	a1
#define len	a2
#define i	v0

.set noreorder 
.set noat

# len must be >1

LEAF(Vadd)
    ld	    t3, 0(addend)
    addiu   t0, zero, 1
    ld	    t2, 0(accum)
    mtm0    t0

    daddiu  i, len, -1	    # i=len-1
    
    daddiu  addend, addend, 8
    daddiu  accum, accum, 8    

    ld	    t1, 0(addend)
    vmulu   t3, t3, t2


VADD$L1: 
    ld	    t2, 0(accum)
    daddiu  i, i, -1
    
    daddiu  accum, accum, 8
    daddiu  addend, addend, 8

    sd	    t3, -16(accum)
    vmulu   t3, t1, t2

    bne	    i, zero, VADD$L1
    ld	    t1, 0(addend)


VADD$DONE:
    jr	    ra
    sd	    t3, -8(accum)
END(Vadd)


# Since I don't support signed vmul's for subtract I will
# skip over all trailing 0's and then do an add of 1's complement

LEAF(Vsub)
    daddiu  i, len, 	    # i=len
    daddiu  t8, zero, -1

VSUB$L1:
    ld	    t1, 0(addend)
    daddiu  i, i, -1
    bltz    i, VSUB$DONE
    daddiu  addend,addend, 8
    ld	    t3, 0(accum)
    beq	    t1, zero, VSUB$L1
    daddiu  accum,accum, 8

    ld	    t3, -8(addend)
    daddiu  t0, zero, 1
    ld	    t2, -8(accum)
    mtm0    t0
    xor	    t3, t3, t8
    daddiu  t3, t3, 1    

    ld	    t1, 0(addend)
    vmulu   t3, t3, t2
    beq     i, zero, VSUB$DONE
    nop


VSUB$L2: 
    ld	    t2, 0(accum)
    daddiu  i, i, -1
    
    daddiu  accum, accum, 8
    daddiu  addend, addend, 8

    xor	    t1, t1, t8

    sd	    t3, -16(accum)
    vmulu   t3, t1, t2

    bne	    i, zero, VSUB$L2
    ld	    t1, 0(addend)


VSUB$DONE:
    jr	    ra
    sd	    t3, -8(accum)
END(Vsub)

#undef accum
#undef addend
#undef len
#undef i



#define product		a0
#define mpcand		a1
#define mplier		a2
#define modulus		a3
#define recip		t0
#define len		t1
#define i               v0
#define init		t8

# len will be * 192b

LEAF(MontMul)
# stash all the registers
    daddiu  sp, sp, -40        # be sure to match this at epilog
    sd      mpcand, 0(sp)
    sd      product, 8(sp)
    sd      len, 16(sp)
    sd      modulus, 24(sp)

# for l=len-1; l>=0; l--)
    daddiu  init, zero, 0

MM$L3:
    ld	    t2, 0(mplier)
    ld	    t3, 8(mplier)
    ld	    t4, 16(mplier)
    mtm0    t2
    ld	    t2, 0(product)
    daddiu  mplier, mplier, 24
    mtm1    t3
    ld	    t3, 0(mpcand)
    mtm2    t4
    ld	    i, 16(sp)		# i=len
    and	    t4, t2, init
    ld	    t2, 8(product)

    v3mulu  t5, t3, t4
    daddiu  i, i, -1		# i=len-1

    ld	    t3, 8(mpcand)
    and	    t4, t2, init
    ld	    t2, 16(product)
    v3mulu  t6, t3, t4

    ld	    t3, 16(mpcand)
    sd	    t5, 0(product)
    and	    t4, t2, init
    v3mulu  t7, t3, t4    
    daddiu  product, product, 24
    ld	    t2, 0(product)
    daddiu  mpcand, mpcand, 24
    sd	    t6, -16(product)

MM$L1:
LOOP1:
    ld	    t3, 0(mpcand)
    and	    t4, t2, init
    ld	    t2, 8(product)
    daddiu  product, product, 24
    v3mulu  t5, t3, t4

    ld	    t3, 8(mpcand)
    and	    t4, t2, init
    ld	    t2, -8(product)
    daddiu  mpcand, mpcand, 24
    sd	    t7, -32(product)
    v3mulu  t6, t3, t4

    ld	    t3, -8(mpcand)
    and	    t4, t2, init
    ld	    t2, 0(product)
    daddiu  i, i, -1
    sd	    t5, -24(product)
    v3mulu  t7, t3, t4

    #bne    i, zero, MM$L1
    bgt    i, zero, MM$L1
    sd	    t6, -16(product)

POSTLOOP1:    
    v3mulu  t5, zero, zero
    sd	    t7, -8(product)
    v3mulu  t6, zero, zero
    daddiu  init, zero, -1
    ld	    mpcand, 0(sp)
    v3mulu  t7, zero, zero
    sd	    t5, 0(product)
    ld	    t2, 0(recip)
    sd	    t6, 8(product)
    ld	    t3, 8(recip)
    mtm0    t2
    sd	    t7, 16(product)
    ld	    t4, 16(recip)
    mtm1    t3
    #restore temporary variables
    ld	    product, 8(sp)
    mtm2    t4

    ld	    t2, 0(product)
    ld	    t3, 8(product)
    ld	    t4, 16(product)
    v3mulu  t2, t2, zero
    v3mulu  t3, t3, zero
    v3mulu  t4, t4, zero

    mtm0    t2
    ld	    t2, 0(modulus)
    mtm1    t3
    ld	    t3, 0(product)
    mtm2    t4

    v3mulu  zero, t2, t3
    ld	    t2, 8(modulus)
    ld	    t3, 8(product)
    v3mulu  zero, t2, t3
    ld	    t2, 16(modulus)
    ld	    t3, 16(product)
    daddiu  modulus, modulus, 24
    daddiu  product, product, 24
    v3mulu  zero, t2, t3

    ld	    i, 16(sp)
    daddiu  i, i, -1		# i=len-3

    ld	    t2, 0(product)
    ld	    t3, 0(modulus)
    v3mulu  t5, t3, t2
    daddiu  product, product, 8
    daddiu  modulus, modulus, 8

    ld	    t2, 0(product)
    ld	    t3, 0(modulus)
    v3mulu  t6, t3, t2
    daddiu  product, product, 8
    daddiu  modulus, modulus, 8

    ld	    t2, 0(product)
    ld	    t3, 0(modulus)
    sd	    t5, -40(product)
    v3mulu  t7, t3, t2
    daddiu  product, product, 8
    daddiu  modulus, modulus, 8
    ld	    t2, 0(product)
    sd	    t6, -40(product)

LOOP2:
MM$L2:
    ld	    t3, 0(modulus)
    or	    t4, zero, t2
    ld	    t2, 8(product)
    daddiu  product, product, 24
    v3mulu  t5, t3, t4

    ld	    t3, 8(modulus)
    or	    t4, zero, t2
    ld	    t2, -8(product)
    daddiu  modulus, modulus, 24
    sd	    t7, -56(product)
    v3mulu  t6, t3, t4

    ld	    t3, -8(modulus)
    or	    t4, zero, t2
    ld	    t2, 0(product)
    daddiu  i, i, -1
    sd	    t5, -48(product)
    v3mulu  t7, t3, t4
    #bne	    i, zero, MM$L2
    bgt    i, zero, MM$L2
    sd	    t6, -40(product)

    sd	    t7, -32(product)

    
POSTLOOP2:
    ld	    t2, 0(product)
    ld	    t3, 8(product)
    ld	    t4, 16(product)
    v3mulu  t5, zero, t2
    ld	    modulus, 24(sp)
    
    v3mulu  t6, zero, t3
    sd	    t5, -24(product)
    v3mulu  t7, zero, t4
    or	    t3, zero, product
    sd	    t6, -16(product)
    sd	    t7, -8(product)
    v3mulu  t6, zero, zero
    ld	    product, 8(sp)
    daddiu  len, len, -1

    #bne	    len, zero, MM$L3
    bgt    len, zero, MM$L3
    sd	    t6, 0(t3)

MM$DONE:    
    jr      ra
    daddiu  sp, sp, 40
END(MontMul)




// len will be *64
LEAF(MontMul576)
# stash all the registers
    daddiu  sp, sp, -64        # be sure to match this at epilog    
    sd	    s0,	0(sp)
    or	    s0, zero, zero
    sd	    s1,	8(sp)
    or	    s1, zero, zero
    sd	    s2,	16(sp)
    or	    s2, zero, zero
    sd	    s3,	24(sp)
    or	    s3, zero, zero
    sd	    s4,	32(sp)
    or	    s4, zero, zero
    sd	    s5,	40(sp)
    or	    s5, zero, zero
    sd	    s6,	48(sp)
    or	    s6, zero, zero
    sd	    s7,	56(sp)
    or	    s7, zero, zero

    ld	    t5, 0(mplier)
    daddiu  len, zero, 9
    ld	    recip, 0(recip)
    or	    t8, zero, zero
    ld	    t2, 0(mpcand)
    mtm0    t5
    ld	    t3, 8(mpcand)
    or	    t9, zero, zero

.align 3
MM576$L3:
    ld	    t4, 16(mpcand)
    vmulu   s0, t2, s0
    ld	    t2, 24(mpcand)
    vmulu   s1, t3, s1
    ld	    t3, 32(mpcand)
    vmulu   s2, t4, s2
    ld	    t4, 40(mpcand)
    vmulu   s3, t2, s3
    ld	    t2, 48(mpcand)
    vmulu   s4, t3, s4
    ld	    t3, 56(mpcand)
    vmulu   s5, t4, s5
    ld	    t4, 64(mpcand)
    vmulu   s6, t2, s6
    vmulu   s7, t3, s7
    vmulu   t8, t4, t8
    vmulu   t9, zero, zero
    
    mtm0    recip
    ld	    t2, 0(modulus)
    vmm0    t5, s0, zero
    ld	    t3, 8(modulus)
    daddiu  mplier, mplier, 8
    daddiu  len, len, -1

    ld	    t4, 16(modulus)
    vmulu   zero, t2, s0
    ld	    t2, 24(modulus)
    vmulu   s0, t3, s1
    ld	    t3, 32(modulus)
    vmulu   s1, t4, s2
    ld	    t4, 40(modulus)
    vmulu   s2, t2, s3
    ld	    t2, 48(modulus)
    vmulu   s3, t3, s4
    ld	    t3, 56(modulus)
    vmulu   s4, t4, s5
    ld	    t4, 64(modulus)
    vmulu   s5, t2, s6
    ld	    t5, 0(mplier)
    vmulu   s6, t3, s7
    ld	    t2, 0(mpcand)
    vmulu   s7, t4, t8
    ld	    t3, 8(mpcand)
    vmulu   t8, zero, t9
    sd	    s0,	0(product)
    vmulu   t9, zero, zero
    bne	    len, zero, MM576$L3
    mtm0    t5

    sd	    s1,	8(product)
    sd	    s2,	16(product)
    sd	    s3,	24(product)
    sd	    s4,	32(product)
    sd	    s5,	40(product)
    sd	    s6,	48(product)
    sd	    s7,	56(product)
    sd	    t8,	64(product)
    sd	    t9,	72(product)


MM576$DONE: 
    ld	    s0,	0(sp)
    ld	    s1,	8(sp)
    ld	    s2,	16(sp)
    ld	    s3,	24(sp)
    ld	    s4, 32(sp)
    ld	    s5,	40(sp)
    ld	    s6,	48(sp)
    ld	    s7,	56(sp)
    jr      ra
    daddiu  sp, sp, 64
END(MontMul576)

