/***********************license start***************
 * Copyright (c) 2003-2010  Cavium Inc. (support@cavium.com). All rights
 * reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.

 *   * Neither the name of Cavium Inc. nor the names of
 *     its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.

 * This Software, including technical data, may be subject to U.S. export  control
 * laws, including the U.S. Export Administration Act and its  associated
 * regulations, and may be subject to export or import  regulations in other
 * countries.

 * TO THE MAXIMUM EXTENT PERMITTED BY LAW, THE SOFTWARE IS PROVIDED "AS IS"
 * AND WITH ALL FAULTS AND CAVIUM INC. MAKES NO PROMISES, REPRESENTATIONS OR
 * WARRANTIES, EITHER EXPRESS, IMPLIED, STATUTORY, OR OTHERWISE, WITH RESPECT TO
 * THE SOFTWARE, INCLUDING ITS CONDITION, ITS CONFORMITY TO ANY REPRESENTATION OR
 * DESCRIPTION, OR THE EXISTENCE OF ANY LATENT OR PATENT DEFECTS, AND CAVIUM
 * SPECIFICALLY DISCLAIMS ALL IMPLIED (IF ANY) WARRANTIES OF TITLE,
 * MERCHANTABILITY, NONINFRINGEMENT, FITNESS FOR A PARTICULAR PURPOSE, LACK OF
 * VIRUSES, ACCURACY OR COMPLETENESS, QUIET ENJOYMENT, QUIET POSSESSION OR
 * CORRESPONDENCE TO DESCRIPTION. THE ENTIRE  RISK ARISING OUT OF USE OR
 * PERFORMANCE OF THE SOFTWARE LIES WITH YOU.
 ***********************license end**************************************/

/**
 * @file
 *
 * Main Octeon executive header file (This should be the second header
 * file included by an application).
 *
 * <hr>$Revision: 70030 $<hr>
*/
#ifndef __CVMX_H__
# define __CVMX_H__

/*
 * bit map for octeon features
 */
# define FEATURE_MAP_SIZE        128

/*
 * Control whether simple executive applications use 1-1 TLB mappings to
 * access physical ** memory addresses.  This must be disabled to allow large
 * programs that use more than ** the 0x10000000 - 0x20000000 virtual address
 * range.
 */
# ifndef CVMX_USE_1_TO_1_TLB_MAPPINGS
#  define CVMX_USE_1_TO_1_TLB_MAPPINGS 1
# endif

# ifndef CVMX_ENABLE_PARAMETER_CHECKING
#  define CVMX_ENABLE_PARAMETER_CHECKING 1
# endif

# ifndef CVMX_ENABLE_DEBUG_PRINTS
#  define CVMX_ENABLE_DEBUG_PRINTS 1
# endif

#ifdef  __cplusplus
extern "C" {
#endif

# include "ciscossl-cvmx-platform.h"
# include "ciscossl-cvmx-asm.h"
# include "octeon-model.h"

typedef struct cvm_rsa_key {
    uint64_t *mod;
    uint64_t *privexp;
    uint64_t *pubexp;
    uint64_t *expp;
    uint64_t *expq;
    uint64_t *p;
    uint64_t *q;
    uint64_t *coeff;
    int len;
    int eplen;
    int eqlen;
} cvm_rsa_key_t;

int cvm_ModExpCrt(uint64_t * product, uint64_t * base,
                  cvm_rsa_key_t * rkey, int nonwalign);

#ifdef  __cplusplus
}
#endif

/* the following are the structs and defines to access rng register */

typedef enum {
    CVMX_MIPS_SPACE_XKSEG = 3LL,
    CVMX_MIPS_SPACE_XKPHYS = 2LL,
    CVMX_MIPS_SPACE_XSSEG = 1LL,
    CVMX_MIPS_SPACE_XUSEG = 0LL
} cvmx_mips_space_t;

# define CVMX_IO_SEG CVMX_MIPS_SPACE_XKPHYS
# define CVMX_ADD_SEG(segment, add)          ((((uint64_t)segment) << 62) | (add))
# define CVMX_ADD_IO_SEG(add)                CVMX_ADD_SEG(CVMX_IO_SEG, (add))
# define CVMX_OCT_DID_RNG 8ULL

# define CASTPTR(type, v) ((type *)(long)(v))
/**
 * Builds a memory address for I/O based on the Major and Sub DID.
 *
 * @param major_did 5 bit major did
 * @param sub_did   3 bit sub did
 * @return I/O base address
 */
static inline uint64_t cvmx_build_io_address(uint64_t major_did,
                                             uint64_t sub_did)
{
    return ((0x1ull << 48) | (major_did << 43) | (sub_did << 40));
}

/**
 * cvmx_rnm_ctl_status
 *
 * RNM_CTL_STATUS = RNM's Control/Status Register
 *
 * The RNM's interrupt enable register.
 */
union cvmx_rnm_ctl_status {
    uint64_t u64;
    struct cvmx_rnm_ctl_status_s {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_12_63:52;
        uint64_t dis_mak:1;                         /**< Disable use of Master AES KEY */
        uint64_t eer_lck:1;                         /**< Encryption enable register locked */
        uint64_t eer_val:1;                         /**< Dormant encryption key match */
        uint64_t ent_sel:4;                         /**< ? */
        uint64_t exp_ent:1;                         /**< Exported entropy enable for random number generator */
        uint64_t rng_rst:1;                         /**< Reset RNG as core reset. */
        uint64_t rnm_rst:1;                         /**< Reset the RNM as core reset except for register
                                                         logic. */
        uint64_t rng_en:1;                          /**< Enable the output of the RNG. */
        uint64_t ent_en:1;                          /**< Entropy enable for random number generator. */
# else
        uint64_t ent_en:1;
        uint64_t rng_en:1;
        uint64_t rnm_rst:1;
        uint64_t rng_rst:1;
        uint64_t exp_ent:1;
        uint64_t ent_sel:4;
        uint64_t eer_val:1;
        uint64_t eer_lck:1;
        uint64_t dis_mak:1;
        uint64_t reserved_12_63:52;
# endif
    } s;
    struct cvmx_rnm_ctl_status_cn30xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_4_63:60;
        uint64_t rng_rst:1;                         /**< Reset RNG as core reset. */
        uint64_t rnm_rst:1;                         /**< Reset the RNM as core reset except for register
                                                         logic. */
        uint64_t rng_en:1;                          /**< Enable the output of the RNG. */
        uint64_t ent_en:1;                          /**< Entropy enable for random number generator. */
# else
        uint64_t ent_en:1;
        uint64_t rng_en:1;
        uint64_t rnm_rst:1;
        uint64_t rng_rst:1;
        uint64_t reserved_4_63:60;
# endif
    } cn30xx;
    struct cvmx_rnm_ctl_status_cn30xx cn31xx;
    struct cvmx_rnm_ctl_status_cn30xx cn38xx;
    struct cvmx_rnm_ctl_status_cn30xx cn38xxp2;
    struct cvmx_rnm_ctl_status_cn50xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_9_63:55;
        uint64_t ent_sel:4;                         /**< ? */
        uint64_t exp_ent:1;                         /**< Exported entropy enable for random number generator */
        uint64_t rng_rst:1;                         /**< Reset RNG as core reset. */
        uint64_t rnm_rst:1;                         /**< Reset the RNM as core reset except for register
                                                         logic. */
        uint64_t rng_en:1;                          /**< Enable the output of the RNG. */
        uint64_t ent_en:1;                          /**< Entropy enable for random number generator. */
# else
        uint64_t ent_en:1;
        uint64_t rng_en:1;
        uint64_t rnm_rst:1;
        uint64_t rng_rst:1;
        uint64_t exp_ent:1;
        uint64_t ent_sel:4;
        uint64_t reserved_9_63:55;
# endif
    } cn50xx;
    struct cvmx_rnm_ctl_status_cn50xx cn52xx;
    struct cvmx_rnm_ctl_status_cn50xx cn52xxp1;
    struct cvmx_rnm_ctl_status_cn50xx cn56xx;
    struct cvmx_rnm_ctl_status_cn50xx cn56xxp1;
    struct cvmx_rnm_ctl_status_cn50xx cn58xx;
    struct cvmx_rnm_ctl_status_cn50xx cn58xxp1;
    struct cvmx_rnm_ctl_status_s cn61xx;
    struct cvmx_rnm_ctl_status_cn63xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_11_63:53;
        uint64_t eer_lck:1;                         /**< Encryption enable register locked */
        uint64_t eer_val:1;                         /**< Dormant encryption key match */
        uint64_t ent_sel:4;                         /**< ? */
        uint64_t exp_ent:1;                         /**< Exported entropy enable for random number generator */
        uint64_t rng_rst:1;                         /**< Reset RNG as core reset. */
        uint64_t rnm_rst:1;                         /**< Reset the RNM as core reset except for register
                                                         logic. */
        uint64_t rng_en:1;                          /**< Enable the output of the RNG. */
        uint64_t ent_en:1;                          /**< Entropy enable for random number generator. */
# else
        uint64_t ent_en:1;
        uint64_t rng_en:1;
        uint64_t rnm_rst:1;
        uint64_t rng_rst:1;
        uint64_t exp_ent:1;
        uint64_t ent_sel:4;
        uint64_t eer_val:1;
        uint64_t eer_lck:1;
        uint64_t reserved_11_63:53;
# endif
    } cn63xx;
    struct cvmx_rnm_ctl_status_cn63xx cn63xxp1;
    struct cvmx_rnm_ctl_status_s cn66xx;
    struct cvmx_rnm_ctl_status_cn63xx cn68xx;
    struct cvmx_rnm_ctl_status_cn63xx cn68xxp1;
    struct cvmx_rnm_ctl_status_s cnf71xx;
};
typedef union cvmx_rnm_ctl_status cvmx_rnm_ctl_status_t;

union cvmx_mio_fus_rcmd {
    uint64_t u64;
    struct cvmx_mio_fus_rcmd_s {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_24_63:40;
        uint64_t dat:8;                             /**< 8bits of fuse data */
        uint64_t reserved_13_15:3;
        uint64_t pend:1;                            /**< SW sets this bit on a write to start FUSE read
                                                         operation.  HW clears when read is complete and
                                                         the DAT is valid */
        uint64_t reserved_9_11:3;
        uint64_t efuse:1;                           /**< When set, return data from the efuse storage
                                                         rather than the local storage */
        uint64_t addr:8;                            /**< The byte address of the fuse to read */
# else
        uint64_t addr:8;
        uint64_t efuse:1;
        uint64_t reserved_9_11:3;
        uint64_t pend:1;
        uint64_t reserved_13_15:3;
        uint64_t dat:8;
        uint64_t reserved_24_63:40;
# endif
    } s;
    struct cvmx_mio_fus_rcmd_cn30xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_24_63:40;
        uint64_t dat:8;                             /**< 8bits of fuse data */
        uint64_t reserved_13_15:3;
        uint64_t pend:1;                            /**< SW sets this bit on a write to start FUSE read
                                                         operation.  HW clears when read is complete and
                                                         the DAT is valid */
        uint64_t reserved_9_11:3;
        uint64_t efuse:1;                           /**< When set, return data from the efuse storage
                                                         rather than the local storage for the 320 HW fuses */
        uint64_t reserved_7_7:1;
        uint64_t addr:7;                            /**< The byte address of the fuse to read */
# else
        uint64_t addr:7;
        uint64_t reserved_7_7:1;
        uint64_t efuse:1;
        uint64_t reserved_9_11:3;
        uint64_t pend:1;
        uint64_t reserved_13_15:3;
        uint64_t dat:8;
        uint64_t reserved_24_63:40;
# endif
    } cn30xx;
    struct cvmx_mio_fus_rcmd_cn30xx cn31xx;
    struct cvmx_mio_fus_rcmd_cn30xx cn38xx;
    struct cvmx_mio_fus_rcmd_cn30xx cn38xxp2;
    struct cvmx_mio_fus_rcmd_cn30xx cn50xx;
    struct cvmx_mio_fus_rcmd_s cn52xx;
    struct cvmx_mio_fus_rcmd_s cn52xxp1;
    struct cvmx_mio_fus_rcmd_s cn56xx;
    struct cvmx_mio_fus_rcmd_s cn56xxp1;
    struct cvmx_mio_fus_rcmd_cn30xx cn58xx;
    struct cvmx_mio_fus_rcmd_cn30xx cn58xxp1;
    struct cvmx_mio_fus_rcmd_s cn61xx;
    struct cvmx_mio_fus_rcmd_s cn63xx;
    struct cvmx_mio_fus_rcmd_s cn63xxp1;
    struct cvmx_mio_fus_rcmd_s cn66xx;
    struct cvmx_mio_fus_rcmd_s cn68xx;
    struct cvmx_mio_fus_rcmd_s cn68xxp1;
    struct cvmx_mio_fus_rcmd_s cnf71xx;
};
typedef union cvmx_mio_fus_rcmd cvmx_mio_fus_rcmd_t;

union cvmx_mio_fus_dat2 {
    uint64_t u64;
    struct cvmx_mio_fus_dat2_s {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_48_63:16;
        uint64_t fus118:1;                          /**< Ignore Authentik disable */
        uint64_t rom_info:10;                       /**< Fuse information - ROM info */
        uint64_t power_limit:2;                     /**< Fuse information - Power limit */
        uint64_t dorm_crypto:1;                     /**< Fuse information - See NOCRYPTO */
        uint64_t fus318:1;                          /**< Reserved */
        uint64_t raid_en:1;                         /**< Fuse information - RAID enabled */
        uint64_t reserved_30_31:2;
        uint64_t nokasu:1;                          /**< Fuse information - Disable Kasumi */
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - DORM_CRYPTO and NOCRYPTO
                                                         together to select 1 of 4 mutually-exclusive
                                                         modes:
                                                         DORM_CRYPT=0,NOCRYPTO=0 AES/DES/HASH enabled
                                                         DORM_CRYPT=0,NOCRYPTO=1 AES/DES/HASH disable
                                                         DORM_CRYPT=1,NOCRYPTO=0 Dormant Encryption enable
                                                         DORM_CRYPT=1,NOCRYPTO=1 Authenik mode */
        uint64_t rst_sht:1;                         /**< Fuse information - When set, use short reset count */
        uint64_t bist_dis:1;                        /**< Fuse information - BIST Disable */
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t reserved_0_15:16;
# else
        uint64_t reserved_0_15:16;
        uint64_t chip_id:8;
        uint64_t bist_dis:1;
        uint64_t rst_sht:1;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t nokasu:1;
        uint64_t reserved_30_31:2;
        uint64_t raid_en:1;
        uint64_t fus318:1;
        uint64_t dorm_crypto:1;
        uint64_t power_limit:2;
        uint64_t rom_info:10;
        uint64_t fus118:1;
        uint64_t reserved_48_63:16;
# endif
    } s;
    struct cvmx_mio_fus_dat2_cn30xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_29_63:35;
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - AES/DES/HASH disable */
        uint64_t rst_sht:1;                         /**< Fuse information - When set, use short reset count */
        uint64_t bist_dis:1;                        /**< Fuse information - BIST Disable */
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t pll_off:4;                         /**< Fuse information - core pll offset
                                                         Used to compute the base offset for the core pll.
                                                         the offset will be (PLL_OFF ^ 8)
                                                         Note, these fuses can only be set from laser fuse */
        uint64_t reserved_1_11:11;
        uint64_t pp_dis:1;                          /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:1;
        uint64_t reserved_1_11:11;
        uint64_t pll_off:4;
        uint64_t chip_id:8;
        uint64_t bist_dis:1;
        uint64_t rst_sht:1;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t reserved_29_63:35;
# endif
    } cn30xx;
    struct cvmx_mio_fus_dat2_cn31xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_29_63:35;
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - AES/DES/HASH disable */
        uint64_t rst_sht:1;                         /**< Fuse information - When set, use short reset count */
        uint64_t bist_dis:1;                        /**< Fuse information - BIST Disable */
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t pll_off:4;                         /**< Fuse information - core pll offset
                                                         Used to compute the base offset for the core pll.
                                                         the offset will be (PLL_OFF ^ 8)
                                                         Note, these fuses can only be set from laser fuse */
        uint64_t reserved_2_11:10;
        uint64_t pp_dis:2;                          /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:2;
        uint64_t reserved_2_11:10;
        uint64_t pll_off:4;
        uint64_t chip_id:8;
        uint64_t bist_dis:1;
        uint64_t rst_sht:1;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t reserved_29_63:35;
# endif
    } cn31xx;
    struct cvmx_mio_fus_dat2_cn38xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_29_63:35;
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2)
                                                         (PASS2 Only) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable
                                                         (PASS2 Only) */
        uint64_t nocrypto:1;                        /**< Fuse information - AES/DES/HASH disable
                                                         (PASS2 Only) */
        uint64_t rst_sht:1;                         /**< Fuse information - When set, use short reset count */
        uint64_t bist_dis:1;                        /**< Fuse information - BIST Disable */
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t pp_dis:16;                         /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:16;
        uint64_t chip_id:8;
        uint64_t bist_dis:1;
        uint64_t rst_sht:1;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t reserved_29_63:35;
# endif
    } cn38xx;
    struct cvmx_mio_fus_dat2_cn38xx cn38xxp2;
    struct cvmx_mio_fus_dat2_cn50xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_34_63:30;
        uint64_t fus318:1;                          /**< Fuse information - a copy of fuse318 */
        uint64_t raid_en:1;                         /**< Fuse information - RAID enabled
                                                         (5020 does not have RAID co-processor) */
        uint64_t reserved_30_31:2;
        uint64_t nokasu:1;                          /**< Fuse information - Disable Kasumi */
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2)
                                                         (5020 does not have DFA co-processor) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - AES/DES/HASH disable */
        uint64_t rst_sht:1;                         /**< Fuse information - When set, use short reset count */
        uint64_t bist_dis:1;                        /**< Fuse information - BIST Disable */
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t reserved_2_15:14;
        uint64_t pp_dis:2;                          /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:2;
        uint64_t reserved_2_15:14;
        uint64_t chip_id:8;
        uint64_t bist_dis:1;
        uint64_t rst_sht:1;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t nokasu:1;
        uint64_t reserved_30_31:2;
        uint64_t raid_en:1;
        uint64_t fus318:1;
        uint64_t reserved_34_63:30;
# endif
    } cn50xx;
    struct cvmx_mio_fus_dat2_cn52xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_34_63:30;
        uint64_t fus318:1;                          /**< Fuse information - a copy of fuse318 */
        uint64_t raid_en:1;                         /**< Fuse information - RAID enabled */
        uint64_t reserved_30_31:2;
        uint64_t nokasu:1;                          /**< Fuse information - Disable Kasumi */
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - AES/DES/HASH disable */
        uint64_t rst_sht:1;                         /**< Fuse information - When set, use short reset count */
        uint64_t bist_dis:1;                        /**< Fuse information - BIST Disable */
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t reserved_4_15:12;
        uint64_t pp_dis:4;                          /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:4;
        uint64_t reserved_4_15:12;
        uint64_t chip_id:8;
        uint64_t bist_dis:1;
        uint64_t rst_sht:1;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t nokasu:1;
        uint64_t reserved_30_31:2;
        uint64_t raid_en:1;
        uint64_t fus318:1;
        uint64_t reserved_34_63:30;
# endif
    } cn52xx;
    struct cvmx_mio_fus_dat2_cn52xx cn52xxp1;
    struct cvmx_mio_fus_dat2_cn56xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_34_63:30;
        uint64_t fus318:1;                          /**< Fuse information - a copy of fuse318 */
        uint64_t raid_en:1;                         /**< Fuse information - RAID enabled */
        uint64_t reserved_30_31:2;
        uint64_t nokasu:1;                          /**< Fuse information - Disable Kasumi */
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - AES/DES/HASH disable */
        uint64_t rst_sht:1;                         /**< Fuse information - When set, use short reset count */
        uint64_t bist_dis:1;                        /**< Fuse information - BIST Disable */
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t reserved_12_15:4;
        uint64_t pp_dis:12;                         /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:12;
        uint64_t reserved_12_15:4;
        uint64_t chip_id:8;
        uint64_t bist_dis:1;
        uint64_t rst_sht:1;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t nokasu:1;
        uint64_t reserved_30_31:2;
        uint64_t raid_en:1;
        uint64_t fus318:1;
        uint64_t reserved_34_63:30;
# endif
    } cn56xx;
    struct cvmx_mio_fus_dat2_cn56xx cn56xxp1;
    struct cvmx_mio_fus_dat2_cn58xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_30_63:34;
        uint64_t nokasu:1;                          /**< Fuse information - Disable Kasumi */
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - AES/DES/HASH disable */
        uint64_t rst_sht:1;                         /**< Fuse information - When set, use short reset count */
        uint64_t bist_dis:1;                        /**< Fuse information - BIST Disable */
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t pp_dis:16;                         /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:16;
        uint64_t chip_id:8;
        uint64_t bist_dis:1;
        uint64_t rst_sht:1;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t nokasu:1;
        uint64_t reserved_30_63:34;
# endif
    } cn58xx;
    struct cvmx_mio_fus_dat2_cn58xx cn58xxp1;
    struct cvmx_mio_fus_dat2_cn61xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_48_63:16;
        uint64_t fus118:1;                          /**< Ignore Authentik disable */
        uint64_t rom_info:10;                       /**< Fuse information - ROM info */
        uint64_t power_limit:2;                     /**< Fuse information - Power limit */
        uint64_t dorm_crypto:1;                     /**< Fuse information - See NOCRYPTO */
        uint64_t fus318:1;                          /**< Reserved */
        uint64_t raid_en:1;                         /**< Fuse information - RAID enabled */
        uint64_t reserved_29_31:3;
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - DORM_CRYPTO and NOCRYPTO
                                                         together to select 1 of 4 mutually-exclusive
                                                         modes:
                                                         DORM_CRYPT=0,NOCRYPTO=0 AES/DES/HASH enabled
                                                         DORM_CRYPT=0,NOCRYPTO=1 AES/DES/HASH disable
                                                         DORM_CRYPT=1,NOCRYPTO=0 Dormant Encryption enable
                                                         DORM_CRYPT=1,NOCRYPTO=1 Authenik mode */
        uint64_t reserved_24_25:2;
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t reserved_4_15:12;
        uint64_t pp_dis:4;                          /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:4;
        uint64_t reserved_4_15:12;
        uint64_t chip_id:8;
        uint64_t reserved_24_25:2;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t reserved_29_31:3;
        uint64_t raid_en:1;
        uint64_t fus318:1;
        uint64_t dorm_crypto:1;
        uint64_t power_limit:2;
        uint64_t rom_info:10;
        uint64_t fus118:1;
        uint64_t reserved_48_63:16;
# endif
    } cn61xx;
    struct cvmx_mio_fus_dat2_cn63xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_35_63:29;
        uint64_t dorm_crypto:1;                     /**< Fuse information - Dormant Encryption enable */
        uint64_t fus318:1;                          /**< Reserved */
        uint64_t raid_en:1;                         /**< Fuse information - RAID enabled */
        uint64_t reserved_29_31:3;
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - AES/DES/HASH disable */
        uint64_t reserved_24_25:2;
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t reserved_6_15:10;
        uint64_t pp_dis:6;                          /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:6;
        uint64_t reserved_6_15:10;
        uint64_t chip_id:8;
        uint64_t reserved_24_25:2;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t reserved_29_31:3;
        uint64_t raid_en:1;
        uint64_t fus318:1;
        uint64_t dorm_crypto:1;
        uint64_t reserved_35_63:29;
# endif
    } cn63xx;
    struct cvmx_mio_fus_dat2_cn63xx cn63xxp1;
    struct cvmx_mio_fus_dat2_cn66xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_48_63:16;
        uint64_t fus118:1;                          /**< Ignore Authentik disable */
        uint64_t rom_info:10;                       /**< Fuse information - ROM info */
        uint64_t power_limit:2;                     /**< Fuse information - Power limit */
        uint64_t dorm_crypto:1;                     /**< Fuse information - See NOCRYPTO */
        uint64_t fus318:1;                          /**< Reserved */
        uint64_t raid_en:1;                         /**< Fuse information - RAID enabled */
        uint64_t reserved_29_31:3;
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - DORM_CRYPTO and NOCRYPTO
                                                         together to select 1 of 4 mutually-exclusive
                                                         modes:
                                                         DORM_CRYPT=0,NOCRYPTO=0 AES/DES/HASH enabled
                                                         DORM_CRYPT=0,NOCRYPTO=1 AES/DES/HASH disable
                                                         DORM_CRYPT=1,NOCRYPTO=0 Dormant Encryption enable
                                                         DORM_CRYPT=1,NOCRYPTO=1 Authenik mode */
        uint64_t reserved_24_25:2;
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t reserved_10_15:6;
        uint64_t pp_dis:10;                         /**< Fuse information - PP_DISABLES */
# else
        uint64_t pp_dis:10;
        uint64_t reserved_10_15:6;
        uint64_t chip_id:8;
        uint64_t reserved_24_25:2;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t reserved_29_31:3;
        uint64_t raid_en:1;
        uint64_t fus318:1;
        uint64_t dorm_crypto:1;
        uint64_t power_limit:2;
        uint64_t rom_info:10;
        uint64_t fus118:1;
        uint64_t reserved_48_63:16;
# endif
    } cn66xx;
    struct cvmx_mio_fus_dat2_cn68xx {
# ifdef __BIG_ENDIAN_BITFIELD
        uint64_t reserved_37_63:27;
        uint64_t power_limit:2;                     /**< Fuse information - Power limit */
        uint64_t dorm_crypto:1;                     /**< Fuse information - Dormant Encryption enable */
        uint64_t fus318:1;                          /**< Reserved */
        uint64_t raid_en:1;                         /**< Fuse information - RAID enabled */
        uint64_t reserved_29_31:3;
        uint64_t nodfa_cp2:1;                       /**< Fuse information - DFA Disable (CP2) */
        uint64_t nomul:1;                           /**< Fuse information - VMUL disable */
        uint64_t nocrypto:1;                        /**< Fuse information - AES/DES/HASH disable */
        uint64_t reserved_24_25:2;
        uint64_t chip_id:8;                         /**< Fuse information - CHIP_ID */
        uint64_t reserved_0_15:16;
# else
        uint64_t reserved_0_15:16;
        uint64_t chip_id:8;
        uint64_t reserved_24_25:2;
        uint64_t nocrypto:1;
        uint64_t nomul:1;
        uint64_t nodfa_cp2:1;
        uint64_t reserved_29_31:3;
        uint64_t raid_en:1;
        uint64_t fus318:1;
        uint64_t dorm_crypto:1;
        uint64_t power_limit:2;
        uint64_t reserved_37_63:27;
# endif
    } cn68xx;
    struct cvmx_mio_fus_dat2_cn68xx cn68xxp1;
    struct cvmx_mio_fus_dat2_cn61xx cnf71xx;
};
typedef union cvmx_mio_fus_dat2 cvmx_mio_fus_dat2_t;

# define CVMX_RNM_CTL_STATUS (CVMX_ADD_IO_SEG(0x0001180040000000ull))
# define CVMX_MIO_FUS_DAT2 (CVMX_ADD_IO_SEG(0x0001180000001410ull))
# define CVMX_MIO_FUS_RCMD (CVMX_ADD_IO_SEG(0x0001180000001500ull))

# define CVMX_MIO_BOOT_BIST_STAT (CVMX_ADD_IO_SEG(0x00011800000000F8ull))

# ifndef OCTEON_N32
#  define CVMX_BUILD_READ64(TYPE, LT)                                     \
static inline TYPE##_t cvmx_read64_##TYPE(uint64_t addr)                \
{                                                                       \
    return *CASTPTR(volatile TYPE##_t, addr);                           \
}
# else
/*
 * The N32 ABI passes all 64bit quantities in a single register, so it is
 * possible to use the arguments directly. We have to use inline assembly for
 * the actual store since a pointer would truncate the address
 */
#  define CVMX_BUILD_READ64(TYPE, LT)                                     \
static inline TYPE##_t cvmx_read64_##TYPE(uint64_t addr)                \
{                                                                       \
    TYPE##_t val;                                                       \
    asm volatile (LT " %[v], 0(%[c])": [v] "=r" (val) : [c] "r" (addr));\
    return val;                                                         \
}
# endif

CVMX_BUILD_READ64(int64, "ld");
CVMX_BUILD_READ64(int32, "lw");
CVMX_BUILD_READ64(int16, "lh");
CVMX_BUILD_READ64(int8, "lb");
CVMX_BUILD_READ64(uint64, "ld");
CVMX_BUILD_READ64(uint32, "lw");
CVMX_BUILD_READ64(uint16, "lhu");
CVMX_BUILD_READ64(uint8, "lbu");

# ifndef OCTEON_N32
#  define CVMX_BUILD_WRITE64(TYPE, ST)                                    \
static inline void cvmx_write64_##TYPE(uint64_t addr, TYPE##_t val)     \
{                                                                       \
    *CASTPTR(volatile TYPE##_t, addr) = val;                            \
}
# else

/*
 * The N32 ABI passes all 64bit quantities in a single register, so it is
 * possible to use the arguments directly. We have to use inline assembly for
 * the actual store since a pointer would truncate the address
 */
#  define CVMX_BUILD_WRITE64(TYPE, ST)                                    \
static inline void cvmx_write64_##TYPE(uint64_t addr, TYPE##_t val)     \
{                                                                       \
    asm volatile (ST " %[v], 0(%[c])" ::[v] "r" (val), [c] "r" (addr)); \
}
# endif

CVMX_BUILD_WRITE64(int64, "sd");
CVMX_BUILD_WRITE64(int32, "sw");
CVMX_BUILD_WRITE64(int16, "sh");
CVMX_BUILD_WRITE64(int8, "sb");
CVMX_BUILD_WRITE64(uint64, "sd");
CVMX_BUILD_WRITE64(uint32, "sw");
CVMX_BUILD_WRITE64(uint16, "sh");
CVMX_BUILD_WRITE64(uint8, "sb");

# define CVMX_RNG_LOAD_ADDRESS   CVMX_ADD_IO_SEG(cvmx_build_io_address(CVMX_OCT_DID_RNG, 0))
static inline uint64_t cvmx_read_csr(uint64_t csr_addr)
{
    return cvmx_read64_uint64(csr_addr);
}

static inline void cvmx_write_csr(uint64_t csr_addr, uint64_t val)
{
    cvmx_write64_uint64(csr_addr, val);

    /*
     * Perform an immediate read after every write to an RSL register to
     * force the write to complete. It doesn't matter what RSL read we do, so
     * we choose CVMX_MIO_BOOT_BIST_STAT because it is fast and harmless
     */
    if (((csr_addr >> 40) & 0x7ffff) == (0x118)) {
        cvmx_read64_uint64(CVMX_MIO_BOOT_BIST_STAT);
    }
}

/**
 * Reads 64 bits of random data from Random number generator
 *
 * @return random data
 */

/**
 * Read a byte of fuse data
 * @param byte_addr   address to read
 *
 * @return fuse value: 0 or 1
 */
static inline uint8_t cvmx_fuse_read_byte(int byte_addr)
{
    cvmx_mio_fus_rcmd_t read_cmd;

    read_cmd.u64 = 0;
    read_cmd.s.addr = byte_addr;
    read_cmd.s.pend = 1;
    cvmx_write_csr(CVMX_MIO_FUS_RCMD, read_cmd.u64);
    while ((read_cmd.u64 =
            cvmx_read_csr(CVMX_MIO_FUS_RCMD))&&read_cmd.s.pend) ;
    return (read_cmd.s.dat);
}

/**
 * Read a single fuse bit
 *
 * @param fuse   Fuse number (0-1024)
 *
 * @return fuse value: 0 or 1
 */
static inline int cvmx_fuse_read(int fuse)
{
    return ((cvmx_fuse_read_byte(fuse >> 3) >> (fuse & 0x7)) & 1);
}

#define OCTEON_IS_OCTEON1()	OCTEON_IS_MODEL(OCTEON_CN3XXX)
#define OCTEON_IS_OCTEONPLUS()	OCTEON_IS_MODEL(OCTEON_CN5XXX)
#define OCTEON_IS_OCTEON1PLUS()	(OCTEON_IS_OCTEON1() || OCTEON_IS_OCTEONPLUS())

static inline int octeon_has_feature_OCTEON_FEATURE_CRYPTO(void)
{
	if (!OCTEON_IS_OCTEON1PLUS()) {	/* OCTEON II and later */
		cvmx_mio_fus_dat2_t fus_2;
		fus_2.u64 = cvmx_read_csr(CVMX_MIO_FUS_DAT2);
		if (fus_2.s.nocrypto || fus_2.s.nomul) {
			return 0;
		} else if (!fus_2.s.dorm_crypto) {
			return 1;
		} else {
			cvmx_rnm_ctl_status_t st;
			st.u64 = cvmx_read_csr(CVMX_RNM_CTL_STATUS);
			return st.s.eer_val;
		}
	} else {
		return !cvmx_fuse_read(90);
	}
}

static inline int octeon_has_feature_OCTEON_FEATURE_DORM_CRYPTO(void)
{
	if (!OCTEON_IS_OCTEON1PLUS()) {	/* OCTEON II and later */
		cvmx_mio_fus_dat2_t fus_2;
		fus_2.u64 = cvmx_read_csr(CVMX_MIO_FUS_DAT2);
		return !fus_2.s.nocrypto && !fus_2.s.nomul && fus_2.s.dorm_crypto;
	} else {
		return 0;
	}
}


#endif                          /* __CVMX_H__ */
