/* -----------------------------------------------------------------
 * crypto/des/asm/des_enc-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <openssl/crypto.h>
#include "des_locl.h"

#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif

/*
 * DES_encrypt1 - encrypt/decrypt a single block using DES
 */
void DESOCT_encrypt1(DES_LONG *data, DES_key_schedule *ks, int enc)
{
    register uint64_t inp;
    unsigned char my_data[16];
    unsigned char *in = &my_data[0];
    unsigned char *out = &my_data[0];
    uint64_t *sao = (uint64_t *) my_data;

    l2c(data[0], in);
    l2c(data[1], in);

    inp = *sao;

    CVMX_MT_3DES_KEY (cvmx_htobe64(ks->cvmkey), 0);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks->cvmkey), 1);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks->cvmkey), 2);

    if (DES_ENCRYPT == enc) {
        CVMX_MT_3DES_ENC(cvmx_htobe64(inp));
        CVMX_MF_3DES_RESULT(inp);
    } else {
        CVMX_MT_3DES_DEC(cvmx_htobe64(inp));
        CVMX_MF_3DES_RESULT(inp);
    }
    *sao = cvmx_betoh64(inp);
    c2l(out, data[0]);
    c2l(out, data[1]);
}

/*
 * DES_decrypt3 - decrypt a single block using 3DES
 */
void DESOCT_decrypt3(DES_LONG *data, DES_key_schedule *ks1,
                     DES_key_schedule *ks2, DES_key_schedule *ks3)
{
    register uint64_t inp = *(uint64_t *) data;
    unsigned char my_data[16];
    unsigned char *in = &my_data[0];
    unsigned char *out = &my_data[0];
    uint64_t *sao = (uint64_t *) my_data;

    l2c(data[0], in);
    l2c(data[1], in);

    inp = *sao;
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks1->cvmkey), 0);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks2->cvmkey), 1);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks3->cvmkey), 2);


    CVMX_MT_3DES_DEC(cvmx_htobe64(inp));
    CVMX_MF_3DES_RESULT(inp);
    *sao = cvmx_betoh64(inp);
    c2l(out, data[0]);
    c2l(out, data[1]);
}

/*
 * DES_encrypt3 - encrypt a single block using 3DES
 */
void DESOCT_encrypt3(DES_LONG *data, DES_key_schedule *ks1,
                     DES_key_schedule *ks2, DES_key_schedule *ks3)
{
    register uint64_t inp;
    unsigned char my_data[16];
    unsigned char *in = &my_data[0];
    unsigned char *out = &my_data[0];
    uint64_t *sao = (uint64_t *) my_data;

    l2c(data[0], in);
    l2c(data[1], in);

    inp = *sao;

    CVMX_MT_3DES_KEY (cvmx_htobe64(ks1->cvmkey), 0);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks2->cvmkey), 1);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks3->cvmkey), 2);

    CVMX_MT_3DES_ENC(cvmx_htobe64(inp));
    CVMX_MF_3DES_RESULT(inp);
    *sao = cvmx_betoh64(inp);
    c2l(out, data[0]);
    c2l(out, data[1]);
}

#ifndef DES_DEFAULT_OPTIONS

# undef CBC_ENC_C__DONT_UPDATE_IV
# include "asm/ncbc_enc-octeon.c"/* DES_ncbc_encrypt */

/*
 * DES_ede3_ncbc_encrypt - encrypts/decrypts using the
 * cipher-block-chaining (CBC) mode of 3DES
 */
void
DESOCT_ede3_cbc_encrypt(const unsigned char *input, unsigned char *output,
                        long length, DES_key_schedule *ks1,
                        DES_key_schedule *ks2, DES_key_schedule *ks3,
                        DES_cblock *ivec, int enc)
{
    uint64_t *iv;
    uint64_t *inp, *outp;
    uint64_t in64 = 0, i0, r0;
    uint64_t out64 = 0;

    iv = (uint64_t *) & (*ivec)[0];
    inp = (uint64_t *) input;
    outp = (uint64_t *) output;
    i0 = inp[0];

    /* Initialise the keys */
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks1->cvmkey), 0);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks2->cvmkey), 1);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks3->cvmkey), 2);
    if (DES_ENCRYPT == enc) {
        /* Encrypt */
        /* Initialise IV */
        CVMX_MT_3DES_IV (cvmx_htobe64(*iv));
        if (length >= 16) {
            CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(i0));
            length -= 8;
            inp++;
            outp++;
            if (length >= 8) {
                i0 = inp[0];
                CVMX_MF_3DES_RESULT(r0);
                CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(i0));
                while (1) {
                    outp[-1] = cvmx_betoh64(r0);
                    length -= 8;
                    inp++;
                    outp++;
                    if (length < 8) {
                        break;
                    }
                    i0 = inp[0];
                    CVMX_MF_3DES_RESULT(r0);
                    CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(i0));
                }
            }
            CVMX_MF_3DES_RESULT(r0);
            outp[-1] = cvmx_betoh64(r0);
        }
        if (length) {
            in64 = 0;
            if (length > 8) {
                CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(i0));
                CVMX_MF_3DES_RESULT(out64);
		*((uint64_t *) outp) = cvmx_betoh64(out64);
                outp++;
                inp++;
                length -= 8;
                memcpy((void *)&in64, (void *)inp, length);
            } else {
                memcpy((void *)&in64, (void *)inp, length);
            }
            CVMX_MT_3DES_ENC_CBC (cvmx_htobe64(in64));
            CVMX_MF_3DES_RESULT(out64);
	    *((uint64_t *) outp) = cvmx_betoh64(out64);
	    *iv = *(uint64_t *) outp;
        } else {
	    *iv = *(uint64_t *) (outp - 1);
        }
    } else {
        /* Decrypt */
        /* Initialise IV */
        CVMX_MT_3DES_IV (cvmx_htobe64(*iv));
        if (length >= 16) {
            CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(i0));
            length -= 8;
            inp++;
            outp++;
            if (length >= 8) {
                i0 = inp[0];
                CVMX_MF_3DES_RESULT(r0);
                CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(i0));
                while (1) {
                    outp[-1] = cvmx_betoh64(r0);
                    length -= 8;
                    inp++;
                    outp++;
                    if (length < 8) {
                        break;
                    }
                    i0 = inp[0];
                    CVMX_MF_3DES_RESULT(r0);
                    CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(i0));
                }
                CVMX_MF_3DES_RESULT(r0);
                *iv = *(uint64_t *) (inp - 1);
                outp[-1] = cvmx_betoh64(r0);
            }
        }
        if (length > 8) {
	    in64 = 0;
            *iv = *(uint64_t *) inp;
            memcpy((void *)&in64, (void *)inp, 8);
	    CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(in64));
	    CVMX_MF_3DES_RESULT(out64);
	    *((uint64_t *) outp) = cvmx_betoh64(out64);
            outp++;
            inp++;
            length -= 8;

            memcpy((void *)&in64, (void *)inp, 8);
	    CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(in64));
	    CVMX_MF_3DES_RESULT(out64);
	    *((uint64_t *) outp) = cvmx_betoh64(out64);
        } else {
	    if (length) {
	        in64 = 0;
		*iv = *(uint64_t *) inp;
		memcpy ((void *) &in64, (void *) inp, 8);
		CVMX_MT_3DES_DEC_CBC (cvmx_htobe64(in64));
		CVMX_MF_3DES_RESULT (out64);
		*((uint64_t *) outp) = cvmx_betoh64(out64);
            }
        }
    }
}
#endif                          /* DES_DEFAULT_OPTIONS */
