/* -----------------------------------------------------------------
 * crypto/des/asm/ecb3_enc-octeon.c - Cavium Octeon specific hardware assist
 * functionality that is to reside within the FIPS canister.
 *
 * January 2013
 *
 * Copyright (c) 2013 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include "des_locl.h"
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif

/*
 * DES_ecb3_encrypt - encrypts/decrypts the input block by using
 * three-key Triple-DES encryption in ECB mode.
 */
void DESOCT_ecb3_encrypt(const_DES_cblock *input, DES_cblock *output,
                         DES_key_schedule *ks1, DES_key_schedule *ks2,
                         DES_key_schedule *ks3, int enc)
{
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks1->cvmkey), 0);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks2->cvmkey), 1);
    CVMX_MT_3DES_KEY (cvmx_htobe64(ks3->cvmkey), 2);

    if (DES_ENCRYPT == enc) {
        register uint64_t inp = *(uint64_t *) input;
        CVMX_MT_3DES_ENC (cvmx_htobe64(inp));
        CVMX_MF_3DES_RESULT(inp);
        *(uint64_t *)output = cvmx_betoh64(inp);
    } else {
        register uint64_t inp = *(uint64_t *) input;
        CVMX_MT_3DES_DEC (cvmx_htobe64(inp));
        CVMX_MF_3DES_RESULT(inp);
        *(uint64_t *)output = cvmx_betoh64(inp);
    }
}
