#ifndef _MIPS_ARCH_OCTEON3
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <signal.h>
#include <crypto.h>
#include "ciscossl-cvmx.h"
#else
#define USE_RUNTIME_MODEL_CHECKS 1
#include <stddef.h>
#include <stdlib.h>
#include <stdint.h>
#include "cvmx.h"
#endif
#include <openssl/crypto.h>

#ifdef OCTEON_HWA_CLEANSE
static unsigned char octeon_rng_enable = 0;
#endif
unsigned int OPENSSL_octcap_P = 0;
CVMX_SHARED uint32_t cvmx_app_init_processor_id;
static uint32_t ciscossl_cvmx_app_init_processor_id;

#define FEATURE_MAP_SIZE	128
CVMX_SHARED uint8_t octeon_feature_map[FEATURE_MAP_SIZE] __attribute__ ((aligned(128)));


/**
 * Returns the Octeon processor ID.
 *
 * @return Octeon processor ID from COP0
 */
int32_t ciscossl_cvmx_get_proc_id(void)
{
#ifndef OCTEON_SYS_SE
    return ciscossl_cvmx_app_init_processor_id;
#else

    uint32_t id;
 asm("mfc0 %0, $15,0":"=r"(id));
    return id;
#endif
}



static inline int cvmx_octeon_crypto_present(void)
{
    if (getenv("OPENSSL_OCTEON_NOCRYPTO")) {
        fprintf(stderr, "\nOcteon artificially running without hardware assist\n");
        return 0;
    }

    return (octeon_has_feature_OCTEON_FEATURE_CRYPTO() ||
	    octeon_has_feature_OCTEON_FEATURE_DORM_CRYPTO());
}

#if defined(__GNUC__) && __GNUC__>=2
void OPENSSL_cpuid_setup(void) __attribute__ ((constructor));
#endif
void OPENSSL_cpuid_setup(void)
{
#ifndef OPENSSL_SYS_SE
    cvmx_linux_enable_xkphys_access(0);
#endif
    if (cvmx_octeon_crypto_present()) {
        OPENSSL_octcap_P |= OCTEON_HW_ASSIST;
    }
}

