/*------------------------------------------------------------------
 * fips/embedsig.c - Utility to embed signatures into files on disk
 * 
 * August, 2012
 *
 * Copyright (c) 2011,2012 by cisco Systems, Inc.
 * All rights reserved.
 *------------------------------------------------------------------
 */

#include <Windows.h>
#include <stdio.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include <time.h>
#include <openssl/hmac.h>
#include <openssl/sha.h>
#include <openssl/crypto.h>
#include "fips.h"
//  Compile and link using the following:
//  cl /Fembedsig.obj -I..\inc32 /MD -W3 -Gs0 -Gy -nologo /Zi -c embedsig.c
//  link /nologo /subsystem:console /opt:ref /out:embedsig.exe embedsig.obj ..\out32dll\libeay32.lib 

static unsigned char *dllptr;
static HANDLE hMMFile = NULL;
static PIMAGE_DOS_HEADER dosHeader = NULL;
extern int fips_selftest_fail;

#define MakePtr( cast, ptr, addValue) (cast)((DWORD)(ptr) + (DWORD)(addValue))

unsigned char* parseDll (PIMAGE_DOS_HEADER dosHeader)
{
    PIMAGE_NT_HEADERS pNTHeader;
    DWORD base = (DWORD)dosHeader;
    unsigned i;
    PIMAGE_SECTION_HEADER section;

    pNTHeader = MakePtr(PIMAGE_NT_HEADERS, dosHeader, dosHeader->e_lfanew);

    if (pNTHeader->Signature != IMAGE_NT_SIGNATURE) {
        printf("\nInvalid NT header signature\n");
        return NULL;
    }

    section = IMAGE_FIRST_SECTION(pNTHeader);

    for ( i=1; i <= pNTHeader->FileHeader.NumberOfSections; i++, section++ ){
        if (!strncmp((char*)section->Name, "fsig", 4)) {
            return ((unsigned char *)(section->PointerToRawData + dllptr));
        }
    }

    printf("Could not find fsig section\n");
    return NULL;
}

unsigned char* get_fips_sig_offset(HANDLE hFile){
    unsigned char* fsig_offset = NULL;
    
    hMMFile = CreateFileMapping(hFile, NULL,
        PAGE_READWRITE, 0, 0, NULL);
    if (hMMFile == NULL) {
        printf("\nUnable to map file\n");
    }

    if(hMMFile != NULL){
        dllptr = (unsigned char *) MapViewOfFile(hMMFile,
            FILE_MAP_WRITE, 0, 0, 0 );
        if (dllptr == NULL) {
            printf("\nUnable to MapViewOfFile\n");        
        }
    }
    if(dllptr != NULL){
        dosHeader = (PIMAGE_DOS_HEADER) dllptr;
        if (dosHeader->e_magic == IMAGE_DOS_SIGNATURE) {
            fsig_offset = parseDll(dosHeader);
        } else {
            printf("\nUnable to get IMAGE_DOS_SIGNATURE\n");            
        }
    }
    return fsig_offset;
}


int main(int argc, char *argv[])
{
    HANDLE hFile = NULL;
    const unsigned char *fsig_raw_offset = NULL;
    DWORD le;
    unsigned i;

    unsigned char hashSig[SHA_DIGEST_LENGTH] = {'\0'};

    fips_selftest_fail = 0;

    hFile = CreateFile (argv[1], GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE) {
        printf("\nunable to open file\n");
        le = GetLastError();
        printf("last error=%d\n", le);
    } else {
        fsig_raw_offset = get_fips_sig_offset(hFile);
    }
    
    if(fsig_raw_offset != NULL && FIPS_calc_disk_hash_win32(hFile,hashSig)){

        printf("\nEmbeding this hash into the dll: ");
    	for (i=0;i<20;i++) {
	    	printf("%02x ", hashSig[i]);
	    }
        printf("\n");
        //write the signature to the FSIG section
        memcpy((void*)fsig_raw_offset, hashSig, SHA_DIGEST_LENGTH);
    } else {
        printf("\nUnable to calculate disk hash\n");
        le = GetLastError();
        printf("last error=%d\n", le);
    }

    
    if(dllptr != NULL){
        UnmapViewOfFile(dllptr);
    }
    if(hMMFile != NULL){
        CloseHandle(hMMFile);
    }
    if(hFile != NULL){
        CloseHandle(hFile);
    }



    return 0;
}
