/* ====================================================================
 * Copyright (c) 2011 The OpenSSL Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.openssl.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    openssl-core@openssl.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.openssl.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#define OPENSSL_FIPSAPI

#include <openssl/crypto.h>
#include <openssl/rand.h>
#include <openssl/fips_rand.h>
#include <openssl/err.h>
#include <openssl/bio.h>
#include <openssl/hmac.h>
#include <openssl/rsa.h>
#include <openssl/dsa.h>
#include <openssl/ec.h>
#include <string.h>
#include <limits.h>
#include "fips_locl.h"
#include "fips_auth.h"

#ifdef OPENSSL_FIPS

#include <openssl/fips.h>

#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#define atox(c) ((c)>='a'?((c)-'a'+10):((c)>='A'?(c)-'A'+10:(c)-'0'))

static int fips_auth_fail = 0;
int fips_mode = 0;
int cc_mode = 0;

static int fips_is_owning_thread(void);
static int fips_set_owning_thread(void);
static int fips_clear_owning_thread(void);
static unsigned char *fips_signature_witness(void);

#define fips_w_lock()	CRYPTO_w_lock(CRYPTO_LOCK_FIPS)
#define fips_w_unlock()	CRYPTO_w_unlock(CRYPTO_LOCK_FIPS)
#define fips_r_lock()	CRYPTO_r_lock(CRYPTO_LOCK_FIPS)
#define fips_r_unlock()	CRYPTO_r_unlock(CRYPTO_LOCK_FIPS)

static void fips_set_mode(int onoff)
{
    int owning_thread = fips_is_owning_thread();

    if (!owning_thread) fips_w_lock();
    fips_mode = onoff;
    if (!owning_thread) fips_w_unlock();
}

static void cc_set_mode(int onoff)
{
    int owning_thread = fips_is_owning_thread();

    if (!owning_thread) fips_w_lock();
    cc_mode = onoff;
    if (!owning_thread) fips_w_unlock();
}

int FIPS_module_mode(void)
{
    return (fips_mode);
}


int FIPS_module_cc_mode(void)
{
    int ret = 0;
    int owning_thread = fips_is_owning_thread();

    if (!owning_thread) fips_r_lock();
    ret = cc_mode;
    if (!owning_thread) fips_r_unlock();
    return ret;
}

int FIPS_selftest_failed(void)
{
    return (fips_post_selftest_failed());
}

/* Selftest failure fatal exit routine. This will be called
 * during *any* cryptographic operation. It has the minimum
 * overhead possible to avoid too big a performance hit.
 */

void FIPS_selftest_check(void)
    {
    if (fips_post_selftest_failed())
	{
	OpenSSLDie(__FILE__,__LINE__, "FATAL FIPS SELFTEST FAILURE");
	}
    }

void fips_set_selftest_fail(void)
    {
	fips_post_set_selftest_failed();
    }

extern const void         *FIPS_text_start(),  *FIPS_text_end();
extern const unsigned char FIPS_rodata_start[], FIPS_rodata_end[];
unsigned char              FIPS_signature [20] = { 0, 0xff };
#if defined WIN32_DISK_HASH_ENABLED && defined OPENSSL_SYSNAME_WIN32
#pragma section("fsig", read)
__declspec(allocate("fsig")) unsigned char FIPS_signature_disk[20] = "FIPSSIGNOTSETxxxxxx";
#endif
__fips_constseg

static const char          FIPS_hmac_key[]="etaonrishdlcupfm";

#if defined WIN32_DISK_HASH_ENABLED && defined OPENSSL_SYSNAME_WIN32
//#include <string.h>
//#include <stdlib.h>
#include <windows.h>
#include <openssl/sha.h>

#define NOTOKAY 0
#define OK 1


static const unsigned char *tsect_raw_offset=NULL;
static size_t tsect_raw_size=0;
static const unsigned char *rosect_raw_offset=NULL;
static size_t rosect_raw_size=0;
static unsigned char *dllptr=NULL;
static HANDLE hMMFile=NULL;

#define MakePtr( cast, ptr, addValue) (cast)((DWORD)(ptr) + (DWORD)(addValue))

//
// Dump the section table from a PE file or an OBJ
//
static int get_sections_and_offsets(PIMAGE_SECTION_HEADER section,
    unsigned cSections)
{
    unsigned i;

    for ( i=1; i <= cSections; i++, section++ )
    {
        if (!strncmp((char*)section->Name, ".text", 5)) {
            tsect_raw_offset = (const unsigned char *)(section->PointerToRawData + dllptr);
            tsect_raw_size = (size_t)section->SizeOfRawData;
        }
        if (!strncmp((char*)section->Name, ".rdata", 6)) {
            rosect_raw_offset = (const unsigned char *)(section->PointerToRawData + dllptr);
            rosect_raw_size = (size_t)section->SizeOfRawData;
        }
    }

    // some sanity to make sure that all our offsets and sizes are valid
    if( tsect_raw_offset == NULL || tsect_raw_size == 0  ||
        rosect_raw_offset == NULL || rosect_raw_size == 0){
        FIPSerr(FIPS_F_GET_SECTIONS_AND_OFFSETS,FIPS_R_UNABLE_TO_GET_SECTION_OFFSETS_AND_SIZES);
        return NOTOKAY;
    }

    return OK;

}

static void FIPS_unmap_dll()
{
    UnmapViewOfFile(dllptr);
    CloseHandle(hMMFile);
}

static int FIPS_map_dll(HANDLE dll_name)
{
    PIMAGE_DOS_HEADER dosHeader;

    // Map file
    hMMFile = CreateFileMapping(dll_name, NULL,
        PAGE_READONLY, 0, 0, NULL);
    if (hMMFile == NULL) {
		FIPSerr(FIPS_F_FIPS_MAP_DLL,FIPS_R_UNABLE_TO_MAP_FILE);
        return NOTOKAY;
    }

    dllptr = (unsigned char *) MapViewOfFile(hMMFile,
        FILE_MAP_READ, 0, 0, 0 );
    if (dllptr == NULL) {
		FIPSerr(FIPS_F_FIPS_MAP_DLL,FIPS_R_UNABLE_TO_MAP_FILE);
        return NOTOKAY;
    }

    // Ensure that the image is correct
    dosHeader = (PIMAGE_DOS_HEADER) dllptr;
    if (dosHeader->e_magic == IMAGE_DOS_SIGNATURE) {
        PIMAGE_NT_HEADERS pNTHeader;
        DWORD base = (DWORD)dosHeader;

        pNTHeader = MakePtr(PIMAGE_NT_HEADERS, dosHeader, dosHeader->e_lfanew);

        if (pNTHeader->Signature != IMAGE_NT_SIGNATURE) {
			FIPSerr(FIPS_F_FIPS_MAP_DLL,FIPS_R_HEADER_SIGNATURE_INCORRECT);
            return NOTOKAY;
        }
        // Find the sections and offsets
        return get_sections_and_offsets (IMAGE_FIRST_SECTION(pNTHeader),
            pNTHeader->FileHeader.NumberOfSections);
    }

	FIPSerr(FIPS_F_FIPS_MAP_DLL,FIPS_R_UNABLE_TO_MAP_FILE);
    return NOTOKAY;
}

static int FIPS_calc_hash_from_offsets(unsigned char* hmac_out_buf){
    HMAC_CTX c;
    unsigned char sig[SHA_DIGEST_LENGTH];
    unsigned int len =0;
    int retval = OK;

    // Assumes that sanity checks have been made on the varioius offsets

    HMAC_CTX_init(&c);
    do{
        retval = HMAC_Init(&c,FIPS_hmac_key,(int)strlen(FIPS_hmac_key),EVP_sha1());
        if(!retval){
            break;
        }
        retval = HMAC_Update(&c,tsect_raw_offset,tsect_raw_size);
        if(!retval){
            break;
        }
        retval = HMAC_Update(&c,rosect_raw_offset,rosect_raw_size);
        if(!retval){
            break;
        }
        retval = HMAC_Final(&c,sig,&len);
        if(!retval){
            break;
        }
    } while(0);

    // If sucessful, set the sig. Otherwise set the sig to 0
    if(retval){
        memcpy(hmac_out_buf,sig,SHA_DIGEST_LENGTH);
    } else {
        memset(hmac_out_buf,'\0',SHA_DIGEST_LENGTH);
    }

    return retval;

}

int FIPS_calc_disk_hash_win32(HANDLE dll_name, unsigned char* hmac_out_buf){

    int retval = 0;

    //Map the DLL to memory
    if(FIPS_map_dll(dll_name) == OK){
        //Now that we have the file mapped, calculate the hash
        retval = FIPS_calc_hash_from_offsets(hmac_out_buf);
    }
    FIPS_unmap_dll();

    return retval;
}
#else
int FIPS_calc_disk_hash_win32(){
    FIPSerr(FIPS_F_FIPS_CALC_DISK_HASH_WIN32,FIPS_R_DISK_HASHING_NOT_ENABLED);
    return 0;
}
#endif

unsigned int FIPS_incore_fingerprint(unsigned char *sig,unsigned int len)
    {
    const unsigned char *p1 = FIPS_text_start();
    const unsigned char *p2 = FIPS_text_end();
    const unsigned char *p3 = FIPS_rodata_start;
    const unsigned char *p4 = FIPS_rodata_end;
    HMAC_CTX c;

    HMAC_CTX_init(&c);
    HMAC_Init(&c,FIPS_hmac_key,(int)strlen(FIPS_hmac_key),EVP_sha1());

    /* detect overlapping regions */
    if (p1<=p3 && p2>=p3)
	p3=p1, p4=p2>p4?p2:p4, p1=NULL, p2=NULL;
    else if (p3<=p1 && p4>=p1)
	p3=p3, p4=p2>p4?p2:p4, p1=NULL, p2=NULL;

    if (p1)
	HMAC_Update(&c,p1,(size_t)p2-(size_t)p1);

    if (FIPS_signature>=p3 && FIPS_signature<p4)
	{
	/* "punch" hole */
	HMAC_Update(&c,p3,(size_t)FIPS_signature-(size_t)p3);
	p3 = FIPS_signature+sizeof(FIPS_signature);
	if (p3<p4)
	    HMAC_Update(&c,p3,(size_t)p4-(size_t)p3);
	}
    else
	HMAC_Update(&c,p3,(size_t)p4-(size_t)p3);

    if (!fips_post_corrupt(FIPS_TEST_INTEGRITY, 0, NULL))
	HMAC_Update(&c, (unsigned char *)FIPS_hmac_key, 1);

    if (!HMAC_Final(&c,sig,&len)) {
	return 0;
    }
    HMAC_CTX_cleanup(&c);
	
    return len;
    }

#if defined WIN32_DISK_HASH_ENABLED && defined OPENSSL_SYSNAME_WIN32
extern char * fips_get_lib_name(void);
unsigned int FIPS_incore_fingerprint_disk(unsigned char *sig,unsigned int len)
    {
    HINSTANCE libeay_dll = NULL;
    PIMAGE_DOS_HEADER dos_header = NULL;
    char fileName[MAX_PATH]={'\0'};

    libeay_dll = GetModuleHandle(fips_get_lib_name());
    if(libeay_dll != NULL){

        HANDLE hFile;
        int ret_len=0;

        // Need to open the file on disk to hash
        GetModuleFileName(libeay_dll, fileName, MAX_PATH);
        hFile = CreateFile (fileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
        if (hFile == INVALID_HANDLE_VALUE) {
			FIPSerr(FIPS_F_FIPS_INCORE_FINGERPRINT_DISK,FIPS_R_UNABLE_TO_OPEN_FILE);
            memset(sig,0,SHA_DIGEST_LENGTH);            
        } else {
            if(FIPS_calc_disk_hash_win32(hFile,sig)){
                ret_len = SHA_DIGEST_LENGTH;
            }
        }
        CloseHandle(hFile);

        return ret_len;
    }
    return 0;
    }
#endif

int FIPS_check_incore_fingerprint()
    {
    unsigned char sig[EVP_MAX_MD_SIZE];
    unsigned int len;
    unsigned int bad_sig = 0;
    int rv = 0;

#if defined(__sgi) && (defined(__mips) || defined(mips))
    extern int __dso_displacement[];
#else
    extern int OPENSSL_NONPIC_relocated;
#endif

    if (!fips_post_started(FIPS_TEST_INTEGRITY, 0, NULL))
	return 1;

    if (FIPS_text_start()==NULL)
	{
	FIPSerr(FIPS_F_FIPS_CHECK_INCORE_FINGERPRINT,FIPS_R_UNSUPPORTED_PLATFORM);
	goto err;
	}
	
#ifdef FIPS_INCORE_ALT_METHODS_ENABLED
    len=FIPS_incore_fingerprint_disk(sig,sizeof(sig));
    if (len!=sizeof(FIPS_signature_disk) ||
        memcmp(FIPS_signature_disk,sig,sizeof(FIPS_signature_disk))){
        bad_sig = 1;
    }
#else
    len=FIPS_incore_fingerprint(sig,sizeof(sig));
    if (len!=sizeof(FIPS_signature) ||
        memcmp(FIPS_signature,sig,sizeof(FIPS_signature)))
    {
        bad_sig = 1;
    }
#endif

    if (bad_sig)
	{
	if (FIPS_signature>=FIPS_rodata_start && FIPS_signature<FIPS_rodata_end)
	    FIPSerr(FIPS_F_FIPS_CHECK_INCORE_FINGERPRINT,FIPS_R_FINGERPRINT_DOES_NOT_MATCH_SEGMENT_ALIASING);
#if defined(__sgi) && (defined(__mips) || defined(mips))
	else if (__dso_displacement!=NULL)
#else
	else if (OPENSSL_NONPIC_relocated)
#endif
	    FIPSerr(FIPS_F_FIPS_CHECK_INCORE_FINGERPRINT,FIPS_R_FINGERPRINT_DOES_NOT_MATCH_NONPIC_RELOCATED);
	else
		FIPSerr(FIPS_F_FIPS_CHECK_INCORE_FINGERPRINT,FIPS_R_FINGERPRINT_DOES_NOT_MATCH);
#ifdef FIPS_INCORE_ALT_METHODS_ENABLED
			FIPSerr(FIPS_F_FIPS_CHECK_INCORE_FINGERPRINT,FIPS_R_MEMORY_HASHING_MAY_CAUSE_FINGERPRINT_MISMATCH_FAILURES_IN_DISK_HASH_IMAGE);
#endif
	    
#ifdef OPENSSL_FIPS_DEBUGGER
    	rv = 1;
#endif
	goto err;
	}
    rv = 1;
    err:
    if (rv == 0)
    	fips_post_failed(FIPS_TEST_INTEGRITY, 0, NULL);
    else
	if (!fips_post_success(FIPS_TEST_INTEGRITY, 0, NULL))
		return 0;
    return rv;
    }

static int fips_asc_check(const unsigned char *sig, const char *asc_sig)
    {
    char tsig[20];
    const char *p;
    int i;
    if (strlen(asc_sig) != 40)
	return 0;
    for (i = 0, p = asc_sig; i < 20; i++, p += 2)
	tsig[i] = (atox(p[0]) << 4) | atox(p[1]);
    if (memcmp(tsig, sig, 20))
	return 0;
    return 1;
    }

static int fips_check_auth(const char *auth)
    {
    unsigned char auth_hmac[20];
    unsigned int hmac_len;
    if (fips_auth_fail)
	return 0;
    if (strlen(auth) < FIPS_AUTH_MIN_LEN)
	return 0;
    if (!HMAC(EVP_sha1(), FIPS_AUTH_KEY, (int)strlen(FIPS_AUTH_KEY),
		(unsigned char *)auth, strlen(auth), auth_hmac, &hmac_len))
	return 0;
    if (hmac_len != sizeof(auth_hmac))
	return 0;

    if (fips_asc_check(auth_hmac, FIPS_AUTH_CRYPTO_OFFICER))
	return 1;

    if (fips_asc_check(auth_hmac, FIPS_AUTH_CRYPTO_USER))
	return 1;

    return 0;
    }
	
    

int FIPS_module_mode_set(int onoff, const char *auth)
    {
    int ret = 0;

    fips_w_lock();
    fips_set_owning_thread();

    if(onoff) {

	/*
	 * Check if the POST has completed.  We check this in the rare and unlikely
	 * event the user is invoking FIPS_module_mode_set() prior to the POST
	 * completing, which could potentially allow the FOM to be used prior
	 * to POST success.
	 */
	if (!fips_post_selftest_done()) {
	    FIPSerr(FIPS_F_FIPS_MODULE_MODE_SET,FIPS_R_SELFTEST_FAILED);
	    OpenSSLDie(__FILE__,__LINE__, "Attempting to use FIPS mode before POST has finished");	  
	    ret = 0;
	    goto end;
	}

	/* 
	 * Check if the POST has passed 
	 */
	if (fips_post_selftest_failed()) {
	    FIPSerr(FIPS_F_FIPS_MODULE_MODE_SET,FIPS_R_SELFTEST_FAILED);
	    OpenSSLDie(__FILE__,__LINE__, "Attempting to use FIPS mode after the POST has failed");	  
	    ret = 0;
	    goto end;
	}

    	if (!fips_check_auth(auth))
	    {
	    fips_auth_fail = 1;
	    FIPSerr(FIPS_F_FIPS_MODULE_MODE_SET,FIPS_R_AUTHENTICATION_FAILURE);
	    return 0;
	    }

	/* Don't go into FIPS mode twice, just so we can do automagic
	   seeding */
	if(FIPS_module_mode())
	    {
	    FIPSerr(FIPS_F_FIPS_MODULE_MODE_SET,FIPS_R_FIPS_MODE_ALREADY_SET);
	    ret = 0;
	    goto end;
	    }

#ifdef OPENSSL_IA32_SSE2
	{
	extern unsigned int OPENSSL_ia32cap_P[2];
	if ((OPENSSL_ia32cap_P[0] & (1<<25|1<<26)) != (1<<25|1<<26))
	    {
	    FIPSerr(FIPS_F_FIPS_MODULE_MODE_SET,FIPS_R_UNSUPPORTED_PLATFORM);
	    ret = 0;
	    goto end;
	    }
	OPENSSL_ia32cap_P[0] |= (1<<28);	/* set "shared cache"	*/
	OPENSSL_ia32cap_P[1] &= ~(1<<(60-32));	/* clear AVX		*/
	}
#endif

	if(fips_signature_witness() != FIPS_signature)
	    {
	    FIPSerr(FIPS_F_FIPS_MODULE_MODE_SET,FIPS_R_CONTRADICTING_EVIDENCE);
	    fips_post_set_selftest_failed();
	    ret = 0;
	    goto end;
	    }
	fips_set_mode(1);
    } else {
	fips_set_mode(0);
	/* exit CC mode as well if set */
	cc_set_mode(0);
    }
    ret = 1;
end:
    fips_clear_owning_thread();
    fips_w_unlock();
    return ret;
    }

int FIPS_module_cc_mode_set(int onoff)
    {
    int ret = 0;

    fips_w_lock();
    fips_set_owning_thread();

    if(onoff) {
        /* don't allow mode set if already set */
	if(FIPS_module_cc_mode())
	    {
	    FIPSerr(FIPS_F_FIPS_MODULE_CC_MODE_SET,FIPS_R_CC_MODE_ALREADY_SET);
	    ret = 0;
	    goto end;
	    }

	/* CC mode set only allowed when already in FIPS mode */
	if(!FIPS_module_mode())
	    {
	    FIPSerr(FIPS_F_FIPS_MODULE_CC_MODE_SET,FIPS_R_FIPS_MODE_NOT_SET);
	    ret = 0;
	    goto end;
	    }

	cc_set_mode(1);
    } else {
	cc_set_mode(0);
    }
    ret = 1;
end:
    fips_clear_owning_thread();
    fips_w_unlock();
    return ret;
    }

static CRYPTO_THREADID fips_thread;
static int fips_thread_set = 0;

static int fips_is_owning_thread(void)
{
	int ret = 0;

	CRYPTO_r_lock(CRYPTO_LOCK_FIPS2);
	if (fips_thread_set) {
			CRYPTO_THREADID cur;
			CRYPTO_THREADID_current(&cur);
			if (!CRYPTO_THREADID_cmp(&cur, &fips_thread))
				ret = 1;
		}
	CRYPTO_r_unlock(CRYPTO_LOCK_FIPS2);
	return ret;
}

int fips_set_owning_thread(void)
{
	int ret = 0;

	CRYPTO_w_lock(CRYPTO_LOCK_FIPS2);
	if (!fips_thread_set) {
		CRYPTO_THREADID_current(&fips_thread);
		ret = 1;
		fips_thread_set = 1;
		}
	CRYPTO_w_unlock(CRYPTO_LOCK_FIPS2);
	return ret;
}

int fips_clear_owning_thread(void)
{
	int ret = 0;

	CRYPTO_w_lock(CRYPTO_LOCK_FIPS2);
	if (fips_thread_set) {
		CRYPTO_THREADID cur;
		CRYPTO_THREADID_current(&cur);
		if (!CRYPTO_THREADID_cmp(&cur, &fips_thread))
			fips_thread_set = 0;
		}
	CRYPTO_w_unlock(CRYPTO_LOCK_FIPS2);
	return ret;
}

unsigned char *fips_signature_witness(void)
	{
	extern unsigned char FIPS_signature[];
	return FIPS_signature;
	}

unsigned long FIPS_module_version(void)
	{
	return FIPS_MODULE_VERSION_NUMBER;
	}

const char *FIPS_module_version_text(void)
	{
	return FIPS_MODULE_VERSION_TEXT;
	}

#if 0
/* The purpose of this is to ensure the error code exists and the function
 * name is to keep the error checking script quiet
 */
void hash_final(void)
	{
	FIPSerr(FIPS_F_HASH_FINAL,FIPS_R_NON_FIPS_METHOD);
	}
#endif


#endif
