/***********************license start***************
 * Copyright (c) 2003-2016  Cavium Inc. (support@cavium.com). All rights
 * reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.

 *   * Neither the name of Cavium Inc. nor the names of
 *     its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.

 * This Software, including technical data, may be subject to U.S. export  control
 * laws, including the U.S. Export Administration Act and its  associated
 * regulations, and may be subject to export or import  regulations in other
 * countries.

 * TO THE MAXIMUM EXTENT PERMITTED BY LAW, THE SOFTWARE IS PROVIDED "AS IS"
 * AND WITH ALL FAULTS AND CAVIUM INC. MAKES NO PROMISES, REPRESENTATIONS OR
 * WARRANTIES, EITHER EXPRESS, IMPLIED, STATUTORY, OR OTHERWISE, WITH RESPECT TO
 * THE SOFTWARE, INCLUDING ITS CONDITION, ITS CONFORMITY TO ANY REPRESENTATION OR
 * DESCRIPTION, OR THE EXISTENCE OF ANY LATENT OR PATENT DEFECTS, AND CAVIUM
 * SPECIFICALLY DISCLAIMS ALL IMPLIED (IF ANY) WARRANTIES OF TITLE,
 * MERCHANTABILITY, NONINFRINGEMENT, FITNESS FOR A PARTICULAR PURPOSE, LACK OF
 * VIRUSES, ACCURACY OR COMPLETENESS, QUIET ENJOYMENT, QUIET POSSESSION OR
 * CORRESPONDENCE TO DESCRIPTION. THE ENTIRE  RISK ARISING OUT OF USE OR
 * PERFORMANCE OF THE SOFTWARE LIES WITH YOU.
 ***********************license end**************************************/

#include <openssl/crypto.h>
#ifndef _MIPS_ARCH_OCTEON3
#include "ciscossl-cvmx.h"
#else
#include "cvmx.h"
#endif
#include "string.h"
#include <openssl/hmac.h>
#include <openssl/aes.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include "octeon_papi.h"
#ifdef OCTEON_SINGLE_PASS
#ifdef DEBUG
static char *cipher_strings[] = {
   [OCT_DES_64]       = "DES_64",
   [OCT_DES_40]       = "DES_40",
   [OCT_EXP_DES_64]   = "EXP_DES_64",
   [OCT_3DES]         = "3DES",
   [OCT_AES_128]      = "AES_128",
   [OCT_AES_256]      = "AES_256"
};
#endif

static uint8_t CRY_HSH_PAD1[48] = {
    0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36,
    0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36,
    0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36,
    0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36,
    0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36,
    0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36 } ;

static uint8_t CRY_HSH_PAD2[48] = {
    0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c,
    0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c,
    0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c,
    0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c,
    0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c,
    0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, 0x5c, } ;

static uint8_t CRY_HMAC_PAD1[48] = {
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0),
    (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), (0x36 ^ 0x0), } ;

static uint8_t CRY_HMAC_PAD2[48] = {
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0),
    (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), (0x5c ^ 0x0), } ;

static uint64_t *CRY_HSH_PAD1_64 = (uint64_t *)CRY_HSH_PAD1;
static uint32_t *CRY_HSH_PAD1_32 = (uint32_t *)CRY_HSH_PAD1;
static uint64_t *CRY_HMAC_PAD1_64 = (uint64_t *)CRY_HMAC_PAD1;
static uint64_t *CRY_HSH_PAD2_64 = (uint64_t *)CRY_HSH_PAD2;
static uint32_t *CRY_HSH_PAD2_32 = (uint32_t *)CRY_HSH_PAD2;
static uint64_t *CRY_HMAC_PAD2_64 = (uint64_t *)CRY_HMAC_PAD2;

#if 0
static uint16_t *CRY_HSH_PAD1_16 = (uint16_t *)CRY_HSH_PAD1;
static uint32_t *CRY_HMAC_PAD1_32 = (uint32_t *)CRY_HMAC_PAD1;
static uint16_t *CRY_HMAC_PAD1_16 = (uint16_t *)CRY_HMAC_PAD1;
static uint16_t *CRY_HSH_PAD2_16 = (uint16_t *)CRY_HSH_PAD2;
static uint32_t *CRY_HMAC_PAD2_32 = (uint32_t *)CRY_HMAC_PAD2;
static uint16_t *CRY_HMAC_PAD2_16 = (uint16_t *)CRY_HMAC_PAD2;
#endif

/*Start of Common functions*/
static inline uint64_t swap64 (uint64_t v)
{
  return ((v >> 56) |
    (((v >> 48) & 0xfful) << 8) |
    (((v >> 40) & 0xfful) << 16) |
    (((v >> 32) & 0xfful) << 24) |
    (((v >> 24) & 0xfful) << 32) |
    (((v >> 16) & 0xfful) << 40) |
    (((v >> 8) & 0xfful) << 48) | (((v >> 0) & 0xfful) << 56));
}

#ifdef DEBUG
void _hexdump(unsigned char *buffer, int len)
{
    int i;

	if (!buffer)
		return;

    for(i=0;i<len;i++)
    {
        cavium_dbgprint("%02X ", buffer[i]);
        if (i && ((i%8) == 7))
                cavium_dbgprint("\n");
    }

    cavium_dbgprint("\n");
}
#endif

typedef struct iovec_record_process
{
	uint16_t hash_req_bytes;
	uint16_t crypto_itr;
	uint32_t hash_size;
	uint16_t hash_itr;
	uint16_t hash_rem_bytes;
} iovec_record_process_t;

/*******************************************************************************
 *                                                                             *
 *      Decrypts block of SHA1_CBLOCK (64) bytes of received encrypted record  *
 *  and feeds the decrypted block to digest engine.                            * 
 *  Protocol : SSL-3.0                                                         *
 *  Cipher   : AES-128/ AES-256                                                *
 *  Digest   : SHA1                                                            *
 *                                                                             *
 * Input:-                                                                     *
 *                                                                             *
 *  in_chunk_cur  : Pointer to start of encrypted payload                      *
 *  out_chunk_cur : Pointer to store decrypted data                            *
 *  hash_bal      : Initial 8 byte block to be to fed to digest engine         *
 *  mode          : Mode of Operation (DEFAULT/ENCRYPTION/HMAC/DIGEST)         *
 *                                                                             *
 * Return:                                                                     *
 *   Excess data (8bytes)   decrypted unused in DIGEST generation.             *
 *                                                                             *
 *******************************************************************************/

static
inline uint64_t block_processing_aes_sha1_dec_ssl30(uint8_t *in_chunk_cur, uint8_t * out_chunk_cur, uint64_t hash_bal, int mode)
{
	register uint64_t h0, h1, h2, h3;
	register uint64_t ic0, ic1, ic2, ic3;
	register uint64_t oc0, oc1, oc2, oc3;

	h0 = hash_bal;

	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 0);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 8);

	CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 16);
	CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 24);

        if(!(mode & 0x02)) {
	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic0));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic1));

	    CVMX_MF_AES_RESULT(oc0, 0);
	    CVMX_MF_AES_RESULT(oc1, 1);

	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic2));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic3));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 0);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 8);
        }
    if( mode == EX_MODE_DEFAULT) {
	    h0 = h0 | (oc0 >> 56);
	    h1 = ((oc0) << 8) | (oc1 >> 56);
	    h2 = (oc1) << 8;
	    CVMX_MT_HSH_DAT(h0, 0);
	    CVMX_MT_HSH_DAT(h1, 1);
    } else {
        CVMX_MT_HSH_DAT(h0, 0);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic0), 1);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic1), 2);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic2), 3);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic3), 4);
    }


#ifdef DEBUG
	cavium_dbgprint("Block Start\n");
	cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", h0, h1); 
	cavium_dbgprint("Decrypted Output : %016llx \t %016llx\n", oc0, oc1);
#endif

	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 32);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 40);

        if(!(mode & 0x02)) {
	    CVMX_MF_AES_RESULT(oc2, 0);
	    CVMX_MF_AES_RESULT(oc3, 1);

	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic0));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic1));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 16);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 24);

	    h2 = h2 | (oc2 >> 56);
	    h3 = ((oc2)<< 8) | (oc3 >> 56);
	    h0 = (oc3) << 8;
        }

	CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 48);
	CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 56);

    if(mode == EX_MODE_DEFAULT) {
	    CVMX_MT_HSH_DAT(h2, 2);
	    CVMX_MT_HSH_DAT(h3, 3);
    } else {
	    CVMX_MT_HSH_DAT(cvmx_htobe64(ic0), 5);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(ic1), 6);
    }

#ifdef DEBUG
	cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", h2, h3); 
	cavium_dbgprint("Decrypted Output : %016llx \t %016llx\n", oc2, oc3);
#endif
     if(!(mode & 0x02)) {
	    CVMX_MF_AES_RESULT(oc0, 0);
	    CVMX_MF_AES_RESULT(oc1, 1);

	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic2));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic3));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 32);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 40);

	    h0 = h0 | (oc0 >> 56);
	    h1 = ((oc0) << 8) | (oc1 >> 56);
	    h2 = (oc1) << 8;
    }

    if(mode == EX_MODE_DEFAULT) { 
	CVMX_MT_HSH_DAT(h0, 4);
	CVMX_MT_HSH_DAT(h1, 5);
    } else if(mode & 0x02) {
        CVMX_MT_HSH_STARTSHA(ic2);
        return ic3;
    }

#ifdef DEBUG
	cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", h0, h1); 
	cavium_dbgprint("Decrypted Output : %016llx \t %016llx\n", oc0, oc1);
#endif
	CVMX_MF_AES_RESULT(oc2, 0);
	CVMX_MF_AES_RESULT(oc3, 1);

	h2 = h2 | (oc2 >> 56);
	h3 = (((oc2) << 8)) | (oc3 >> 56);
	h0 = (oc3) << 8;

	CVMX_MT_HSH_DAT(h2, 6);
	CVMX_MT_HSH_STARTSHA(h3);

	CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 48);
	CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 56);

#ifdef DEBUG
	cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", h2, h3); 
	cavium_dbgprint("Decrypted Output : %016llx \t %016llx \n", oc2, oc3);
	cavium_dbgprint("Block End\n");
#endif

	return h0;
}

/*******************************************************************************
 *                                                                             *
 *      Decrypts block of SHA1_CBLOCK (64) bytes of received encrypted record  *
 *  and feeds the decrypted block to digest engine.                            * 
 *  Protocol : SSL-3.0                                                         *
 *  Cipher   : DES-CBC / DES3-CBC                                              *
 *  Digest   : SHA1                                                            *
 *                                                                             *
 * Input:-                                                                     *
 *                                                                             *
 *  in_chunk_cur  : Pointer to start of encrypted payload                      *
 *  out_chunk_cur : Pointer to store decrypted data                            *
 *  hash_bal      : Initial 8 byte block to be to fed to digest engine         *
 *  mode          : Mode of Operation (DEFAULT/ENCRYPTION/HMAC/DIGEST)         *
 *                                                                             *
 * Return:                                                                     *
 *   Excess data (8bytes)   decrypted unused in DIGEST generation.             *
 *                                                                             *
 *******************************************************************************/

static
inline uint64_t block_processing_des_sha1_dec_ssl30(uint8_t *in_chunk_cur, uint8_t * out_chunk_cur, uint64_t hash_bal,uint8_t mode)
{
	register uint64_t h0, h1, h2, h3;
	register uint64_t ic0, ic1, ic2, ic3;
	register uint64_t oc0, oc1, oc2, oc3;

	h0 = hash_bal;

	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 0);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 8);

	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic0));

	CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 16);
	CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 24);

    if(mode == EX_MODE_DEFAULT || mode == EX_MODE_ENCRYPT) {
	    CVMX_MF_3DES_RESULT(oc0); // 1

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic1));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 0);
	    h0 = h0 | (oc0 >> 56);

	    CVMX_MF_3DES_RESULT(oc1); // 2

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic2));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 8);
	    h1 = ((oc0) << 8) | (oc1 >> 56);

	    CVMX_MF_3DES_RESULT(oc2); // 3

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic3));

	    h2 = (oc1) << 8;

	    CVMX_MT_HSH_DAT(h0, 0);
	    CVMX_MT_HSH_DAT(h1, 1);
    } else {
        CVMX_MT_HSH_DAT(h0, 0);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic0), 1);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic1), 2);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic2), 3);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic3), 4);
    }
#ifdef DEBUG
	cavium_dbgprint("Block Start\n");
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h0, h1, __LINE__); 
#endif

	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 32);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 40);

    if(mode == EX_MODE_ENCRYPT || mode == EX_MODE_DEFAULT) {
	CVMX_MF_3DES_RESULT(oc3); // 4

	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic0));

	CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 16);
	CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 24);

	h2 = h2 | (oc2 >> 56);
	h3 = ((oc2)<< 8) | (oc3 >> 56);

	CVMX_MF_3DES_RESULT(oc0); // 5

	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic1));


	h0 = (oc3) << 8;
    } else {
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic0), 5);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic1), 6);
    }

	CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 48);
	CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 56);

    if(mode == EX_MODE_DEFAULT || mode == EX_MODE_ENCRYPT) {
	    CVMX_MT_HSH_DAT(h2, 2);
	    CVMX_MT_HSH_DAT(h3, 3);
#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h2, h3, __LINE__); 
#endif
	    CVMX_MF_3DES_RESULT(oc1); // 6

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic2));

	    h0 = h0 | (oc0 >> 56);
	    h1 = ((oc0) << 8) | (oc1 >> 56);

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 32);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 40);

	    h2 = (oc1) << 8;

	    CVMX_MF_3DES_RESULT(oc2); // 7

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic3));

	    CVMX_MT_HSH_DAT(h0, 4);
	    CVMX_MT_HSH_DAT(h1, 5);
    } else {
        CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ic2));
        hash_bal = cvmx_htobe64(ic3);
        return hash_bal;
    }


#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h0, h1, __LINE__); 
#endif
	CVMX_MF_3DES_RESULT(oc3);

	h2 = h2 | (oc2 >> 56);
	h3 = (((oc2) << 8)) | (oc3 >> 56);
	h0 = (oc3) << 8;

	CVMX_MT_HSH_DAT(h2, 6);
	CVMX_MT_HSH_STARTSHA(h3);

	CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 48);
	CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 56);

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h2, h3, __LINE__); 
	cavium_dbgprint("Block End\n");
#endif

	return h0;
}

/*******************************************************************************
 *                                                                             *
 *      Decrypts block of SHA1_CBLOCK (64) bytes of received encrypted record  *
 *  and feeds the decrypted block to digest engine.                            * 
 *  Protocol : TLS-1.0 / TLS-1.1 / TLS-1.2                                     *
 *  Cipher   : AES-128/ AES-256                                                *
 *  Digest   : SHA1                                                            *
 *                                                                             *
 * Input:-                                                                     *
 *                                                                             *
 *  in_chunk_cur  : Pointer to start of encrypted payload                      *
 *  out_chunk_cur : Pointer to store decrypted data                            *
 *  hash_bal      : Initial 8 byte block to be to fed to digest engine         *
 *  mode          : Mode of Operation (DEFAULT/ENCRYPTION/HMAC/DIGEST)         *
 *                                                                             *
 * Return:                                                                     *
 *   Excess data (8bytes)   decrypted unused in DIGEST generation.             *
 *                                                                             *
 *******************************************************************************/

static
inline uint8_t* block_processing_aes_sha1_dec_tls10(uint8_t *in_chunk_cur, uint8_t * out_chunk_cur, uint8_t * hash_bal,uint8_t mode)
{
	register uint64_t h0, h1, h2 =0, h3=0;
	register uint64_t ic0, ic1, ic2, ic3;
	register uint64_t oc0, oc1, oc2, oc3;

	CVMX_LOADUNA_INT64(h0, hash_bal, 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(h0), 0);

#ifdef DEBUG
	cavium_dbgprint("Block Start\n");
	cavium_dbgprint("sha inp bytes : %016llx (%d)\n", h0, __LINE__); 
#endif
	CVMX_LOADUNA_INT64(h0, hash_bal, 8); // we need only 5 bytes
    h0 = cvmx_htobe64(h0);
    if(mode == EX_MODE_DEFAULT) 
	    h0 = h0 & 0xffffffffff000000;
        
	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 0);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 8);

	CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 16);
	CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 24);

    if(mode == EX_MODE_DEFAULT || mode == EX_MODE_ENCRYPT) {
	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic0));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic1));

	    CVMX_MF_AES_RESULT(oc0, 0);
	    CVMX_MF_AES_RESULT(oc1, 1);

	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic2));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic3));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 0);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 8);

	    h0 = h0 | (oc0 >> 40);
	    h1 = ((oc0) << 24) | (oc1 >> 40);
	    h2 = (oc1) << 24;

	    CVMX_MT_HSH_DAT(h0, 1);
	    CVMX_MT_HSH_DAT(h1, 2);
    } else {
        CVMX_MT_HSH_DAT(h0, 1);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic0), 2);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic1), 3);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic2), 4);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic3), 5);
    }
#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h0, h1, __LINE__); 
#endif

	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 32);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 40);

    if(mode == EX_MODE_DEFAULT || mode == EX_MODE_ENCRYPT) {
	    CVMX_MF_AES_RESULT(oc2, 0);
	    CVMX_MF_AES_RESULT(oc3, 1);

	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic0));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic1));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 16);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 24);

	    h2 = h2 | (oc2 >> 40);
	    h3 = ((oc2)<< 24) | (oc3 >> 40);
	    h0 = (oc3) << 24;
    }
	CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 48);
	CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 56);

    if(mode & 0x02) {
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic0), 6);
        CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ic1));
        CVMX_STOREUNA_INT64(ic2, hash_bal, 0);
        CVMX_STOREUNA_INT64(ic3, hash_bal, 8);
        return hash_bal;
    }


	CVMX_MT_HSH_DAT(h2, 3);
	CVMX_MT_HSH_DAT(h3, 4);

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h2, h3, __LINE__); 
#endif
	CVMX_MF_AES_RESULT(oc0, 0);
	CVMX_MF_AES_RESULT(oc1, 1);

	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic2));
	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic3));

	CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 32);
	CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 40);

	h0 = h0 | (oc0 >> 40);
	h1 = ((oc0) << 24) | (oc1 >> 40);
	h2 = (oc1) << 24;

	CVMX_MT_HSH_DAT(h0, 5);
	CVMX_MT_HSH_DAT(h1, 6);

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h0, h1, __LINE__); 
#endif
	CVMX_MF_AES_RESULT(oc2, 0);
	CVMX_MF_AES_RESULT(oc3, 1);

	h2 = h2 | (oc2 >> 40);
	h3 = (((oc2) << 24)) | (oc3 >> 40);
	h0 = (oc3) << 24;

	CVMX_MT_HSH_STARTSHA(h2);

	CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 48);
	CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 56);

	CVMX_STOREUNA_INT64(cvmx_htobe64(h3), hash_bal, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(h0), hash_bal, 8);

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx (%d)\n", h2, __LINE__); 
	cavium_dbgprint("Block End\n");
#endif

	return hash_bal;
}



/*******************************************************************************
 *                                                                             *
 *      Decrypts block of SHA1_CBLOCK (64) bytes of received encrypted record  *
 *  and feeds the decrypted block to digest engine.                            * 
 *  Protocol : TLS-1.0 / TLS-1.1 / TLS-1.2                                     *
 *  Cipher   : DES-CBC / DES3-CBC                                              *
 *  Digest   : SHA1                                                            *
 *                                                                             *
 * Input:-                                                                     *
 *                                                                             *
 *  in_chunk_cur  : Pointer to start of encrypted payload                      *
 *  out_chunk_cur : Pointer to store decrypted data                            *
 *  hash_bal      : Initial 8 byte block to be to fed to digest engine         *
 *  mode          : Mode of Operation (DEFAULT/ENCRYPTION/HMAC/DIGEST)         *
 *                                                                             *
 * Return:                                                                     *
 *   Excess data (8bytes)   decrypted unused in DIGEST generation.             *
 *                                                                             *
 *******************************************************************************/

static
inline uint8_t * block_processing_des_sha1_dec_tls10(uint8_t *in_chunk_cur, uint8_t * out_chunk_cur, uint8_t *hash_bal, uint8_t mode)
{
	register uint64_t h0, h1, h2, h3;
	register uint64_t ic0, ic1, ic2, ic3;
	register uint64_t oc0, oc1, oc2, oc3;

	CVMX_LOADUNA_INT64(h0, hash_bal, 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(h0), 0);

#ifdef DEBUG
	cavium_dbgprint("Block Start\n");
	cavium_dbgprint("sha inp bytes : %016llx(%d)\n", h0, __LINE__); 
#endif

	CVMX_LOADUNA_INT64(h0, hash_bal, 8); // we need only 5 bytes
    h0 = cvmx_htobe64(h0);
    if(mode == EX_MODE_DEFAULT)
    	h0 = h0 & 0xffffffffff000000;

	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 0);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 8);

    if(mode == EX_MODE_DEFAULT || mode == EX_MODE_ENCRYPT) 
	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic0));

	    CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 16);
	    CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 24);

    if(mode == EX_MODE_DEFAULT || mode == EX_MODE_ENCRYPT) {
	    CVMX_MF_3DES_RESULT(oc0); // 1

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic1));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 0);
	    h0 = h0 | (oc0 >> 40);

	    CVMX_MF_3DES_RESULT(oc1); // 2

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic2));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 8);
	    h1 = ((oc0) << 24) | (oc1 >> 40);

	    CVMX_MF_3DES_RESULT(oc2); // 3

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic3));

	    h2 = (oc1) << 24;

	    CVMX_MT_HSH_DAT(h0, 1);
	    CVMX_MT_HSH_DAT(h1, 2);
    } else if(mode & 0x02) { 
        CVMX_MT_HSH_DAT(h0, 1);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic0), 2);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic1), 3);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic2), 4);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic3), 5);
    }

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h0, h1, __LINE__); 
#endif

	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 32);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 40);

    if(mode == EX_MODE_DEFAULT || mode == EX_MODE_ENCRYPT) {
	    CVMX_MF_3DES_RESULT(oc3); // 4

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic0));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 16);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 24);

	    h2 = h2 | (oc2 >> 40);
	    h3 = ((oc2)<< 24) | (oc3 >> 40);

	    CVMX_MF_3DES_RESULT(oc0); // 5

	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic1));


	    h0 = (oc3) << 24;
    }

	    CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 48);
	    CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 56);
    if(mode == EX_MODE_DEFAULT) {
	    CVMX_MT_HSH_DAT(h2, 3);
	    CVMX_MT_HSH_DAT(h3, 4);
    } else {
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic0), 6);
        CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ic1));
        CVMX_STOREUNA_INT64(ic2, hash_bal, 0);
        CVMX_STOREUNA_INT64(ic3, hash_bal, 8);
        return hash_bal;
    }

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h2, h3, __LINE__); 
#endif
	CVMX_MF_3DES_RESULT(oc1); // 6

	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic2));

	h0 = h0 | (oc0 >> 40);
	h1 = ((oc0) << 24) | (oc1 >> 40);

	CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 32);
	CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 40);

	h2 = (oc1) << 24;

	CVMX_MF_3DES_RESULT(oc2); // 7

	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(ic3));

	CVMX_MT_HSH_DAT(h0, 5);
	CVMX_MT_HSH_DAT(h1, 6);

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes: %016llx %016llx (%d)\n", h0, h1, __LINE__); 
#endif
	CVMX_MF_3DES_RESULT(oc3);

	h2 = h2 | (oc2 >> 40);
	h3 = (((oc2) << 24)) | (oc3 >> 40);
	h0 = (oc3) << 24;

	CVMX_MT_HSH_STARTSHA(h2);

	CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 48);
	CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 56);

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes: %016llx(%d)\n", h2, __LINE__); 
	cavium_dbgprint("Block End\n");
#endif

	CVMX_STOREUNA_INT64(cvmx_htobe64(h3), hash_bal, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(h0), hash_bal, 8);

	return hash_bal;
}

/*******************************************************************************
 *                                                                             *
 *      Decrypts block of SHA256_CBLOCK (64) bytes of received encrypted       *
 *  record and feeds the decrypted block to digest engine.                     * 
 *  Protocol : TLS-1.2                                                         *
 *  Cipher   : AES-128 / AES-256                                               *
 *  Digest   : SHA256                                                          *
 *                                                                             *
 * Input:-                                                                     *
 *                                                                             *
 *  in_chunk_cur  : Pointer to start of encrypted payload                      *
 *  out_chunk_cur : Pointer to store decrypted data                            *
 *  hash_bal      : Initial 8 byte block to be to fed to digest engine         *
 *  mode          : Mode of Operation (DEFAULT/ENCRYPTION/HMAC/DIGEST)         *
 *                                                                             *
 * Return:                                                                     *
 *   Excess data (8bytes)   decrypted unused in DIGEST generation.             *
 *                                                                             *
 *******************************************************************************/

static
inline uint8_t* block_processing_aes_sha256_dec_tls12(uint8_t *in_chunk_cur, uint8_t * out_chunk_cur, uint8_t * hash_bal,
        uint8_t mode)
{
	register uint64_t h0, h1, h2, h3;
	register uint64_t ic0, ic1, ic2, ic3;
	register uint64_t oc0, oc1, oc2, oc3;

	CVMX_LOADUNA_INT64(h0, hash_bal, 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(h0), 0);
#ifdef DEBUG
	cavium_dbgprint("Block Start\n");
	cavium_dbgprint("sha inp bytes : %016llx (%d)\n", h0, __LINE__); 
#endif
	CVMX_LOADUNA_INT64(h0, hash_bal, 8); // we need only 5 bytes
    if(mode == EX_MODE_DEFAULT) {
        h0 = cvmx_htobe64(h0);
	    h0 = h0 & 0xffffffffff000000;
    } else 
        CVMX_MT_HSH_DAT(cvmx_htobe64(h0),1);
	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 0);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 8);

	CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 16);
	CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 24);

    if(mode == EX_MODE_DEFAULT || mode == EX_MODE_ENCRYPT) {
	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic0));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic1));

	    CVMX_MF_AES_RESULT(oc0, 0);
	    CVMX_MF_AES_RESULT(oc1, 1);

	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic2));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic3));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 0);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 8);

	    h0 = h0 | (oc0 >> 40);
	    h1 = ((oc0) << 24) | (oc1 >> 40);
	    h2 = (oc1) << 24;
	    CVMX_MT_HSH_DAT(h0, 1);
	    CVMX_MT_HSH_DAT(h1, 2);
    } else {
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic0),2);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic1),3);
    }

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h0, h1, __LINE__); 
#endif

	CVMX_LOADUNA_INT64(ic0, in_chunk_cur, 32);
	CVMX_LOADUNA_INT64(ic1, in_chunk_cur, 40);

    if (mode == EX_MODE_DEFAULT || mode == EX_MODE_ENCRYPT) {
	    CVMX_MF_AES_RESULT(oc2, 0);
	    CVMX_MF_AES_RESULT(oc3, 1);

	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic0));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic1));

	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 16);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 24);

	    h2 = h2 | (oc2 >> 40);
	    h3 = ((oc2)<< 24) | (oc3 >> 40);
	    h0 = (oc3) << 24;
    } else {
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic2),4);
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic3),5);
    }
	CVMX_LOADUNA_INT64(ic2, in_chunk_cur, 48);
	CVMX_LOADUNA_INT64(ic3, in_chunk_cur, 56);

    if(mode == EX_MODE_DEFAULT) {
	    CVMX_MT_HSH_DAT(h2, 3);
	    CVMX_MT_HSH_DAT(h3, 4);
    } else {
        CVMX_MT_HSH_DAT(cvmx_htobe64(ic0),6);
        CVMX_MT_HSH_STARTSHA256(cvmx_htobe64(ic1));
    }

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h2, h3, __LINE__); 
#endif
	CVMX_MF_AES_RESULT(oc0, 0);
	CVMX_MF_AES_RESULT(oc1, 1);

	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(ic2));
	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(ic3));

	CVMX_STOREUNA_INT64(cvmx_htobe64(oc0), out_chunk_cur, 32);
	CVMX_STOREUNA_INT64(cvmx_htobe64(oc1), out_chunk_cur, 40);

	h0 = h0 | (oc0 >> 40);
	h1 = ((oc0) << 24) | (oc1 >> 40);
	h2 = (oc1) << 24;

    if(mode == EX_MODE_DEFAULT) {
	    CVMX_MT_HSH_DAT(h0, 5);
	    CVMX_MT_HSH_DAT(h1, 6);
    }

#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx %016llx (%d)\n", h0, h1, __LINE__); 
#endif
	CVMX_MF_AES_RESULT(oc2, 0);
	CVMX_MF_AES_RESULT(oc3, 1);

	h2 = h2 | (oc2 >> 40);
	h3 = (((oc2) << 24)) | (oc3 >> 40);
	h0 = (oc3) << 24;

    if(mode == EX_MODE_DEFAULT) {
    	CVMX_MT_HSH_STARTSHA256(h2);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(h3), hash_bal, 0);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(h0), hash_bal, 8);
    } else {
        CVMX_STOREUNA_INT64(ic2,hash_bal, 0);
        CVMX_STOREUNA_INT64(ic3, hash_bal, 8);
    }

	CVMX_STOREUNA_INT64(cvmx_htobe64(oc2), out_chunk_cur, 48);
	CVMX_STOREUNA_INT64(cvmx_htobe64(oc3), out_chunk_cur, 56);


#ifdef DEBUG
	cavium_dbgprint("sha inp bytes : %016llx (%d)\n", h2, __LINE__); 
	cavium_dbgprint("Block End\n");
#endif

	return hash_bal;
}



/*Start of encrypt functions*/

/*******************************************************************************
 *     Common stages of Operation                                              *
 * Encryption: -                                                               *
 *  i) Length of data to be encrypted is greater than or equal to DIGEST block *
 *     length, SHA_CBLOCK/SHA256_CBLOCK                                        *
 *           a) Feed digest block size of plain data as multiple 8 byte words  *
 *              to cipher encrypt engine.                                      *
 *           b) Feed BLOCK size of data to digest engine and generate partial  *
 *              hash.                                                          *
 * ii) Lenght of data to be encrypted is lesser than DIGEST block length       *
 *           a) Encrypt remaining plain data.                                  *
 *           b) Feed remaining data to digest engine.                          *
 *           c) Generate inner hash and feed to generate outer hash.           *
 *           d) Calculate padding length and add padding bytes.                *
 *           e) Encrypt HMAC+Padding bytes and append to encrypted plain data. *
 *                                                                             *
 * Decryption:-                                                                *
 *  i) Decrypt last block of data to find padding length                       *
 * ii) Calculate plain data length to generate TLS header                      *
 *        Encrypted data length - MAC size - Pad length                        *
 *iii) Length of plain data to be decrypted is more than DIGEST block length   *
 *           a) Decrypt input payload as 8 byte block and feed to digest       *
 *              engine.                                                        *
 * iv) Length of plain data to be decrypted is less than digest block length   *
 *           a) Decrypt remaining plain data and feed to digest engine         *
 *           b) Decrypt remaining data ( MAC + Padding)                        *
 *           c) Generate inner hash and feed to calculate outer hash           *
 *           d) Compare decrypted HMAC with generated outer HMAC               *
 *                                                                             *
 *******************************************************************************/



/*******************************************************************************
 *                                                                             *
 *      Generates SSL-3.0 SHA1 HMAC, adds padding bytes and  encrypts payload  *
 *  data and stores the result in output iovec buffer. Initialisation vector   *
 *  is updated after each encryption.Initialisation vector to be used for      *
 *  encrytion is taken from EVP_AESOCT_HMAC_SHA1 structure passed to the       *
 *  function. Key and MAC secret to be used for cipher encryption and HMAC     *
 *  generation respectively are also taken from the key structure passed.      *
 *  Sequence number and header components, record type and length of record,   *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the SSL-3.0 record header.          *
 *  Protocol : SSl-3.0                                                         *
 *  Cipher   : AES128-CBC / AES256-CBC                                         *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing encryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to encrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store encryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Lenght of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/
int32_t oct_aes_sha1_enc_ssl30(EVP_AESOCT_HMAC_SHA1 *key,
		struct iovec_cav *inp_vec, uint16_t inp_cnt,
		struct iovec_cav *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t p0, p1, p2, p3, p4, p5;
	uint8_t hash_buf[128] = {0}, enc_buf[128] = {0}, final_buf[128] = {0};
	uint8_t idx = 0, cipher_type = 0;
	uint16_t inp_len = 0;
	int32_t enc_rem_num = 0, hash_rem_num = 0, enc_copied_bytes = 0, hash_copied_bytes = 0, eout_space_left = 0;
	int32_t keylength = 1, fill_bytes = 0, padlen = 0, next_output_iovec = 0, start_inner_hash = 0;
	uint64_t tmpvar = 0, tmpvar2 = 0, tmpvar3 = 0, tmpvar4 = 0, seq = 0, total_bits = 0;
	uint8_t *mac_secret = NULL, *hash_in = NULL, *hash_end = NULL, *enc_in = NULL, *enc_end = NULL, *enc_out = NULL;
	uint64_t *enc_key = NULL, *ptr = NULL, *enc_iv = NULL;
    uint16_t *hash_buf_ptr_16 = NULL;
    uint32_t *hash_buf_ptr_32 = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct ssl3_rec_hdr *in_rec_hdr = NULL;


	in_rec_hdr = (struct ssl3_rec_hdr *)&key->aux.tls_aad_8[8];
	inp_len = cvmx_htobe16(in_rec_hdr->length);

	enc_key = (uint64_t *)key->cipher_key;
    enc_iv  = (uint64_t *)key->cipher_iv; 
	mac_secret = key->mac_secret; 
	seq = cvmx_htobe64(key->aux.tls_aad_64[0]);

	cipher_type = key->cipher_type;
	switch(cipher_type)
	{
		case OCT_AES_128:
			keylength = 1;
			break;

		case OCT_AES_256:
			keylength = 3;
			break;
		default:
			break;
	}

#ifdef DEBUG
	cavium_dbgprint("rec length: %d, seq: %llx\n", in_rec_hdr->length, seq);
	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
#endif

    if(key->ex_mode == EX_MODE_DIGEST)
        total_bits = inp_len * 8;
    else if (key->ex_mode == EX_MODE_HMAC)
        total_bits = (HMAC_SHA1_KEY_LEN + 40 + inp_len)*8;
    else
	    total_bits = (HMAC_SHA1_KEY_LEN + 40 + 8 + 1 + 2 + inp_len) * 8; /* mac_secret + pad1 + seq + rec_type + rec_len */ 
	tmpvar2 = (uint64_t) cvmx_htobe32(*((uint32_t *)&mac_secret[16]));
	tmpvar2 = (tmpvar2 << 32) | CRY_HSH_PAD1_32[0];
    if(key->ex_mode == EX_MODE_DEFAULT)
	    tmpvar3 = (CRY_HSH_PAD1_64[0] & 0xFFFFFFFF00000000) | upper_32_bits(seq);
    else if(key->ex_mode == EX_MODE_HMAC) {
            if(inp_len >= 4) {
                tmpvar3 = (CRY_HSH_PAD1_64[0] & 0xFFFFFFFF00000000) |cvmx_htobe32(*((uint32_t *)inp_vec[0].ptr));
                inp_vec[0].ptr += 4;
                inp_vec[0].size -=4;
                inp_len -=4;
            } else {
                tmpvar3 = (CRY_HSH_PAD1_64[0] & 0xFFFFFFFF00000000);
                memcpy(((uint8_t *)&tmpvar3)+4, inp_vec[0].ptr, inp_len);
                inp_vec[0].ptr += inp_len;
                inp_vec[0].size -= inp_len;
                *(((uint8_t *)&tmpvar3)+ 4 + inp_len) = 0x80;
                inp_len -=0;
            }
        }


	/* assume AES keylen 128 bit */
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[1]), 1);
	if(keylength == 3)
	{
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[2]), 2);
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[3]), 3);
	}
	else
	{
		CVMX_MT_AES_KEY((uint64_t)0, 2);
		CVMX_MT_AES_KEY((uint64_t)0, 3);
	}
	CVMX_MT_AES_KEYLENGTH(keylength);

	/* Init AES iv */
	CVMX_MT_AES_IV(cvmx_htobe64(enc_iv[0]), 0);
	CVMX_MT_AES_IV(cvmx_htobe64(enc_iv[1]), 1);

	/* The first hash (mac secret (20 bytes) + pad_1 (40 bytes) + upper 4 bytes of sequence number */
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT0), 0);
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT1), 1);
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT2), 2);
    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_MT_HSH_DAT(cvmx_htobe64(((uint64_t *) mac_secret)[0]), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(((uint64_t *) mac_secret)[1]), 1);
	    CVMX_MT_HSH_DAT(tmpvar2, 2);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(CRY_HSH_PAD1_64[0]), 3);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(CRY_HSH_PAD1_64[0]), 4);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(CRY_HSH_PAD1_64[0]), 5);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(CRY_HSH_PAD1_64[0]), 6);
	    CVMX_MT_HSH_STARTSHA(tmpvar3);
    }

	CVMX_PREFETCH0(inp_vec[idx].ptr);


    if(key->ex_mode == EX_MODE_DEFAULT) {
	    /* store lower 4 bytes of sequence number + rec_type + rec_len */
        hash_buf_ptr_32 = (uint32_t *)hash_buf;
	    hash_buf_ptr_32[0] = cvmx_htobe32(lower_32_bits(seq));
	    hash_buf[4] = in_rec_hdr->type;
        hash_buf_ptr_16 = (uint16_t *)&hash_buf[5];
	    hash_buf_ptr_16[0] = cvmx_htobe16(inp_len);
	    hash_copied_bytes = 7; /* lower 32 bit of seq num (4 byte) + record type (1 byte) + recore length (2 bytes) */
    }
    else if(key->ex_mode & 0x02){
        if(inp_vec[0].size > 8) {
            memcpy(hash_buf,inp_vec[idx].ptr,8);
            hash_copied_bytes = 8;
            inp_vec[idx].ptr += 8;
            inp_vec[idx].size -= 8;
        } else {
            memcpy(hash_buf,inp_vec[idx].ptr,inp_vec[0].size);
            hash_copied_bytes = inp_vec[0].size;
            inp_vec[idx].ptr += inp_vec[0].size;
            inp_vec[idx].size -= inp_vec[0].size;
        }
    }
	enc_copied_bytes = 0;

	hash_in = inp_vec[idx].ptr ;
	hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
	hash_rem_num = (hash_end - hash_in) + hash_copied_bytes;

	enc_in = inp_vec[idx].ptr ;
	enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
	enc_rem_num = (enc_end - enc_in);

	enc_out = outp_vec[idx].ptr ;

	if (hash_rem_num <= SHA_CBLOCK || key->ex_mode == EX_MODE_ENCRYPT) {
		ptr = (uint64_t *)(hash_buf + hash_copied_bytes);	
		CVMX_STOREUNA_INT64(0, ptr, 0);
		CVMX_STOREUNA_INT64(0, ptr, 8);
		CVMX_STOREUNA_INT64(0, ptr, 16);
		CVMX_STOREUNA_INT64(0, ptr, 24);
		CVMX_STOREUNA_INT64(0, ptr, 32);
		CVMX_STOREUNA_INT64(0, ptr, 40);
		CVMX_STOREUNA_INT64(0, ptr, 48);
		CVMX_STOREUNA_INT64(0, ptr, 56);

		memcpy(hash_buf + hash_copied_bytes, hash_in, inp_vec[idx].size);	
		hash_copied_bytes += (inp_vec[idx].size);
		hash_buf[hash_copied_bytes] = 0x80;
#ifdef DEBUG
		cavium_dbgprint("input hash data\n");
		_hexdump(hash_buf, 64);
#endif
        if(key->ex_mode != EX_MODE_ENCRYPT)
    		SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
		hash_copied_bytes = 0;
		hash_rem_num = 0;

		while (enc_rem_num >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	
			CVMX_LOADUNA_INT64(p1, enc_in, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			enc_in += AES_BLOCK_SIZE;
			enc_rem_num -= AES_BLOCK_SIZE;

			CVMX_MF_AES_RESULT(p0, 0);	
			CVMX_MF_AES_RESULT(p1, 1);	

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

			enc_out += AES_BLOCK_SIZE;
		}

		memcpy(enc_buf, enc_in, enc_rem_num);
		enc_copied_bytes = enc_rem_num;

		goto get_inner_hash;
	}

	/* inp_vec[0] has payload(inp_vec[0].size - sizeof(struct ssl3_rec_hdr)) > 57 bytes */
    if(key->ex_mode & 0x02)
	    ptr = (uint64_t *)(hash_in);
    else {
	    ptr = (uint64_t *)(hash_in + 1);
	    hash_buf[hash_copied_bytes] = *hash_in;
    }
	CVMX_LOADUNA_INT64(p0, ptr, 0);
	CVMX_LOADUNA_INT64(p1, ptr, 8);
	CVMX_LOADUNA_INT64(p2, ptr, 16);
	CVMX_LOADUNA_INT64(p3, ptr, 24);
	CVMX_LOADUNA_INT64(p4, ptr, 32);
	CVMX_LOADUNA_INT64(p5, ptr, 40);

    hash_buf_ptr_64 = (uint64_t *)hash_buf;
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 1);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 4);
	CVMX_LOADUNA_INT64(p0, ptr, 48);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 6);

	CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p0));

    if(key->ex_mode & 0x02)
        hash_in += 56;
    else
	    hash_in += 57;
	hash_rem_num = (hash_end - hash_in);
	hash_copied_bytes = 0;

	/* SHA_CBLOCK chunks in 1st iovec */
	while ((hash_in + SHA_CBLOCK) <= hash_end) {
		/* load 1st aes bytes */
		CVMX_LOADUNA_INT64(p0, enc_in, 0);	
		CVMX_LOADUNA_INT64(p1, enc_in, 8);
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 0);
		CVMX_LOADUNA_INT64(p1, hash_in, 8);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

		/* get 2nd aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 16);
		CVMX_LOADUNA_INT64(p1, enc_in, 24);

		/* 1st aes result */
		CVMX_MF_AES_RESULT(p2, 0);	
		CVMX_MF_AES_RESULT(p3, 1);	

		/* load 2nd aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 16);
		CVMX_LOADUNA_INT64(p1, hash_in, 24);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

		/* store 1st aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 0);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

		/* get 3rd aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 32);
		CVMX_LOADUNA_INT64(p1, enc_in, 40);

		/* 2nd aes result */
		CVMX_MF_AES_RESULT(p2, 0);	
		CVMX_MF_AES_RESULT(p3, 1);	

		/* load 3rd aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 32);
		CVMX_LOADUNA_INT64(p1, hash_in, 40);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

		/* store 2nd aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 16);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

		/* get 4th aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 48);
		CVMX_LOADUNA_INT64(p1, enc_in, 56);

		/* get 3rd aes result */
		CVMX_MF_AES_RESULT(p2, 0);
		CVMX_MF_AES_RESULT(p3, 1);

		/* load 4th aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 48);
		CVMX_LOADUNA_INT64(p1, hash_in, 56);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		hash_in += SHA_CBLOCK;
		enc_in += SHA_CBLOCK;
		hash_rem_num -= SHA_CBLOCK;
		enc_rem_num -= SHA_CBLOCK;

		/* store 3rd aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 32);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

		/* get 4th aes result */
		CVMX_MF_AES_RESULT(p2, 0);
		CVMX_MF_AES_RESULT(p3, 1);

		CVMX_PREFETCH(enc_in, SHA_CBLOCK);

		/* store 4th aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 48);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);

		enc_out += SHA_CBLOCK;
	}

	/* for each iovec */
	while (idx++ < inp_cnt) {
		/* for each SHA_CBLOCK in iovec */
		while ((hash_in + SHA_CBLOCK) <= hash_end) {
			/* load 1st aes bytes */
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	
			CVMX_LOADUNA_INT64(p1, enc_in, 8);
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 0);
			CVMX_LOADUNA_INT64(p1, hash_in, 8);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

			/* get 2nd aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 16);
			CVMX_LOADUNA_INT64(p1, enc_in, 24);

			/* 1st aes result */
			CVMX_MF_AES_RESULT(p2, 0);	
			CVMX_MF_AES_RESULT(p3, 1);	

			/* load 2nd aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 16);
			CVMX_LOADUNA_INT64(p1, hash_in, 24);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

			/* store 1st aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

			/* get 3rd aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 32);
			CVMX_LOADUNA_INT64(p1, enc_in, 40);

			/* 2nd aes result */
			CVMX_MF_AES_RESULT(p2, 0);	
			CVMX_MF_AES_RESULT(p3, 1);	

			/* load 3rd aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 32);
			CVMX_LOADUNA_INT64(p1, hash_in, 40);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

			/* store 2nd aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 16);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

			/* get 4th aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 48);
			CVMX_LOADUNA_INT64(p1, enc_in, 56);

			/* get 3rd aes result */
			CVMX_MF_AES_RESULT(p2, 0);
			CVMX_MF_AES_RESULT(p3, 1);

			/* load 4th aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 48);
			CVMX_LOADUNA_INT64(p1, hash_in, 56);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

			hash_in += SHA_CBLOCK;
			enc_in += SHA_CBLOCK;
			hash_rem_num -= SHA_CBLOCK;
			enc_rem_num -= SHA_CBLOCK;

			/* store 3rd aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 32);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

			/* get 4th aes result */
			CVMX_MF_AES_RESULT(p2, 0);
			CVMX_MF_AES_RESULT(p3, 1);

			CVMX_PREFETCH(enc_in, SHA_CBLOCK);

			/* store 4th aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 48);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);

			enc_out += SHA_CBLOCK;
		} /* for each SHA_CBLOCK in iovec */

		hash_rem_num = (hash_end - hash_in);
		enc_rem_num = (hash_end - enc_in);

#ifdef DEBUG
		cavium_dbgprint("hash_rem_num: %d, enc_rem_num: %d (%d)\n", hash_rem_num, enc_rem_num, __LINE__);
#endif
		/* copy the remaining sha inp bytes */
		memcpy(hash_buf, hash_in, hash_rem_num);
		hash_copied_bytes = hash_rem_num;
		hash_rem_num = 0;

		if (idx < inp_cnt) {
			/* try fill the sha block, get bytes from next iovec */
			hash_in = inp_vec[idx].ptr;
			hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
			hash_rem_num = (hash_end - hash_in);
		}
#ifdef DEBUG
		cavium_dbgprint("hash_rem_num: %d, enc_rem_num: %d (%d)\n", hash_rem_num, enc_rem_num, __LINE__);
#endif

		/* less bytes than needed, try to finish inner hash */
		if (hash_rem_num < (SHA_CBLOCK - hash_copied_bytes)) {
			memcpy(hash_buf + hash_copied_bytes, hash_in, hash_rem_num);
			hash_copied_bytes += hash_rem_num;
			hash_buf[hash_copied_bytes] = 0x80;

			memset(hash_buf + hash_copied_bytes + 1, 0, (SHA_CBLOCK - hash_copied_bytes - 1));

			SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
			hash_rem_num = 0;
			hash_copied_bytes = 0;

			/* complete aes bytes in previous iovec */
			while (enc_rem_num >= AES_BLOCK_SIZE) {
				CVMX_LOADUNA_INT64(p0, enc_in, 0);
				CVMX_LOADUNA_INT64(p1, enc_in, 8);

				CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
				CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

				enc_in += AES_BLOCK_SIZE;
				enc_rem_num -= AES_BLOCK_SIZE;

				CVMX_MF_AES_RESULT(p0, 0);
				CVMX_MF_AES_RESULT(p1, 1);

				CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
				CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

				enc_out += AES_BLOCK_SIZE;
			}

			if (enc_rem_num) {
				/* copy the remaining aes bytes from previous iovec */
				memcpy(enc_buf, enc_in, enc_rem_num);
				enc_copied_bytes = enc_rem_num;
				eout_space_left = enc_rem_num;
				enc_rem_num = 0;

				if (idx < inp_cnt) {
					/* move to next iovec */
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					enc_rem_num = (enc_end - enc_in);
				}

				/* fill bytes from next iovec */
				if (enc_rem_num <= (AES_BLOCK_SIZE - enc_copied_bytes)) {
					memcpy(enc_buf + enc_copied_bytes, enc_in, enc_rem_num);
					enc_copied_bytes += enc_rem_num;
					enc_rem_num = 0;
					if (idx < outp_cnt)
						next_output_iovec = 1;
				} else {
					fill_bytes = (AES_BLOCK_SIZE - enc_copied_bytes);
					uint64_t enc_result[2];
					memcpy(enc_buf + enc_copied_bytes, enc_in, fill_bytes);
					enc_copied_bytes = AES_BLOCK_SIZE;
					enc_rem_num -= fill_bytes;

					CVMX_LOADUNA_INT64(p0, enc_buf, 0);
					CVMX_LOADUNA_INT64(p1, enc_buf, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

					enc_in += fill_bytes;

					CVMX_MF_AES_RESULT(enc_result[0], 0);
					CVMX_MF_AES_RESULT(enc_result[1], 1);
                    enc_result[0] = cvmx_htobe64(enc_result[0]);
                    enc_result[1] = cvmx_htobe64(enc_result[0]);

					/* copy into the space left */
					memcpy(enc_out, enc_result, eout_space_left);	

					/* move to next output iovec */
					enc_out = outp_vec[idx].ptr;
					memcpy(enc_out, ((uint8_t *)enc_result + eout_space_left), fill_bytes);
					enc_out += fill_bytes;
					eout_space_left = 0;
					enc_copied_bytes = 0;

					/* encrypt AES_BLOCK_SIZE chunks if any */
					while (enc_rem_num >= AES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);
						CVMX_LOADUNA_INT64(p1, enc_in, 8);

						CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
						CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

						enc_in += AES_BLOCK_SIZE;
						enc_rem_num -= AES_BLOCK_SIZE;

						CVMX_MF_AES_RESULT(p0, 0);
						CVMX_MF_AES_RESULT(p1, 1);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
						CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	

						enc_out += AES_BLOCK_SIZE;
					}
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;
				}
			} else {
				if (idx < inp_cnt) {
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					enc_rem_num = (enc_end - enc_in);
				}

				if (idx < outp_cnt)
					enc_out = outp_vec[idx].ptr;

				while (enc_rem_num >= AES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);
					CVMX_LOADUNA_INT64(p1, enc_in, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

					enc_in += AES_BLOCK_SIZE;
					enc_rem_num -= AES_BLOCK_SIZE;

					CVMX_MF_AES_RESULT(p0, 0);
					CVMX_MF_AES_RESULT(p1, 1);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
					CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	

					enc_out += AES_BLOCK_SIZE;
				}
				if (enc_in) {
					memcpy(enc_buf, enc_in, enc_rem_num);	
					enc_copied_bytes = enc_rem_num;
					enc_rem_num = 0;
				}
			}
			goto get_inner_hash;
		}

		/* more than fill bytes in next iovec */
		fill_bytes = SHA_CBLOCK - hash_copied_bytes;
		memcpy(hash_buf + hash_copied_bytes, hash_in, fill_bytes);
		hash_copied_bytes += fill_bytes;

		hash_in += fill_bytes;
		hash_rem_num -= fill_bytes;

		ptr = (uint64_t *)hash_buf;
		CVMX_LOADUNA_INT64(p0, ptr, 0);
		CVMX_LOADUNA_INT64(p1, ptr, 8);
		CVMX_LOADUNA_INT64(p2, ptr, 16);
		CVMX_LOADUNA_INT64(p3, ptr, 24);
		CVMX_LOADUNA_INT64(p4, ptr, 32);
		CVMX_LOADUNA_INT64(p5, ptr, 40);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
		CVMX_LOADUNA_INT64(p0, ptr, 48);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_LOADUNA_INT64(p1, ptr, 56);

		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		while (enc_rem_num >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);
			CVMX_LOADUNA_INT64(p1, enc_in, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			enc_in += AES_BLOCK_SIZE;
			enc_rem_num -= AES_BLOCK_SIZE;

			CVMX_MF_AES_RESULT(p0, 0);
			CVMX_MF_AES_RESULT(p1, 1);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

			enc_out += AES_BLOCK_SIZE;
		}

		/* if sha remain bytes are less than SHA_CBLOCK, finish inner hash */
		if (hash_rem_num < SHA_CBLOCK) {
			memcpy(hash_buf, hash_in, hash_rem_num);	
			hash_copied_bytes = hash_rem_num;
			hash_buf[hash_copied_bytes] = 0x80;

			memset(hash_buf + hash_copied_bytes + 1, 0, (SHA_CBLOCK - hash_copied_bytes - 1));

			SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
			hash_rem_num = 0;
			hash_copied_bytes = 0;

			start_inner_hash = 1;
		}

		if (enc_rem_num) {
			uint64_t enc_result[2];

			memcpy(enc_buf, enc_in, enc_rem_num);
			enc_copied_bytes = enc_rem_num;
			eout_space_left = enc_rem_num;

			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;	
			enc_rem_num = (enc_end - enc_in);

			if (enc_rem_num <= (AES_BLOCK_SIZE - enc_copied_bytes)) {
				/* copy the remaining aes bytes */
				memcpy(enc_buf + enc_copied_bytes, enc_in, enc_rem_num);
				enc_copied_bytes += enc_rem_num;
				enc_rem_num = 0;
				if (idx < outp_cnt)
					next_output_iovec = 1;

				goto get_inner_hash;
			} else {
				fill_bytes = (AES_BLOCK_SIZE - enc_copied_bytes);

				/* chance that aes bytes are crossing sha inp bytes */
				if ((enc_rem_num - fill_bytes) < hash_rem_num) {
					ptr = (uint64_t *)hash_in;	
					CVMX_LOADUNA_INT64(p0, ptr, 0);
					CVMX_LOADUNA_INT64(p1, ptr, 8);
					CVMX_LOADUNA_INT64(p2, ptr, 16);
					CVMX_LOADUNA_INT64(p3, ptr, 24);
					CVMX_LOADUNA_INT64(p4, ptr, 32);
					CVMX_LOADUNA_INT64(p5, ptr, 40);

					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
					CVMX_LOADUNA_INT64(p0, ptr, 48);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
					CVMX_LOADUNA_INT64(p1, ptr, 56);

					CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

					hash_in += SHA_CBLOCK;
					hash_rem_num -= SHA_CBLOCK;
				}

				memcpy(enc_buf + enc_copied_bytes, enc_in, fill_bytes); 
				enc_copied_bytes += fill_bytes;

				CVMX_LOADUNA_INT64(p0, enc_buf, 0);
				CVMX_LOADUNA_INT64(p1, enc_buf, 8);

				CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
				CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

				enc_in += fill_bytes;
				enc_rem_num -= fill_bytes;

				CVMX_MF_AES_RESULT(enc_result[0], 0);
				CVMX_MF_AES_RESULT(enc_result[1], 1);
                enc_result[0] = cvmx_htobe64(enc_result[0]);
                enc_result[1] = cvmx_htobe64(enc_result[1]);

				memcpy(enc_out, enc_result, eout_space_left);

				/* move to next output iovec */
				enc_out = outp_vec[idx].ptr;
				memcpy(enc_out, ((uint8_t *)enc_result + eout_space_left), fill_bytes);
				enc_out += fill_bytes;

				eout_space_left = 0;
				enc_copied_bytes = 0;

				if (start_inner_hash) {
					/* encrypt AES_BLOCK_SIZE chunks if any */
					while (enc_rem_num >= AES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);
						CVMX_LOADUNA_INT64(p1, enc_in, 8);

						CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
						CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

						enc_in += AES_BLOCK_SIZE;
						enc_rem_num -= AES_BLOCK_SIZE;

						CVMX_MF_AES_RESULT(p0, 0);
						CVMX_MF_AES_RESULT(p1, 1);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
						CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	

						enc_out += AES_BLOCK_SIZE;
					}
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;

					goto get_inner_hash;
				}
			}
		} else {
			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
			enc_rem_num = (enc_end - enc_in);

			if (idx < outp_cnt)
				enc_out = outp_vec[idx].ptr;

			if (start_inner_hash) {
				while (enc_rem_num >= AES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);
					CVMX_LOADUNA_INT64(p1, enc_in, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

					enc_in += AES_BLOCK_SIZE;
					enc_rem_num -= AES_BLOCK_SIZE;

					CVMX_MF_AES_RESULT(p0, 0);
					CVMX_MF_AES_RESULT(p1, 1);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
					CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	

					enc_out += AES_BLOCK_SIZE;
				}

				if (enc_in) {
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;
					enc_rem_num = 0;
				}

				goto get_inner_hash;
			}
		}
	} /* for each iovec */

    if(key->ex_mode == EX_MODE_ENCRYPT) {
        return 1;
    }
get_inner_hash:
	/* get inner hash */
	CVMX_MF_HSH_IV(tmpvar, 0);
	CVMX_MF_HSH_IV(tmpvar2, 1);
	CVMX_MF_HSH_IV(tmpvar3, 2);
    if(key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar), outp_vec[0].ptr, 0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar2), outp_vec[0].ptr, 8);
        CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(tmpvar3)), outp_vec[0].ptr, 16);
        return 1;
    }
        

#ifdef DEBUG
	cavium_dbgprint("hash_copied_bytes: %d, hash_rem_num: %d (%d)\n", hash_copied_bytes, hash_rem_num, __LINE__);
	_hexdump(hash_buf, hash_copied_bytes);
	cavium_dbgprint("enc_copied_bytes: %d, enc_rem_num: %d (%d)\n", enc_copied_bytes, enc_rem_num, __LINE__);
	_hexdump(enc_buf, enc_copied_bytes);;
	cavium_dbgprint("Inner Hash\n");
	_hexdump(&tmpvar, 8);
	_hexdump(&tmpvar2, 8);
	_hexdump(&tmpvar3, 4);
#endif

	/* start outer hash (mac secret + pad_2 + inner hash) */
	tmpvar3 = tmpvar3 | 0x80000000;

	tmpvar4 = CRY_HSH_PAD2_64[0];
	tmpvar4 = (tmpvar4 << 32) | ((tmpvar >> 32) & 0xFFFFFFFFULL);
	ptr = (uint64_t *) CRY_HSH_PAD2;

	/* SHA init */
	CVMX_MT_HSH_IV(CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV(CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV(CRY_SHA1_INIT2, 2);

	CVMX_MT_HSH_DAT(cvmx_htobe64(((uint64_t *)mac_secret)[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(((uint64_t *)mac_secret)[1]), 1);
	CVMX_MT_HSH_DAT(((uint64_t)cvmx_htobe32(*((uint32_t *)&mac_secret[16]))) << 32 | CRY_HSH_PAD2_32[0], 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(*ptr), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(*ptr), 4);
	CVMX_MT_HSH_DAT(cvmx_htobe64(*ptr), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(*ptr), 6);

	CVMX_MT_HSH_STARTSHA(tmpvar4);

	ptr = (uint64_t *)&enc_buf[enc_copied_bytes];

	tmpvar4 = (tmpvar << 32 ) | (tmpvar2 >> 32);
	tmpvar2 = (tmpvar2 << 32) | (tmpvar3 >> 32);

	CVMX_MT_HSH_DAT(tmpvar4, 0 );
	CVMX_MT_HSH_DAT(tmpvar2, 1 );

	CVMX_MT_HSH_DAT(CRY_HSH_BLK1, 2);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 3);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 4);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 5);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 6);

	CVMX_MT_HSH_STARTSHA(SSL_CRY_SHA1_SSL3_LEN * 8);

	enc_copied_bytes += SHA_DIGEST_LENGTH;

	/* calculate pad length */
	padlen = AES_BLOCK_SIZE - ((enc_copied_bytes) % AES_BLOCK_SIZE);

	/* add pad bytes pattren (padlen - 1) */
	memset(enc_buf + enc_copied_bytes, padlen - 1, padlen);
	enc_copied_bytes += padlen;

	/* get outer hash */
	CVMX_MF_HSH_IV(p0, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p0), ptr, 0);
	CVMX_MF_HSH_IV(p1, 1);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p1), ptr, 8);
	CVMX_MF_HSH_IV(tmpvar, 2);

	CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(tmpvar)), ((uint32_t *)ptr) + 4, 0);
    if(key->ex_mode == EX_MODE_HMAC){
        memcpy(outp_vec[0].ptr,ptr,SHA_DIGEST_LENGTH);
        return 1;
    }

#ifdef DEBUG
	cavium_dbgprint("Outer Hash\n");
	_hexdump(ptr, SHA_DIGEST_LENGTH);
#endif
	uint8_t *out_enc_ptr = final_buf;
	uint8_t *in_enc_ptr = enc_buf;
	enc_rem_num = (enc_copied_bytes - eout_space_left);

	while (enc_copied_bytes >= AES_BLOCK_SIZE) {
		CVMX_LOADUNA_INT64(p0, in_enc_ptr, 0);
		CVMX_LOADUNA_INT64(p1, in_enc_ptr, 8);

		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		in_enc_ptr += AES_BLOCK_SIZE;
		enc_copied_bytes -= AES_BLOCK_SIZE;	

		CVMX_MF_AES_RESULT(p0, 0);
		CVMX_MF_AES_RESULT(p1, 1);

		CVMX_STOREUNA_INT64(cvmx_htobe64(p0), out_enc_ptr, 0);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p1), out_enc_ptr, 8);

		out_enc_ptr += AES_BLOCK_SIZE;
	}
	if (enc_copied_bytes)
		cavium_dbgprint("ERROR: enc_copied_bytes: %d\n", enc_copied_bytes);

	memcpy(enc_out, final_buf, eout_space_left);

	if (next_output_iovec)
		enc_out = outp_vec[idx].ptr;			
	else
		enc_out += eout_space_left;

	if (enc_out)
		memcpy(enc_out, final_buf+ eout_space_left, enc_rem_num);

	CVMX_MF_AES_IV(enc_iv[0], 0);
	CVMX_MF_AES_IV(enc_iv[1], 1);
    enc_iv[0] = cvmx_htobe64(enc_iv[0]);
    enc_iv[1] = cvmx_htobe64(enc_iv[1]);

	//update ssl output header
	in_rec_hdr->length = ROUNDUP16(inp_len + SHA_DIGEST_LENGTH + 1);
	*len = in_rec_hdr->length;


	return 1;
}

/*******************************************************************************
 *                                                                             *
 *      Generates SSL-3.0 SHA1 HMAC, adds padding bytes and  encrypts payload  *
 *  data and stores the result in output iovec buffer. Initialisation vector   *
 *  is updated after each encryption. Initialisation vector to be used for     *
 *  encrytion is taken from EVP_DESOCT_HMAC_SHA1 structure passed to the       *
 *  function. Key and MAC secret to be used for cipher encryption and HMAC     *
 *  generation respectively are also taken from the key structure passed.      *
 *  Sequence number and header components, record type and length of record,   *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the SSL-3.0 record header.          *
 *  Protocol : SSl-3.0                                                         *
 *  Cipher   : DES-CBC / DES3-CBC                                              *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing encryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to encrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store encryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_des_sha1_enc_ssl30(EVP_DESOCT_HMAC_SHA1 *key,
		struct iovec_cav *inp_vec, uint16_t inp_cnt,
		struct iovec_cav *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t p0, p1, p2, p3, p4, p5;
	uint8_t hash_buf[128] = {0}, enc_buf[128] = {0}, final_buf[128] = {0};
	uint16_t inp_len;
	int32_t enc_rem_num = 0, hash_rem_num = 0, enc_copied_bytes = 0, hash_copied_bytes = 0, eout_space_left = 0, idx = 0;
	int32_t fill_bytes = 0, padlen = 0, next_output_iovec = 0, start_inner_hash = 0;
	uint64_t tmpvar = 0, tmpvar2 = 0, tmpvar3 = 0, tmpvar4 = 0, seq = 0, total_bits = 0;
	uint8_t *mac_secret = NULL, *key_data = NULL, *hash_in = NULL, *hash_end = NULL, *enc_in = NULL, *enc_end = NULL, *enc_out = NULL;
	uint64_t *enckey = NULL, *ptr = NULL;
    uint16_t *hash_buf_ptr_16 = NULL;
    uint32_t *hash_buf_ptr_32 = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct ssl3_rec_hdr *in_rec_hdr = NULL;

	in_rec_hdr = (struct ssl3_rec_hdr *)&key->aux.tls_aad_8[8];
	inp_len = cvmx_htobe16(in_rec_hdr->length);

	mac_secret = key->mac_secret;
	seq = cvmx_htobe64(key->aux.tls_aad_64[0]);

#ifdef DEBUG
	cavium_dbgprint("rec length: %d, seq: %llx\n", in_rec_hdr->length, seq);
	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
#endif

	tmpvar2 = (uint64_t) cvmx_htobe32(*((uint32_t *)&mac_secret[16]));
	tmpvar2 = (tmpvar2 << 32) | CRY_HSH_PAD1_32[0];
    if(key->ex_mode == EX_MODE_DEFAULT)
    	tmpvar3 = (CRY_HSH_PAD1_64[0] & 0xFFFFFFFF00000000) | upper_32_bits(seq);
    else if(key->ex_mode == EX_MODE_HMAC) {
            if(inp_len >= 4) {
                tmpvar3 = (CRY_HSH_PAD1_64[0] & 0xFFFFFFFF00000000) |cvmx_htobe32(*((uint32_t *)inp_vec[0].ptr));
                inp_vec[0].ptr += 4;
                inp_vec[0].size -=4;
            } else {
                tmpvar3 = (CRY_HSH_PAD1_64[0] & 0xFFFFFFFF00000000);
                memcpy(((uint8_t *)&tmpvar3)+4, inp_vec[0].ptr, inp_len);
                inp_vec[0].ptr += inp_len;
                inp_vec[0].size -= inp_len;
                *(((uint8_t *)&tmpvar3)+inp_len + 4) = 0x80;
            }
        }


	key_data = key->cipher_iv;


	enckey = (uint64_t *)key->cipher_key;
	if(key->cipher_type == OCT_DES_64 || key->cipher_type == OCT_DES_40)
	{
		enckey[1] = enckey[0];
		enckey[2] = enckey[0];
	}

	/* initialize encryption key */
	CVMX_MT_3DES_KEY(cvmx_htobe64(enckey[0]), 0);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enckey[1]), 1);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enckey[2]), 2);

	/* Init encryption iv */
	CVMX_MT_3DES_IV(cvmx_htobe64(*(uint64_t *)key_data));

	/* The first hash (mac secret (20 bytes) + pad_1 (40 bytes) + upper 4 bytes of sequence number */
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT0), 0);
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT1), 1);
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT2), 2);
    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_MT_HSH_DAT(cvmx_htobe64(((uint64_t *) mac_secret)[0]), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(((uint64_t *) mac_secret)[1]), 1);
	    CVMX_MT_HSH_DAT(tmpvar2, 2);
	    CVMX_MT_HSH_DAT(CRY_HSH_PAD1_64[0], 3);
	    CVMX_MT_HSH_DAT(CRY_HSH_PAD1_64[0], 4);
	    CVMX_MT_HSH_DAT(CRY_HSH_PAD1_64[0], 5);
	    CVMX_MT_HSH_DAT(CRY_HSH_PAD1_64[0], 6);

	    CVMX_MT_HSH_STARTSHA(tmpvar3);
    }
	CVMX_PREFETCH0(inp_vec[idx].ptr);

    if(key->ex_mode == EX_MODE_DEFAULT)
    	total_bits = (HMAC_SHA1_KEY_LEN + 40 + 8 + 1 + 2 + inp_len) * 8; /* mac_secret + pad1 + seq + rec_type + rec_len */ 
    else if (key->ex_mode == EX_MODE_HMAC)
    	total_bits = (HMAC_SHA1_KEY_LEN + 40 + inp_len) * 8; /* mac_secret + pad1 + rec_len */ 
    else
    	total_bits =  inp_len * 8;

	/* store lower 4 bytes of sequence number + rec_type + rec_len */
    if(key->ex_mode & 0x02) {
        if(inp_vec[0].size >= 8) {
            memcpy(hash_buf,inp_vec[idx].ptr,8);
            inp_vec[idx].size -= 8;
            inp_vec[idx].ptr +=8;
            hash_copied_bytes = 8;
        } else {
            memcpy(hash_buf, inp_vec[0].ptr, inp_vec[0].size);
            hash_copied_bytes = inp_vec[0].size;
            inp_vec[0].ptr += inp_vec[0].size;
            inp_vec[0].size -= inp_vec[0].size;
        }
    }else {
        hash_buf_ptr_32 = (uint32_t *)hash_buf;
	    hash_buf_ptr_32[0] = cvmx_htobe32(lower_32_bits(seq));
	    hash_buf[4] = in_rec_hdr->type;
        hash_buf_ptr_16 = (uint16_t *)&hash_buf[5];
	    hash_buf_ptr_16[0] = cvmx_htobe16(inp_len);
	    hash_copied_bytes = 7; /* lower 32 bit of seq num (4 byte) + record type (1 byte) + recore length (2 bytes) */
    }

	enc_copied_bytes = 0;

	hash_in = inp_vec[idx].ptr ;
	hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
	hash_rem_num = (hash_end - hash_in) + hash_copied_bytes;

	enc_in = inp_vec[idx].ptr;
	enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
	enc_rem_num = (enc_end - enc_in);

	enc_out = outp_vec[idx].ptr ;

	if (hash_rem_num <= SHA_CBLOCK || key->ex_mode == EX_MODE_ENCRYPT) {
		ptr = (uint64_t *)(hash_buf + hash_copied_bytes);	
		CVMX_STOREUNA_INT64(0, ptr, 0);
		CVMX_STOREUNA_INT64(0, ptr, 8);
		CVMX_STOREUNA_INT64(0, ptr, 16);
		CVMX_STOREUNA_INT64(0, ptr, 24);
		CVMX_STOREUNA_INT64(0, ptr, 32);
		CVMX_STOREUNA_INT64(0, ptr, 40);
		CVMX_STOREUNA_INT64(0, ptr, 48);
		CVMX_STOREUNA_INT64(0, ptr, 56);

		memcpy(hash_buf + hash_copied_bytes, hash_in, inp_vec[idx].size );	
		hash_copied_bytes += (inp_vec[idx].size);
		hash_buf[hash_copied_bytes] = 0x80;

        if(key->ex_mode != EX_MODE_ENCRYPT)
    		SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
		hash_copied_bytes = 0;
		hash_rem_num = 0;

		while (enc_rem_num >= DES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	

			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

			enc_in += DES_BLOCK_SIZE;
			enc_rem_num -= DES_BLOCK_SIZE;

			CVMX_MF_3DES_RESULT(p0);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);

			enc_out += DES_BLOCK_SIZE;
		}

		memcpy(enc_buf, enc_in, enc_rem_num);
		enc_copied_bytes = enc_rem_num;

		goto get_inner_hash;
	}

	/* inp_vec[0] has payload(inp_vec[0].size - sizeof(struct ssl3_rec_hdr)) > 57 bytes */
    if(key->ex_mode & 0x02)
        ptr = (uint64_t *)hash_in;
    else {
	    ptr = (uint64_t *)(hash_in + 1);
	    hash_buf[hash_copied_bytes] = *hash_in; /* 1st byte from the payload */
    }
	CVMX_LOADUNA_INT64(p0, ptr, 0);
	CVMX_LOADUNA_INT64(p1, ptr, 8);
	CVMX_LOADUNA_INT64(p2, ptr, 16);
	CVMX_LOADUNA_INT64(p3, ptr, 24);
	CVMX_LOADUNA_INT64(p4, ptr, 32);
	CVMX_LOADUNA_INT64(p5, ptr, 40);

    hash_buf_ptr_64 = (uint64_t *)hash_buf;
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 1);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 4);
	CVMX_LOADUNA_INT64(p0, ptr, 48);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 6);

	CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p0));

    if(key->ex_mode & 0x02) 
        hash_in += 56;
    else
	    hash_in += 57;
	hash_rem_num = (hash_end - hash_in);
	hash_copied_bytes = 0;

	/* SHA_CBLOCK chunks in 1st iovec */
	while ((hash_in + SHA_CBLOCK) <= hash_end) {
		/* get and load 1st des bytes */
		CVMX_LOADUNA_INT64(p2, enc_in, 0);	
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 0);
		CVMX_LOADUNA_INT64(p1, hash_in, 8);

		/* get 2nd des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 8);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

		/* 1st des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 2nd des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 1st des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 0);

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 16);
		CVMX_LOADUNA_INT64(p1, hash_in, 24);

		/* get 3rd des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 16);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

		/* 2nd des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 3rd des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 2nd des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 32);
		CVMX_LOADUNA_INT64(p1, hash_in, 40);

		/* get 4th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 24);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

		/* 3rd des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 4th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 3rd des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 16);

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 48);
		CVMX_LOADUNA_INT64(p1, hash_in, 56);

		/* 4th des result */
		CVMX_MF_3DES_RESULT(p3);	

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		/* get 5th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 32);

		/* store 4th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

		/* load 5th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* get 6th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 40);

		/* 5th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 6th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* get 7th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 48);

		/* store 5th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 32);

		/* 6th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 7th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* get 8th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 56);

		/* store 6th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

		/* 7th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 8th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 7th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 48);

		hash_in += SHA_CBLOCK;
		enc_in += SHA_CBLOCK;
		enc_rem_num -= SHA_CBLOCK;
		hash_rem_num -= SHA_CBLOCK;

		CVMX_PREFETCH(enc_in, SHA_CBLOCK);

		/* 8th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* store 8th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);

		enc_out += SHA_CBLOCK;
	}

	/* for each iovec */
	while (idx++ < inp_cnt) {
		/* for each SHA_CBLOCK in iovec */
		while ((hash_in + SHA_CBLOCK) <= hash_end) {
			/* get and load 1st des bytes */
			CVMX_LOADUNA_INT64(p2, enc_in, 0);	
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 0);
			CVMX_LOADUNA_INT64(p1, hash_in, 8);

			/* get 2nd des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 8);

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

			/* 1st des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 2nd des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 1st des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 0);

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 16);
			CVMX_LOADUNA_INT64(p1, hash_in, 24);

			/* get 3rd des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 16);

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

			/* 2nd des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 3rd des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 2nd des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 32);
			CVMX_LOADUNA_INT64(p1, hash_in, 40);

			/* get 4th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 24);

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

			/* 3rd des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 4th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 3rd des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 16);

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 48);
			CVMX_LOADUNA_INT64(p1, hash_in, 56);

			/* 4th des result */
			CVMX_MF_3DES_RESULT(p3);	

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

			/* get 5th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 32);

			/* store 4th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

			/* load 5th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* get 6th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 40);

			/* 5th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 6th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* get 7th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 48);

			/* store 5th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 32);

			/* 6th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 7th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* get 8th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 56);

			/* store 6th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

			/* 7th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 8th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 7th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 48);

			hash_in += SHA_CBLOCK;
			enc_in += SHA_CBLOCK;
			enc_rem_num -= SHA_CBLOCK;
			hash_rem_num -= SHA_CBLOCK;

			CVMX_PREFETCH(enc_in, SHA_CBLOCK);

			/* 8th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* store 8th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);

			enc_out += SHA_CBLOCK;
		} /* for each SHA_CBLOCK in iovec */

		hash_rem_num = (hash_end - hash_in);
		enc_rem_num = (hash_end - enc_in);

		/* copy the remaining hash bytes */
		memcpy(hash_buf, hash_in, hash_rem_num);
		hash_copied_bytes = hash_rem_num;
		hash_rem_num = 0;

		if (idx < inp_cnt) {
			/* try fill the sha block, get bytes from next iovec */
			hash_in = inp_vec[idx].ptr;
			hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
			hash_rem_num = (hash_end - hash_in);
		}
#ifdef DEBUG
		cavium_dbgprint("hash_rem_num: %d, enc_rem_num: %d (%d)\n", hash_rem_num, enc_rem_num, __LINE__);
#endif

		/* less bytes than needed, try to finish inner hash */
		if (hash_rem_num < (SHA_CBLOCK - hash_copied_bytes)) {
			memcpy(hash_buf + hash_copied_bytes, hash_in, hash_rem_num);
			hash_copied_bytes += hash_rem_num;
			hash_buf[hash_copied_bytes] = 0x80;

			memset(hash_buf + hash_copied_bytes + 1, 0, (SHA_CBLOCK - hash_copied_bytes - 1));

			SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
			hash_rem_num = 0;
			hash_copied_bytes = 0;

			/* complete enc bytes in previous iovec */
			while (enc_rem_num >= DES_BLOCK_SIZE) {
				CVMX_LOADUNA_INT64(p0, enc_in, 0);

				CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

				enc_in += DES_BLOCK_SIZE;
				enc_rem_num -= DES_BLOCK_SIZE;

				CVMX_MF_3DES_RESULT(p0);

				CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);

				enc_out += DES_BLOCK_SIZE;
			}

			if (enc_rem_num) {
				/* copy the remaining enc bytes from previous iovec */
				memcpy(enc_buf, enc_in, enc_rem_num);
				enc_copied_bytes = enc_rem_num;
				eout_space_left = enc_rem_num;
				enc_rem_num = 0;

				if (idx < inp_cnt) {
					/* move to next iovec */
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					enc_rem_num = (enc_end - enc_in);
				}

				/* fill bytes from next iovec */
				if (enc_rem_num <= (DES_BLOCK_SIZE - enc_copied_bytes)) {
					memcpy(enc_buf + enc_copied_bytes, enc_in, enc_rem_num);
					enc_copied_bytes += enc_rem_num;
					enc_rem_num = 0;
					if (idx < outp_cnt)
						next_output_iovec = 1;
				} else {
					uint64_t enc_result;

					fill_bytes = (DES_BLOCK_SIZE - enc_copied_bytes);
					memcpy(enc_buf + enc_copied_bytes, enc_in, fill_bytes);
					enc_copied_bytes = DES_BLOCK_SIZE;
					enc_rem_num -= fill_bytes;

					CVMX_LOADUNA_INT64(p0, enc_buf, 0);

					CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

					enc_in += fill_bytes;

					CVMX_MF_3DES_RESULT(enc_result);
                    enc_result = cvmx_htobe64(enc_result);

					/* copy into the space left */
					memcpy(enc_out, &enc_result, eout_space_left);	

					/* move to next output iovec */
					enc_out = outp_vec[idx].ptr;
					memcpy(enc_out, ((uint8_t *)&enc_result + eout_space_left), fill_bytes);
					enc_out += fill_bytes;
					eout_space_left = 0;
					enc_copied_bytes = 0;

					/* encrypt DES_BLOCK_SIZE chunks if any */
					while (enc_rem_num >= DES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);

						CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

						enc_in += DES_BLOCK_SIZE;
						enc_rem_num -= DES_BLOCK_SIZE;

						CVMX_MF_3DES_RESULT(p0);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

						enc_out += DES_BLOCK_SIZE;
					}
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;
				}
			} else {
				if (idx < inp_cnt) {
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					enc_rem_num = (enc_end - enc_in);
				}

				if (idx < outp_cnt)
					enc_out = outp_vec[idx].ptr;

				while (enc_rem_num >= DES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);

					CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

					enc_in += DES_BLOCK_SIZE;
					enc_rem_num -= DES_BLOCK_SIZE;

					CVMX_MF_3DES_RESULT(p0);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

					enc_out += DES_BLOCK_SIZE;
				}
				if (enc_in) {
					memcpy(enc_buf, enc_in, enc_rem_num);	
					enc_copied_bytes = enc_rem_num;
					enc_rem_num = 0;
				}
			}
			goto get_inner_hash;
		}

		/* more than fill bytes in next iovec */
		fill_bytes = SHA_CBLOCK - hash_copied_bytes;
		memcpy(hash_buf + hash_copied_bytes, hash_in, fill_bytes);
		hash_copied_bytes += fill_bytes;

		hash_in += fill_bytes;
		hash_rem_num -= fill_bytes;

		ptr = (uint64_t *)hash_buf;
		CVMX_LOADUNA_INT64(p0, ptr, 0);
		CVMX_LOADUNA_INT64(p1, ptr, 8);
		CVMX_LOADUNA_INT64(p2, ptr, 16);
		CVMX_LOADUNA_INT64(p3, ptr, 24);
		CVMX_LOADUNA_INT64(p4, ptr, 32);
		CVMX_LOADUNA_INT64(p5, ptr, 40);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
		CVMX_LOADUNA_INT64(p0, ptr, 48);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_LOADUNA_INT64(p1, ptr, 56);

		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		while (enc_rem_num >= DES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);

			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

			enc_in += DES_BLOCK_SIZE;
			enc_rem_num -= DES_BLOCK_SIZE;

			CVMX_MF_3DES_RESULT(p0);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);

			enc_out += DES_BLOCK_SIZE;
		}

		/* if sha remain bytes are less than SHA_CBLOCK, finish inner hash */
		if (hash_rem_num < SHA_CBLOCK) {
			memcpy(hash_buf, hash_in, hash_rem_num);	
			hash_copied_bytes = hash_rem_num;
			hash_buf[hash_copied_bytes] = 0x80;

			memset(hash_buf + hash_copied_bytes + 1, 0, (SHA_CBLOCK - hash_copied_bytes - 1));

			SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
			hash_rem_num = 0;
			hash_copied_bytes = 0;

			start_inner_hash = 1;
		}

		if (enc_rem_num) {
			uint64_t enc_result;

			memcpy(enc_buf, enc_in, enc_rem_num);
			enc_copied_bytes = enc_rem_num;
			eout_space_left = enc_rem_num;

			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;	
			enc_rem_num = (enc_end - enc_in);

			if (enc_rem_num <= (DES_BLOCK_SIZE - enc_copied_bytes)) {
				/* copy the remaining enc bytes */
				memcpy(enc_buf + enc_copied_bytes, enc_in, enc_rem_num);
				enc_copied_bytes += enc_rem_num;
				enc_rem_num = 0;
				if (idx < outp_cnt)
					next_output_iovec = 1;

				goto get_inner_hash;
			} else {
				fill_bytes = (DES_BLOCK_SIZE - enc_copied_bytes);

				/* chance that enc bytes are crossing hash bytes */
				if ((enc_rem_num - fill_bytes) < hash_rem_num) {
					ptr = (uint64_t *)hash_in;	
					CVMX_LOADUNA_INT64(p0, ptr, 0);
					CVMX_LOADUNA_INT64(p1, ptr, 8);
					CVMX_LOADUNA_INT64(p2, ptr, 16);
					CVMX_LOADUNA_INT64(p3, ptr, 24);
					CVMX_LOADUNA_INT64(p4, ptr, 32);
					CVMX_LOADUNA_INT64(p5, ptr, 40);

					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
					CVMX_LOADUNA_INT64(p0, ptr, 48);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
					CVMX_LOADUNA_INT64(p1, ptr, 56);

					CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

					hash_in += SHA_CBLOCK;
					hash_rem_num -= SHA_CBLOCK;
				}

				memcpy(enc_buf + enc_copied_bytes, enc_in, fill_bytes); 
				enc_copied_bytes += fill_bytes;

				CVMX_LOADUNA_INT64(p0, enc_buf, 0);

				CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

				enc_in += fill_bytes;
				enc_rem_num -= fill_bytes;

				CVMX_MF_3DES_RESULT(enc_result);
                enc_result  = cvmx_htobe64(enc_result);

				memcpy(enc_out, &enc_result, eout_space_left);

				/* move to next output iovec */
				enc_out = outp_vec[idx].ptr;
				memcpy(enc_out, ((uint8_t *)&enc_result + eout_space_left), fill_bytes);
				enc_out += fill_bytes;

				eout_space_left = 0;
				enc_copied_bytes = 0;

				if (start_inner_hash) {
					/* encrypt DES_BLOCK_SIZE chunks if any */
					while (enc_rem_num >= DES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);

						CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

						enc_in += DES_BLOCK_SIZE;
						enc_rem_num -= DES_BLOCK_SIZE;

						CVMX_MF_3DES_RESULT(p0);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

						enc_out += DES_BLOCK_SIZE;
					}
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;

					goto get_inner_hash;
				}
			}
		} else {
			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
			enc_rem_num = (enc_end - enc_in);

			if (idx < outp_cnt)
				enc_out = outp_vec[idx].ptr;

			if (start_inner_hash) {
				while (enc_rem_num >= DES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);

					CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

					enc_in += DES_BLOCK_SIZE;
					enc_rem_num -= DES_BLOCK_SIZE;

					CVMX_MF_3DES_RESULT(p0);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

					enc_out += DES_BLOCK_SIZE;
				}

				if (enc_in) {
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;
					enc_rem_num = 0;
				}

				goto get_inner_hash;
			}
		}
	} /* for each iovec */

get_inner_hash:
	/* get inner hash */
	CVMX_MF_HSH_IV(tmpvar, 0);
	CVMX_MF_HSH_IV(tmpvar2, 1);
	CVMX_MF_HSH_IV(tmpvar3, 2);

    if(key->ex_mode == EX_MODE_ENCRYPT) {
        return 1;
    } else if(key->ex_mode == EX_MODE_DIGEST) {
		CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar), outp_vec[0].ptr, 0);	
		CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar2), outp_vec[0].ptr, 8);	
		CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(tmpvar3)), outp_vec[0].ptr, 16);	
        return 1;
    }
#ifdef DEBUG
	cavium_dbgprint("hash_copied_bytes: %d, hash_rem_num: %d (%d)\n", hash_copied_bytes, hash_rem_num, __LINE__);
	_hexdump(hash_buf, hash_copied_bytes);
	cavium_dbgprint("enc_copied_bytes: %d, enc_rem_num: %d (%d)\n", enc_copied_bytes, enc_rem_num, __LINE__);
	_hexdump(enc_buf, enc_copied_bytes);;
	cavium_dbgprint("Inner Hash : %s\n", __func__);
	_hexdump(&tmpvar, 8);
	_hexdump(&tmpvar2, 8);
	_hexdump(&tmpvar3, 8);
#endif

	/* start outer hash (mac secret + pad_2 + inner hash) */
	tmpvar3 = tmpvar3 | 0x80000000;

	tmpvar4 = CRY_HSH_PAD2_64[0];
	tmpvar4 = (tmpvar4 << 32) | ((tmpvar >> 32) & 0xFFFFFFFFULL);
	ptr = &CRY_HSH_PAD2_64[0];

	/* SHA init */
	CVMX_MT_HSH_IV(CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV(CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV(CRY_SHA1_INIT2, 2);

	CVMX_MT_HSH_DAT(cvmx_htobe64(((uint64_t *)mac_secret)[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(((uint64_t *)mac_secret)[1]), 1);
	CVMX_MT_HSH_DAT(((uint64_t)cvmx_htobe32(*((uint32_t *)&mac_secret[16]))) << 32 | CRY_HSH_PAD2_32[0], 2);
	CVMX_MT_HSH_DAT(*ptr, 3);
	CVMX_MT_HSH_DAT(*ptr, 4);
	CVMX_MT_HSH_DAT(*ptr, 5);
	CVMX_MT_HSH_DAT(*ptr, 6);

	CVMX_MT_HSH_STARTSHA(tmpvar4);

	ptr = (uint64_t *)&enc_buf[enc_copied_bytes];

	tmpvar4 = (tmpvar << 32 ) | (tmpvar2 >> 32);
	tmpvar2 = (tmpvar2 << 32) | (tmpvar3 >> 32);

	CVMX_MT_HSH_DAT(tmpvar4, 0 );
	CVMX_MT_HSH_DAT(tmpvar2, 1 );
	CVMX_MT_HSH_DAT(CRY_HSH_BLK1, 2);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 3);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 4);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 5);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 6);

	CVMX_MT_HSH_STARTSHA(SSL_CRY_SHA1_SSL3_LEN * 8);

	enc_copied_bytes += SHA_DIGEST_LENGTH;

	/* calculate pad length */
	padlen = DES_BLOCK_SIZE - ((enc_copied_bytes) % DES_BLOCK_SIZE);

	/* add pad bytes pattren (padlen - 1) */
	memset(enc_buf + enc_copied_bytes, padlen - 1, padlen);
	enc_copied_bytes += padlen;

	/* get outer hash */
	CVMX_MF_HSH_IV(p0, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p0), ptr, 0);
	CVMX_MF_HSH_IV(p1, 1);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p1), ptr, 8);
	CVMX_MF_HSH_IV(tmpvar, 2);

	CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(tmpvar)), ((uint32_t *)ptr) + 4, 0);
    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr,ptr,SHA_DIGEST_LENGTH);
        return 1;
    }
#ifdef DEBUG
	cavium_dbgprint("Outer Hash : %s\n", __func__);
	_hexdump(ptr, 24);
#endif

	uint8_t *out_enc_ptr = final_buf;
	uint8_t *in_enc_ptr = enc_buf;
	enc_rem_num = (enc_copied_bytes - eout_space_left);

	while (enc_copied_bytes >= DES_BLOCK_SIZE) {
		CVMX_LOADUNA_INT64(p0, in_enc_ptr, 0);

		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

		in_enc_ptr += DES_BLOCK_SIZE;
		enc_copied_bytes -= DES_BLOCK_SIZE;	

		CVMX_MF_3DES_RESULT(p0);

		CVMX_STOREUNA_INT64(cvmx_htobe64(p0), out_enc_ptr, 0);

		out_enc_ptr += DES_BLOCK_SIZE;
	}
	if (enc_copied_bytes)
		cavium_dbgprint("ERROR: enc_copied_bytes: %d\n", enc_copied_bytes);

	memcpy(enc_out, final_buf, eout_space_left);

	if (next_output_iovec)
		enc_out = outp_vec[idx].ptr;			
	else
		enc_out += eout_space_left;

	if (enc_out)
		memcpy(enc_out, final_buf + eout_space_left, enc_rem_num);

	CVMX_MF_3DES_IV(*(uint64_t *)key_data);
    *(uint64_t *)&key_data[0] = cvmx_htobe64(*(uint64_t *)&key_data[0]); 

	//update ssl output header
	
	*len = ROUNDUP8(inp_len + SHA_DIGEST_LENGTH + 1);



	return 1;
}



/*Start of decrypt functions*/

/*******************************************************************************
 *                                                                             *
 *    Decrypts input record and generates SSL-3.0 SHA1 HMAC; authenticates     *
 *  record and stores result in output iovec buffer. Initialisation vector is  *
 *  updated after each decryption. Initialisation vector to be used for        *
 *  decrytion is taken from EVP_AESOCT_HMAC_SHA1 structure passed to the       *
 *  function. Key and MAC secret to be used for cipher decryption and HMAC     *
 *  generation respectively are also taken from the key structure passed.      *
 *  Sequence number and header components, record type and length of record,   *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the SSL-3.0 record header.          *
 *  Protocol : SSl-3.0                                                         *
 *  Cipher   : DES-CBC / DES3-CBC                                              *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing decryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to decrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store decryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/
int32_t oct_aes_sha1_dec_ssl30(EVP_AESOCT_HMAC_SHA1 *key,
		struct iovec_cav *inp_vec, uint16_t inp_cnt,
		struct iovec_cav *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t r0, r1, r2, r3, r4, r5;
	uint8_t i = 0, keylength = 1, inner_hash_done = 0, output_overlap = 0, hash_buf[128] = {0}, hash_rem_buf[128] = {0}, buf[128] = {0};
	uint16_t raw_data_len = 0, pad_data_len = 0, hash_input_rem_len = 0, crypt_input_rem_len = 0, cipher_type = 0;
	uint32_t enc_rem_cur_chunk = 0, overlap_buf_size = 0, inner_hash_total_bytes = 0, prev_iovec_rem_bytes = 0,ret_val = 1;
	uint64_t seq_num = 0, hash_bal = 0 ;
	uint8_t *mac_secret = NULL, *key_data = NULL, *recv_hmac_ptr = NULL, *in_chunk_cur = NULL, *out_chunk_cur = NULL, *enc_rem_ptr = NULL;
	uint64_t *enc_key = NULL;
    uint16_t *hash_buf_ptr_16 = NULL;
    uint32_t *hash_buf_ptr_32 = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct ssl3_rec_hdr *in_rec_hdr = NULL;
	struct iovec_cav hash_overlap_buf;
	hash_overlap_buf.size = 0;

#ifdef DEBUG
	uint8_t idx = 0;
	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
	for (idx = 0; idx < inp_cnt; idx++)
	{
		cavium_dbgprint("inp_vec[%d].ptr: %p, inp_vec[%d].size: %d\n", idx, inp_vec[idx].ptr, idx, inp_vec[idx].size);
	}

	for (idx = 0; idx < outp_cnt; idx++)
	{
		cavium_dbgprint("outp_vec[%d].ptr: %p, outp_vec[%d].size: %d\n", idx, outp_vec[idx].ptr, idx, outp_vec[idx].size);
	}
#endif
	cipher_type = key->cipher_type;

	in_rec_hdr = (struct ssl3_rec_hdr *)&(key->aux.tls_aad_8[8]);
	crypt_input_rem_len = cvmx_htobe16(in_rec_hdr->length);

   	if ((crypt_input_rem_len % AES_BLOCK_SIZE) && (key->ex_mode == EX_MODE_DEFAULT || key->ex_mode == EX_MODE_ENCRYPT))
	{
      		cavium_dbgprint(
        		"%s:%d Not properly padded data is received for cipher %s\n",
              		__FUNCTION__,__LINE__, cipher_strings[cipher_type]);
      		ret_val = 0;
      		return ret_val;
   	}

	if(cipher_type == OCT_AES_256)
	{
		keylength = 3;
	}

	seq_num = cvmx_htobe64(key->aux.tls_aad_64[0]);
	enc_key = (uint64_t *)key->cipher_key;
	mac_secret = key->mac_secret;
	key_data = key->cipher_iv;

	/*Load AES key*/
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[1]), 1);
	if(keylength == 3)
	{
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[2]), 2);
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[3]), 3);
	}
	else
	{
		CVMX_MT_AES_KEY((uint64_t)0, 2);
		CVMX_MT_AES_KEY((uint64_t)0, 3);
	}
	CVMX_MT_AES_KEYLENGTH(keylength);




	/* Pad data can not be more than AES_BLOCK_SIZE
	 * As per AES CBC implementation, previous aes block result is fed as IV for current aes block
	 * So decrypt last 32 bytes to know PAD length*/


	// calculate pad length by decrypting last 32 bytes
	// if the last iovec buffer contains data >= 32
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    if(inp_vec[inp_cnt - 1].size >= 2*AES_BLOCK_SIZE)
	    {
	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 32, 0);
	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 24, 0);

	    	CVMX_MT_AES_IV(cvmx_htobe64(r0), 0);
	    	CVMX_MT_AES_IV(cvmx_htobe64(r1), 1);

	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 16, 0);
	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 8, 0);
	    }
	    else // last 32 bytes data is overlapped
	    {
	    	// X + Y = 32
	    	// X bytes from iovec[inp_cnt - 2]
	    	// Y bytes from iovec[inp_cnt - 1]

	    	uint8_t padarray[32];
	    	memcpy(padarray, inp_vec[inp_cnt - 2].ptr + (inp_vec[inp_cnt - 2].size - (32 - inp_vec[inp_cnt - 1].size)),
                    (32 - inp_vec[inp_cnt - 1].size));
	    	memcpy(padarray + (32 - inp_vec[inp_cnt - 1].size), inp_vec[inp_cnt - 1].ptr, inp_vec[inp_cnt - 1].size);

	    	CVMX_LOADUNA_INT64(r0, padarray, 0);
	    	CVMX_LOADUNA_INT64(r1, padarray, 8);

	    	CVMX_MT_AES_IV(cvmx_htobe64(r0), 0);
	    	CVMX_MT_AES_IV(cvmx_htobe64(r1), 1);

	    	CVMX_LOADUNA_INT64(r0, padarray, 16);
	    	CVMX_LOADUNA_INT64(r1, padarray, 24);

	    }

	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));


	    CVMX_MF_AES_RESULT(r0, 0);
	    CVMX_MF_AES_RESULT(r1, 1);

	    pad_data_len = (r1 & 0xff ) + 1;

	    if (pad_data_len > AES_BLOCK_SIZE)
	    {
	    	cavium_dbgprint(
	    		"%s:%d Not properly padded data is received for cipher %s\n",
	    		__FUNCTION__,__LINE__, cipher_strings[cipher_type]);
	    	ret_val = 0;
	    	return ret_val;
	    }

	    raw_data_len = crypt_input_rem_len - SHA_DIGEST_LENGTH - pad_data_len;
    } else 
        raw_data_len = crypt_input_rem_len;

#ifdef DEBUG
	cavium_dbgprint("raw enc data len: %d (%d)\n", raw_data_len, __LINE__);
	cavium_dbgprint("hmac digest length  : %d (%d)\n", SHA_DIGEST_LENGTH, __LINE__);
	cavium_dbgprint("pad data len: %d (%d)\n", pad_data_len, __LINE__);
	cavium_dbgprint("total enc input len : %d (%d)\n", crypt_input_rem_len, __LINE__);
#endif

	CVMX_LOADUNA_INT64(r2, key_data, 0);
	CVMX_LOADUNA_INT64(r3, key_data, 8);
	CVMX_MT_AES_IV(cvmx_htobe64(r2), 0); // key_data[0]
	CVMX_MT_AES_IV(cvmx_htobe64(r3), 1); // key_data[1]

	//ssl 3.0 HMAC = hash(mac_secret + CRY_HSH_PAD2 + hash(mac_secret + CRY_HSH_PAD1 + seq_num + rec.type + rec.len + input_data))

	// sha1 first block - 20 bytes mac_secret + 40 bytes pad1 + upper 4 bytes seq_num
    if(key->ex_mode == EX_MODE_DIGEST)
    	inner_hash_total_bytes = raw_data_len;
    if(key->ex_mode == EX_MODE_HMAC)
    	inner_hash_total_bytes = SHA_DIGEST_LENGTH + 40 + raw_data_len;
    if(key->ex_mode == EX_MODE_DEFAULT)
    	inner_hash_total_bytes = SHA_DIGEST_LENGTH + 40 + 8 + 1 + 2 + raw_data_len;
	CVMX_STOREUNA_INT64(*((uint64_t *)mac_secret), hash_buf, 0);
	CVMX_STOREUNA_INT64(*((uint64_t *)(mac_secret + 8)), hash_buf, 8);
	CVMX_STOREUNA_INT32(*((uint32_t *)(mac_secret + 16)), hash_buf, 16);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 20);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 28);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 36);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 44);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 52);
    if(key->ex_mode == EX_MODE_HMAC) {
        if(raw_data_len >= 4) {
            CVMX_STOREUNA_INT32(*((uint32_t *)inp_vec[0].ptr), hash_buf, 60);
            inp_vec[0].ptr += 4;
            inp_vec[0].size -= 4;
            raw_data_len -= 4;
       } else {
            uint32_t datz = 0;
            memcpy((uint8_t *)&datz, inp_vec[0].ptr, raw_data_len);
            inp_vec[0].ptr += raw_data_len;
            inp_vec[0].size -= raw_data_len;
            (((uint8_t *)&datz)[raw_data_len]) = 0x80;
            raw_data_len = 0;
            CVMX_STOREUNA_INT32(*((uint32_t *)&datz), hash_buf, 60);
       }
    }
    else if(key->ex_mode == EX_MODE_DEFAULT) {
       hash_buf_ptr_32 = (uint32_t *)(hash_buf+60);
	   hash_buf_ptr_32[0] = cvmx_htobe32((uint32_t)((seq_num&0xFFFFFFFF00000000)>>32));
    }

	CVMX_MT_HSH_IV( ((uint64_t)CRY_SHA1_INIT0), 0);
	CVMX_MT_HSH_IV( ((uint64_t)CRY_SHA1_INIT1), 1);
	CVMX_MT_HSH_IV( ((uint64_t)CRY_SHA1_INIT2), 2);


    hash_buf_ptr_64 = (uint64_t *)hash_buf;
    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[1]), 1);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[2]), 2);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[3]), 3);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[4]), 4);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[5]), 5);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[6]), 6);

	    CVMX_MT_HSH_STARTSHA(cvmx_htobe64(hash_buf_ptr_64[7]));
    }
#ifdef DEBUG
	cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", *(uint64_t *)(hash_buf), *(uint64_t *)(hash_buf + 8));
	cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", *(uint64_t *)(hash_buf + 16), *(uint64_t *)(hash_buf + 24));
	cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", *(uint64_t *)(hash_buf + 32), *(uint64_t *)(hash_buf + 40));
	cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", *(uint64_t *)(hash_buf + 48), *(uint64_t *)(hash_buf + 56));
#endif

    if(key->ex_mode == EX_MODE_DEFAULT) {
        hash_buf_ptr_32 = (uint32_t *)(hash_buf+64);
        hash_buf_ptr_32[0] = cvmx_htobe32((uint32_t)(seq_num&0xFFFFFFFF));
	    hash_buf[68] = in_rec_hdr->type;
        hash_buf_ptr_16 = (uint16_t *)&hash_buf[69];
	    hash_buf_ptr_16[0]  = cvmx_htobe16(raw_data_len);
	    CVMX_LOADUNA_INT64(hash_bal, hash_buf, 64);
        hash_bal = cvmx_htobe64(hash_bal);
	    hash_bal &= bit_start7;
    } else if(key->ex_mode & 0x02){
        if(raw_data_len > 8) {
            memcpy(hash_buf+64, inp_vec[0].ptr, 8);
            CVMX_LOADUNA_INT64(hash_bal, hash_buf, 64);
            hash_bal = cvmx_htobe64(hash_bal);
            inp_vec[0].ptr += 8;
            inp_vec[0].size -= 8;
        } else {
            memcpy(hash_buf+64, inp_vec[0].ptr, raw_data_len);
            inp_vec[0].ptr += raw_data_len;
            inp_vec[0].size -= raw_data_len;
            CVMX_LOADUNA_INT64(hash_bal, hash_buf, 64);
            hash_bal = cvmx_htobe64(hash_bal);
        }
    } else
        hash_bal = 0;

#ifdef DEBUG
	cavium_dbgprint("hash_bal : %016llx\n", hash_bal);
#endif


	hash_input_rem_len = raw_data_len;

	if(inp_cnt == 1)
	{
		in_chunk_cur = inp_vec[0].ptr;
        if(key->ex_mode & 0x02)
            out_chunk_cur = inp_vec[0].ptr;
        else
		    out_chunk_cur = outp_vec[0].ptr;

		while(hash_input_rem_len >= 64)
		{
			hash_bal = block_processing_aes_sha1_dec_ssl30(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);

#ifdef DEBUG
			cavium_dbgprint("hash_bal : %016llx \n", hash_bal);
#endif
			in_chunk_cur += 64;
			out_chunk_cur += 64;

			hash_input_rem_len -= 64;
			crypt_input_rem_len -= 64;
		}
		recv_hmac_ptr = out_chunk_cur;

		// remaining encrypted input data is compulsory a multiple of 16
		// hash computation is done on encrypted remaining raw data
		// decryption is done on encrypted (remaining raw data + hmac + pad data)
        if((key->ex_mode == EX_MODE_ENCRYPT && crypt_input_rem_len)|| 
                key->ex_mode == EX_MODE_DEFAULT) {
		    do
		    {
		    	CVMX_LOADUNA_INT64(r0, in_chunk_cur, 0);
		    	CVMX_LOADUNA_INT64(r1, in_chunk_cur, 8);

		    	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
		    	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

		    	in_chunk_cur += 16;
		    	crypt_input_rem_len -= 16;

		    	CVMX_MF_AES_RESULT(r2, 0);
		    	CVMX_MF_AES_RESULT(r3, 1);

#ifdef DEBUG
			cavium_dbgprint("Decrypted Output : %016llx \t %016llx\n", r2, r3);
#endif
	    		CVMX_STOREUNA_INT64(cvmx_htobe64(r2), out_chunk_cur, 0);
	    		CVMX_STOREUNA_INT64(cvmx_htobe64(r3), out_chunk_cur, 8);

	    		out_chunk_cur += 16;

	    	} while(crypt_input_rem_len);
        }
        if(key->ex_mode == EX_MODE_ENCRYPT)
            return 1;

		// Along with 7 bytes hash_bal, hash_input_rem_len should be given to hash engine as input
		CVMX_STOREUNA_INT64(cvmx_htobe64(hash_bal), hash_rem_buf, 0);

		if(hash_input_rem_len &&  key->ex_mode == EX_MODE_DEFAULT) {
			memcpy(hash_rem_buf + 7, recv_hmac_ptr, hash_input_rem_len);
		} else if((key->ex_mode & 0x02)) {
            if(hash_input_rem_len > 8) 
    			memcpy(hash_rem_buf + 8, recv_hmac_ptr, hash_input_rem_len-8);
    		 hash_rem_buf[hash_input_rem_len] = 0x80;
        }

        if (key->ex_mode == EX_MODE_DEFAULT)
		    hash_rem_buf[7 + hash_input_rem_len] = 0x80;
		recv_hmac_ptr += hash_input_rem_len;

		uint8_t *ptr = hash_rem_buf;
		if((((hash_input_rem_len + 7) >= 56) && (key->ex_mode == EX_MODE_DEFAULT)) ||
          ((hash_input_rem_len  >= 56) && (key->ex_mode == EX_MODE_DIGEST || key->ex_mode == EX_MODE_HMAC)))

		{
			CVMX_LOADUNA_INT64(r0, ptr, 0);
			CVMX_LOADUNA_INT64(r1, ptr, 8);
#ifdef DEBUG
			cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", r0, r1); 
#endif
			CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);
			CVMX_LOADUNA_INT64(r2, ptr, 16);
			CVMX_LOADUNA_INT64(r3, ptr, 24);
#ifdef DEBUG
			cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", r2, r3); 
#endif
			CVMX_MT_HSH_DAT(cvmx_htobe64(r2), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 3);
			CVMX_LOADUNA_INT64(r0, ptr, 32);
			CVMX_LOADUNA_INT64(r1, ptr, 40);
#ifdef DEBUG
			cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", r0, r1); 
#endif
			CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 5);
			CVMX_LOADUNA_INT64(r2, ptr, 48);
			CVMX_LOADUNA_INT64(r3, ptr, 56);
#ifdef DEBUG
			cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", r2, r3); 
#endif
			CVMX_MT_HSH_DAT(cvmx_htobe64(r2), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(r3));

			ptr += 64;
		}
		CVMX_LOADUNA_INT64(r0, ptr, 0);
		CVMX_LOADUNA_INT64(r1, ptr, 8);
#ifdef DEBUG
		cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", r0, r1); 
#endif
		CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);
		CVMX_LOADUNA_INT64(r2, ptr, 16);
		CVMX_LOADUNA_INT64(r3, ptr, 24);
#ifdef DEBUG
		cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", r2, r3); 
#endif
		CVMX_MT_HSH_DAT(cvmx_htobe64(r2), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 3);
		CVMX_LOADUNA_INT64(r0, ptr, 32);
		CVMX_LOADUNA_INT64(r1, ptr, 40);
#ifdef DEBUG
		cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", r0, r1); 
#endif
		CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 5);
		CVMX_LOADUNA_INT64(r2, ptr, 48);
#ifdef DEBUG
		cavium_dbgprint("SHA1 Input : %016llx \t %016llx\n", r2, (uint64_t)(inner_hash_total_bytes * 8)); 
#endif
		CVMX_MT_HSH_DAT(cvmx_htobe64(r2), 6);
		CVMX_MT_HSH_STARTSHA(inner_hash_total_bytes * 8);
	}
	else if(inp_cnt > 1)
	{
		hash_overlap_buf.ptr = buf;
		hash_overlap_buf.size = 0;
		iovec_record_process_t *iovec_stat = NULL;
		iovec_stat = malloc(inp_cnt * sizeof(iovec_record_process_t));
		memset(iovec_stat, 0, inp_cnt * sizeof(iovec_record_process_t));

		prev_iovec_rem_bytes = 0;
		// populate loop counters
		while(hash_input_rem_len && (i < inp_cnt))
		{
			if(hash_input_rem_len >= inp_vec[i].size)
			{
				iovec_stat[i].hash_size = inp_vec[i].size;
				hash_input_rem_len -= inp_vec[i].size;
			}
			else
			{
				iovec_stat[i].hash_size = hash_input_rem_len;
				hash_input_rem_len = 0;
			}

			if(prev_iovec_rem_bytes)
			{
				iovec_stat[i].hash_req_bytes = 64 - prev_iovec_rem_bytes;
				// in case of last chunk not holding enough bytes to fill 64 byte block
				if(iovec_stat[i].hash_size < iovec_stat[i].hash_req_bytes)
				{
					iovec_stat[i].hash_req_bytes = iovec_stat[i].hash_size;
				}
			}
			else
			{
				iovec_stat[i].hash_req_bytes = 0;
			}
			iovec_stat[i].crypto_itr = (iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) / AES_BLOCK_SIZE;
			iovec_stat[i].hash_itr = iovec_stat[i].crypto_itr / 4;

			iovec_stat[i].hash_rem_bytes = ((iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) - (iovec_stat[i].hash_itr * 64));
			prev_iovec_rem_bytes = iovec_stat[i].hash_rem_bytes;

			i++;
		}
		uint16_t num_hash_iovec = i;
		// need to update enc_digest_ptr
		if(iovec_stat[i-1].hash_size == inp_vec[i-1].size)
		{
			enc_rem_ptr = inp_vec[i].ptr;
		}
		else
		{
			enc_rem_ptr = inp_vec[i-1].ptr + iovec_stat[i-1].hash_size;
			enc_rem_cur_chunk = inp_vec[i-1].size - iovec_stat[i-1].hash_size;

		}

#ifdef DEBUG
		for(i = 0; i < inp_cnt; i++)
		{
			cavium_dbgprint("Chunk : %d\n", i+1);
			cavium_dbgprint("hash_req_bytes : %d\n",iovec_stat[i].hash_req_bytes);
			cavium_dbgprint("crypto_itr : %d\n", iovec_stat[i].crypto_itr);
			cavium_dbgprint("hash_itr : %d\n", iovec_stat[i].hash_itr);
			cavium_dbgprint("hash_rem_bytes : %d\n",iovec_stat[i].hash_rem_bytes);
		}
#endif

		for(i = 0; i < num_hash_iovec; i++)
		{
			in_chunk_cur = inp_vec[i].ptr;
			out_chunk_cur = outp_vec[i].ptr;

			if(iovec_stat[i].hash_req_bytes)
			{	
				memcpy(hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, 
						in_chunk_cur, 
						iovec_stat[i].hash_req_bytes);
				in_chunk_cur += iovec_stat[i].hash_req_bytes;

				if((iovec_stat[i-1].hash_rem_bytes + iovec_stat[i].hash_req_bytes) < 64)
				{
					hash_overlap_buf.size += iovec_stat[i].hash_req_bytes;
					out_chunk_cur = outp_vec[i-1].ptr + iovec_stat[i-1].hash_req_bytes + (64 * iovec_stat[i-1].hash_itr);
					output_overlap = 1;
					continue;
				}
				else
				{
					hash_bal = block_processing_aes_sha1_dec_ssl30(hash_overlap_buf.ptr, hash_overlap_buf.ptr, hash_bal, key->ex_mode);

					memcpy(outp_vec[i-1].ptr + (outp_vec[i-1].size - iovec_stat[i-1].hash_rem_bytes),
							hash_overlap_buf.ptr, 
							iovec_stat[i-1].hash_rem_bytes);
					memcpy(out_chunk_cur, 
							hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, 
							iovec_stat[i].hash_req_bytes);

					out_chunk_cur += iovec_stat[i].hash_req_bytes;
					hash_overlap_buf.size = 0;
				}
			}

			register uint8_t hash_itr = iovec_stat[i].hash_itr;
			while(hash_itr)
			{
				hash_bal = block_processing_aes_sha1_dec_ssl30(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);
#ifdef DEBUG
				cavium_dbgprint("hash_bal : %016llx \n", hash_bal);
#endif

				in_chunk_cur += 64;
				out_chunk_cur += 64;

				hash_itr--;
			}

			if(iovec_stat[i].hash_rem_bytes)
			{
				//copy rem bytes to hash overlap buf
				memcpy(hash_overlap_buf.ptr, in_chunk_cur, iovec_stat[i].hash_rem_bytes);
				in_chunk_cur += iovec_stat[i].hash_rem_bytes;
				hash_overlap_buf.size = iovec_stat[i].hash_rem_bytes;
			}
		}

#ifdef DEBUG
		cavium_dbgprint("hash_overlap_buf.size : %d \n", hash_overlap_buf.size);
#endif
		//obviously this case gets hit all the time
		if(inner_hash_done == 0)
		{
			//if(hash_overlap_buf.size)
			{
				uint8_t *ptr = NULL;
				overlap_buf_size = hash_overlap_buf.size;
				memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size, enc_rem_ptr, enc_rem_cur_chunk);
				overlap_buf_size += enc_rem_cur_chunk;
#ifdef DEBUG
				cavium_dbgprint("enc_rem_cur_chunk : %d\n", enc_rem_cur_chunk);
#endif
				if(enc_rem_cur_chunk < (SHA_DIGEST_LENGTH + pad_data_len))
				{
					memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size + enc_rem_cur_chunk, 
							inp_vec[inp_cnt-1].ptr, inp_vec[inp_cnt-1].size);

					overlap_buf_size += inp_vec[inp_cnt-1].size;
#ifdef DEBUG
					cavium_dbgprint("overlap_buf_size : %d\n", overlap_buf_size);
#endif
				}
				ptr = hash_overlap_buf.ptr;
				uint16_t num_itr = 0;
				num_itr = overlap_buf_size / AES_BLOCK_SIZE;
				while(num_itr)
				{
					CVMX_LOADUNA_INT64(r0, ptr, 0);
					CVMX_LOADUNA_INT64(r1, ptr, 8);

					CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
					CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

					CVMX_MF_AES_RESULT(r2, 0);
					CVMX_MF_AES_RESULT(r3, 1);

#ifdef DEBUG
					cavium_dbgprint("r2 : %llx  r3 : %llx\n", r2, r3);
#endif
					CVMX_STOREUNA_INT64(cvmx_htobe64(r2), ptr, 0);
					CVMX_STOREUNA_INT64(cvmx_htobe64(r3), ptr, 8);

					ptr += 16;
					num_itr--;
				}

				if(hash_overlap_buf.size)
				{
					if(output_overlap)
					{
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, iovec_stat[num_hash_iovec - 2].hash_rem_bytes);
						memcpy(outp_vec[inp_cnt -1].ptr, hash_overlap_buf.ptr + iovec_stat[num_hash_iovec - 2].hash_rem_bytes, iovec_stat[num_hash_iovec - 1].hash_req_bytes);
					}
					else
					{
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, hash_overlap_buf.size);
					}
				}
			}

			memset(hash_rem_buf, 0, 128);
            hash_buf_ptr_64 = (uint64_t *) hash_rem_buf;
			hash_buf_ptr_64[0] = hash_bal;
			memcpy(hash_rem_buf + 7, hash_overlap_buf.ptr, hash_overlap_buf.size);
			hash_rem_buf[hash_overlap_buf.size + 7] = 0x80;
			uint64_t *ptr = (uint64_t *)hash_rem_buf;
#ifdef DEBUG
			cavium_dbgprint("hash rem buf\n");
			_hexdump(hash_rem_buf, 128);
#endif
			if((hash_overlap_buf.size + 7) >= 56)
			{
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
				CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

				ptr += 8;
			}
			inner_hash_total_bytes = 20 + 40 + 8 + 1 + 2 + raw_data_len;
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(inner_hash_total_bytes * 8);

			inner_hash_done = 1;
		}
		free(iovec_stat);
	}

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

	CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 60);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 68);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 76);
#ifdef DEBUG
	cavium_dbgprint("%s : Calculated Inner Hash\n", __func__);
	cavium_dbgprint("r0 : %llx\n", r0);
	cavium_dbgprint("r1 : %llx\n", r1);
	cavium_dbgprint("r2 : %llx\n", r2 & 0xffffffff00000000ull);
#endif
    if(key->ex_mode == EX_MODE_DIGEST) {
        memcpy(outp_vec[0].ptr, hash_buf+60, 20);
        return 1;
    } else if (key->ex_mode == EX_MODE_ENCRYPT)
        return 1;

	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 20);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 28);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 36);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 44);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 52);
	hash_buf[80] = 0x80;


	//calculate outer hash
	CVMX_MT_HSH_IV(CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV(CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV(CRY_SHA1_INIT2, 2);

	//outer_hash_total_bytes = 20 (MAC SECRET) + 40 (PAD2) + 20 (INNER HMAC);

    hash_buf_ptr_64 = (uint64_t *)hash_buf;
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[1]), 1);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[2]), 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[3]), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[4]), 4);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[5]), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[6]), 6);

	CVMX_MT_HSH_STARTSHA(cvmx_htobe64(hash_buf_ptr_64[7]));


	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[8]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[9]), 1);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK1, 2);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 3);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 4);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 5);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 6);

	CVMX_MT_HSH_STARTSHA(HMAC_SHA1_OUTER_BITS);

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);
	r2 = r2 & 0xFFFFFFFF00000000ull;


#ifdef DEBUG
	cavium_dbgprint("%s : Calculated Outer Hash\n", __func__);
	_hexdump(hash_buf, SHA_DIGEST_LENGTH);

#endif

	//Compare calculated HMAC with received HMAC
	if(inp_cnt == 1)
	{
#ifdef DEBUG
		cavium_dbgprint("%s : Received outer hash\n", __func__);
		_hexdump(recv_hmac_ptr, SHA_DIGEST_LENGTH);
#endif

		CVMX_LOADUNA_INT64(r3, recv_hmac_ptr, 0);
		CVMX_LOADUNA_INT64(r4, recv_hmac_ptr, 8);
		CVMX_LOADUNA_INT64(r5, recv_hmac_ptr, 16);
        r3 = cvmx_htobe64(r3);
        r4 = cvmx_htobe64(r4);
        r5 = cvmx_htobe64(r5);
		r5 = r5 & 0xFFFFFFFF00000000ull;
		if((r0 != r3) || (r1 != r4) || (r2 != r5))
		{
			cavium_dbgprint(
				"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
			ret_val = 0;
		}
	}
	else
	{
		CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 0);
		CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 8);
		CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 16);

		if(memcmp(hash_buf, hash_overlap_buf.ptr + hash_overlap_buf.size, SHA_DIGEST_LENGTH))
		{
			cavium_dbgprint(
				"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
			ret_val = 0;
		}
	}
    if(key->ex_mode == EX_MODE_HMAC) {
		CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 0);
		CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 8);
		CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 16);
        memcpy(outp_vec[0].ptr, hash_buf, SHA_DIGEST_LENGTH);
        return 1;
    }

	CVMX_MF_AES_IV(r0, 0);
	CVMX_MF_AES_IV(r1, 1);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r0), key_data, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), key_data, 8);

	CVMX_STOREUNA_INT16(raw_data_len, len, 0);

	return ret_val;
}

/*******************************************************************************
 *                                                                             *
 *    Decrypts input record and generates SSL-3.0 SHA1 HMAC; authenticates     *
 *  record and stores result in output iovec buffer. Initialisation vector is  *
 *  updated after each decryption. Initialisation vector to be used for        *
 *  decrytion is taken from EVP_DESOCT_HMAC_SHA1 structure passed to the       *
 *  function. Key and MAC secret to be used for cipher decryption and HMAC     *
 *  generation respectively are also taken from the key structure passed.      *
 *  Sequence number and header components, record type and length of record,   *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the SSL-3.0 record header.          *
 *  Protocol : SSl-3.0                                                         *
 *  Cipher   : DES-CBC / DES3-CBC                                              *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing decryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to decrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store decryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_des_sha1_dec_ssl30(EVP_DESOCT_HMAC_SHA1 *key,
		struct iovec_cav *inp_vec, uint16_t inp_cnt,
		struct iovec_cav *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t r0, r1, r2;

	uint8_t padarray[2*16] = {0}, hash_buf[128] = {0}, hash_rem_buf[128] = {0}, buf[128] = {0};
	uint8_t des = 0, inner_hash_done = 0, output_overlap = 0;
	uint16_t  cipher_type = 0, enc_data_len = 0, raw_data_len = 0, pad_data_len = 0, hash_input_rem_len = 0;
	uint32_t overlap_buf_size = 0, enc_rem_cur_chunk = 0, inner_hash_total_bytes = 0, outer_hash_total_bytes = 0, prev_iovec_rem_bytes = 0;
	int32_t i = 0;
	uint64_t seq_num = 0, hash_bal = 0, ret_val = 1;
	struct iovec_cav hash_overlap_buf;
	hash_overlap_buf.size = 0;

	uint8_t *out_ptr = NULL, *mac_secret = NULL, *in_chunk_cur = NULL, *out_chunk_cur = NULL, *enc_rem_ptr = NULL;
	uint64_t *enc_key = NULL, *key_data = NULL;
    uint16_t *hash_buf_ptr_16 = NULL;
    uint32_t *hash_buf_ptr_32 = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct ssl3_rec_hdr *in_rec_hdr = NULL;

#ifdef DEBUG
	uint32_t itr = 0;

	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
	for (itr = 0; itr < inp_cnt; itr++)
	{
		cavium_dbgprint("inp_vec[%d].ptr: %p, inp_vec[%d].size: %d\n", itr, inp_vec[itr].ptr, itr, inp_vec[itr].size);
	}

	for (itr = 0; itr < outp_cnt; itr++)
	{
		cavium_dbgprint("outp_vec[%d].ptr: %p, outp_vec[%d].size: %d\n", itr, outp_vec[itr].ptr, itr, outp_vec[itr].size);
	}
#endif

	cipher_type = key->cipher_type;

	in_rec_hdr = (struct ssl3_rec_hdr *)&key->aux.tls_aad_8[8];
	enc_data_len = cvmx_htobe16(in_rec_hdr->length);

#ifdef DEBUG
	cavium_dbgprint("enc_data_len: %d (%d)\n", enc_data_len, __LINE__);
#endif

	if ((enc_data_len % DES_BLOCK_SIZE) && (key->ex_mode == EX_MODE_DEFAULT || key->ex_mode == EX_MODE_ENCRYPT))
	{
		cavium_dbgprint(
			"%s:%d Not properly padded data is received for cipher %s\n",
			__FUNCTION__,__LINE__, cipher_strings[cipher_type]);
		ret_val = 0;
		return ret_val;
	}

	switch(cipher_type)
	{
		case OCT_DES_40:
		case OCT_DES_64:
			des = 1;
			break;

		case OCT_3DES:
			break;
		default:
			break;
	}

	seq_num = cvmx_htobe64(key->aux.tls_aad_64[0]);
	enc_key = (uint64_t *)key->cipher_key;
	mac_secret = key->mac_secret;
	key_data = (uint64_t *)key->cipher_iv;


	/*Load DES key*/
	if(des == 1)
	{
		enc_key[1] = enc_key[0];
		enc_key[2] = enc_key[0];
	}
	CVMX_MT_3DES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enc_key[1]), 1);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enc_key[2]), 2);

	/* Pad data can not be more than DES_BLOCK_SIZE
	 * As per DES CBC implementation, previous aes block result is fed as IV for current des block
	 * So decrypt last 32 bytes to know PAD length*/


	// calculate pad length by decrypting last 16 bytes
	// if the last iovec buffer contains data >= 16
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    if(inp_vec[inp_cnt - 1].size >= 2*DES_BLOCK_SIZE)
	    {
	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 16, 0);

	    	CVMX_MT_3DES_IV(cvmx_htobe64(r0));

	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 8, 0);

	    	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r1));

	    	CVMX_MF_3DES_RESULT(r0);

	    	pad_data_len = (r0 & 0xff ) + 1;
	    }
	    else
	    {
	    	// X + Y = 16
	    	// X bytes from iovec[inp_cnt - 2]
	    	// Y bytes from iovec[inp_cnt - 1]
	    	memcpy(padarray, inp_vec[inp_cnt - 2].ptr + (inp_vec[inp_cnt - 2].size - (16 - inp_vec[inp_cnt - 1].size)), (16 - inp_vec[inp_cnt - 1].size));
	    	memcpy(padarray + (16 - inp_vec[inp_cnt - 1].size), inp_vec[inp_cnt - 1].ptr, inp_vec[inp_cnt - 1].size);
	    	CVMX_LOADUNA_INT64(r0, padarray, 0);
	    	CVMX_LOADUNA_INT64(r1, padarray, 8);

	    	CVMX_MT_3DES_IV(cvmx_htobe64(r0));

	    	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r1));

	    	CVMX_MF_3DES_RESULT(r0);

	    	pad_data_len = (r0 & 0xff ) + 1;
	    }

#ifdef DEBUG
	cavium_dbgprint("pad_data_len: %d (%d)\n", pad_data_len, __LINE__);
#endif

	    if (pad_data_len > DES_BLOCK_SIZE)
	    {
	    	cavium_dbgprint(
	    		"%s:%d Not properly padded data is received for cipher %s\n",
	    		__FUNCTION__,__LINE__, cipher_strings[cipher_type]);
	    	ret_val = 0;
	    	return ret_val;
	    }
    }


	CVMX_MT_3DES_IV(cvmx_htobe64(key_data[0]));

    if(key->ex_mode == EX_MODE_DEFAULT) 
    	raw_data_len = enc_data_len - SHA_DIGEST_LENGTH - pad_data_len;
    else
        raw_data_len = enc_data_len;
#ifdef DEBUG
	cavium_dbgprint("raw_data_len: %d (%d)\n", raw_data_len, __LINE__);
#endif

    if(key->ex_mode == EX_MODE_DIGEST)
	    inner_hash_total_bytes =  raw_data_len;
    else if(key->ex_mode == EX_MODE_HMAC)
	    inner_hash_total_bytes = 20 + 40 + raw_data_len;
    else if(key->ex_mode == EX_MODE_DEFAULT)
	    inner_hash_total_bytes = 20 + 40 + 8 + 1 + 2 + raw_data_len;
	// first chunk - 20+40+4

	CVMX_STOREUNA_INT64(*((uint64_t *)mac_secret), hash_buf, 0);
	CVMX_STOREUNA_INT64(*((uint64_t *)(mac_secret + 8)), hash_buf, 8);
	CVMX_STOREUNA_INT32(*((uint32_t *)(mac_secret + 16)), hash_buf, 16);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 20);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 28);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 36);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 44);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD1_64[0], hash_buf, 52);
    if(key->ex_mode == EX_MODE_DEFAULT)
    	*((uint32_t *)(hash_buf+60)) = cvmx_htobe32((uint32_t)((seq_num&0xFFFFFFFF00000000)>>32));
     else if(key->ex_mode == EX_MODE_HMAC) {
        if(raw_data_len >= 4) {
            CVMX_STOREUNA_INT32(*((uint32_t *)inp_vec[0].ptr), hash_buf, 60);
            inp_vec[0].ptr += 4;
            inp_vec[0].size -= 4;
            raw_data_len -=4;
       } else {
            uint32_t datz = 0;
            memcpy((uint8_t *)&datz, inp_vec[0].ptr, raw_data_len);
            inp_vec[0].ptr += raw_data_len;
            inp_vec[0].size -= raw_data_len;
           (((uint8_t *)&datz)[raw_data_len]) = 0x80;
            raw_data_len = 0;
            CVMX_STOREUNA_INT32(*((uint32_t *)&datz), hash_buf, 60);
       }
    }

	CVMX_MT_HSH_IV( ((uint64_t)CRY_SHA1_INIT0), 0);
	CVMX_MT_HSH_IV( ((uint64_t)CRY_SHA1_INIT1), 1);
	CVMX_MT_HSH_IV( ((uint64_t)CRY_SHA1_INIT2), 2);

    hash_buf_ptr_64 = (uint64_t *)hash_buf;
    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[1]), 1);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[2]), 2);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[3]), 3);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[4]), 4);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[5]), 5);
	    CVMX_MT_HSH_DAT( cvmx_htobe64(hash_buf_ptr_64[6]), 6);

	    CVMX_MT_HSH_STARTSHA( cvmx_htobe64(hash_buf_ptr_64[7]));
    }
    if(key->ex_mode == EX_MODE_DEFAULT) {
        hash_buf_ptr_32 = (uint32_t *)hash_buf;
	    hash_buf_ptr_32[0] =  cvmx_htobe32((uint32_t)(seq_num&0xFFFFFFFF));
	    hash_buf[4] = in_rec_hdr->type;
        hash_buf_ptr_16 = (uint16_t *)&hash_buf[5];
	    hash_buf_ptr_16[0]  = cvmx_htobe16(raw_data_len);
        hash_buf[7] = 0x00;
	    CVMX_LOADUNA_INT64(hash_bal, hash_buf, 0);
        hash_bal = cvmx_htobe64(hash_bal);
	    hash_bal &= bit_start7;
    } else if (key->ex_mode & 0x02) {
        if(raw_data_len > 8) {
            CVMX_LOADUNA_INT64(hash_bal, inp_vec[0].ptr, 0);
            hash_bal = cvmx_htobe64(hash_bal);
            inp_vec[0].ptr += 8;
            inp_vec[0].size -=8;
        } else {
            memcpy(&hash_bal, inp_vec[0].ptr, raw_data_len);
            hash_bal = cvmx_htobe64(hash_bal);
            inp_vec[0].ptr += raw_data_len;
            inp_vec[0].size -= raw_data_len;
        }
    } else
        hash_bal = 0;

#ifdef DEBUG
	cavium_dbgprint("hash bal : %llx\n", hash_bal);
#endif

	hash_input_rem_len = raw_data_len;

	if(inp_cnt == 1)
	{
		in_chunk_cur = inp_vec[0].ptr;
        if(key->ex_mode & 0x02)
            out_chunk_cur = inp_vec[0].ptr;
        else
		    out_chunk_cur = outp_vec[0].ptr;

		while(hash_input_rem_len >= 64)
		{
			hash_bal = block_processing_des_sha1_dec_ssl30(in_chunk_cur, out_chunk_cur, hash_bal,key->ex_mode);

#ifdef DEBUG
			cavium_dbgprint("hash_bal : %016llx \n", hash_bal);
#endif
			in_chunk_cur += 64;
			out_chunk_cur += 64;

			hash_input_rem_len -= 64;
		}
		out_ptr = out_chunk_cur;
#ifdef DEBUG
		cavium_dbgprint("hash_bal : %016llx \n", hash_bal);
#endif
		// remaining data is compulsory a multiple of 8
        if((key->ex_mode == EX_MODE_ENCRYPT && hash_input_rem_len)|| 
            key->ex_mode == EX_MODE_DEFAULT) {
		    do
		    {
		    	CVMX_LOADUNA_INT64(r0, in_chunk_cur, 0);

		    	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r0));

		    	in_chunk_cur += 8;

		    	CVMX_MF_3DES_RESULT(r1);
#ifdef DEBUG
			cavium_dbgprint("r1 : %llx\n", r1);
#endif
		    	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), out_chunk_cur, 0);

		    	out_chunk_cur += 8;

		    }while(((inp_vec[0].ptr + inp_vec[0].size) - in_chunk_cur) > 0);
        }
		// Along with 7 bytes hash balance, hash_input_rem_len 
		// should be given to hash engine as input
		CVMX_STOREUNA_INT64(cvmx_htobe64(hash_bal), hash_rem_buf, 0);

		if(hash_input_rem_len)
		{
            if(key->ex_mode & 0x02) {
                if(hash_input_rem_len > 8)
                    memcpy(hash_rem_buf+8, out_ptr, hash_input_rem_len-8);
            } else if(key->ex_mode == EX_MODE_DEFAULT) {
    			memcpy(hash_rem_buf + 7, out_ptr, hash_input_rem_len);
            }
		}
        if (key->ex_mode == EX_MODE_DEFAULT)
	        hash_rem_buf[7 + hash_input_rem_len] = 0x80;
        else if(key->ex_mode & 0x02) 
    	     hash_rem_buf[ hash_input_rem_len] = 0x80;

		uint64_t * ptr = (uint64_t *)hash_rem_buf;
		if(((hash_input_rem_len + 7) >= 56) ||
            ((key->ex_mode & 0x02) && (hash_input_rem_len >= 56)))
		{
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

			ptr += 8;
		}
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
		CVMX_MT_HSH_STARTSHA(inner_hash_total_bytes * 8);
	}
	else if(inp_cnt > 1)
	{
		hash_overlap_buf.ptr = buf;
		hash_overlap_buf.size = 0;
		iovec_record_process_t *iovec_stat = NULL;
		iovec_stat = malloc(inp_cnt * sizeof(iovec_record_process_t));
		memset(iovec_stat, 0, inp_cnt * sizeof(iovec_record_process_t));

		prev_iovec_rem_bytes = 0;
		// populate loop counters
		while(hash_input_rem_len && (i < inp_cnt))
		{
			if(hash_input_rem_len >= inp_vec[i].size)
			{
				iovec_stat[i].hash_size = inp_vec[i].size;
				hash_input_rem_len -= inp_vec[i].size;
			}
			else
			{
				iovec_stat[i].hash_size = hash_input_rem_len;
				hash_input_rem_len = 0;
			}

			if(prev_iovec_rem_bytes)
			{
				iovec_stat[i].hash_req_bytes = 64 - prev_iovec_rem_bytes;
				// in case of last chunk not holding enough bytes to fill 64 byte block
				if(iovec_stat[i].hash_size < iovec_stat[i].hash_req_bytes)
				{
					iovec_stat[i].hash_req_bytes = iovec_stat[i].hash_size;
				}
			}
			else
			{
				iovec_stat[i].hash_req_bytes = 0;
			}
			iovec_stat[i].crypto_itr = (iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) / DES_BLOCK_SIZE;
			iovec_stat[i].hash_itr = iovec_stat[i].crypto_itr / 8;

			iovec_stat[i].hash_rem_bytes = ((iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) - (iovec_stat[i].hash_itr * 64));
			prev_iovec_rem_bytes = iovec_stat[i].hash_rem_bytes;

			i++;
		}
		uint16_t num_hash_iovec = i;
		// need to update enc_digest_ptr
		if(iovec_stat[i-1].hash_size == inp_vec[i-1].size)
		{
			enc_rem_ptr = inp_vec[i].ptr;
		}
		else
		{
			enc_rem_ptr = inp_vec[i-1].ptr + iovec_stat[i-1].hash_size;
			enc_rem_cur_chunk = inp_vec[i-1].size - iovec_stat[i-1].hash_size;

		}

#ifdef DEBUG
		for(i = 0; i < inp_cnt; i++)
		{
			cavium_dbgprint("Chunk : %d\n", i+1);
			cavium_dbgprint("hash_req_bytes : %d\n",iovec_stat[i].hash_req_bytes);
			cavium_dbgprint("crypto_itr : %d\n", iovec_stat[i].crypto_itr);
			cavium_dbgprint("hash_itr : %d\n", iovec_stat[i].hash_itr);
			cavium_dbgprint("hash_rem_bytes : %d\n",iovec_stat[i].hash_rem_bytes);
		}
#endif

		for(i = 0; i < num_hash_iovec; i++)
		{
			in_chunk_cur = inp_vec[i].ptr;
			out_chunk_cur = outp_vec[i].ptr;

			if(iovec_stat[i].hash_req_bytes)
			{	
				memcpy(hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, 
						in_chunk_cur, 
						iovec_stat[i].hash_req_bytes);
				in_chunk_cur += iovec_stat[i].hash_req_bytes;

				if((iovec_stat[i-1].hash_rem_bytes + iovec_stat[i].hash_req_bytes) < 64)
				{
					hash_overlap_buf.size += iovec_stat[i].hash_req_bytes;
					out_chunk_cur = outp_vec[i-1].ptr + iovec_stat[i-1].hash_req_bytes + (64 * iovec_stat[i-1].hash_itr);
					output_overlap = 1;
					continue;
				}
				else
				{
					hash_bal = block_processing_des_sha1_dec_ssl30(hash_overlap_buf.ptr, hash_overlap_buf.ptr, hash_bal, key->ex_mode);

					memcpy(outp_vec[i-1].ptr + (outp_vec[i-1].size - iovec_stat[i-1].hash_rem_bytes),
							hash_overlap_buf.ptr, 
							iovec_stat[i-1].hash_rem_bytes);
					memcpy(out_chunk_cur, 
							hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, 
							iovec_stat[i].hash_req_bytes);

					out_chunk_cur += iovec_stat[i].hash_req_bytes;
					hash_overlap_buf.size = 0;
				}
			}

			register uint8_t hash_itr = iovec_stat[i].hash_itr;
			while(hash_itr)
			{
				hash_bal = block_processing_des_sha1_dec_ssl30(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);
#ifdef DEBUG
				cavium_dbgprint("hash_bal : %016llx \n", hash_bal);
#endif

				in_chunk_cur += 64;
				out_chunk_cur += 64;

				hash_itr--;
			}

			if(iovec_stat[i].hash_rem_bytes)
			{
				//copy rem bytes to hash overlap buf
				memcpy(hash_overlap_buf.ptr, in_chunk_cur, iovec_stat[i].hash_rem_bytes);
				in_chunk_cur += iovec_stat[i].hash_rem_bytes;
				hash_overlap_buf.size = iovec_stat[i].hash_rem_bytes;
			}
		}

#ifdef DEBUG
		cavium_dbgprint("hash_overlap_buf.size : %d \n", hash_overlap_buf.size);
#endif
		//obviously this case gets hit all the time
		if(inner_hash_done == 0)
		{
			//if(hash_overlap_buf.size)
			{
				uint8_t *ptr = NULL;
				overlap_buf_size = hash_overlap_buf.size;
				memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size, enc_rem_ptr, enc_rem_cur_chunk);
				overlap_buf_size += enc_rem_cur_chunk;
#ifdef DEBUG
				cavium_dbgprint("enc_rem_cur_chunk : %d\n", enc_rem_cur_chunk);
#endif
				if(enc_rem_cur_chunk < (SHA_DIGEST_LENGTH + pad_data_len))
				{
					memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size + enc_rem_cur_chunk, 
							inp_vec[inp_cnt-1].ptr, inp_vec[inp_cnt-1].size);

					overlap_buf_size += inp_vec[inp_cnt-1].size;
#ifdef DEBUG
					cavium_dbgprint("overlap_buf_size : %d\n", overlap_buf_size);
#endif
				}
				ptr = hash_overlap_buf.ptr;
				uint16_t num_itr = 0;
				num_itr = overlap_buf_size / DES_BLOCK_SIZE;
				while(num_itr)
				{
					CVMX_LOADUNA_INT64(r0, ptr, 0);

					CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r0));

					num_itr--;

					CVMX_MF_3DES_RESULT(r1);

#ifdef DEBUG
					cavium_dbgprint("r1 : %llx\n", r1);
#endif
					CVMX_STOREUNA_INT64(cvmx_htobe64(r1), ptr, 0);
					ptr += 8;

				}

				if(hash_overlap_buf.size)
				{
					if(output_overlap)
					{
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, iovec_stat[num_hash_iovec - 2].hash_rem_bytes);
						memcpy(outp_vec[inp_cnt -1].ptr, hash_overlap_buf.ptr + iovec_stat[num_hash_iovec - 2].hash_rem_bytes, iovec_stat[num_hash_iovec - 1].hash_req_bytes);
					}
					else
					{
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, hash_overlap_buf.size);
					}
				}
			}

			memset(hash_rem_buf, 0, 128);
            hash_buf_ptr_64 = (uint64_t *)hash_rem_buf;
			hash_buf_ptr_64[0] = hash_bal;
			memcpy(hash_rem_buf + 7, hash_overlap_buf.ptr, hash_overlap_buf.size);
			hash_rem_buf[hash_overlap_buf.size + 7] = 0x80;
			uint64_t *ptr = (uint64_t *)hash_rem_buf;
#ifdef DEBUG
			cavium_dbgprint("hash rem buf\n");
			_hexdump(hash_rem_buf, 128);
#endif
			if((hash_overlap_buf.size + 7) >= 56)
			{
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
				CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

				ptr += 8;
			}
			inner_hash_total_bytes = 20 + 40 + 8 + 1 + 2 + raw_data_len;
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(inner_hash_total_bytes * 8);

			inner_hash_done = 1;
		}
		free(iovec_stat);
	}

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

	CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 60);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 68);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 76);
#ifdef DEBUG
	cavium_dbgprint("Inner Hash Calculated : %s\n", __func__);
	cavium_dbgprint("r0 : %llx\n", r0);
	cavium_dbgprint("r1 : %llx\n", r1);
	cavium_dbgprint("r2 : %llx\n", r2);
#endif
    
    if(key->ex_mode == EX_MODE_DIGEST) {
        memcpy(outp_vec[0].ptr, hash_buf+60, SHA_DIGEST_LENGTH);
        return 1;
    } else if (key->ex_mode == EX_MODE_ENCRYPT)
        return 1;


	CVMX_STOREUNA_INT64(*((uint64_t *)mac_secret), hash_buf, 0);
	CVMX_STOREUNA_INT64(*((uint64_t *)(mac_secret + 8)), hash_buf, 8);
	CVMX_STOREUNA_INT32(*((uint32_t *)(mac_secret + 16)), hash_buf, 16);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 20);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 28);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 36);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 44);
	CVMX_STOREUNA_INT64(CRY_HSH_PAD2_64[0], hash_buf, 52);
	hash_buf[80] = 0x80;


	//calculate outer hash
	CVMX_MT_HSH_IV(CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV(CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV(CRY_SHA1_INIT2, 2);

	outer_hash_total_bytes = 20 + 40 + 20;

    hash_buf_ptr_64 = (uint64_t *)hash_buf;
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[1]), 1);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[2]), 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[3]), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[4]), 4);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[5]), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[6]), 6);

	CVMX_MT_HSH_STARTSHA(cvmx_htobe64(hash_buf_ptr_64[7]));

	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[8]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[9]), 1);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK1, 2);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 3);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 4);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 5);
	CVMX_MT_HSH_DAT(CRY_HSH_BLK2, 6);

	CVMX_MT_HSH_STARTSHA(outer_hash_total_bytes * 8);

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

	CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 8);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 16);

#ifdef DEBUG
	cavium_dbgprint("Outer Hash Calculated : %s\n", __func__);
	_hexdump(hash_buf, SHA_DIGEST_LENGTH);
#endif

    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr, hash_buf, SHA_DIGEST_LENGTH);
        return 1;
    }
	//Compare calculated HMAC with received HMAC
	if(inp_cnt == 1)
	{
        hash_buf_ptr_64 = (uint64_t *)hash_buf;
        hash_buf_ptr_32 = (uint32_t *)(hash_buf+16);
		if((hash_buf_ptr_64[0] != (((uint64_t *)(out_ptr + hash_input_rem_len))[0])) 		|| 
			(hash_buf_ptr_64[1] != (((uint64_t *)(out_ptr + hash_input_rem_len))[1])) 	|| 
			(hash_buf_ptr_32[0] != (((uint32_t *)(out_ptr + hash_input_rem_len + 16))[0])))
		{
			ret_val = 0;
			cavium_dbgprint(
				"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
			cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
			cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(out_ptr + hash_input_rem_len))[0]));
			cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(out_ptr + hash_input_rem_len))[1]));
			cavium_dbgprint("3. :%x  : %x\n", ((uint32_t *)(hash_buf + 16))[0], (((uint32_t *)(out_ptr + hash_input_rem_len + 16))[0]));
#endif
		}
	}
	else
	{
		if(memcmp(hash_buf, hash_overlap_buf.ptr + hash_overlap_buf.size, SHA_DIGEST_LENGTH))
		{
			ret_val = 0;
			cavium_dbgprint(
				"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
			cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
			cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[0]));
			cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[1]));
			cavium_dbgprint("3. :%x  : %x\n", ((uint32_t *)(hash_buf + 16))[0], (((uint32_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size + 16))[0]));
#endif
		}
	}


	CVMX_MF_3DES_IV(key_data[0]);
    key_data[0] = cvmx_htobe64(key_data[0]);


	*len = raw_data_len;
	return ret_val;
}


/* tls 1.0 encrypt and decrypt functions */

/*******************************************************************************
 *                                                                             *
 *      Generates TLS-1.0 SHA1 HMAC, adds padding bytes and  encrypts payload  *
 *  data and stores the result in output iovec buffer. Initialisation vector   * 
 *  be used for encrytion is taken from EVP_AESOCT_HMAC_SHA1 structure passed  *
 *  to the function. Initialisation vector is updated after each encryption.   *
 *  Key and MAC secret to be used for cipher encryption and HMAC               *
 *  generation respectively are also taken from the key structure passed.      *
 *  Sequence number and record header to be used for used for HMAC generation  *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the TLS-1.0 record header.          *
 *  Protocol : TLS-1.0                                                         *
 *  Cipher   : AES128-CBC / AES256-CBC                                         *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing encryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to encrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store encryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_aes_sha1_enc_tls10(EVP_AESOCT_HMAC_SHA1 *key,
		struct iovec_cav *inp_vec, uint16_t inp_cnt,
		struct iovec_cav *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t p0, p1, p2, p3, p4, p5, r0, r1;
	uint8_t sha1_rem[128] = {0};
	uint8_t aes_rem[128] = {0};
	uint8_t aes_out[128] = {0};
	uint8_t idx = 0;
	uint16_t inp_len = 0, cipher_type = 0;
	uint32_t aes_rem_num = 0, sha1_rem_num = 0;
	int32_t aes_out_len = 0, keylength = 1, eout_space_left = 0, aes_copied_bytes = 0, sha1_copied_bytes = 0;
	int32_t fill_bytes = 0, next_output_iovec = 0, start_inner_hash = 0;
	uint64_t tmpvar = 0, tmpvar2 = 0, total_bits = 0, seq = cvmx_htobe64(key->aux.tls_aad_64[0]);
      	uint8_t *mac_secret = NULL, *key_data = NULL, *hash_in = NULL, *hash_end = NULL, *enc_in = NULL, *enc_end = NULL, *enc_out = NULL;
	uint64_t *ptr = NULL, *enc_key = (uint64_t *)key->cipher_key;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct tls_rec_hdr *in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];

#ifdef DEBUG
	cavium_dbgprint("rec length: %d, seq: %llx\n", in_rec_hdr->length, seq);
	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
#endif
	mac_secret = key->mac_secret;

	cipher_type = key->cipher_type;
	switch(cipher_type)
	{
		case OCT_AES_128:
			keylength = 1;
			break;

		case OCT_AES_256:
			keylength = 3;
			break;
		default:
			break;
	}


	key_data = key->cipher_iv;


	/* assume AES keylen 128 bit */
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[1]), 1);
	if(keylength == 3)
	{
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[2]), 2);
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[3]), 3);
	}
	else
	{
		CVMX_MT_AES_KEY((uint64_t)0, 2);
		CVMX_MT_AES_KEY((uint64_t)0, 3);
	}
	CVMX_MT_AES_KEYLENGTH(keylength);
	
	/* Init AES iv */
	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)key_data), 0);
	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)(&key_data[8])), 1);

	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

	CVMX_LOADUNA_INT64(r0, mac_secret, 0);
	CVMX_LOADUNA_INT64(r1, mac_secret, 8);
	
	r0 = r0 ^ (0x3636363636363636);
	r1 = r1 ^ (0x3636363636363636);

    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);

	    r0 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	    r0 = r0 << 32;
	    r0 = r0 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(r0, 2);

	    CVMX_LOADUNA_INT64(r1, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(r1, 3);
	    CVMX_MT_HSH_DAT(r1, 4);
	    CVMX_MT_HSH_DAT(r1, 5);
	    CVMX_MT_HSH_DAT(r1, 6);
	    CVMX_MT_HSH_STARTSHA(r1);
    }
	CVMX_PREFETCH0(inp_vec[idx].ptr);

	inp_len = cvmx_htobe16(in_rec_hdr->length);
    if(key->ex_mode == EX_MODE_DEFAULT) 
    	total_bits = ((SHA_CBLOCK + 8 + 1 + 2 + 2 + inp_len) * 8);
    else if (key->ex_mode == EX_MODE_HMAC)
    	total_bits = ((SHA_CBLOCK + inp_len) * 8);
    else if (key->ex_mode == EX_MODE_DIGEST)
    	total_bits = (inp_len * 8);

	/*store 8 bytes seq_num , 1 byte record type, 2 bytes record version, 2 bytes record length*/
    if(key->ex_mode == EX_MODE_DEFAULT ) {
        hash_buf_ptr_64 = (uint64_t *)sha1_rem;
        hash_buf_ptr_64[0] = cvmx_htobe64(seq);
	    CVMX_STOREUNA_INT32(key->aux.tls_aad_32[2], sha1_rem, 8); //12
	    sha1_rem[12] = key->aux.tls_aad_8[12]; //13
	    sha1_copied_bytes = 13; /* seq num (8 bytes) + record type (1 byte) + record version (2 bytes) + record length (2 bytes) */
    } else if(key->ex_mode & 0x02) {
        if(inp_len >= 16) {
            memcpy(sha1_rem, inp_vec[0].ptr,16);
            inp_vec[0].ptr += 16;
            inp_vec[0].size -= 16;
            sha1_copied_bytes = 16;
        } else {
            memcpy(sha1_rem, inp_vec[0].ptr, inp_len);
            inp_vec[0].ptr += inp_len;
            inp_vec[0].size -= inp_len;
            sha1_copied_bytes = inp_len;
        }
    }

	aes_copied_bytes = 0;

	hash_in = inp_vec[idx].ptr;
	hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
	sha1_rem_num = (hash_end - hash_in) + sha1_copied_bytes;

	enc_in = inp_vec[idx].ptr;
	enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
	aes_rem_num = (enc_end - enc_in);

	enc_out = outp_vec[idx].ptr ;

	if (sha1_rem_num <= SHA_CBLOCK || key->ex_mode == EX_MODE_ENCRYPT) {
		ptr = (uint64_t *)(sha1_rem + sha1_copied_bytes);	
		CVMX_STOREUNA_INT64(0, ptr, 0);
		CVMX_STOREUNA_INT64(0, ptr, 8);
		CVMX_STOREUNA_INT64(0, ptr, 16);
		CVMX_STOREUNA_INT64(0, ptr, 24);
		CVMX_STOREUNA_INT64(0, ptr, 32);
		CVMX_STOREUNA_INT64(0, ptr, 40);
		CVMX_STOREUNA_INT64(0, ptr, 48);
		CVMX_STOREUNA_INT64(0, ptr, 56);

		memcpy(sha1_rem + sha1_copied_bytes, hash_in, inp_vec[idx].size);	
		sha1_copied_bytes += (inp_vec[idx].size);
		sha1_rem[sha1_copied_bytes] = 0x80;

        if(key->ex_mode != EX_MODE_ENCRYPT)
    		SHA1_FINISH(sha1_rem, sha1_copied_bytes, total_bits);
		sha1_copied_bytes = 0;
		sha1_rem_num = 0;

		while (aes_rem_num >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	
			CVMX_LOADUNA_INT64(p1, enc_in, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
			
			enc_in += AES_BLOCK_SIZE;
			aes_rem_num -= AES_BLOCK_SIZE;

			CVMX_MF_AES_RESULT(p0, 0);	
			CVMX_MF_AES_RESULT(p1, 1);	

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

			enc_out += AES_BLOCK_SIZE;
		}

		memcpy(aes_rem, enc_in, aes_rem_num);
		aes_copied_bytes = aes_rem_num;

		goto get_inner_hash;
	}

    if(key->ex_mode == EX_MODE_DEFAULT) {
	    /* inp_vec[0] has payload(inp_vec[0].size - sizeof(struct tls_rec_hdr)) > 51 bytes */
	    /*sha1 copied bytes = 13, so pushing 3 more bytes from hash_in to sha1_rem*/
	    sha1_rem[sha1_copied_bytes] = hash_in[0];
	    sha1_rem[sha1_copied_bytes + 1] = hash_in[1];
	    sha1_rem[sha1_copied_bytes + 2] = hash_in[2];
	    ptr = (uint64_t *)(hash_in + 3);
    } else {
        ptr = (uint64_t * )hash_in;
    }
	CVMX_LOADUNA_INT64(p0, ptr, 0);
	CVMX_LOADUNA_INT64(p1, ptr, 8);
	CVMX_LOADUNA_INT64(p2, ptr, 16);
	CVMX_LOADUNA_INT64(p3, ptr, 24);
	CVMX_LOADUNA_INT64(p4, ptr, 32);
	
    hash_buf_ptr_64 = (uint64_t *)sha1_rem;
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[1]), 1);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 4);
       	CVMX_LOADUNA_INT64(p0, ptr, 40);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 6);

	CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p0));

    if(key->ex_mode == EX_MODE_DEFAULT)
    	hash_in += 51;
    else
        hash_in += 48;
	sha1_rem_num = (hash_end - hash_in);
	sha1_copied_bytes = 0;
	
	/* SHA_CBLOCK chunks in 1st iovec */

	while ((hash_in + SHA_CBLOCK) <= hash_end) {
		/* load 1st aes bytes */
		CVMX_LOADUNA_INT64(p0, enc_in, 0);	
		CVMX_LOADUNA_INT64(p1, enc_in, 8);
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 0);
		CVMX_LOADUNA_INT64(p1, hash_in, 8);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		
		/* get 2nd aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 16);
		CVMX_LOADUNA_INT64(p1, enc_in, 24);

		/* 1st aes result */
		CVMX_MF_AES_RESULT(p2, 0);	
		CVMX_MF_AES_RESULT(p3, 1);	

		/* load 2nd aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 16);
		CVMX_LOADUNA_INT64(p1, hash_in, 24);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

		/* store 1st aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 0);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

		/* get 3rd aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 32);
		CVMX_LOADUNA_INT64(p1, enc_in, 40);
		
		/* 2nd aes result */
		CVMX_MF_AES_RESULT(p2, 0);	
		CVMX_MF_AES_RESULT(p3, 1);	

		/* load 3rd aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 32);
		CVMX_LOADUNA_INT64(p1, hash_in, 40);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

		/* store 2nd aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 16);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

		/* get 4th aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 48);
		CVMX_LOADUNA_INT64(p1, enc_in, 56);
		
		/* get 3rd aes result */
		CVMX_MF_AES_RESULT(p2, 0);
		CVMX_MF_AES_RESULT(p3, 1);

		/* load 4th aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 48);
		CVMX_LOADUNA_INT64(p1, hash_in, 56);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		hash_in += SHA_CBLOCK;
		enc_in += SHA_CBLOCK;
		sha1_rem_num -= SHA_CBLOCK;
		aes_rem_num -= SHA_CBLOCK;

		/* store 3rd aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 32);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

		/* get 4th aes result */
		CVMX_MF_AES_RESULT(p2, 0);
		CVMX_MF_AES_RESULT(p3, 1);
			
		CVMX_PREFETCH(enc_in, SHA_CBLOCK);

		/* store 4th aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 48);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);
		
		enc_out += SHA_CBLOCK;
	}

	/* for each iovec */
	while (idx++ < inp_cnt) {
		/* for each SHA_CBLOCK in iovec */
		while ((hash_in + SHA_CBLOCK) <= hash_end) {
			/* load 1st aes bytes */
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	
			CVMX_LOADUNA_INT64(p1, enc_in, 8);
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 0);
			CVMX_LOADUNA_INT64(p1, hash_in, 8);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		
			/* get 2nd aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 16);
			CVMX_LOADUNA_INT64(p1, enc_in, 24);

			/* 1st aes result */
			CVMX_MF_AES_RESULT(p2, 0);	
			CVMX_MF_AES_RESULT(p3, 1);	

			/* load 2nd aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 16);
			CVMX_LOADUNA_INT64(p1, hash_in, 24);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

			/* store 1st aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

			/* get 3rd aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 32);
			CVMX_LOADUNA_INT64(p1, enc_in, 40);
		
			/* 2nd aes result */
			CVMX_MF_AES_RESULT(p2, 0);	
			CVMX_MF_AES_RESULT(p3, 1);	

			/* load 3rd aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 32);
			CVMX_LOADUNA_INT64(p1, hash_in, 40);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

			/* store 2nd aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 16);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

			/* get 4th aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 48);
			CVMX_LOADUNA_INT64(p1, enc_in, 56);
		
			/* get 3rd aes result */
			CVMX_MF_AES_RESULT(p2, 0);
			CVMX_MF_AES_RESULT(p3, 1);

			/* load 4th aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 48);
			CVMX_LOADUNA_INT64(p1, hash_in, 56);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

			hash_in += SHA_CBLOCK;
			enc_in += SHA_CBLOCK;
			sha1_rem_num -= SHA_CBLOCK;
			aes_rem_num -= SHA_CBLOCK;

			/* store 3rd aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 32);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

			/* get 4th aes result */
			CVMX_MF_AES_RESULT(p2, 0);
			CVMX_MF_AES_RESULT(p3, 1);
			
			CVMX_PREFETCH(enc_in, SHA_CBLOCK);

			/* store 4th aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 48);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);
		
			enc_out += SHA_CBLOCK;
		} /* for each SHA_CBLOCK in iovec */

		sha1_rem_num = (hash_end - hash_in);
		aes_rem_num = (hash_end - enc_in);

#ifdef DEBUG
		cavium_dbgprint("sha1_rem_num: %d, aes_rem_num: %d (%d)\n", sha1_rem_num, aes_rem_num, __LINE__);
#endif
		/* copy the remaining sha1 bytes */
		memcpy(sha1_rem, hash_in, sha1_rem_num);
		sha1_copied_bytes = sha1_rem_num;
		sha1_rem_num = 0;

		if (idx < inp_cnt) {
			/* try fill the sha1 block, get bytes from next iovec */
			hash_in = inp_vec[idx].ptr;
			hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
			sha1_rem_num = (hash_end - hash_in);
		}
#ifdef DEBUG
		cavium_dbgprint("sha1_rem_num: %d, aes_rem_num: %d (%d)\n", sha1_rem_num, aes_rem_num, __LINE__);
#endif

		/* less bytes than needed, try to finish inner hash */
		if (sha1_rem_num < (SHA_CBLOCK - sha1_copied_bytes)) {
			memcpy(sha1_rem + sha1_copied_bytes, hash_in, sha1_rem_num);
			sha1_copied_bytes += sha1_rem_num;
			sha1_rem[sha1_copied_bytes] = 0x80;
			
			memset(sha1_rem + sha1_copied_bytes + 1, 0, (SHA_CBLOCK - sha1_copied_bytes - 1));

			SHA1_FINISH(sha1_rem, sha1_copied_bytes, total_bits);
			sha1_rem_num = 0;
			sha1_copied_bytes = 0;

			/* complete aes bytes in previous iovec */
			while (aes_rem_num >= AES_BLOCK_SIZE) {
				CVMX_LOADUNA_INT64(p0, enc_in, 0);
				CVMX_LOADUNA_INT64(p1, enc_in, 8);

				CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
				CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

				enc_in += AES_BLOCK_SIZE;
				aes_rem_num -= AES_BLOCK_SIZE;

				CVMX_MF_AES_RESULT(p0, 0);
				CVMX_MF_AES_RESULT(p1, 1);

				CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
				CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

				enc_out += AES_BLOCK_SIZE;
			}

			if (aes_rem_num) {
				/* copy the remaining aes bytes from previous iovec */
				memcpy(aes_rem, enc_in, aes_rem_num);
				aes_copied_bytes = aes_rem_num;
				eout_space_left = aes_rem_num;
				aes_rem_num = 0;

				if (idx < inp_cnt) {
					/* move to next iovec */
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					aes_rem_num = (enc_end - enc_in);
				}

				/* fill bytes from next iovec */
				if (aes_rem_num <= (AES_BLOCK_SIZE - aes_copied_bytes)) {
					memcpy(aes_rem + aes_copied_bytes, enc_in, aes_rem_num);
					aes_copied_bytes += aes_rem_num;
					aes_rem_num = 0;
					if (idx < outp_cnt)
						next_output_iovec = 1;
				} else {
					fill_bytes = (AES_BLOCK_SIZE - aes_copied_bytes);
					uint64_t enc_result[2];
					memcpy(aes_rem + aes_copied_bytes, enc_in, fill_bytes);
					aes_copied_bytes = AES_BLOCK_SIZE;
					aes_rem_num -= fill_bytes;

					CVMX_LOADUNA_INT64(p0, aes_rem, 0);
					CVMX_LOADUNA_INT64(p1, aes_rem, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

					enc_in += fill_bytes;

					CVMX_MF_AES_RESULT(enc_result[0], 0);
					CVMX_MF_AES_RESULT(enc_result[1], 1);

                    enc_result[0] = cvmx_htobe64(enc_result[0]);
                    enc_result[1] = cvmx_htobe64(enc_result[1]);
					/* copy into the space left */
					memcpy(enc_out, enc_result, eout_space_left);	

					/* move to next output iovec */
					enc_out = outp_vec[idx].ptr;
					memcpy(enc_out, ((uint8_t *)enc_result + eout_space_left), fill_bytes);
					enc_out += fill_bytes;
					eout_space_left = 0;
					aes_copied_bytes = 0;

					/* encrypt AES_BLOCK_SIZE chunks if any */
					while (aes_rem_num >= AES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);
						CVMX_LOADUNA_INT64(p1, enc_in, 8);

						CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
						CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
						enc_in += AES_BLOCK_SIZE;
						aes_rem_num -= AES_BLOCK_SIZE;

						CVMX_MF_AES_RESULT(p0, 0);
						CVMX_MF_AES_RESULT(p1, 1);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
						CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
						enc_out += AES_BLOCK_SIZE;
					}
					memcpy(aes_rem, enc_in, aes_rem_num);
					aes_copied_bytes = aes_rem_num;
				}
			} else {
				if (idx < inp_cnt) {
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					aes_rem_num = (enc_end - enc_in);
				}
			
				if (idx < outp_cnt)
					enc_out = outp_vec[idx].ptr;

				while (aes_rem_num >= AES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);
					CVMX_LOADUNA_INT64(p1, enc_in, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
					enc_in += AES_BLOCK_SIZE;
					aes_rem_num -= AES_BLOCK_SIZE;

					CVMX_MF_AES_RESULT(p0, 0);
					CVMX_MF_AES_RESULT(p1, 1);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
					CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
					enc_out += AES_BLOCK_SIZE;
				}
				if (enc_in) {
					memcpy(aes_rem, enc_in, aes_rem_num);	
					aes_copied_bytes = aes_rem_num;
					aes_rem_num = 0;
				}
			}
			goto get_inner_hash;
		}

		/* more than fill bytes in next iovec */
		fill_bytes = SHA_CBLOCK - sha1_copied_bytes;
		memcpy(sha1_rem + sha1_copied_bytes, hash_in, fill_bytes);
		sha1_copied_bytes += fill_bytes;

		hash_in += fill_bytes;
		sha1_rem_num -= fill_bytes;

		ptr = (uint64_t *)sha1_rem;
		CVMX_LOADUNA_INT64(p0, ptr, 0);
		CVMX_LOADUNA_INT64(p1, ptr, 8);
		CVMX_LOADUNA_INT64(p2, ptr, 16);
		CVMX_LOADUNA_INT64(p3, ptr, 24);
		CVMX_LOADUNA_INT64(p4, ptr, 32);
		CVMX_LOADUNA_INT64(p5, ptr, 40);
	
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
		CVMX_LOADUNA_INT64(p0, ptr, 48);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_LOADUNA_INT64(p1, ptr, 56);
		
		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));
		
		while (aes_rem_num >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);
			CVMX_LOADUNA_INT64(p1, enc_in, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			enc_in += AES_BLOCK_SIZE;
			aes_rem_num -= AES_BLOCK_SIZE;

			CVMX_MF_AES_RESULT(p0, 0);
			CVMX_MF_AES_RESULT(p1, 1);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

			enc_out += AES_BLOCK_SIZE;
		}

		/* if sha1 remain bytes are less than SHA_CBLOCK, finish inner hash */
		if (sha1_rem_num < SHA_CBLOCK) {
			memcpy(sha1_rem, hash_in, sha1_rem_num);	
			sha1_copied_bytes = sha1_rem_num;
			sha1_rem[sha1_copied_bytes] = 0x80;

			memset(sha1_rem + sha1_copied_bytes + 1, 0, (SHA_CBLOCK - sha1_copied_bytes - 1));

			SHA1_FINISH(sha1_rem, sha1_copied_bytes, total_bits);
			sha1_rem_num = 0;
			sha1_copied_bytes = 0;

			start_inner_hash = 1;
		}

		if (aes_rem_num) {
			uint64_t enc_result[2];

			memcpy(aes_rem, enc_in, aes_rem_num);
			aes_copied_bytes = aes_rem_num;
			eout_space_left = aes_rem_num;

			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;	
			aes_rem_num = (enc_end - enc_in);

			if (aes_rem_num <= (AES_BLOCK_SIZE - aes_copied_bytes)) {
				/* copy the remaining aes bytes */
				memcpy(aes_rem + aes_copied_bytes, enc_in, aes_rem_num);
				aes_copied_bytes += aes_rem_num;
				aes_rem_num = 0;
				if (idx < outp_cnt)
					next_output_iovec = 1;

				goto get_inner_hash;
			} else {
				fill_bytes = (AES_BLOCK_SIZE - aes_copied_bytes);

				/* chance that aes bytes are crossing sha1 bytes */
				if ((aes_rem_num - fill_bytes) < sha1_rem_num) {
					ptr = (uint64_t *)hash_in;	
					CVMX_LOADUNA_INT64(p0, ptr, 0);
					CVMX_LOADUNA_INT64(p1, ptr, 8);
					CVMX_LOADUNA_INT64(p2, ptr, 16);
					CVMX_LOADUNA_INT64(p3, ptr, 24);
					CVMX_LOADUNA_INT64(p4, ptr, 32);
					CVMX_LOADUNA_INT64(p5, ptr, 40);
	
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
					CVMX_LOADUNA_INT64(p0, ptr, 48);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
					CVMX_LOADUNA_INT64(p1, ptr, 56);
		
					CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

					hash_in += SHA_CBLOCK;
					sha1_rem_num -= SHA_CBLOCK;
				}

				memcpy(aes_rem + aes_copied_bytes, enc_in, fill_bytes); 
				aes_copied_bytes += fill_bytes;

				CVMX_LOADUNA_INT64(p0, aes_rem, 0);
				CVMX_LOADUNA_INT64(p1, aes_rem, 8);

				CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
				CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

				enc_in += fill_bytes;
				aes_rem_num -= fill_bytes;

				CVMX_MF_AES_RESULT(enc_result[0], 0);
				CVMX_MF_AES_RESULT(enc_result[1], 1);
                enc_result[0] = cvmx_htobe64(enc_result[0]);
                enc_result[1] = cvmx_htobe64(enc_result[1]);
			
				memcpy(enc_out, enc_result, eout_space_left);

				/* move to next output iovec */
				enc_out = outp_vec[idx].ptr;
				memcpy(enc_out, ((uint8_t *)enc_result + eout_space_left), fill_bytes);
				enc_out += fill_bytes;

				eout_space_left = 0;
				aes_copied_bytes = 0;

				if (start_inner_hash) {
					/* encrypt AES_BLOCK_SIZE chunks if any */
					while (aes_rem_num >= AES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);
						CVMX_LOADUNA_INT64(p1, enc_in, 8);

						CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
						CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
						enc_in += AES_BLOCK_SIZE;
						aes_rem_num -= AES_BLOCK_SIZE;

						CVMX_MF_AES_RESULT(p0, 0);
						CVMX_MF_AES_RESULT(p1, 1);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
						CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
						enc_out += AES_BLOCK_SIZE;
					}
					memcpy(aes_rem, enc_in, aes_rem_num);
					aes_copied_bytes = aes_rem_num;

					goto get_inner_hash;
				}
			}
		} else {
			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
			aes_rem_num = (enc_end - enc_in);

			if (idx < outp_cnt)
				enc_out = outp_vec[idx].ptr;

			if (start_inner_hash) {
				while (aes_rem_num >= AES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);
					CVMX_LOADUNA_INT64(p1, enc_in, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
					enc_in += AES_BLOCK_SIZE;
					aes_rem_num -= AES_BLOCK_SIZE;

					CVMX_MF_AES_RESULT(p0, 0);
					CVMX_MF_AES_RESULT(p1, 1);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
					CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
					enc_out += AES_BLOCK_SIZE;
				}

				if (enc_in) {
					memcpy(aes_rem, enc_in, aes_rem_num);
					aes_copied_bytes = aes_rem_num;
					aes_rem_num = 0;
				}

				goto get_inner_hash;
			}
		}
	} /* for each iovec */

get_inner_hash:
	/* get inner hash */
	CVMX_MF_HSH_IV(tmpvar, 0);
	CVMX_MF_HSH_IV(tmpvar2, 1);
	CVMX_MF_HSH_IV(r0, 2);

    if(key->ex_mode == EX_MODE_ENCRYPT) {
        return 1;
    } else if(key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar), outp_vec[0].ptr, 0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar2),outp_vec[0].ptr, 8);
        CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(r0)),outp_vec[0].ptr,16);
        return 1;
    }
#ifdef DEBUG
	cavium_dbgprint("sha1_copied_bytes: %d, sha1_rem_num: %d (%d)\n", sha1_copied_bytes, sha1_rem_num, __LINE__);
	_hexdump(sha1_rem, sha1_copied_bytes);
	cavium_dbgprint("aes_copied_bytes: %d, aes_rem_num: %d (%d)\n", aes_copied_bytes, aes_rem_num, __LINE__);
	_hexdump(aes_rem, aes_copied_bytes);;
	cavium_dbgprint("Inner Hash\n");
	cavium_dbgprint("%llx\n", tmpvar);
	cavium_dbgprint("%llx\n", tmpvar2);
	cavium_dbgprint("%llx\n", r0);
#endif

	/* start outer hash ((mac secret ^ pad_2) + inner hash) */


	/* SHA init */
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

			CVMX_LOADUNA_INT64(r1, mac_secret, 0);
			r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
			CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 0);
			CVMX_LOADUNA_INT64(r1, mac_secret, 8);
			r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
			CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);

			r1 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
			r1 = r1 << 32;
			r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
			CVMX_MT_HSH_DAT(r1, 2);
			
			CVMX_LOADUNA_INT64(r1, CRY_HMAC_PAD2_64, 0);
			CVMX_MT_HSH_DAT(r1, 3);
			CVMX_MT_HSH_DAT(r1, 4);
			CVMX_MT_HSH_DAT(r1, 5);
			CVMX_MT_HSH_DAT(r1, 6);
			CVMX_MT_HSH_STARTSHA(r1);


	ptr = (uint64_t *)&aes_rem[aes_copied_bytes];


	CVMX_MT_HSH_DAT(tmpvar, 0);
	CVMX_MT_HSH_DAT(tmpvar2, 1);
	r0 = r0 & 0xFFFFFFFF00000000ULL;
	r0 = r0 | 0x0000000080000000ULL;

	CVMX_MT_HSH_DAT(r0, 2);
	CVMX_MT_HSH_DATZ(3);
	CVMX_MT_HSH_DATZ(4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	total_bits = ((SHA_CBLOCK + SHA_DIGEST_LENGTH) * 8);
	CVMX_MT_HSH_STARTSHA(total_bits);
      
	aes_copied_bytes += SHA_DIGEST_LENGTH;

	/* calculate pad length */
      	tmpvar = AES_BLOCK_SIZE - ((aes_copied_bytes) % AES_BLOCK_SIZE);
	
	/* add pad bytes pattren (padlen - 1) */
      	memset(aes_rem + aes_copied_bytes, tmpvar - 1, tmpvar);
	aes_copied_bytes += tmpvar;

	/* get outer hash */
	CVMX_MF_HSH_IV(p0, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p0), ptr, 0);
	CVMX_MF_HSH_IV(p1, 1);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p1), ptr, 8);
	CVMX_MF_HSH_IV(r0, 2);
	CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(r0)), ptr, 16);
#ifdef DEBUG
	cavium_dbgprint("Outer Hash\n");
	cavium_dbgprint("%llx\n", p0);
	cavium_dbgprint("%llx\n", p1);
    cavium_dbgprint("%llx\n", r0);
#endif
    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr,ptr,SHA_DIGEST_LENGTH);
        return 1;
    }
	if (aes_copied_bytes) {
		uint8_t *out_aes_ptr = aes_out;
		uint8_t *in_aes_ptr = aes_rem;
		aes_out_len = aes_copied_bytes;

		while (aes_copied_bytes >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, in_aes_ptr, 0);
			CVMX_LOADUNA_INT64(p1, in_aes_ptr, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			in_aes_ptr += AES_BLOCK_SIZE;
			aes_copied_bytes -= AES_BLOCK_SIZE;	

			CVMX_MF_AES_RESULT(p0, 0);
			CVMX_MF_AES_RESULT(p1, 1);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), out_aes_ptr, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), out_aes_ptr, 8);
			
			out_aes_ptr += AES_BLOCK_SIZE;
		}
		if (aes_copied_bytes)
			cavium_dbgprint("ERROR: aes_copied_bytes: %d\n", aes_copied_bytes);
	}	

	memcpy(enc_out, aes_out, eout_space_left);

	if (next_output_iovec)
		enc_out = outp_vec[idx].ptr;			
	else
		enc_out += eout_space_left;

	aes_rem_num = aes_out_len - eout_space_left;

	if (enc_out)
		memcpy(enc_out, aes_out + eout_space_left, aes_rem_num);

      CVMX_MF_AES_IV(*(uint64_t *)key_data, 0);
      CVMX_MF_AES_IV(*(uint64_t *)(&key_data[8]), 1);
      *((uint64_t *)key_data) = cvmx_htobe64(*((uint64_t *)key_data));
      *((uint64_t *)&key_data[8]) = cvmx_htobe64(*((uint64_t *)&key_data[8]));


	
#ifdef DEBUG	
        _hexdump(outp_vec[0].ptr, cal_enc_len);
#endif

	return 1;
}


/*******************************************************************************
 *                                                                             *
 *      Generates TLS-1.0 SHA1 HMAC, adds padding bytes and  encrypts payload  *
 *  data and stores the result in output iovec buffer. Initialisation vector   * 
 *  be used for encrytion is taken from EVP_DESOCT_HMAC_SHA1 structure passed  *
 *  to the function. Initialisation vector is updated after each encryption.   *
 *  Key and MAC secret to be used for cipher encryption and HMAC               *
 *  generation respectively are also taken from the key structure passed.      *
 *  Sequence number and record header to be used for used for HMAC generation  *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the TLS-1.0 record header.          *
 *  Protocol : TLS-1.0                                                         *
 *  Cipher   : DES- CBC / DES3-CBC                                             *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing encryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to encrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store encryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_des_sha1_enc_tls10(EVP_DESOCT_HMAC_SHA1 *key,
		struct iovec_cav *inp_vec, uint16_t inp_cnt,
		struct iovec_cav *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t p0, p1, p2, p3, p4, p5;
	uint8_t hash_buf[128] = {0}, enc_buf[128] = {0}, final_buf[128] = {0};
	uint16_t inp_len = 0;
	uint64_t tmpvar = 0, tmpvar2 = 0, seq = 0, total_bits = 0;
	int32_t enc_rem_num = 0, hash_rem_num = 0,enc_copied_bytes = 0, hash_copied_bytes = 0, eout_space_left = 0;
	int32_t idx = 0, fill_bytes = 0, padlen = 0, next_output_iovec = 0, start_inner_hash = 0;

	uint8_t *mac_secret = NULL, *key_data = NULL, *hash_in = NULL, *hash_end = NULL, *enc_in = NULL, *enc_end = NULL, *enc_out = NULL;
	uint64_t *enckey = NULL, *ptr = NULL;
	struct tls_rec_hdr *in_rec_hdr = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];
	inp_len = cvmx_htobe16(in_rec_hdr->length);

	mac_secret = key->mac_secret;
	seq = cvmx_htobe64(key->aux.tls_aad_64[0]);

#ifdef DEBUG
	cavium_dbgprint("rec length: %d, seq: %llx\n", in_rec_hdr->length, seq);
	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
#endif

	key_data = key->cipher_iv;


	enckey = (uint64_t *)key->cipher_key;

#ifdef DEBUG
	cavium_dbgprint("enckey\n");
	_hexdump(enckey, 24);
	cavium_dbgprint("iv\n");
	_hexdump(key_data, 8);
#endif
	if(key->cipher_type == OCT_DES_64 || key->cipher_type == OCT_DES_40)
	{
		enckey[1] = enckey[0];
		enckey[2] = enckey[0];
	}
	/* initialize encryption key */
	CVMX_MT_3DES_KEY(cvmx_htobe64(enckey[0]), 0);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enckey[1]), 1);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enckey[2]), 2);

	/* Init encryption iv */
	CVMX_MT_3DES_IV(cvmx_htobe64(*(uint64_t *)key_data));

	/* The first hash (mac secret ^ pad_1 (64 bytes) */
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT0), 0);
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT1), 1);
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT2), 2);

	CVMX_LOADUNA_INT64(p0, mac_secret, 0);
	CVMX_LOADUNA_INT64(p1, mac_secret, 8);
	p0 = p0 ^ (0x3636363636363636);
	p1 = p1 ^ (0x3636363636363636);

    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

	    p0 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	    p0 = p0 << 32;
	    p0 = p0 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(p0, 2);

	    CVMX_LOADUNA_INT64(p1, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(p1, 3);
	    CVMX_MT_HSH_DAT(p1, 4);
	    CVMX_MT_HSH_DAT(p1, 5);
	    CVMX_MT_HSH_DAT(p1, 6);
	    CVMX_MT_HSH_STARTSHA(p1);
    }
	CVMX_PREFETCH0(inp_vec[idx].ptr);

    if(key->ex_mode == EX_MODE_DEFAULT)
    	total_bits = ((SHA_CBLOCK + 8 + 1 + 2 + 2 + inp_len) * 8); 
    else if (key->ex_mode == EX_MODE_HMAC)
    	total_bits = ((SHA_CBLOCK + inp_len) * 8); 
    else if (key->ex_mode == EX_MODE_DIGEST)
    	total_bits = inp_len * 8; 

	/* store sequence number + rec_type + rec_len */
    if(key->ex_mode & 0x02) {
        if(inp_len > 16) {
            memcpy(hash_buf,inp_vec[0].ptr,16);
            inp_vec[0].ptr += 16;
            inp_vec[0].size -= 16;
            hash_copied_bytes = 16;
        } else {
            memcpy(hash_buf,inp_vec[0].ptr,inp_len);
            inp_vec[0].ptr += inp_len;
            inp_vec[0].size -= inp_len;
            hash_copied_bytes = inp_len;
        }
    } else {
        hash_buf_ptr_64 = (uint64_t *)hash_buf;
	    hash_buf_ptr_64[0] = cvmx_htobe64(seq);
	    CVMX_STOREUNA_INT32(key->aux.tls_aad_32[2], hash_buf, 8); //12
	    hash_buf[12] = key->aux.tls_aad_8[12]; //13
	    hash_copied_bytes = 13; /* seq num (8 bytes) + record type (1 byte) + record version (2 bytes) + record length (2 bytes) */
    }
	enc_copied_bytes = 0;

	hash_in = inp_vec[idx].ptr ;
	hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
	hash_rem_num = (hash_end - hash_in) + hash_copied_bytes;

	enc_in = inp_vec[idx].ptr ;
	enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
	enc_rem_num = (enc_end - enc_in);

	enc_out = outp_vec[idx].ptr ;

	if (hash_rem_num <= SHA_CBLOCK || key->ex_mode == EX_MODE_ENCRYPT) {
		ptr = (uint64_t *)(hash_buf + hash_copied_bytes);	
		CVMX_STOREUNA_INT64(0, ptr, 0);
		CVMX_STOREUNA_INT64(0, ptr, 8);
		CVMX_STOREUNA_INT64(0, ptr, 16);
		CVMX_STOREUNA_INT64(0, ptr, 24);
		CVMX_STOREUNA_INT64(0, ptr, 32);
		CVMX_STOREUNA_INT64(0, ptr, 40);
		CVMX_STOREUNA_INT64(0, ptr, 48);
		CVMX_STOREUNA_INT64(0, ptr, 56);

		memcpy(hash_buf + hash_copied_bytes, hash_in, inp_vec[idx].size);	
		hash_copied_bytes += (inp_vec[idx].size);
		hash_buf[hash_copied_bytes] = 0x80;

        if(key->ex_mode != EX_MODE_ENCRYPT)
    		SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
		hash_copied_bytes = 0;
		hash_rem_num = 0;

		while (enc_rem_num >= DES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	

			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

			enc_in += DES_BLOCK_SIZE;
			enc_rem_num -= DES_BLOCK_SIZE;

			CVMX_MF_3DES_RESULT(p0);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);

			enc_out += DES_BLOCK_SIZE;
		}

		memcpy(enc_buf, enc_in, enc_rem_num);
		enc_copied_bytes = enc_rem_num;

		goto get_inner_hash;
	}

	/* inp_vec[0] has payload(inp_vec[0].size - sizeof(struct tls_rec_hdr)) > 51 bytes */
	/*hash copied bytes = 13, so pushing 3 more bytes from hash_in to sha1_rem*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    hash_buf[hash_copied_bytes] = hash_in[0];
	    hash_buf[hash_copied_bytes + 1] = hash_in[1];
	    hash_buf[hash_copied_bytes + 2] = hash_in[2];
	    ptr = (uint64_t *)(hash_in + 3);
    } else { 
        ptr = (uint64_t *) hash_in;
    }

	CVMX_LOADUNA_INT64(p0, ptr, 0);
	CVMX_LOADUNA_INT64(p1, ptr, 8);
	CVMX_LOADUNA_INT64(p2, ptr, 16);
	CVMX_LOADUNA_INT64(p3, ptr, 24);
	CVMX_LOADUNA_INT64(p4, ptr, 32);

    hash_buf_ptr_64 = (uint64_t *)hash_buf;
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[1]), 1);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 4);
	CVMX_LOADUNA_INT64(p0, ptr, 40);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 6);

	CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p0));

    if(key->ex_mode == EX_MODE_DEFAULT)
    	hash_in += 51;
    else
        hash_in += 48;
	hash_rem_num = (hash_end - hash_in);
	hash_copied_bytes = 0;

	/* SHA_CBLOCK chunks in 1st iovec */
	while ((hash_in + SHA_CBLOCK) <= hash_end) {
		/* get and load 1st des bytes */
		CVMX_LOADUNA_INT64(p2, enc_in, 0);	
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 0);
		CVMX_LOADUNA_INT64(p1, hash_in, 8);

		/* get 2nd des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 8);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

		/* 1st des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 2nd des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 1st des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 0);

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 16);
		CVMX_LOADUNA_INT64(p1, hash_in, 24);

		/* get 3rd des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 16);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

		/* 2nd des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 3rd des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 2nd des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 32);
		CVMX_LOADUNA_INT64(p1, hash_in, 40);

		/* get 4th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 24);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

		/* 3rd des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 4th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 3rd des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 16);

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 48);
		CVMX_LOADUNA_INT64(p1, hash_in, 56);

		/* 4th des result */
		CVMX_MF_3DES_RESULT(p3);	

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		/* get 5th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 32);

		/* store 4th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

		/* load 5th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* get 6th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 40);

		/* 5th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 6th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* get 7th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 48);

		/* store 5th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 32);

		/* 6th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 7th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* get 8th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 56);

		/* store 6th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

		/* 7th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 8th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 7th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 48);

		hash_in += SHA_CBLOCK;
		enc_in += SHA_CBLOCK;
		enc_rem_num -= SHA_CBLOCK;
		hash_rem_num -= SHA_CBLOCK;

		CVMX_PREFETCH(enc_in, SHA_CBLOCK);

		/* 8th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* store 8th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);

		enc_out += SHA_CBLOCK;
	}

	/* for each iovec */
	while (idx++ < inp_cnt) {
		/* for each SHA_CBLOCK in iovec */
		while ((hash_in + SHA_CBLOCK) <= hash_end) {
			/* get and load 1st des bytes */
			CVMX_LOADUNA_INT64(p2, enc_in, 0);	
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 0);
			CVMX_LOADUNA_INT64(p1, hash_in, 8);

			/* get 2nd des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 8);

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

			/* 1st des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 2nd des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 1st des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 0);

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 16);
			CVMX_LOADUNA_INT64(p1, hash_in, 24);

			/* get 3rd des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 16);

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

			/* 2nd des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 3rd des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 2nd des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 32);
			CVMX_LOADUNA_INT64(p1, hash_in, 40);

			/* get 4th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 24);

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

			/* 3rd des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 4th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 3rd des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 16);

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 48);
			CVMX_LOADUNA_INT64(p1, hash_in, 56);

			/* 4th des result */
			CVMX_MF_3DES_RESULT(p3);	

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

			/* get 5th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 32);

			/* store 4th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

			/* load 5th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* get 6th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 40);

			/* 5th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 6th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* get 7th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 48);

			/* store 5th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 32);

			/* 6th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 7th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* get 8th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 56);

			/* store 6th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

			/* 7th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 8th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 7th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 48);

			hash_in += SHA_CBLOCK;
			enc_in += SHA_CBLOCK;
			enc_rem_num -= SHA_CBLOCK;
			hash_rem_num -= SHA_CBLOCK;

			CVMX_PREFETCH(enc_in, SHA_CBLOCK);

			/* 8th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* store 8th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);

			enc_out += SHA_CBLOCK;
		} /* for each SHA_CBLOCK in iovec */

		hash_rem_num = (hash_end - hash_in);
		enc_rem_num = (hash_end - enc_in);

#ifdef DEBUG
		cavium_dbgprint("hash_rem_num: %d, enc_rem_num: %d (%d)\n", hash_rem_num, enc_rem_num, __LINE__);
#endif
		/* copy the remaining hash bytes */
		memcpy(hash_buf, hash_in, hash_rem_num);
		hash_copied_bytes = hash_rem_num;
		hash_rem_num = 0;

		if (idx < inp_cnt) {
			/* try fill the sha1 block, get bytes from next iovec */
			hash_in = inp_vec[idx].ptr;
			hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
			hash_rem_num = (hash_end - hash_in);
		}
#ifdef DEBUG
		cavium_dbgprint("hash_rem_num: %d, enc_rem_num: %d (%d)\n", hash_rem_num, enc_rem_num, __LINE__);
#endif

		/* less bytes than needed, try to finish inner hash */
		if (hash_rem_num < (SHA_CBLOCK - hash_copied_bytes)) {
			memcpy(hash_buf + hash_copied_bytes, hash_in, hash_rem_num);
			hash_copied_bytes += hash_rem_num;
			hash_buf[hash_copied_bytes] = 0x80;

			memset(hash_buf + hash_copied_bytes + 1, 0, (SHA_CBLOCK - hash_copied_bytes - 1));

			SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
			hash_rem_num = 0;
			hash_copied_bytes = 0;

			/* complete enc bytes in previous iovec */
			while (enc_rem_num >= DES_BLOCK_SIZE) {
				CVMX_LOADUNA_INT64(p0, enc_in, 0);

				CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

				enc_in += DES_BLOCK_SIZE;
				enc_rem_num -= DES_BLOCK_SIZE;

				CVMX_MF_3DES_RESULT(p0);

				CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);

				enc_out += DES_BLOCK_SIZE;
			}

			if (enc_rem_num) {
				/* copy the remaining enc bytes from previous iovec */
				memcpy(enc_buf, enc_in, enc_rem_num);
				enc_copied_bytes = enc_rem_num;
				eout_space_left = enc_rem_num;
				enc_rem_num = 0;

				if (idx < inp_cnt) {
					/* move to next iovec */
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					enc_rem_num = (enc_end - enc_in);
				}

				/* fill bytes from next iovec */
				if (enc_rem_num <= (DES_BLOCK_SIZE - enc_copied_bytes)) {
					memcpy(enc_buf + enc_copied_bytes, enc_in, enc_rem_num);
					enc_copied_bytes += enc_rem_num;
					enc_rem_num = 0;
					if (idx < outp_cnt)
						next_output_iovec = 1;
				} else {
					uint64_t enc_result;

					fill_bytes = (DES_BLOCK_SIZE - enc_copied_bytes);
					memcpy(enc_buf + enc_copied_bytes, enc_in, fill_bytes);
					enc_copied_bytes = DES_BLOCK_SIZE;
					enc_rem_num -= fill_bytes;

					CVMX_LOADUNA_INT64(p0, enc_buf, 0);

					CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

					enc_in += fill_bytes;

					CVMX_MF_3DES_RESULT(enc_result);
                    enc_result = cvmx_htobe64(enc_result);

					/* copy into the space left */
					memcpy(enc_out, &enc_result, eout_space_left);	

					/* move to next output iovec */
					enc_out = outp_vec[idx].ptr;
					memcpy(enc_out, ((uint8_t *)&enc_result + eout_space_left), fill_bytes);
					enc_out += fill_bytes;
					eout_space_left = 0;
					enc_copied_bytes = 0;

					/* encrypt DES_BLOCK_SIZE chunks if any */
					while (enc_rem_num >= DES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);

						CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

						enc_in += DES_BLOCK_SIZE;
						enc_rem_num -= DES_BLOCK_SIZE;

						CVMX_MF_3DES_RESULT(p0);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

						enc_out += DES_BLOCK_SIZE;
					}
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;
				}
			} else {
				if (idx < inp_cnt) {
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					enc_rem_num = (enc_end - enc_in);
				}

				if (idx < outp_cnt)
					enc_out = outp_vec[idx].ptr;

				while (enc_rem_num >= DES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);

					CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

					enc_in += DES_BLOCK_SIZE;
					enc_rem_num -= DES_BLOCK_SIZE;

					CVMX_MF_3DES_RESULT(p0);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

					enc_out += DES_BLOCK_SIZE;
				}
				if (enc_in) {
					memcpy(enc_buf, enc_in, enc_rem_num);	
					enc_copied_bytes = enc_rem_num;
					enc_rem_num = 0;
				}
			}
			goto get_inner_hash;
		}

		/* more than fill bytes in next iovec */
		fill_bytes = SHA_CBLOCK - hash_copied_bytes;
		memcpy(hash_buf + hash_copied_bytes, hash_in, fill_bytes);
		hash_copied_bytes += fill_bytes;

		hash_in += fill_bytes;
		hash_rem_num -= fill_bytes;

		ptr = (uint64_t *)hash_buf;
		CVMX_LOADUNA_INT64(p0, ptr, 0);
		CVMX_LOADUNA_INT64(p1, ptr, 8);
		CVMX_LOADUNA_INT64(p2, ptr, 16);
		CVMX_LOADUNA_INT64(p3, ptr, 24);
		CVMX_LOADUNA_INT64(p4, ptr, 32);
		CVMX_LOADUNA_INT64(p5, ptr, 40);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
		CVMX_LOADUNA_INT64(p0, ptr, 48);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_LOADUNA_INT64(p1, ptr, 56);

		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		while (enc_rem_num >= DES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);

			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

			enc_in += DES_BLOCK_SIZE;
			enc_rem_num -= DES_BLOCK_SIZE;

			CVMX_MF_3DES_RESULT(p0);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);

			enc_out += DES_BLOCK_SIZE;
		}

		/* if sha1 remain bytes are less than SHA_CBLOCK, finish inner hash */
		if (hash_rem_num < SHA_CBLOCK) {
			memcpy(hash_buf, hash_in, hash_rem_num);	
			hash_copied_bytes = hash_rem_num;
			hash_buf[hash_copied_bytes] = 0x80;

			memset(hash_buf + hash_copied_bytes + 1, 0, (SHA_CBLOCK - hash_copied_bytes - 1));

			SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
			hash_rem_num = 0;
			hash_copied_bytes = 0;

			start_inner_hash = 1;
		}

		if (enc_rem_num) {
			uint64_t enc_result;

			memcpy(enc_buf, enc_in, enc_rem_num);
			enc_copied_bytes = enc_rem_num;
			eout_space_left = enc_rem_num;

			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;	
			enc_rem_num = (enc_end - enc_in);

			if (enc_rem_num <= (DES_BLOCK_SIZE - enc_copied_bytes)) {
				/* copy the remaining enc bytes */
				memcpy(enc_buf + enc_copied_bytes, enc_in, enc_rem_num);
				enc_copied_bytes += enc_rem_num;
				enc_rem_num = 0;
				if (idx < outp_cnt)
					next_output_iovec = 1;

				goto get_inner_hash;
			} else {
				fill_bytes = (DES_BLOCK_SIZE - enc_copied_bytes);

				/* chance that enc bytes are crossing hash bytes */
				if ((enc_rem_num - fill_bytes) < hash_rem_num) {
					ptr = (uint64_t *)hash_in;	
					CVMX_LOADUNA_INT64(p0, ptr, 0);
					CVMX_LOADUNA_INT64(p1, ptr, 8);
					CVMX_LOADUNA_INT64(p2, ptr, 16);
					CVMX_LOADUNA_INT64(p3, ptr, 24);
					CVMX_LOADUNA_INT64(p4, ptr, 32);
					CVMX_LOADUNA_INT64(p5, ptr, 40);

					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
					CVMX_LOADUNA_INT64(p0, ptr, 48);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
					CVMX_LOADUNA_INT64(p1, ptr, 56);

					CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

					hash_in += SHA_CBLOCK;
					hash_rem_num -= SHA_CBLOCK;
				}

				memcpy(enc_buf + enc_copied_bytes, enc_in, fill_bytes); 
				enc_copied_bytes += fill_bytes;

				CVMX_LOADUNA_INT64(p0, enc_buf, 0);

				CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

				enc_in += fill_bytes;
				enc_rem_num -= fill_bytes;

				CVMX_MF_3DES_RESULT(enc_result);
                enc_result = cvmx_htobe64(enc_result);

				memcpy(enc_out, &enc_result, eout_space_left);

				/* move to next output iovec */
				enc_out = outp_vec[idx].ptr;
				memcpy(enc_out, ((uint8_t *)&enc_result + eout_space_left), fill_bytes);
				enc_out += fill_bytes;

				eout_space_left = 0;
				enc_copied_bytes = 0;

				if (start_inner_hash) {
					/* encrypt DES_BLOCK_SIZE chunks if any */
					while (enc_rem_num >= DES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);

						CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

						enc_in += DES_BLOCK_SIZE;
						enc_rem_num -= DES_BLOCK_SIZE;

						CVMX_MF_3DES_RESULT(p0);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

						enc_out += DES_BLOCK_SIZE;
					}
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;

					goto get_inner_hash;
				}
			}
		} else {
			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
			enc_rem_num = (enc_end - enc_in);

			if (idx < outp_cnt)
				enc_out = outp_vec[idx].ptr;

			if (start_inner_hash) {
				while (enc_rem_num >= DES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);

					CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

					enc_in += DES_BLOCK_SIZE;
					enc_rem_num -= DES_BLOCK_SIZE;

					CVMX_MF_3DES_RESULT(p0);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

					enc_out += DES_BLOCK_SIZE;
				}

				if (enc_in) {
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;
					enc_rem_num = 0;
				}

				goto get_inner_hash;
			}
		}
	} /* for each iovec */

get_inner_hash:
	/* get inner hash */
	CVMX_MF_HSH_IV(tmpvar, 0);
	CVMX_MF_HSH_IV(tmpvar2, 1);
	CVMX_MF_HSH_IV(p0, 2);
    
    if(key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar),outp_vec[0].ptr,0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar2),outp_vec[0].ptr,8);
        CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(p0)),outp_vec[0].ptr,16);
        return 1;
    } else if(key->ex_mode == EX_MODE_ENCRYPT) {
        return 1;
    }

#ifdef DEBUG
	cavium_dbgprint("hash_copied_bytes: %d, hash_rem_num: %d (%d)\n", hash_copied_bytes, hash_rem_num, __LINE__);
	_hexdump(hash_buf, hash_copied_bytes);
	cavium_dbgprint("enc_copied_bytes: %d, enc_rem_num: %d (%d)\n", enc_copied_bytes, enc_rem_num, __LINE__);
	_hexdump(enc_buf, enc_copied_bytes);;
	cavium_dbgprint("innerhash\n");
	_hexdump(&tmpvar, 8);
	_hexdump(&tmpvar2, 8);
	cavium_dbgprint("%llx\n", p0);
#endif

	/* start outer hash (mac secret ^ pad_2 + inner hash) */

	/* SHA init */
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);


	CVMX_LOADUNA_INT64(p1, mac_secret, 0);
	p1 = p1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 0);
	CVMX_LOADUNA_INT64(p1, mac_secret, 8);
	p1 = p1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

	p1 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	p1 = p1 << 32;
	p1 = p1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(p1, 2);

	CVMX_LOADUNA_INT64(p1, CRY_HMAC_PAD2_64, 0);
	CVMX_MT_HSH_DAT(p1, 3);
	CVMX_MT_HSH_DAT(p1, 4);
	CVMX_MT_HSH_DAT(p1, 5);
	CVMX_MT_HSH_DAT(p1, 6);
	CVMX_MT_HSH_STARTSHA(p1);


	ptr = (uint64_t *)&enc_buf[enc_copied_bytes];


	CVMX_MT_HSH_DAT(tmpvar, 0 );
	CVMX_MT_HSH_DAT(tmpvar2, 1 );
	p0 = p0 & 0xFFFFFFFF00000000ULL;
	p0 = p0 | 0x0000000080000000ULL;

	CVMX_MT_HSH_DAT(p0, 2);
	CVMX_MT_HSH_DATZ(3);
	CVMX_MT_HSH_DATZ(4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	total_bits = ((SHA_CBLOCK + SHA_DIGEST_LENGTH) * 8);
	CVMX_MT_HSH_STARTSHA(total_bits);

	enc_copied_bytes += SHA_DIGEST_LENGTH;

	/* calculate pad length */
	padlen = DES_BLOCK_SIZE - ((enc_copied_bytes) % DES_BLOCK_SIZE);

	/* add pad bytes pattren (padlen - 1) */
	memset(enc_buf + enc_copied_bytes, padlen - 1, padlen);
	enc_copied_bytes += padlen;

	/* get outer hash */
	CVMX_MF_HSH_IV(p0, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p0), ptr, 0);
	CVMX_MF_HSH_IV(p1, 1);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p1), ptr, 8);
	CVMX_MF_HSH_IV(p0, 2);
	CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(p0)), ptr, 16);

#ifdef DEBUG
	cavium_dbgprint("Outer Hash\n");
	_hexdump(ptr, SHA_DIGEST_LENGTH);
#endif
    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr,ptr,SHA_DIGEST_LENGTH);
        return 1;
    }
	uint8_t *out_enc_ptr = final_buf;
	uint8_t *in_enc_ptr = enc_buf;
	enc_rem_num = (enc_copied_bytes - eout_space_left);

#ifdef DEBUG
	cavium_dbgprint("enc_copied_bytes: %d eout_space_left: %d (%d)\n", enc_copied_bytes, eout_space_left, __LINE__);
	_hexdump(enc_buf, enc_copied_bytes);
#endif
	while (enc_copied_bytes >= DES_BLOCK_SIZE) {
		CVMX_LOADUNA_INT64(p0, in_enc_ptr, 0);

		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

		in_enc_ptr += DES_BLOCK_SIZE;
		enc_copied_bytes -= DES_BLOCK_SIZE;	

		CVMX_MF_3DES_RESULT(p0);
#ifdef DEBUG
		cavium_dbgprint("out: %llx\n", p0);
#endif

		CVMX_STOREUNA_INT64(cvmx_htobe64(p0), out_enc_ptr, 0);

		out_enc_ptr += DES_BLOCK_SIZE;
	}
	if (enc_copied_bytes)
		cavium_dbgprint("ERROR: enc_copied_bytes: %d\n", enc_copied_bytes);

	memcpy(enc_out, final_buf, eout_space_left);

	if (next_output_iovec)
		enc_out = outp_vec[idx].ptr;			
	else
		enc_out += eout_space_left;

	if (enc_out)
		memcpy(enc_out, final_buf + eout_space_left, enc_rem_num);

	CVMX_MF_3DES_IV(*(uint64_t *)key_data);
    *(uint64_t *)key_data = cvmx_htobe64(*(uint64_t *)key_data);

	//update ssl output header



	return 1;
}


/*******************************************************************************
 *                                                                             *
 *    Decrypts input record and generates TLS-1.0 SHA1 HMAC; authenticates     *
 *  record and stores result in output iovec buffer. Initialisation vector to  *
 *  be used for decrytion is taken from EVP_AESOCT_HMAC_SHA1 structure passed  *
 *  to the function. Key and MAC secret to be used for cipher decryption and   *
 *  HMAC generation respectively are  taken from the key structure passed.     *
 *  Sequence number and record header to be used while generating record HMAC  *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the TLS-1.0 record header.          *
 *  Initialisation vector is updated after each decryption.                    *
 *                                                                             *
 *  Protocol : TLS-1.0                                                         *
 *  Cipher   : AES128-CBC / AES256-CBC                                         *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing decryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to decrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store decryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_aes_sha1_dec_tls10(EVP_AESOCT_HMAC_SHA1 *key,
		struct iovec_cav *inp_vec, uint16_t inp_cnt,
		struct iovec_cav *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t r0, r1, r2, r3;

	uint8_t padarray[32] = {0}, hash_buf[128] = {0}, hash_rem_buf[128] = {0}, buf[128] = {0};
	uint8_t  keylength = 0, inner_hash_done = 0, output_overlap = 0, mac_size = 0, cipher_type = 0;
	int32_t enc_data_len = 0, raw_data_len = 0, pad_data_len = 0, hash_input_rem_len = 0;
	int32_t overlap_buf_size = 0, enc_rem_cur_chunk = 0, inner_hash_total_bytes = 0, outer_hash_total_bytes = 0, prev_iovec_rem_bytes = 0, ret_val = 1;
	int32_t i = 0;
	uint64_t seq_num = 0, inner_hash_total_bits = 0, outer_hash_total_bits = 0;
	uint8_t *hash_bal = NULL, *out_ptr = NULL, *mac_secret = NULL, *key_data = NULL;
	uint8_t *in_chunk_cur = NULL, *out_chunk_cur = NULL, *enc_rem_ptr = NULL;
	struct tls_rec_hdr *in_rec_hdr = NULL;
	uint64_t *enc_key = NULL;
    uint16_t *hash_buf_ptr_16 = NULL;
    uint32_t *hash_buf_ptr_32 = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct iovec_cav hash_overlap_buf;
	hash_overlap_buf.size = 0;

#ifdef DEBUG
	uint32_t itr = 0;

	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
	for (itr = 0; itr < inp_cnt; itr++)
	{
		cavium_dbgprint("inp_vec[%d].ptr: %p, inp_vec[%d].size: %d\n", itr, inp_vec[itr].ptr, itr, inp_vec[itr].size);
	}

	for (itr = 0; itr < outp_cnt; itr++)
	{
		cavium_dbgprint("outp_vec[%d].ptr: %p, outp_vec[%d].size: %d\n", itr, outp_vec[itr].ptr, itr, outp_vec[itr].size);
	}
#endif

	in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];
	enc_data_len = cvmx_htobe16(in_rec_hdr->length);

#ifdef DEBUG
	cavium_dbgprint("enc_data_len: %d (%d)\n", enc_data_len, __LINE__);
#endif

	cipher_type = key->cipher_type;

	if ((enc_data_len % AES_BLOCK_SIZE) && (key->ex_mode == EX_MODE_DEFAULT || key->ex_mode == EX_MODE_ENCRYPT))
	{
		cavium_dbgprint(
			"%s:%d Not properly padded data is received for cipher %s\n",
             __FUNCTION__,__LINE__, cipher_strings[cipher_type]);
		ret_val = 0;
		return ret_val;
	}

	switch(cipher_type)
	{
		case OCT_AES_128:
			keylength = 1;
			break;

		case OCT_AES_256:
			keylength = 3;
			break;
		default:
			break;
	}

	mac_size = SHA_DIGEST_LENGTH;

	seq_num = cvmx_htobe64(key->aux.tls_aad_64[0]);
	enc_key = (uint64_t *)key->cipher_key;
	mac_secret = key->mac_secret;
	key_data = key->cipher_iv;

	/*Load AES key*/
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[1]), 1);
	if(keylength == 3)
	{
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[2]), 2);
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[3]), 3);
	}
	else
	{
		CVMX_MT_AES_KEY((uint64_t)0, 2);
		CVMX_MT_AES_KEY((uint64_t)0, 3);
	}
	CVMX_MT_AES_KEYLENGTH(keylength);


	/* Pad data can not be more than AES_BLOCK_SIZE
	 * As per AES CBC implementation, previous aes block result is fed as IV for current aes block
	 * So decrypt last 32 bytes to know PAD length*/


	// calculate pad length by decrypting last 32 bytes
	// if the last iovec buffer contains data >= 32
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    if(inp_vec[inp_cnt - 1].size >= 2*AES_BLOCK_SIZE)
	    {
	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 32, 0);
	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 24, 0);

	    	CVMX_MT_AES_IV(cvmx_htobe64(r0), 0);
	    	CVMX_MT_AES_IV(cvmx_htobe64(r1), 1);

	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 16, 0);
	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 8, 0);

	    	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
	    	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

	    	CVMX_MF_AES_RESULT(r0, 0);
	    	CVMX_MF_AES_RESULT(r1, 1);

	    	pad_data_len = (r1 & 0xff ) + 1;
	    }
	    else
	    {
	    	// X + Y = 32
	    	// X bytes from iovec[inp_cnt - 2]
	    	// Y bytes from iovec[inp_cnt - 1]
	    	memcpy(padarray, inp_vec[inp_cnt - 2].ptr + (inp_vec[inp_cnt - 2].size - (32 - inp_vec[inp_cnt - 1].size)), (32 - inp_vec[inp_cnt - 1].size));
	    	memcpy(padarray + (32 - inp_vec[inp_cnt - 1].size), inp_vec[inp_cnt - 1].ptr, inp_vec[inp_cnt - 1].size);
	    	CVMX_LOADUNA_INT64(r0, padarray, 0);
	    	CVMX_LOADUNA_INT64(r1, padarray, 8);

	    	CVMX_MT_AES_IV(cvmx_htobe64(r0), 0);
	    	CVMX_MT_AES_IV(cvmx_htobe64(r1), 1);

	    	CVMX_LOADUNA_INT64(r0, padarray, 16);
	    	CVMX_LOADUNA_INT64(r1, padarray, 24);

	    	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
	    	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

	    	CVMX_MF_AES_RESULT(r0, 0);
	    	CVMX_MF_AES_RESULT(r1, 1);

	    	pad_data_len = (r1 & 0xff ) + 1;
	    }
    }

#ifdef DEBUG
	cavium_dbgprint("pad_data_len: %d (%d)\n", pad_data_len, __LINE__);
#endif

	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)key_data), 0);
	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)(&key_data[8])), 1);

    if(key->ex_mode == EX_MODE_DEFAULT)
    	raw_data_len = enc_data_len - SHA_DIGEST_LENGTH - pad_data_len;
    else
        raw_data_len =  enc_data_len;

	/* If message length < 0, then we are sure we received error packet,
 	   to avoid timing attack decrypt packet and calculate HMAC and throw error at end */
	if (raw_data_len < 0)
	{
		raw_data_len = enc_data_len - mac_size;
		pad_data_len = 0;
	}

#ifdef DEBUG
	cavium_dbgprint("raw_data_len: %d (%d)\n", raw_data_len, __LINE__);
#endif
	// first chunk - 16 + 48

	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);


    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_LOADUNA_INT64(r0, mac_secret, 0);
	    CVMX_LOADUNA_INT64(r1, mac_secret, 8);

	    r0 = r0 ^ (0x3636363636363636);
	    r1 = r1 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);

	    r0 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	    r0 = r0 << 32;
	    r0 = r0 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(r0, 2);

	    CVMX_LOADUNA_INT64(r1, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(r1, 3);
	    CVMX_MT_HSH_DAT(r1, 4);
	    CVMX_MT_HSH_DAT(r1, 5);
	    CVMX_MT_HSH_DAT(r1, 6);
	    CVMX_MT_HSH_STARTSHA(r1);
    }

    if(key->ex_mode == EX_MODE_DEFAULT) {
	/*8 + 1 + 2 + 2 = 13 bytes are left over*/
	CVMX_STOREUNA_INT64(cvmx_htobe64(seq_num), hash_buf, 0);
	hash_buf[8] = in_rec_hdr->type;
    hash_buf_ptr_16 = (uint16_t *)&hash_buf[9];
	hash_buf_ptr_16[0]  = in_rec_hdr->u.version;
	hash_buf_ptr_16[1]  = cvmx_htobe16(raw_data_len);
	hash_bal = hash_buf;
	// hash bal is 13 bytes
    } else if (key->ex_mode & 0x02) {
        if(raw_data_len > 16) {
            memcpy(hash_buf, inp_vec[0].ptr, 16);
            inp_vec[0].ptr += 16;
            inp_vec[0].size -= 16;
            hash_bal = hash_buf;
        } else {
            memcpy(hash_buf, inp_vec[0].ptr, raw_data_len);
            inp_vec[0].ptr += raw_data_len;
            inp_vec[0].size -= raw_data_len;
            hash_bal = hash_buf;
        }
    } else
        hash_bal = padarray;


	hash_input_rem_len = raw_data_len;

	if(inp_cnt == 1)
	{
		in_chunk_cur = inp_vec[0].ptr;
        if(key->ex_mode & 0x02)
            out_chunk_cur = inp_vec[0].ptr;
    	else
            out_chunk_cur = outp_vec[0].ptr;

		while(hash_input_rem_len >= 64)
		{
			hash_bal = block_processing_aes_sha1_dec_tls10(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);

#ifdef DEBUG
			cavium_dbgprint("hash_bal\n");
			_hexdump(hash_bal, 11);
#endif
			in_chunk_cur += 64;
			out_chunk_cur += 64;

			hash_input_rem_len -= 64;
		}
		out_ptr = out_chunk_cur;
#ifdef DEBUG
		cavium_dbgprint("hash_bal\n");
		_hexdump(hash_bal, 13);
#endif
		// remaining data is compulsory a multiple of 16
        if((key->ex_mode == EX_MODE_ENCRYPT && hash_input_rem_len) || 
                key->ex_mode == EX_MODE_DEFAULT) {
    		do
    		{
    			CVMX_LOADUNA_INT64(r0, in_chunk_cur, 0);
    			CVMX_LOADUNA_INT64(r1, in_chunk_cur, 8);
    
    			CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
    			CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));
    
    			in_chunk_cur += 16;
    
    			CVMX_MF_AES_RESULT(r2, 0);
    			CVMX_MF_AES_RESULT(r3, 1);
    
#ifdef DEBUG
			cavium_dbgprint("r2 : %llx  r3 : %llx\n", r2, r3);
#endif
		    	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), out_chunk_cur, 0);
		    	CVMX_STOREUNA_INT64(cvmx_htobe64(r3), out_chunk_cur, 8);

		    	out_chunk_cur += 16;

		    }while(((inp_vec[0].ptr + inp_vec[0].size) - in_chunk_cur) > 0);
        }
		// Along with 13 bytes hash balance, hash_input_rem_len 
		// should be given to hash engine as input
        if(key->ex_mode & 0x02)
            memcpy(hash_rem_buf, hash_bal,(hash_input_rem_len>16)?16:hash_input_rem_len);
        else if(key->ex_mode == EX_MODE_DEFAULT)
		    memcpy(hash_rem_buf, hash_bal, 13);

		if(hash_input_rem_len) {
            if((key->ex_mode & 0x02) && (hash_input_rem_len > 16)) {
		    	memcpy(hash_rem_buf + 16, out_ptr, hash_input_rem_len-16);
            } else if(key->ex_mode == EX_MODE_DEFAULT){
		    	memcpy(hash_rem_buf + 13, out_ptr, hash_input_rem_len);
            }
		}
        
        if(key->ex_mode & 0x02) 
            hash_rem_buf[hash_input_rem_len] = 0x80;
         else
		    hash_rem_buf[13 + hash_input_rem_len] = 0x80;

#ifdef DEBUG
		cavium_dbgprint("hash_rem_buf\n");
		_hexdump(hash_rem_buf, 64);
#endif
		uint64_t * ptr = (uint64_t *)hash_rem_buf;
		if((((hash_input_rem_len + 13) >= 56) && (key->ex_mode == EX_MODE_DEFAULT)) ||
            ((hash_input_rem_len  >= 56) && (key->ex_mode & 0x02)))
		{
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

			ptr += 8;
		}
        if(key->ex_mode == EX_MODE_DIGEST)
	    	inner_hash_total_bytes = raw_data_len;
        else if(key->ex_mode == EX_MODE_HMAC)
	    	inner_hash_total_bytes = SHA_CBLOCK + raw_data_len;
        else if(key->ex_mode == EX_MODE_DEFAULT)
	    	inner_hash_total_bytes = SHA_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
		inner_hash_total_bits = inner_hash_total_bytes * 8;
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
		CVMX_MT_HSH_STARTSHA(inner_hash_total_bits);
	}
	else if(inp_cnt > 1)
	{
		hash_overlap_buf.ptr = buf;
		hash_overlap_buf.size = 0;
		iovec_record_process_t *iovec_stat = NULL;
		iovec_stat = malloc(inp_cnt * sizeof(iovec_record_process_t));
		memset(iovec_stat, 0, inp_cnt * sizeof(iovec_record_process_t));

		prev_iovec_rem_bytes = 0;
		// populate loop counters
		while(hash_input_rem_len && (i < inp_cnt))
		{
			if(hash_input_rem_len >= inp_vec[i].size)
			{
				iovec_stat[i].hash_size = inp_vec[i].size;
				hash_input_rem_len -= inp_vec[i].size;
			}
			else
			{
				iovec_stat[i].hash_size = hash_input_rem_len;
				hash_input_rem_len = 0;
			}

			if(prev_iovec_rem_bytes)
			{
				iovec_stat[i].hash_req_bytes = 64 - prev_iovec_rem_bytes;
				// in case of last chunk not holding enough bytes to fill 64 byte block
				if(iovec_stat[i].hash_size < iovec_stat[i].hash_req_bytes)
				{
					iovec_stat[i].hash_req_bytes = iovec_stat[i].hash_size;
				}
			}
			else
			{
				iovec_stat[i].hash_req_bytes = 0;
			}
			iovec_stat[i].crypto_itr = (iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) / AES_BLOCK_SIZE;
			iovec_stat[i].hash_itr = iovec_stat[i].crypto_itr / 4;

			iovec_stat[i].hash_rem_bytes = ((iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) - (iovec_stat[i].hash_itr * 64));
			prev_iovec_rem_bytes = iovec_stat[i].hash_rem_bytes;

			i++;
		}
		uint16_t num_hash_iovec = i;
		// need to update enc_digest_ptr
		if(iovec_stat[i-1].hash_size == inp_vec[i-1].size)
		{
			enc_rem_ptr = inp_vec[i].ptr;
		}
		else
		{
			enc_rem_ptr = inp_vec[i-1].ptr + iovec_stat[i-1].hash_size;
			enc_rem_cur_chunk = inp_vec[i-1].size - iovec_stat[i-1].hash_size;

		}

#ifdef DEBUG
		for(i = 0; i < inp_cnt; i++)
		{
			cavium_dbgprint("Chunk : %d\n", i+1);
			cavium_dbgprint("hash_req_bytes : %d\n",iovec_stat[i].hash_req_bytes);
			cavium_dbgprint("crypto_itr : %d\n", iovec_stat[i].crypto_itr);
			cavium_dbgprint("hash_itr : %d\n", iovec_stat[i].hash_itr);
			cavium_dbgprint("hash_rem_bytes : %d\n",iovec_stat[i].hash_rem_bytes);
		}
#endif

		for(i = 0; i < num_hash_iovec; i++)
		{
			in_chunk_cur = inp_vec[i].ptr;
			out_chunk_cur = outp_vec[i].ptr;

			if(iovec_stat[i].hash_req_bytes)
			{	
				memcpy(hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, 
						in_chunk_cur, 
						iovec_stat[i].hash_req_bytes);
				in_chunk_cur += iovec_stat[i].hash_req_bytes;

				if((iovec_stat[i-1].hash_rem_bytes + iovec_stat[i].hash_req_bytes) < 64)
				{
					hash_overlap_buf.size += iovec_stat[i].hash_req_bytes;
					out_chunk_cur = outp_vec[i-1].ptr + iovec_stat[i-1].hash_req_bytes + (64 * iovec_stat[i-1].hash_itr);
					output_overlap = 1;
					continue;
				}
				else
				{
					hash_bal = block_processing_aes_sha1_dec_tls10(hash_overlap_buf.ptr, hash_overlap_buf.ptr, hash_bal, key->ex_mode );
#ifdef DEBUG
					cavium_dbgprint("hash_bal : %d\n", __LINE__);
					_hexdump(hash_bal, 13);
#endif

					memcpy(outp_vec[i-1].ptr + (outp_vec[i-1].size - iovec_stat[i-1].hash_rem_bytes),
							hash_overlap_buf.ptr, 
							iovec_stat[i-1].hash_rem_bytes);
					memcpy(out_chunk_cur, 
							hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, 
							iovec_stat[i].hash_req_bytes);

					out_chunk_cur += iovec_stat[i].hash_req_bytes;
					hash_overlap_buf.size = 0;
				}
			}

			register uint8_t hash_itr = iovec_stat[i].hash_itr;
			while(hash_itr)
			{
				hash_bal = block_processing_aes_sha1_dec_tls10(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);
#ifdef DEBUG
				cavium_dbgprint("hash_bal : %d\n", __LINE__);
				_hexdump(hash_bal, 13);
#endif

				in_chunk_cur += 64;
				out_chunk_cur += 64;

				hash_itr--;
			}

			if(iovec_stat[i].hash_rem_bytes)
			{
				//copy rem bytes to hash overlap buf
				memcpy(hash_overlap_buf.ptr, in_chunk_cur, iovec_stat[i].hash_rem_bytes);
				in_chunk_cur += iovec_stat[i].hash_rem_bytes;
				hash_overlap_buf.size = iovec_stat[i].hash_rem_bytes;
			}
		}

#ifdef DEBUG
		cavium_dbgprint("hash_overlap_buf.size : %d \n", hash_overlap_buf.size);
#endif
		//obviously this case gets hit all the time
		if(inner_hash_done == 0)
		{
			//if(hash_overlap_buf.size)
			{
				uint8_t *ptr = NULL;
				overlap_buf_size = hash_overlap_buf.size;
				memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size, enc_rem_ptr, enc_rem_cur_chunk);
				overlap_buf_size += enc_rem_cur_chunk;
#ifdef DEBUG
				cavium_dbgprint("enc_rem_cur_chunk : %d\n", enc_rem_cur_chunk);
#endif
				if(enc_rem_cur_chunk < (SHA_DIGEST_LENGTH + pad_data_len))
				{
					memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size + enc_rem_cur_chunk, 
							inp_vec[inp_cnt-1].ptr, inp_vec[inp_cnt-1].size);

					overlap_buf_size += inp_vec[inp_cnt-1].size;
#ifdef DEBUG
					cavium_dbgprint("overlap_buf_size : %d\n", overlap_buf_size);
#endif
				}
				ptr = hash_overlap_buf.ptr;
				uint16_t num_itr = 0;
				num_itr = overlap_buf_size / AES_BLOCK_SIZE;
				while(num_itr)
				{
					CVMX_LOADUNA_INT64(r0, ptr, 0);
					CVMX_LOADUNA_INT64(r1, ptr, 8);

					CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
					CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

					CVMX_MF_AES_RESULT(r2, 0);
					CVMX_MF_AES_RESULT(r3, 1);

#ifdef DEBUG
					cavium_dbgprint("r2 : %llx  r3 : %llx\n", r2, r3);
#endif
					CVMX_STOREUNA_INT64(cvmx_htobe64(r2), ptr, 0);
					CVMX_STOREUNA_INT64(cvmx_htobe64(r3), ptr, 8);

					ptr += 16;
					num_itr--;
				}

				if(hash_overlap_buf.size)
				{
					if(output_overlap)
					{
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, iovec_stat[num_hash_iovec - 2].hash_rem_bytes);
						memcpy(outp_vec[inp_cnt -1].ptr, hash_overlap_buf.ptr + iovec_stat[num_hash_iovec - 2].hash_rem_bytes, iovec_stat[num_hash_iovec - 1].hash_req_bytes);
					}
					else
					{
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, hash_overlap_buf.size);
					}
				}
			}

			memset(hash_rem_buf, 0, 128);
			memcpy(hash_rem_buf, hash_bal, 13);
			memcpy(hash_rem_buf + 13, hash_overlap_buf.ptr, hash_overlap_buf.size);
			hash_rem_buf[hash_overlap_buf.size + 13] = 0x80;
			uint64_t *ptr = (uint64_t *)hash_rem_buf;
#ifdef DEBUG
			cavium_dbgprint("hash rem buf\n");
			_hexdump(hash_rem_buf, 128);
#endif
			if((hash_overlap_buf.size + 13) >= 56)
			{
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
				CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

				ptr += 8;
			}
			inner_hash_total_bytes = SHA_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
			inner_hash_total_bits = inner_hash_total_bytes * 8;
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(inner_hash_total_bits);

			inner_hash_done = 1;
		}
		free(iovec_stat);
	}

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

#ifdef DEBUG
	cavium_dbgprint("Inner Hash\n");
	cavium_dbgprint("r0 : %llx\n", r0);
	cavium_dbgprint("r1 : %llx\n", r1);
	cavium_dbgprint("r2 : %llx\n", r2);
#endif
    if(key->ex_mode == EX_MODE_ENCRYPT)
        return 1;
    else if (key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(r0), outp_vec[0].ptr, 0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(r1), outp_vec[0].ptr, 8);
        CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(r2)), outp_vec[0].ptr, 16);
        return 1;
    }

	//calculate outer hash
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

	outer_hash_total_bytes = SHA_CBLOCK + SHA_DIGEST_LENGTH;
	outer_hash_total_bits = outer_hash_total_bytes * 8;

	CVMX_LOADUNA_INT64(r3, mac_secret, 0);
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 0);
	CVMX_LOADUNA_INT64(r3, mac_secret, 8);
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 1);

	r3 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	r3 = r3 << 32;
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(r3, 2);

	CVMX_LOADUNA_INT64(r3, CRY_HMAC_PAD2_64, 0);
	CVMX_MT_HSH_DAT(r3, 3);
	CVMX_MT_HSH_DAT(r3, 4);
	CVMX_MT_HSH_DAT(r3, 5);
	CVMX_MT_HSH_DAT(r3, 6);
	CVMX_MT_HSH_STARTSHA(r3);


	CVMX_MT_HSH_DAT(r0, 0);
	CVMX_MT_HSH_DAT(r1, 1);
	r2 = r2 & 0xFFFFFFFF00000000ULL;
	r2 = r2 | 0x0000000080000000ULL;

	CVMX_MT_HSH_DAT(r2, 2);
	CVMX_MT_HSH_DATZ(3);
	CVMX_MT_HSH_DATZ(4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	CVMX_MT_HSH_STARTSHA(outer_hash_total_bits);

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

	CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 8);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 16);

    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr, hash_buf, SHA_DIGEST_LENGTH);
        return 1;
    }
#ifdef DEBUG
	cavium_dbgprint("Outer Hash Calculated : %s\n", __func__);
	_hexdump(hash_buf, SHA_DIGEST_LENGTH);

#endif


	//Compare Calculated HMAC and Received HMAC
	if(inp_cnt == 1)
	{
        hash_buf_ptr_64 = (uint64_t *)hash_buf;
        hash_buf_ptr_32 = (uint32_t *)&hash_buf[16];
		if((hash_buf_ptr_64[0] != (((uint64_t *)(out_ptr + hash_input_rem_len))[0])) ||
		 	(hash_buf_ptr_64[1] != (((uint64_t *)(out_ptr + hash_input_rem_len))[1])) ||
			 (hash_buf_ptr_32[0] != (((uint32_t *)(out_ptr + hash_input_rem_len + 16))[0])))
		{
			ret_val = 0;
			cavium_dbgprint(
				"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
			cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
			cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(out_ptr + hash_input_rem_len))[0]));
			cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(out_ptr + hash_input_rem_len))[1]));
			cavium_dbgprint("3. :%x  : %x\n", ((uint32_t *)(hash_buf + 16))[0], (((uint32_t *)(out_ptr + hash_input_rem_len + 16))[0]));
#endif
		}
	}
	else
	{
		if(memcmp(hash_buf, hash_overlap_buf.ptr + hash_overlap_buf.size, SHA_DIGEST_LENGTH))
		{
			ret_val = 0;
			cavium_dbgprint(
				"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
			cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
			cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[0]));
			cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[1]));
			cavium_dbgprint("3. :%x  : %x\n", ((uint32_t *)(hash_buf + 16))[0], (((uint32_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size + 16))[0]));
#endif
		}
	}



	CVMX_MF_AES_IV(*(uint64_t *)key_data, 0);
	CVMX_MF_AES_IV(*(uint64_t *)(&key_data[8]), 1);
    *(uint64_t *)key_data = cvmx_htobe64(*(uint64_t *)&key_data[0]);
    *(uint64_t *)&key_data[8] = cvmx_htobe64(*(uint64_t *)&key_data[8]);

	*len = raw_data_len; 
	return ret_val;
}

/*******************************************************************************
 *                                                                             *
 *    Decrypts input record and generates TLS-1.0 SHA1 HMAC; authenticates     *
 *  record and stores result in output iovec buffer. Initialisation vector to  *
 *  be used for decrytion is taken from EVP_DESOCT_HMAC_SHA1 structure passed  *
 *  to the function. Key and MAC secret to be used for cipher decryption and   *
 *  HMAC generation respectively are  taken from the key structure passed.     *
 *  Sequence number and record header to be used while generating record HMAC  *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the TLS-1.0 record header.          *
 *  Initialisation vector is updated after each decryption.                    *
 *                                                                             *
 *  Protocol : TLS-1.0                                                         *
 *  Cipher   : DES-CBC / DES3-CBC                                              *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing decryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to decrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store decryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/


int32_t oct_des_sha1_dec_tls10(EVP_DESOCT_HMAC_SHA1 *key,
		struct iovec_cav *inp_vec, uint16_t inp_cnt,
		struct iovec_cav *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t r0, r1, r2, r3;
	uint8_t padarray[2*16] = {0}, hash_buf[128] = {0}, hash_rem_buf[128] = {0}, buf[128] = {0};
	uint8_t des = 0, inner_hash_done = 0, output_overlap = 0, cipher_type = 0, mac_size = 0;
	int32_t enc_data_len = 0, raw_data_len = 0, pad_data_len = 0, hash_input_rem_len = 0;
	int32_t overlap_buf_size = 0, enc_rem_cur_chunk = 0, inner_hash_total_bytes = 0, outer_hash_total_bytes = 0, prev_iovec_rem_bytes = 0, ret_val = 1;
	int32_t i = 0;
	uint64_t seq_num = 0, inner_hash_total_bits = 0, outer_hash_total_bits = 0;
	uint8_t *hash_bal = NULL, *out_ptr = NULL, *in_chunk_cur = NULL, *out_chunk_cur = NULL, *enc_rem_ptr = NULL, *mac_secret = NULL;
	uint64_t *enc_key = NULL, *key_data = NULL;
    uint16_t *hash_buf_ptr_16 = NULL;
    uint32_t *hash_buf_ptr_32 = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct tls_rec_hdr *in_rec_hdr = NULL;
	struct iovec_cav hash_overlap_buf;
	hash_overlap_buf.size = 0;

#ifdef DEBUG
	uint32_t itr = 0;

	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
	for (itr = 0; itr < inp_cnt; itr++)
	{
		cavium_dbgprint("inp_vec[%d].ptr: %p, inp_vec[%d].size: %d\n", itr, inp_vec[itr].ptr, itr, inp_vec[itr].size);
	}

	for (itr = 0; itr < outp_cnt; itr++)
	{
		cavium_dbgprint("outp_vec[%d].ptr: %p, outp_vec[%d].size: %d\n", itr, outp_vec[itr].ptr, itr, outp_vec[itr].size);
	}
#endif

	in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];
	enc_data_len = cvmx_htobe16(in_rec_hdr->length);

#ifdef DEBUG
	cavium_dbgprint("enc_data_len: %d (%d)\n", enc_data_len, __LINE__);
#endif

	cipher_type = key->cipher_type;

	if ((enc_data_len % DES_BLOCK_SIZE) && (key->ex_mode == EX_MODE_DEFAULT || key->ex_mode == EX_MODE_ENCRYPT))
	{
		cavium_dbgprint(
			"%s:%d Not properly padded data is received for cipher %s\n",
			__FUNCTION__,__LINE__, cipher_strings[cipher_type]);
		ret_val = 0;
		return ret_val;
	}

	switch(cipher_type)
	{
		case OCT_DES_40:
		case OCT_DES_64:
			des = 1;
			break;

		case OCT_3DES:
			break;
		default:
			break;
	}

	mac_size = SHA_DIGEST_LENGTH;

	seq_num = cvmx_htobe64(key->aux.tls_aad_64[0]);
	enc_key = (uint64_t *)key->cipher_key;
	mac_secret = key->mac_secret;
	key_data = (uint64_t *)key->cipher_iv;


	/*Load DES key*/
	if(des == 1)
	{
		enc_key[1] = enc_key[0];
		enc_key[2] = enc_key[0];
	}
	CVMX_MT_3DES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enc_key[1]), 1);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enc_key[2]), 2);

	/* Pad data can not be more than DES_BLOCK_SIZE
	 * As per DES CBC implementation, previous aes block result is fed as IV for current des block
	 * So decrypt last 32 bytes to know PAD length*/


	// calculate pad length by decrypting last 16 bytes
	// if the last iovec buffer contains data >= 16
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    if(inp_vec[inp_cnt - 1].size >= 2*DES_BLOCK_SIZE)
	    {
	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 16, 0);

	    	CVMX_MT_3DES_IV(cvmx_htobe64(r0));

	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 8, 0);

	    	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r1));


	    	CVMX_MF_3DES_RESULT(r0);

	    	pad_data_len = (r0 & 0xff ) + 1;
	    }
	    else
	    {
	    	// X + Y = 16
	    	// X bytes from iovec[inp_cnt - 2]
	    	// Y bytes from iovec[inp_cnt - 1]
	    	memcpy(padarray, inp_vec[inp_cnt - 2].ptr + (inp_vec[inp_cnt - 2].size - (16 - inp_vec[inp_cnt - 1].size)), (16 - inp_vec[inp_cnt - 1].size));
	    	memcpy(padarray + (16 - inp_vec[inp_cnt - 1].size), inp_vec[inp_cnt - 1].ptr, inp_vec[inp_cnt - 1].size);
	    	CVMX_LOADUNA_INT64(r0, padarray, 0);
	    	CVMX_LOADUNA_INT64(r1, padarray, 8);

	    	CVMX_MT_3DES_IV(cvmx_htobe64(r0));

	    	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r1));

	    	CVMX_MF_3DES_RESULT(r0);

	    	pad_data_len = (r0 & 0xff ) + 1;
	    }
    }

#ifdef DEBUG
	cavium_dbgprint("pad_data_len: %d (%d)\n", pad_data_len, __LINE__);
#endif


	CVMX_MT_3DES_IV(cvmx_htobe64(key_data[0]));

    if(key->ex_mode == EX_MODE_DEFAULT)
    	raw_data_len = enc_data_len - mac_size - pad_data_len;
    else
        raw_data_len = enc_data_len;

	/* If message length < 0, then we are sure we received error packet,
 	   to avoid timing attack decrypt packet and calculate HMAC and throw error at end */
	if (raw_data_len < 0)
	{
		raw_data_len = enc_data_len - mac_size;
		pad_data_len = 0;
	}

#ifdef DEBUG
	cavium_dbgprint("raw_data_len: %d (%d)\n", raw_data_len, __LINE__);
#endif
	// first chunk - 16+48

	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);


    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_LOADUNA_INT64(r0, mac_secret, 0);
	    CVMX_LOADUNA_INT64(r1, mac_secret, 8);

	    r0 = r0 ^ (0x3636363636363636);
	    r1 = r1 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);

	    r0 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	    r0 = r0 << 32;
	    r0 = r0 ^ (0x3636363636363636);
	    CVMX_MT_HSH_DAT(r0, 2);

	    CVMX_LOADUNA_INT64(r1, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(r1, 3);
	    CVMX_MT_HSH_DAT(r1, 4);
	    CVMX_MT_HSH_DAT(r1, 5);
	    CVMX_MT_HSH_DAT(r1, 6);
	    CVMX_MT_HSH_STARTSHA(r1);
    }

    if(key->ex_mode == EX_MODE_DEFAULT) {
	    /*8 + 1 + 2 + 2 = 13 bytes are left over*/
	    CVMX_STOREUNA_INT64(cvmx_htobe64(seq_num), hash_buf, 0);
	    hash_buf[8] = in_rec_hdr->type;
        hash_buf_ptr_16 = (uint16_t *)&hash_buf[9];
	    hash_buf_ptr_16[0]  = in_rec_hdr->u.version;
	    hash_buf_ptr_16[1]  = cvmx_htobe16(raw_data_len);
    } else if (key->ex_mode & 0x02) {
        if(raw_data_len > 16) {
            memcpy(hash_buf, inp_vec[0].ptr, 16);
            inp_vec[0].ptr += 16;
            inp_vec[0].size -= 16;
        } else {
            memcpy(hash_buf, inp_vec[0].ptr, raw_data_len);
            inp_vec[0].ptr += raw_data_len;
            inp_vec[0].size -= raw_data_len;
        }
    }
    

	hash_bal = hash_buf;

#ifdef DEBUG
	cavium_dbgprint("hash bal : %d\n", __LINE__);
	_hexdump(hash_bal, 13);
#endif

	hash_input_rem_len = raw_data_len;

	if(inp_cnt == 1)
	{
		in_chunk_cur = inp_vec[0].ptr;
        if(key->ex_mode & 0x02)
		    out_chunk_cur = inp_vec[0].ptr;
        else
		    out_chunk_cur = outp_vec[0].ptr;

		while(hash_input_rem_len >= 64)
		{
			hash_bal = block_processing_des_sha1_dec_tls10(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);

#ifdef DEBUG
			cavium_dbgprint("hash_bal : %d\n", __LINE__);
			_hexdump(hash_bal, 13);
#endif
			in_chunk_cur += 64;
			out_chunk_cur += 64;

			hash_input_rem_len -= 64;
		}
		out_ptr = out_chunk_cur;
#ifdef DEBUG
		cavium_dbgprint("hash_bal : %d \n", __LINE__);
		_hexdump(hash_bal, 13);
#endif
		// remaining data is compulsory a multiple of 8
        if(key->ex_mode == EX_MODE_DEFAULT || 
           (key->ex_mode == EX_MODE_ENCRYPT && hash_input_rem_len)) {
		    do
		    {
		    	CVMX_LOADUNA_INT64(r0, in_chunk_cur, 0);

		    	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r0));

		    	in_chunk_cur += 8;

		    	CVMX_MF_3DES_RESULT(r1);
#ifdef DEBUG
			cavium_dbgprint("r1 : %llx\n", r1);
#endif
	    		CVMX_STOREUNA_INT64(cvmx_htobe64(r1), out_chunk_cur, 0);

	    		out_chunk_cur += 8;

	    	}while(((inp_vec[0].ptr + inp_vec[0].size) - in_chunk_cur) > 0);
        }
		// Along with 13 bytes hash balance, hash_input_rem_len 
		// should be given to hash engine as input
		memset(hash_rem_buf, 0, 128);
        if(key->ex_mode & 0x02)
		    memcpy(hash_rem_buf, hash_bal, (hash_input_rem_len>16)?16:hash_input_rem_len);
        else
		    memcpy(hash_rem_buf, hash_bal, 13);

		if(hash_input_rem_len)
		{
            if(key->ex_mode & 0x02 && hash_input_rem_len > 16) {
			    memcpy(hash_rem_buf + 16, out_ptr, hash_input_rem_len-16);
            } else if(key->ex_mode == EX_MODE_DEFAULT) {
			    memcpy(hash_rem_buf + 13, out_ptr, hash_input_rem_len);
            }
		}
        if(key->ex_mode & 0x02) 
    	    hash_rem_buf[hash_input_rem_len] = 0x80;
         else if (key->ex_mode == EX_MODE_DEFAULT)
		        hash_rem_buf[13 + hash_input_rem_len] = 0x80;

		uint64_t * ptr = (uint64_t *)hash_rem_buf;
#ifdef DEBUG
		cavium_dbgprint("hash_rem_buf\n", __LINE__);
		_hexdump(hash_rem_buf, 128);
#endif
		if((((hash_input_rem_len + 13) >= 56) && (key->ex_mode == EX_MODE_DEFAULT)) ||
           ((hash_input_rem_len >= 56) && (key->ex_mode & 0x02)))
		{
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

			ptr += 8;
		}
#ifdef DEBUG
		cavium_dbgprint("raw_data_length : %d\n", raw_data_len);
#endif
        if(key->ex_mode == EX_MODE_DIGEST)
		    inner_hash_total_bytes = raw_data_len;
        if(key->ex_mode == EX_MODE_HMAC)
		    inner_hash_total_bytes = SHA_CBLOCK + raw_data_len;
        if(key->ex_mode == EX_MODE_DEFAULT)
		    inner_hash_total_bytes = SHA_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
		inner_hash_total_bits = inner_hash_total_bytes * 8;
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
		CVMX_MT_HSH_STARTSHA(inner_hash_total_bits);
	}
	else if(inp_cnt > 1)
	{
		hash_overlap_buf.ptr = buf;
		hash_overlap_buf.size = 0;
		iovec_record_process_t *iovec_stat = NULL;
		iovec_stat = malloc(inp_cnt * sizeof(iovec_record_process_t));
		memset(iovec_stat, 0, inp_cnt * sizeof(iovec_record_process_t));

		prev_iovec_rem_bytes = 0;
		// populate loop counters
		while(hash_input_rem_len && (i < inp_cnt))
		{
			if(hash_input_rem_len >= inp_vec[i].size)
			{
				iovec_stat[i].hash_size = inp_vec[i].size;
				hash_input_rem_len -= inp_vec[i].size;
			}
			else
			{
				iovec_stat[i].hash_size = hash_input_rem_len;
				hash_input_rem_len = 0;
			}

			if(prev_iovec_rem_bytes)
			{
				iovec_stat[i].hash_req_bytes = 64 - prev_iovec_rem_bytes;
				// in case of last chunk not holding enough bytes to fill 64 byte block
				if(iovec_stat[i].hash_size < iovec_stat[i].hash_req_bytes)
				{
					iovec_stat[i].hash_req_bytes = iovec_stat[i].hash_size;
				}
			}
			else
			{
				iovec_stat[i].hash_req_bytes = 0;
			}
			iovec_stat[i].crypto_itr = (iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) / DES_BLOCK_SIZE;
			iovec_stat[i].hash_itr = iovec_stat[i].crypto_itr / 8;

			iovec_stat[i].hash_rem_bytes = ((iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) - (iovec_stat[i].hash_itr * 64));
			prev_iovec_rem_bytes = iovec_stat[i].hash_rem_bytes;

			i++;
		}
		uint16_t num_hash_iovec = i;
		// need to update enc_digest_ptr
		if(iovec_stat[i-1].hash_size == inp_vec[i-1].size)
		{
			enc_rem_ptr = inp_vec[i].ptr;
		}
		else
		{
			enc_rem_ptr = inp_vec[i-1].ptr + iovec_stat[i-1].hash_size;
			enc_rem_cur_chunk = inp_vec[i-1].size - iovec_stat[i-1].hash_size;

		}

#ifdef DEBUG
		for(i = 0; i < inp_cnt; i++)
		{
			cavium_dbgprint("Chunk : %d\n", i+1);
			cavium_dbgprint("hash_req_bytes : %d\n",iovec_stat[i].hash_req_bytes);
			cavium_dbgprint("crypto_itr : %d\n", iovec_stat[i].crypto_itr);
			cavium_dbgprint("hash_itr : %d\n", iovec_stat[i].hash_itr);
			cavium_dbgprint("hash_rem_bytes : %d\n",iovec_stat[i].hash_rem_bytes);
		}
#endif

		for(i = 0; i < num_hash_iovec; i++)
		{
			in_chunk_cur = inp_vec[i].ptr;
			out_chunk_cur = outp_vec[i].ptr;

			if(iovec_stat[i].hash_req_bytes)
			{	
				memcpy(hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, 
						in_chunk_cur, 
						iovec_stat[i].hash_req_bytes);
				in_chunk_cur += iovec_stat[i].hash_req_bytes;

				if((iovec_stat[i-1].hash_rem_bytes + iovec_stat[i].hash_req_bytes) < 64)
				{
					hash_overlap_buf.size += iovec_stat[i].hash_req_bytes;
					out_chunk_cur = outp_vec[i-1].ptr + iovec_stat[i-1].hash_req_bytes + (64 * iovec_stat[i-1].hash_itr);
					output_overlap = 1;
					continue;
				}
				else
				{
					hash_bal = block_processing_des_sha1_dec_tls10(hash_overlap_buf.ptr, hash_overlap_buf.ptr, hash_bal, key->ex_mode);
#ifdef DEBUG
					cavium_dbgprint("hash_bal : %d\n", __LINE__);
					_hexdump(hash_bal, 13);
#endif

					memcpy(outp_vec[i-1].ptr + (outp_vec[i-1].size - iovec_stat[i-1].hash_rem_bytes),
							hash_overlap_buf.ptr, 
							iovec_stat[i-1].hash_rem_bytes);
					memcpy(out_chunk_cur, 
							hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, 
							iovec_stat[i].hash_req_bytes);

					out_chunk_cur += iovec_stat[i].hash_req_bytes;
					hash_overlap_buf.size = 0;
				}
			}

			register uint8_t hash_itr = iovec_stat[i].hash_itr;
			while(hash_itr)
			{
				hash_bal = block_processing_des_sha1_dec_tls10(in_chunk_cur, out_chunk_cur, hash_bal,key->ex_mode);
#ifdef DEBUG
				cavium_dbgprint("hash_bal : %d\n", __LINE__);
				_hexdump(hash_bal, 13);
#endif

				in_chunk_cur += 64;
				out_chunk_cur += 64;

				hash_itr--;
			}

			if(iovec_stat[i].hash_rem_bytes)
			{
				//copy rem bytes to hash overlap buf
				memcpy(hash_overlap_buf.ptr, in_chunk_cur, iovec_stat[i].hash_rem_bytes);
				in_chunk_cur += iovec_stat[i].hash_rem_bytes;
				hash_overlap_buf.size = iovec_stat[i].hash_rem_bytes;
			}
		}

#ifdef DEBUG
		cavium_dbgprint("hash_overlap_buf.size : %d \n", hash_overlap_buf.size);
#endif
		//obviously this case gets hit all the time
		if(inner_hash_done == 0)
		{
			//if(hash_overlap_buf.size)
			{
				uint8_t *ptr = NULL;
				overlap_buf_size = hash_overlap_buf.size;
				memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size, enc_rem_ptr, enc_rem_cur_chunk);
				overlap_buf_size += enc_rem_cur_chunk;
#ifdef DEBUG
				cavium_dbgprint("enc_rem_cur_chunk : %d\n", enc_rem_cur_chunk);
#endif
				if(enc_rem_cur_chunk < (SHA_DIGEST_LENGTH + pad_data_len))
				{
					memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size + enc_rem_cur_chunk, 
							inp_vec[inp_cnt-1].ptr, inp_vec[inp_cnt-1].size);

					overlap_buf_size += inp_vec[inp_cnt-1].size;
#ifdef DEBUG
					cavium_dbgprint("overlap_buf_size : %d\n", overlap_buf_size);
#endif
				}
				ptr = hash_overlap_buf.ptr;
				uint16_t num_itr = 0;
				num_itr = overlap_buf_size / DES_BLOCK_SIZE;
				while(num_itr)
				{
					CVMX_LOADUNA_INT64(r0, ptr, 0);

					CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r0));

					num_itr--;

					CVMX_MF_3DES_RESULT(r1);

#ifdef DEBUG
					cavium_dbgprint("r1 : %llx\n", r1);
#endif
					CVMX_STOREUNA_INT64(cvmx_htobe64(r1), ptr, 0);
					ptr += 8;

				}

				if(hash_overlap_buf.size)
				{
					if(output_overlap)
					{
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, iovec_stat[num_hash_iovec - 2].hash_rem_bytes);
						memcpy(outp_vec[inp_cnt -1].ptr, hash_overlap_buf.ptr + iovec_stat[num_hash_iovec - 2].hash_rem_bytes, iovec_stat[num_hash_iovec - 1].hash_req_bytes);
					}
					else
					{
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, hash_overlap_buf.size);
					}
				}
			}

			memset(hash_rem_buf, 0, 128);
			memcpy(hash_rem_buf, hash_bal, 13);
			memcpy(hash_rem_buf + 13, hash_overlap_buf.ptr, hash_overlap_buf.size);
			hash_rem_buf[hash_overlap_buf.size + 13] = 0x80;
			uint64_t *ptr = (uint64_t *)hash_rem_buf;
#ifdef DEBUG
			cavium_dbgprint("hash rem buf\n");
			_hexdump(hash_rem_buf, 128);
#endif
			if((hash_overlap_buf.size + 13) >= 56)
			{
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
				CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

				ptr += 8;
			}
			inner_hash_total_bytes = SHA_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
			inner_hash_total_bits = inner_hash_total_bytes * 8;
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(inner_hash_total_bits);

			inner_hash_done = 1;
		}
		free(iovec_stat);
	}

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

#ifdef DEBUG
	cavium_dbgprint("Inner Hash \n");
	cavium_dbgprint("r0 : %llx\n", r0);
	cavium_dbgprint("r1 : %llx\n", r1);
	cavium_dbgprint("r2 : %llx\n", r2);
#endif
    if(key->ex_mode == EX_MODE_ENCRYPT)
        return 1;
    else if (key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(r0), outp_vec[0].ptr, 0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(r1), outp_vec[0].ptr, 8);
        CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(r2)), outp_vec[0].ptr, 16);
        return 1;
    }


	//Calculate Outer Hash
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

	outer_hash_total_bytes = SHA_CBLOCK + SHA_DIGEST_LENGTH;
	outer_hash_total_bits = outer_hash_total_bytes * 8;

	CVMX_LOADUNA_INT64(r3, mac_secret, 0);
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 0);
	CVMX_LOADUNA_INT64(r3, mac_secret, 8);
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 1);

	r3 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	r3 = r3 << 32;
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(r3, 2);

	CVMX_LOADUNA_INT64(r3, CRY_HMAC_PAD2_64, 0);
	CVMX_MT_HSH_DAT(r3, 3);
	CVMX_MT_HSH_DAT(r3, 4);
	CVMX_MT_HSH_DAT(r3, 5);
	CVMX_MT_HSH_DAT(r3, 6);
	CVMX_MT_HSH_STARTSHA(r3);


	CVMX_MT_HSH_DAT(r0, 0);
	CVMX_MT_HSH_DAT(r1, 1);
	r2 = r2 & 0xFFFFFFFF00000000ULL;
	r2 = r2 | 0x0000000080000000ULL;

	CVMX_MT_HSH_DAT(r2, 2);
	CVMX_MT_HSH_DATZ(3);
	CVMX_MT_HSH_DATZ(4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	CVMX_MT_HSH_STARTSHA(outer_hash_total_bits);


	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

	CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 8);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 16);

    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr, hash_buf, SHA_DIGEST_LENGTH);
        return 1;
    }
#ifdef DEBUG
	cavium_dbgprint("Outer Hash Calculated : %s\n", __func__);
	_hexdump(hash_buf, SHA_DIGEST_LENGTH);
#endif


	//Compare calculated HMAC with received HMAC
	if(inp_cnt == 1)
	{
        hash_buf_ptr_64 = (uint64_t *)hash_buf;
        hash_buf_ptr_32 = (uint32_t *)&hash_buf[16];
		if((hash_buf_ptr_64[0] != (((uint64_t *)(out_ptr + hash_input_rem_len))[0])) || 
			(hash_buf_ptr_64[1] != (((uint64_t *)(out_ptr + hash_input_rem_len))[1])) || 
			(hash_buf_ptr_32[0] != (((uint32_t *)(out_ptr + hash_input_rem_len + 16))[0])))
		{
			ret_val = 0;
			cavium_dbgprint(
				"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
			cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
			cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(out_ptr + hash_input_rem_len))[0]));
			cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(out_ptr + hash_input_rem_len))[1]));
#endif
		}
	}
	else
	{
		if(memcmp(hash_buf, hash_overlap_buf.ptr + hash_overlap_buf.size, SHA_DIGEST_LENGTH))
		{
			ret_val = 0;
			cavium_dbgprint(
				"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
			cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
			cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[0]));
			cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[1]));
#endif
		}
	}


	CVMX_MF_3DES_IV(key_data[0]);
    key_data[0] = cvmx_htobe64(key_data[0]);


	*len = raw_data_len;
	return ret_val;
}



/*TLS1.1 enc functions*/

/*******************************************************************************
 *                                                                             *
 *      Generates TLS-1.1 / TLS-1.2 SHA1 HMAC, adds padding bytes and encrypts *
 *  payload data and stores the result in output iovec buffer. Initialisation  *
 *  vector  be used for encrytion is taken from the ext_iv_data pointer passed *
 *  to the function. Key and MAC secret to be used for cipher encryption and   *
 *  HMAC generation respectively are taken from the key structure passed.      *
 *  Sequence number and record header to be used for used for HMAC generation  *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the TLS-1.1 / TLS-1.2 record header.*
 *  Protocol : TLS-1.1 / TLS-1.2                                               *
 *  Cipher   : AES128-CBC / AES256-CBC                                         *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing encryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to encrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store encryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  ext_iv_data : Pointer to explicit IV to be used for encryption.            *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_aes_sha1_enc_tls11(EVP_AESOCT_HMAC_SHA1 *key,
		cav_iovec_t *inp_vec, uint16_t inp_cnt,
		cav_iovec_t *outp_vec, uint16_t outp_cnt,
		uint8_t *ext_iv_data, uint16_t *len)
{
	register uint64_t p0, p1, p2, p3, p4, p5, r0, r1;
	uint8_t sha1_rem[128] = {0};
	uint8_t aes_rem[128] = {0};
	uint8_t aes_out[128] = {0};
	uint8_t idx = 0;
	uint16_t inp_len = 0, cipher_type = 0;
	uint32_t aes_rem_num = 0, sha1_rem_num = 0;
	int32_t aes_out_len = 0, keylength = 1, cal_enc_len = 0, eout_space_left = 0, aes_copied_bytes = 0, sha1_copied_bytes = 0;
	int32_t fill_bytes = 0, next_output_iovec = 0, start_inner_hash = 0;
	uint64_t tmpvar = 0, tmpvar2 = 0, total_bits = 0, seq = cvmx_htobe64(key->aux.tls_aad_64[0]); 
      	uint8_t *mac_secret = NULL, *key_data = NULL, *hash_in = NULL, *hash_end = NULL, *enc_in = NULL, *enc_end = NULL, *enc_out = NULL;
	uint64_t *ptr = NULL, *enc_key = (uint64_t *)key->cipher_key;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct tls_rec_hdr *in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];

#ifdef DEBUG
	cavium_dbgprint("rec length: %d, seq: %llx\n", in_rec_hdr->length, seq);
	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
#endif
	mac_secret = key->mac_secret;

	cipher_type = key->cipher_type;
	switch (cipher_type) {
		case OCT_AES_128:
			keylength = 1;
			break;
		case OCT_AES_256:
			keylength = 3;
			break;
		default:
			break;
	}


	key_data = key->cipher_iv;


	/* assume AES keylen 128 bit */
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[1]), 1);
	if (keylength == 3) {
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[2]), 2);
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[3]), 3);
	} else {
		CVMX_MT_AES_KEY((uint64_t)0, 2);
		CVMX_MT_AES_KEY((uint64_t)0, 3);
	}
	CVMX_MT_AES_KEYLENGTH(keylength);
	
	/* Init AES iv */
	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)key_data), 0);
	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)(&key_data[8])), 1);

	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_LOADUNA_INT64(r0, mac_secret, 0);
	    CVMX_LOADUNA_INT64(r1, mac_secret, 8);
	    
	    r0 = r0 ^ (0x3636363636363636);
	    r1 = r1 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);

	    r0 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	    r0 = r0 << 32;
	    r0 = r0 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(r0, 2);

	    CVMX_LOADUNA_INT64(r1, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(r1, 3);
	    CVMX_MT_HSH_DAT(r1, 4);
	    CVMX_MT_HSH_DAT(r1, 5);
	    CVMX_MT_HSH_DAT(r1, 6);
	    CVMX_MT_HSH_STARTSHA(r1);
    }
	CVMX_PREFETCH0(inp_vec[idx].ptr);

	inp_len = cvmx_htobe16(in_rec_hdr->length);
    if(key->ex_mode == EX_MODE_DEFAULT)
    	total_bits = ((SHA_CBLOCK + 8 + 1 + 2 + 2 + inp_len) * 8);
    if(key->ex_mode == EX_MODE_HMAC)
    	total_bits = ((SHA_CBLOCK + inp_len) * 8);
    if(key->ex_mode == EX_MODE_DIGEST)
    	total_bits = inp_len * 8;

	/*store 8 bytes seq_num , 1 byte record type, 2 bytes record version, 2 bytes record length*/
    if(key->ex_mode & 0x02) {
        if(inp_len >= 16) {
            memcpy(sha1_rem,inp_vec[idx].ptr,16);
            inp_vec[0].ptr += 16;
            inp_vec[0].size -= 16;
            sha1_copied_bytes = 16;
        } else {
            memcpy(sha1_rem,inp_vec[idx].ptr,inp_len);
            inp_vec[0].ptr += inp_len;
            inp_vec[0].size -= inp_len;
            sha1_copied_bytes = inp_len;
        }
    } else {
        hash_buf_ptr_64 = (uint64_t *)sha1_rem;
        hash_buf_ptr_64[0] = cvmx_htobe64(seq);
	    CVMX_STOREUNA_INT32(key->aux.tls_aad_32[2], sha1_rem, 8); //12
	    sha1_rem[12] = key->aux.tls_aad_8[12]; //13
	    sha1_copied_bytes = 13; /* seq num (8 bytes) + record type (1 byte) + record version (2 bytes) + record length (2 bytes) */
    }

	aes_copied_bytes = 0;


/*Encrpyt External IV*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    enc_in = ext_iv_data;
	    CVMX_LOADUNA_INT64(p0, enc_in, 0);
	    CVMX_LOADUNA_INT64(p1, enc_in, 8);

	    CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
	    CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

	    enc_out = outp_vec[idx].ptr ;

	    CVMX_MF_AES_RESULT(p0, 0);	
	    CVMX_MF_AES_RESULT(p1, 1);	

	    CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);
    }

	hash_in = inp_vec[idx].ptr;
	hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
	sha1_rem_num = (hash_end - hash_in) + sha1_copied_bytes;

	enc_in = inp_vec[idx].ptr ;
	enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
	aes_rem_num = (enc_end - enc_in);

    if(key->ex_mode == EX_MODE_DEFAULT)
    	enc_out = outp_vec[idx].ptr + AES_BLOCK_SIZE;
    else
        enc_out = outp_vec[idx].ptr;

	if (sha1_rem_num <= SHA_CBLOCK || key->ex_mode == EX_MODE_ENCRYPT) {
		ptr = (uint64_t *)(sha1_rem + sha1_copied_bytes);	
/*memset 64 bytes to 0*/
		CVMX_STOREUNA_INT64(0, ptr, 0);
		CVMX_STOREUNA_INT64(0, ptr, 8);
		CVMX_STOREUNA_INT64(0, ptr, 16);
		CVMX_STOREUNA_INT64(0, ptr, 24);
		CVMX_STOREUNA_INT64(0, ptr, 32);
		CVMX_STOREUNA_INT64(0, ptr, 40);
		CVMX_STOREUNA_INT64(0, ptr, 48);
		CVMX_STOREUNA_INT64(0, ptr, 56);

		memcpy(sha1_rem + sha1_copied_bytes, hash_in, inp_vec[idx].size );	
		sha1_copied_bytes += (inp_vec[idx].size );
		sha1_rem[sha1_copied_bytes] = 0x80;

        if(key->ex_mode != EX_MODE_ENCRYPT)
    		SHA1_FINISH(sha1_rem, sha1_copied_bytes, total_bits);
		sha1_copied_bytes = 0;
		sha1_rem_num = 0;

		while (aes_rem_num >= AES_BLOCK_SIZE && (key->ex_mode == EX_MODE_ENCRYPT || key->ex_mode == EX_MODE_DEFAULT)) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	
			CVMX_LOADUNA_INT64(p1, enc_in, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
			
			enc_in += AES_BLOCK_SIZE;
			aes_rem_num -= AES_BLOCK_SIZE;

			CVMX_MF_AES_RESULT(p0, 0);	
			CVMX_MF_AES_RESULT(p1, 1);	

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

			enc_out += AES_BLOCK_SIZE;
		}

		memcpy(aes_rem, enc_in, aes_rem_num);
		aes_copied_bytes = aes_rem_num;

		goto get_inner_hash;
	}

	/* inp_vec[0] has payload(inp_vec[0].size - sizeof(struct tls_rec_hdr)) > 51 bytes */
	/*sha1 copied bytes = 13, so pushing 3 more bytes from hash_in to sha1_rem*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    sha1_rem[sha1_copied_bytes] = hash_in[0];
	    sha1_rem[sha1_copied_bytes + 1] = hash_in[1];
	    sha1_rem[sha1_copied_bytes + 2] = hash_in[2];
	    ptr = (uint64_t *)(hash_in + 3);
    } else {
        ptr = (uint64_t *)hash_in;
    }
	CVMX_LOADUNA_INT64(p0, ptr, 0);
	CVMX_LOADUNA_INT64(p1, ptr, 8);
	CVMX_LOADUNA_INT64(p2, ptr, 16);
	CVMX_LOADUNA_INT64(p3, ptr, 24);
	CVMX_LOADUNA_INT64(p4, ptr, 32);
	
    hash_buf_ptr_64 = (uint64_t *)sha1_rem;
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[1]), 1);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 4);
       	CVMX_LOADUNA_INT64(p0, ptr, 40);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 6);

	CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p0));

    if(key->ex_mode == EX_MODE_DEFAULT)
    	hash_in += 51;
    else 
        hash_in += 48;
	sha1_rem_num = (hash_end - hash_in);
	sha1_copied_bytes = 0;
	
	/* SHA_CBLOCK chunks in 1st iovec */

	while ((hash_in + SHA_CBLOCK) <= hash_end) {
		/* load 1st aes bytes */
		CVMX_LOADUNA_INT64(p0, enc_in, 0);	
		CVMX_LOADUNA_INT64(p1, enc_in, 8);
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 0);
		CVMX_LOADUNA_INT64(p1, hash_in, 8);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		
		/* get 2nd aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 16);
		CVMX_LOADUNA_INT64(p1, enc_in, 24);

		/* 1st aes result */
		CVMX_MF_AES_RESULT(p2, 0);	
		CVMX_MF_AES_RESULT(p3, 1);	

		/* load 2nd aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 16);
		CVMX_LOADUNA_INT64(p1, hash_in, 24);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

		/* store 1st aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 0);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

		/* get 3rd aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 32);
		CVMX_LOADUNA_INT64(p1, enc_in, 40);
		
		/* 2nd aes result */
		CVMX_MF_AES_RESULT(p2, 0);	
		CVMX_MF_AES_RESULT(p3, 1);	

		/* load 3rd aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 32);
		CVMX_LOADUNA_INT64(p1, hash_in, 40);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

		/* store 2nd aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 16);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

		/* get 4th aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 48);
		CVMX_LOADUNA_INT64(p1, enc_in, 56);
		
		/* get 3rd aes result */
		CVMX_MF_AES_RESULT(p2, 0);
		CVMX_MF_AES_RESULT(p3, 1);

		/* load 4th aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 48);
		CVMX_LOADUNA_INT64(p1, hash_in, 56);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		hash_in += SHA_CBLOCK;
		enc_in += SHA_CBLOCK;
		sha1_rem_num -= SHA_CBLOCK;
		aes_rem_num -= SHA_CBLOCK;

		/* store 3rd aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 32);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

		/* get 4th aes result */
		CVMX_MF_AES_RESULT(p2, 0);
		CVMX_MF_AES_RESULT(p3, 1);
			
		CVMX_PREFETCH(enc_in, SHA_CBLOCK);

		/* store 4th aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 48);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);
		
		enc_out += SHA_CBLOCK;
	}

	/* for each iovec */
	while (idx++ < inp_cnt) {
		/* for each SHA_CBLOCK in iovec */
		while ((hash_in + SHA_CBLOCK) <= hash_end) {
			/* load 1st aes bytes */
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	
			CVMX_LOADUNA_INT64(p1, enc_in, 8);
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 0);
			CVMX_LOADUNA_INT64(p1, hash_in, 8);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		
			/* get 2nd aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 16);
			CVMX_LOADUNA_INT64(p1, enc_in, 24);

			/* 1st aes result */
			CVMX_MF_AES_RESULT(p2, 0);	
			CVMX_MF_AES_RESULT(p3, 1);	

			/* load 2nd aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 16);
			CVMX_LOADUNA_INT64(p1, hash_in, 24);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

			/* store 1st aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

			/* get 3rd aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 32);
			CVMX_LOADUNA_INT64(p1, enc_in, 40);
		
			/* 2nd aes result */
			CVMX_MF_AES_RESULT(p2, 0);	
			CVMX_MF_AES_RESULT(p3, 1);	

			/* load 3rd aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 32);
			CVMX_LOADUNA_INT64(p1, hash_in, 40);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

			/* store 2nd aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 16);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

			/* get 4th aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 48);
			CVMX_LOADUNA_INT64(p1, enc_in, 56);
		
			/* get 3rd aes result */
			CVMX_MF_AES_RESULT(p2, 0);
			CVMX_MF_AES_RESULT(p3, 1);

			/* load 4th aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 48);
			CVMX_LOADUNA_INT64(p1, hash_in, 56);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

			hash_in += SHA_CBLOCK;
			enc_in += SHA_CBLOCK;
			sha1_rem_num -= SHA_CBLOCK;
			aes_rem_num -= SHA_CBLOCK;

			/* store 3rd aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 32);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

			/* get 4th aes result */
			CVMX_MF_AES_RESULT(p2, 0);
			CVMX_MF_AES_RESULT(p3, 1);
			
			CVMX_PREFETCH(enc_in, SHA_CBLOCK);

			/* store 4th aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 48);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);
		
			enc_out += SHA_CBLOCK;
		} /* for each SHA_CBLOCK in iovec */

		sha1_rem_num = (hash_end - hash_in);
		aes_rem_num = (hash_end - enc_in);

#ifdef DEBUG
		cavium_dbgprint("iovec_idx : %d sha1_rem_num: %d, aes_rem_num: %d (%d)\n", idx-1, sha1_rem_num, aes_rem_num, __LINE__);
#endif
		/* copy the remaining sha1 bytes */
		memcpy(sha1_rem, hash_in, sha1_rem_num);
		sha1_copied_bytes = sha1_rem_num;
		sha1_rem_num = 0;

		if (idx < inp_cnt) {
			/* try fill the sha1 block, get bytes from next iovec */
			hash_in = inp_vec[idx].ptr;
			hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
			sha1_rem_num = (hash_end - hash_in);
		}
#ifdef DEBUG
		cavium_dbgprint("iovec_idx : %d sha1_rem_num: %d, aes_rem_num: %d (%d)\n", idx, sha1_rem_num, aes_rem_num, __LINE__);
#endif

		/* less bytes than needed, try to finish inner hash */
		if (sha1_rem_num < (SHA_CBLOCK - sha1_copied_bytes)) {
			memcpy(sha1_rem + sha1_copied_bytes, hash_in, sha1_rem_num);
			sha1_copied_bytes += sha1_rem_num;
			sha1_rem[sha1_copied_bytes] = 0x80;
			
			memset(sha1_rem + sha1_copied_bytes + 1, 0, (SHA_CBLOCK - sha1_copied_bytes - 1));

			SHA1_FINISH(sha1_rem, sha1_copied_bytes, total_bits);
			sha1_rem_num = 0;
			sha1_copied_bytes = 0;

			/* complete aes bytes in previous iovec */
			while (aes_rem_num >= AES_BLOCK_SIZE) {
				CVMX_LOADUNA_INT64(p0, enc_in, 0);
				CVMX_LOADUNA_INT64(p1, enc_in, 8);

				CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
				CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

				enc_in += AES_BLOCK_SIZE;
				aes_rem_num -= AES_BLOCK_SIZE;

				CVMX_MF_AES_RESULT(p0, 0);
				CVMX_MF_AES_RESULT(p1, 1);

				CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
				CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

				enc_out += AES_BLOCK_SIZE;
			}

#ifdef DEBUG
		cavium_dbgprint("iovec_idx : %d sha1_rem_num: %d, aes_rem_num: %d (%d)\n", idx, sha1_rem_num, aes_rem_num, __LINE__);
#endif
			if (aes_rem_num) {
				/* copy the remaining aes bytes from previous iovec */
				memcpy(aes_rem, enc_in, aes_rem_num);
				aes_copied_bytes = aes_rem_num;
				eout_space_left = aes_rem_num;
				aes_rem_num = 0;

				if (idx < inp_cnt) {
					/* move to next iovec */
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					aes_rem_num = (enc_end - enc_in);
				}

				/* fill bytes from next iovec */
				if (aes_rem_num <= (AES_BLOCK_SIZE - aes_copied_bytes)) {
					memcpy(aes_rem + aes_copied_bytes, enc_in, aes_rem_num);
					aes_copied_bytes += aes_rem_num;
					aes_rem_num = 0;
					if (idx < outp_cnt)
						next_output_iovec = 1;
				} else {
					fill_bytes = (AES_BLOCK_SIZE - aes_copied_bytes);
					uint64_t enc_result[2];
					memcpy(aes_rem + aes_copied_bytes, enc_in, fill_bytes);
					aes_copied_bytes = AES_BLOCK_SIZE;
					aes_rem_num -= fill_bytes;

					CVMX_LOADUNA_INT64(p0, aes_rem, 0);
					CVMX_LOADUNA_INT64(p1, aes_rem, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

					enc_in += fill_bytes;

					CVMX_MF_AES_RESULT(enc_result[0], 0);
					CVMX_MF_AES_RESULT(enc_result[1], 1);
                    enc_result[0] = cvmx_htobe64(enc_result[0]);
                    enc_result[1] = cvmx_htobe64(enc_result[1]);

					/* copy into the space left */
					memcpy(enc_out, enc_result, eout_space_left);	

					/* move to next output iovec */
					enc_out = outp_vec[idx].ptr;
					memcpy(enc_out, ((uint8_t *)enc_result + eout_space_left), fill_bytes);
					enc_out += fill_bytes;
					eout_space_left = 0;
					aes_copied_bytes = 0;

					/* encrypt AES_BLOCK_SIZE chunks if any */
					while (aes_rem_num >= AES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);
						CVMX_LOADUNA_INT64(p1, enc_in, 8);

						CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
						CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
						enc_in += AES_BLOCK_SIZE;
						aes_rem_num -= AES_BLOCK_SIZE;

						CVMX_MF_AES_RESULT(p0, 0);
						CVMX_MF_AES_RESULT(p1, 1);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
						CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
						enc_out += AES_BLOCK_SIZE;
					}
					memcpy(aes_rem, enc_in, aes_rem_num);
					aes_copied_bytes = aes_rem_num;
				}
			} else {
				if (idx < inp_cnt) {
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					aes_rem_num = (enc_end - enc_in);
				}
			
				if (idx < outp_cnt)
					enc_out = outp_vec[idx].ptr;

				while (aes_rem_num >= AES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);
					CVMX_LOADUNA_INT64(p1, enc_in, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
					enc_in += AES_BLOCK_SIZE;
					aes_rem_num -= AES_BLOCK_SIZE;

					CVMX_MF_AES_RESULT(p0, 0);
					CVMX_MF_AES_RESULT(p1, 1);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
					CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
					enc_out += AES_BLOCK_SIZE;
				}
				if (enc_in) {
					memcpy(aes_rem, enc_in, aes_rem_num);	
					aes_copied_bytes = aes_rem_num;
					aes_rem_num = 0;
				}
			}
			goto get_inner_hash;
		}

		/* more than fill bytes in next iovec */
		fill_bytes = SHA_CBLOCK - sha1_copied_bytes;
		memcpy(sha1_rem + sha1_copied_bytes, hash_in, fill_bytes);
		sha1_copied_bytes += fill_bytes;

		hash_in += fill_bytes;
		sha1_rem_num -= fill_bytes;

		ptr = (uint64_t *)sha1_rem;
		CVMX_LOADUNA_INT64(p0, ptr, 0);
		CVMX_LOADUNA_INT64(p1, ptr, 8);
		CVMX_LOADUNA_INT64(p2, ptr, 16);
		CVMX_LOADUNA_INT64(p3, ptr, 24);
		CVMX_LOADUNA_INT64(p4, ptr, 32);
		CVMX_LOADUNA_INT64(p5, ptr, 40);
	
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
		CVMX_LOADUNA_INT64(p0, ptr, 48);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_LOADUNA_INT64(p1, ptr, 56);
		
		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));
		
		while (aes_rem_num >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);
			CVMX_LOADUNA_INT64(p1, enc_in, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			enc_in += AES_BLOCK_SIZE;
			aes_rem_num -= AES_BLOCK_SIZE;

			CVMX_MF_AES_RESULT(p0, 0);
			CVMX_MF_AES_RESULT(p1, 1);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

			enc_out += AES_BLOCK_SIZE;
		}

		/* if sha1 remain bytes are less than SHA_CBLOCK, finish inner hash */
		if (sha1_rem_num < SHA_CBLOCK) {
			memcpy(sha1_rem, hash_in, sha1_rem_num);	
			sha1_copied_bytes = sha1_rem_num;
			sha1_rem[sha1_copied_bytes] = 0x80;

			memset(sha1_rem + sha1_copied_bytes + 1, 0, (SHA_CBLOCK - sha1_copied_bytes - 1));

			SHA1_FINISH(sha1_rem, sha1_copied_bytes, total_bits);
			sha1_rem_num = 0;
			sha1_copied_bytes = 0;

			start_inner_hash = 1;
		}

		if (aes_rem_num) {
			uint64_t enc_result[2];

			memcpy(aes_rem, enc_in, aes_rem_num);
			aes_copied_bytes = aes_rem_num;
			eout_space_left = aes_rem_num;

			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;	
			aes_rem_num = (enc_end - enc_in);

			if (aes_rem_num <= (AES_BLOCK_SIZE - aes_copied_bytes)) {
				/* copy the remaining aes bytes */
				memcpy(aes_rem + aes_copied_bytes, enc_in, aes_rem_num);
				aes_copied_bytes += aes_rem_num;
				aes_rem_num = 0;
				if (idx < outp_cnt)
					next_output_iovec = 1;

				goto get_inner_hash;
			} else {
				fill_bytes = (AES_BLOCK_SIZE - aes_copied_bytes);

				/* chance that aes bytes are crossing sha1 bytes */
				if ((aes_rem_num - fill_bytes) < sha1_rem_num) {
					ptr = (uint64_t *)hash_in;	
					CVMX_LOADUNA_INT64(p0, ptr, 0);
					CVMX_LOADUNA_INT64(p1, ptr, 8);
					CVMX_LOADUNA_INT64(p2, ptr, 16);
					CVMX_LOADUNA_INT64(p3, ptr, 24);
					CVMX_LOADUNA_INT64(p4, ptr, 32);
					CVMX_LOADUNA_INT64(p5, ptr, 40);
	
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
					CVMX_LOADUNA_INT64(p0, ptr, 48);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
					CVMX_LOADUNA_INT64(p1, ptr, 56);
		
					CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

					hash_in += SHA_CBLOCK;
					sha1_rem_num -= SHA_CBLOCK;
				}

				memcpy(aes_rem + aes_copied_bytes, enc_in, fill_bytes); 
				aes_copied_bytes += fill_bytes;

				CVMX_LOADUNA_INT64(p0, aes_rem, 0);
				CVMX_LOADUNA_INT64(p1, aes_rem, 8);

				CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
				CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

				enc_in += fill_bytes;
				aes_rem_num -= fill_bytes;

				CVMX_MF_AES_RESULT(enc_result[0], 0);
				CVMX_MF_AES_RESULT(enc_result[1], 1);
                enc_result[0] = cvmx_htobe64(enc_result[0]);
                enc_result[1] = cvmx_htobe64(enc_result[1]);
			
				memcpy(enc_out, enc_result, eout_space_left);

				/* move to next output iovec */
				enc_out = outp_vec[idx].ptr;
				memcpy(enc_out, ((uint8_t *)enc_result + eout_space_left), fill_bytes);
				enc_out += fill_bytes;

				eout_space_left = 0;
				aes_copied_bytes = 0;

				if (start_inner_hash) {
					/* encrypt AES_BLOCK_SIZE chunks if any */
					while (aes_rem_num >= AES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);
						CVMX_LOADUNA_INT64(p1, enc_in, 8);

						CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
						CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
						enc_in += AES_BLOCK_SIZE;
						aes_rem_num -= AES_BLOCK_SIZE;

						CVMX_MF_AES_RESULT(p0, 0);
						CVMX_MF_AES_RESULT(p1, 1);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
						CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
						enc_out += AES_BLOCK_SIZE;
					}
					memcpy(aes_rem, enc_in, aes_rem_num);
					aes_copied_bytes = aes_rem_num;

					goto get_inner_hash;
				}
			}
		} else {
			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
			aes_rem_num = (enc_end - enc_in);

			if (idx < outp_cnt)
				enc_out = outp_vec[idx].ptr;

			if (start_inner_hash) {
				while (aes_rem_num >= AES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);
					CVMX_LOADUNA_INT64(p1, enc_in, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
					enc_in += AES_BLOCK_SIZE;
					aes_rem_num -= AES_BLOCK_SIZE;

					CVMX_MF_AES_RESULT(p0, 0);
					CVMX_MF_AES_RESULT(p1, 1);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
					CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
					enc_out += AES_BLOCK_SIZE;
				}

				if (enc_in) {
					memcpy(aes_rem, enc_in, aes_rem_num);
					aes_copied_bytes = aes_rem_num;
					aes_rem_num = 0;
				}

				goto get_inner_hash;
			}
		}
	} /* for each iovec */

get_inner_hash:
	/* get inner hash */
	CVMX_MF_HSH_IV(tmpvar, 0);
	CVMX_MF_HSH_IV(tmpvar2, 1);
	CVMX_MF_HSH_IV(r0, 2);

    if(key->ex_mode == EX_MODE_ENCRYPT) {
        return 1;
    } else if(key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar),outp_vec[0].ptr, 0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar2),outp_vec[0].ptr, 8);
        CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(r0)),outp_vec[0].ptr, 16);
        return 1;
    }
#ifdef DEBUG
	cavium_dbgprint("sha1_copied_bytes: %d, sha1_rem_num: %d (%d)\n", sha1_copied_bytes, sha1_rem_num, __LINE__);
	_hexdump(sha1_rem, sha1_copied_bytes);
	cavium_dbgprint("aes_copied_bytes: %d, aes_rem_num: %d (%d)\n", aes_copied_bytes, aes_rem_num, __LINE__);
	_hexdump(aes_rem, aes_copied_bytes);;
	cavium_dbgprint("Inner Hash\n");
	cavium_dbgprint("%llx\n", tmpvar);
	cavium_dbgprint("%llx\n", tmpvar2);
	cavium_dbgprint("%llx\n", r0);
#endif

	/* start outer hash ((mac secret ^ pad_2) + inner hash) */


	/* SHA init */
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

	CVMX_LOADUNA_INT64(r1, mac_secret, 0);
	r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 0);
	CVMX_LOADUNA_INT64(r1, mac_secret, 8);
	r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);

	r1 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	r1 = r1 << 32;
	r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(r1, 2);
	
	CVMX_LOADUNA_INT64(r1, CRY_HMAC_PAD2_64, 0);
	CVMX_MT_HSH_DAT(r1, 3);
	CVMX_MT_HSH_DAT(r1, 4);
	CVMX_MT_HSH_DAT(r1, 5);
	CVMX_MT_HSH_DAT(r1, 6);
	CVMX_MT_HSH_STARTSHA(r1);


	ptr = (uint64_t *)&aes_rem[aes_copied_bytes];


	CVMX_MT_HSH_DAT(tmpvar, 0);
	CVMX_MT_HSH_DAT(tmpvar2, 1);
	r0 = r0 & 0xFFFFFFFF00000000ULL;
	r0 = r0 | 0x0000000080000000ULL;

	CVMX_MT_HSH_DAT(r0, 2);
	CVMX_MT_HSH_DATZ(3);
	CVMX_MT_HSH_DATZ(4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	total_bits = ((SHA_CBLOCK + SHA_DIGEST_LENGTH) * 8);
	CVMX_MT_HSH_STARTSHA(total_bits);
      
#ifdef DEBUG
	cavium_dbgprint("Aes copied bytes to aes rem buf : %d\n", aes_copied_bytes);
#endif
	aes_copied_bytes += SHA_DIGEST_LENGTH;

	/* calculate pad length */
      	tmpvar = AES_BLOCK_SIZE - ((aes_copied_bytes) % AES_BLOCK_SIZE);
	
	/* add pad bytes pattren (padlen - 1) */
      	memset(aes_rem + aes_copied_bytes, tmpvar - 1, tmpvar);
	aes_copied_bytes += tmpvar;

	/* get outer hash */
	CVMX_MF_HSH_IV(p0, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p0), ptr, 0);
	CVMX_MF_HSH_IV(p1, 1);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p1), ptr, 8);
	CVMX_MF_HSH_IV(r0, 2);
	CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(r0)), ptr, 16);
    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr, ptr, SHA_DIGEST_LENGTH);
        return 1;
    }
#ifdef DEBUG
	cavium_dbgprint("Outer Hash\n");
	cavium_dbgprint("%llx\n", p0);
	cavium_dbgprint("%llx\n", p1);
	cavium_dbgprint("%llx\n", r0);
#endif
#ifdef DEBUG
	cavium_dbgprint("Aes copied bytes to aes rem buf : %d\n", aes_copied_bytes);
#endif
	if (aes_copied_bytes) {
		uint8_t *out_aes_ptr = aes_out;
		uint8_t *in_aes_ptr = aes_rem;
		aes_out_len = aes_copied_bytes;

		while (aes_copied_bytes >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, in_aes_ptr, 0);
			CVMX_LOADUNA_INT64(p1, in_aes_ptr, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			in_aes_ptr += AES_BLOCK_SIZE;
			aes_copied_bytes -= AES_BLOCK_SIZE;	

			CVMX_MF_AES_RESULT(p0, 0);
			CVMX_MF_AES_RESULT(p1, 1);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), out_aes_ptr, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), out_aes_ptr, 8);
			
			out_aes_ptr += AES_BLOCK_SIZE;
		}
		if (aes_copied_bytes)
			cavium_dbgprint("ERROR: aes_copied_bytes: %d\n", aes_copied_bytes);
	}	

#ifdef DEBUG
	cavium_dbgprint("eout space left in previous output iovec : %d\n", eout_space_left);
#endif
	memcpy(enc_out, aes_out, eout_space_left);

	if (next_output_iovec)
		enc_out = outp_vec[idx].ptr;
	else
		enc_out += eout_space_left;

	aes_rem_num = aes_out_len - eout_space_left;

	if (enc_out)
		memcpy(enc_out, aes_out + eout_space_left, aes_rem_num);


	//update ssl output header
    cal_enc_len = ROUNDUP16(AES_BLOCK_SIZE + inp_len + SHA_DIGEST_LENGTH + 1);

	// External IV is included between Record HDR and Record Data, so increase iovec[0] size
	
#ifdef DEBUG	
	//cavium_dbgprint("out length: %d\n", out_rec_hdr->length);
        _hexdump(outp_vec[0].ptr, cal_enc_len);
#endif

	return cal_enc_len;
}


/*******************************************************************************
 *                                                                             *
 *      Generates TLS-1.1 / TLS-1.2 SHA1 HMAC, adds padding bytes and encrypts *
 *  payload data and stores the result in output iovec buffer. Initialisation  *
 *  vector  be used for encrytion is taken from the ext_iv_data pointer passed *
 *  to the function. Key and MAC secret to be used for cipher encryption and   *
 *  HMAC generation respectively are taken from the key structure passed.      *
 *  Sequence number and record header to be used for used for HMAC generation  *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the TLS-1.1 / TLS-1.2 record header.*
 *  Protocol : TLS-1.1 / TLS-1.2                                               *
 *  Cipher   : DES-CBC / DES3-CBC                                              *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing encryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to encrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store encryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  ext_iv_data : Pointer to explicit IV to be used for encryption.            *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_des_sha1_enc_tls11(EVP_DESOCT_HMAC_SHA1 *key,
		cav_iovec_t *inp_vec, uint16_t inp_cnt,
		cav_iovec_t *outp_vec, uint16_t outp_cnt,
		uint8_t *ext_iv_data,uint16_t *len)
{
	register uint64_t p0, p1, p2, p3, p4, p5;
	uint8_t hash_buf[128] = {0}, enc_buf[128] = {0}, final_buf[128] = {0};
	uint16_t inp_len = 0;
	uint64_t tmpvar = 0, tmpvar2 = 0, seq = 0, total_bits = 0;
	int32_t enc_rem_num = 0, hash_rem_num = 0,enc_copied_bytes = 0, hash_copied_bytes = 0, eout_space_left = 0;
	int32_t idx = 0, fill_bytes = 0, padlen = 0, next_output_iovec = 0, start_inner_hash = 0;

	uint8_t *mac_secret = NULL, *key_data = NULL, *hash_in = NULL, *hash_end = NULL, *enc_in = NULL, *enc_end = NULL, *enc_out = NULL;
	uint64_t *enckey = NULL, *ptr = NULL;
	struct tls_rec_hdr *in_rec_hdr = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];
	inp_len = cvmx_htobe16(in_rec_hdr->length);

	mac_secret = key->mac_secret;
	seq = cvmx_htobe64(key->aux.tls_aad_64[0]);

#ifdef DEBUG
	cavium_dbgprint("rec length: %d, seq: %llx\n", in_rec_hdr->length, seq);
	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
#endif

/*Generate External IV*/

	key_data = key->cipher_iv;


	enckey = (uint64_t *)key->cipher_key;

#ifdef DEBUG
	cavium_dbgprint("enckey\n");
	_hexdump(enckey, 24);
	cavium_dbgprint("iv\n");
	_hexdump(key_data, 8);
#endif
	if(key->cipher_type == OCT_DES_64 || key->cipher_type == OCT_DES_40)
	{
		enckey[1] = enckey[0];
		enckey[2] = enckey[0];
	}
	/* initialize encryption key */
	CVMX_MT_3DES_KEY(cvmx_htobe64(enckey[0]), 0);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enckey[1]), 1);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enckey[2]), 2);

	/* Init encryption iv */
	CVMX_MT_3DES_IV(cvmx_htobe64(*(uint64_t *)key_data));

	/* The first hash (mac secret ^ pad_1 (64 bytes) */
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT0), 0);
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT1), 1);
	CVMX_MT_HSH_IV(((uint64_t)CRY_SHA1_INIT2), 2);

    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_LOADUNA_INT64(p0, mac_secret, 0);
	    CVMX_LOADUNA_INT64(p1, mac_secret, 8);
	    p0 = p0 ^ (0x3636363636363636);
	    p1 = p1 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

	    p0 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	    p0 = p0 << 32;
	    p0 = p0 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(p0, 2);

	    CVMX_LOADUNA_INT64(p1, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(p1, 3);
	    CVMX_MT_HSH_DAT(p1, 4);
	    CVMX_MT_HSH_DAT(p1, 5);
	    CVMX_MT_HSH_DAT(p1, 6);
	    CVMX_MT_HSH_STARTSHA(p1);
    }
	CVMX_PREFETCH0(inp_vec[idx].ptr);

	if(key->ex_mode == EX_MODE_DIGEST) 
        total_bits = inp_len * 8;
    else if (key->ex_mode == EX_MODE_HMAC)
        total_bits = (SHA_CBLOCK + inp_len) * 8;
    else
        total_bits = ((SHA_CBLOCK + 8 + 1 + 2 + 2 + inp_len) * 8); 

    if(key->ex_mode == EX_MODE_DEFAULT) {
	    /* store sequence number + rec_type + rec_len */
        hash_buf_ptr_64 = (uint64_t *)hash_buf;
	    hash_buf_ptr_64[0] = cvmx_htobe64(seq);
	    CVMX_STOREUNA_INT32(key->aux.tls_aad_32[2], hash_buf, 8); //12
	    hash_buf[12] = key->aux.tls_aad_8[12]; //13


	    hash_copied_bytes = 13; /* seq num (8 bytes) + record type (1 byte) + record version (2 bytes) + record length (2 bytes) */
    } else if(key->ex_mode & 0x02){
        if(inp_len >= 16) {
            memcpy(hash_buf,inp_vec[0].ptr, 16);
            inp_vec[0].ptr += 16;
            inp_vec[0].size -= 16;
            hash_copied_bytes= 16;
        } else {
            memcpy(hash_buf,inp_vec[0].ptr, inp_len);
            inp_vec[0].ptr += inp_len;
            inp_vec[0].size -= inp_len;
            hash_copied_bytes= inp_len;
        }
    }
	enc_copied_bytes = 0;

/*Encrpyt External IV*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    enc_in = ext_iv_data;
	    CVMX_LOADUNA_INT64(p0, enc_in, 0);

	    CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

	    enc_out = outp_vec[idx].ptr ;

	    CVMX_MF_3DES_RESULT(p0);	

	    CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
    }

	hash_in = inp_vec[idx].ptr ;
	hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
	hash_rem_num = (hash_end - hash_in) + hash_copied_bytes;

	enc_in = inp_vec[idx].ptr ;
	enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
	enc_rem_num = (enc_end - enc_in);

    if(key->ex_mode == EX_MODE_DEFAULT)
    	enc_out = outp_vec[idx].ptr  + DES_BLOCK_SIZE;
    else 
        enc_out = outp_vec[idx].ptr;

	if (hash_rem_num <= SHA_CBLOCK || key->ex_mode == EX_MODE_ENCRYPT) {
		ptr = (uint64_t *)(hash_buf + hash_copied_bytes);	
		CVMX_STOREUNA_INT64(0, ptr, 0);
		CVMX_STOREUNA_INT64(0, ptr, 8);
		CVMX_STOREUNA_INT64(0, ptr, 16);
		CVMX_STOREUNA_INT64(0, ptr, 24);
		CVMX_STOREUNA_INT64(0, ptr, 32);
		CVMX_STOREUNA_INT64(0, ptr, 40);
		CVMX_STOREUNA_INT64(0, ptr, 48);
		CVMX_STOREUNA_INT64(0, ptr, 56);

		memcpy(hash_buf + hash_copied_bytes, hash_in, inp_vec[idx].size);	
		hash_copied_bytes += (inp_vec[idx].size);
		hash_buf[hash_copied_bytes] = 0x80;

        if(key->ex_mode != EX_MODE_ENCRYPT)
    		SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
		hash_copied_bytes = 0;
		hash_rem_num = 0;

		while (enc_rem_num >= DES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	

			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

			enc_in += DES_BLOCK_SIZE;
			enc_rem_num -= DES_BLOCK_SIZE;

			CVMX_MF_3DES_RESULT(p0);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);

			enc_out += DES_BLOCK_SIZE;
		}

		memcpy(enc_buf, enc_in, enc_rem_num);
		enc_copied_bytes = enc_rem_num;

		goto get_inner_hash;
	}

    if(key->ex_mode == EX_MODE_DIGEST || key->ex_mode == EX_MODE_HMAC)
	    ptr = (uint64_t *)(hash_in);
    else {
	/* inp_vec[0] has payload(inp_vec[0].size - sizeof(struct tls_rec_hdr)) > 51 bytes */
	/*hash copied bytes = 13, so pushing 3 more bytes from hash_in to sha1_rem*/
	    hash_buf[hash_copied_bytes] = hash_in[0];
	    hash_buf[hash_copied_bytes + 1] = hash_in[1];
	    hash_buf[hash_copied_bytes + 2] = hash_in[2];
	    ptr = (uint64_t *)(hash_in + 3);
    }

	CVMX_LOADUNA_INT64(p0, ptr, 0);
	CVMX_LOADUNA_INT64(p1, ptr, 8);
	CVMX_LOADUNA_INT64(p2, ptr, 16);
	CVMX_LOADUNA_INT64(p3, ptr, 24);
	CVMX_LOADUNA_INT64(p4, ptr, 32);

    hash_buf_ptr_64 = (uint64_t *)hash_buf;
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[1]), 1);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 4);
	CVMX_LOADUNA_INT64(p0, ptr, 40);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 6);

	CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p0));


    if(key->ex_mode == EX_MODE_DIGEST || key->ex_mode == EX_MODE_HMAC)
        hash_in += 48;
    else
	    hash_in += 51;
	hash_rem_num = (hash_end - hash_in);
	hash_copied_bytes = 0;

	/* SHA_CBLOCK chunks in 1st iovec */
	while ((hash_in + SHA_CBLOCK) <= hash_end) {
		/* get and load 1st des bytes */
		CVMX_LOADUNA_INT64(p2, enc_in, 0);	
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 0);
		CVMX_LOADUNA_INT64(p1, hash_in, 8);

		/* get 2nd des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 8);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

		/* 1st des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 2nd des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 1st des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 0);

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 16);
		CVMX_LOADUNA_INT64(p1, hash_in, 24);

		/* get 3rd des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 16);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

		/* 2nd des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 3rd des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 2nd des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 32);
		CVMX_LOADUNA_INT64(p1, hash_in, 40);

		/* get 4th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 24);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

		/* 3rd des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 4th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 3rd des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 16);

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 48);
		CVMX_LOADUNA_INT64(p1, hash_in, 56);

		/* 4th des result */
		CVMX_MF_3DES_RESULT(p3);	

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		/* get 5th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 32);

		/* store 4th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

		/* load 5th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* get 6th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 40);

		/* 5th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 6th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* get 7th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 48);

		/* store 5th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 32);

		/* 6th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 7th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* get 8th des input */
		CVMX_LOADUNA_INT64(p2, enc_in, 56);

		/* store 6th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

		/* 7th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* load 8th des bytes */
		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

		/* store 7th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 48);

		hash_in += SHA_CBLOCK;
		enc_in += SHA_CBLOCK;
		enc_rem_num -= SHA_CBLOCK;
		hash_rem_num -= SHA_CBLOCK;

		CVMX_PREFETCH(enc_in, SHA_CBLOCK);

		/* 8th des result */
		CVMX_MF_3DES_RESULT(p3);	

		/* store 8th des result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);

		enc_out += SHA_CBLOCK;
	}

	/* for each iovec */
	while (idx++ < inp_cnt) {
		/* for each SHA_CBLOCK in iovec */
		while ((hash_in + SHA_CBLOCK) <= hash_end) {
			/* get and load 1st des bytes */
			CVMX_LOADUNA_INT64(p2, enc_in, 0);	
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 0);
			CVMX_LOADUNA_INT64(p1, hash_in, 8);

			/* get 2nd des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 8);

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

			/* 1st des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 2nd des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 1st des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 0);

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 16);
			CVMX_LOADUNA_INT64(p1, hash_in, 24);

			/* get 3rd des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 16);

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

			/* 2nd des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 3rd des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 2nd des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 32);
			CVMX_LOADUNA_INT64(p1, hash_in, 40);

			/* get 4th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 24);

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

			/* 3rd des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 4th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 3rd des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 16);

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 48);
			CVMX_LOADUNA_INT64(p1, hash_in, 56);

			/* 4th des result */
			CVMX_MF_3DES_RESULT(p3);	

			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

			/* get 5th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 32);

			/* store 4th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

			/* load 5th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* get 6th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 40);

			/* 5th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 6th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* get 7th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 48);

			/* store 5th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 32);

			/* 6th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 7th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* get 8th des input */
			CVMX_LOADUNA_INT64(p2, enc_in, 56);

			/* store 6th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

			/* 7th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* load 8th des bytes */
			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p2));

			/* store 7th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 48);

			hash_in += SHA_CBLOCK;
			enc_in += SHA_CBLOCK;
			enc_rem_num -= SHA_CBLOCK;
			hash_rem_num -= SHA_CBLOCK;

			CVMX_PREFETCH(enc_in, SHA_CBLOCK);

			/* 8th des result */
			CVMX_MF_3DES_RESULT(p3);	

			/* store 8th des result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);

			enc_out += SHA_CBLOCK;
		} /* for each SHA_CBLOCK in iovec */

		hash_rem_num = (hash_end - hash_in);
		enc_rem_num = (hash_end - enc_in);

#ifdef DEBUG
		cavium_dbgprint("hash_rem_num: %d, enc_rem_num: %d (%d)\n", hash_rem_num, enc_rem_num, __LINE__);
#endif
		/* copy the remaining hash bytes */
		memcpy(hash_buf, hash_in, hash_rem_num);
		hash_copied_bytes = hash_rem_num;
		hash_rem_num = 0;

		if (idx < inp_cnt) {
			/* try fill the sha1 block, get bytes from next iovec */
			hash_in = inp_vec[idx].ptr;
			hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
			hash_rem_num = (hash_end - hash_in);
		}
#ifdef DEBUG
		cavium_dbgprint("hash_rem_num: %d, enc_rem_num: %d (%d)\n", hash_rem_num, enc_rem_num, __LINE__);
#endif

		/* less bytes than needed, try to finish inner hash */
		if (hash_rem_num < (SHA_CBLOCK - hash_copied_bytes)) {
			memcpy(hash_buf + hash_copied_bytes, hash_in, hash_rem_num);
			hash_copied_bytes += hash_rem_num;
			hash_buf[hash_copied_bytes] = 0x80;

			memset(hash_buf + hash_copied_bytes + 1, 0, (SHA_CBLOCK - hash_copied_bytes - 1));

			SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
			hash_rem_num = 0;
			hash_copied_bytes = 0;

			/* complete enc bytes in previous iovec */
			while (enc_rem_num >= DES_BLOCK_SIZE) {
				CVMX_LOADUNA_INT64(p0, enc_in, 0);

				CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

				enc_in += DES_BLOCK_SIZE;
				enc_rem_num -= DES_BLOCK_SIZE;

				CVMX_MF_3DES_RESULT(p0);

				CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);

				enc_out += DES_BLOCK_SIZE;
			}

			if (enc_rem_num) {
				/* copy the remaining enc bytes from previous iovec */
				memcpy(enc_buf, enc_in, enc_rem_num);
				enc_copied_bytes = enc_rem_num;
				eout_space_left = enc_rem_num;
				enc_rem_num = 0;

				if (idx < inp_cnt) {
					/* move to next iovec */
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					enc_rem_num = (enc_end - enc_in);
				}

				/* fill bytes from next iovec */
				if (enc_rem_num <= (DES_BLOCK_SIZE - enc_copied_bytes)) {
					memcpy(enc_buf + enc_copied_bytes, enc_in, enc_rem_num);
					enc_copied_bytes += enc_rem_num;
					enc_rem_num = 0;
					if (idx < outp_cnt)
						next_output_iovec = 1;
				} else {
					uint64_t enc_result;

					fill_bytes = (DES_BLOCK_SIZE - enc_copied_bytes);
					memcpy(enc_buf + enc_copied_bytes, enc_in, fill_bytes);
					enc_copied_bytes = DES_BLOCK_SIZE;
					enc_rem_num -= fill_bytes;

					CVMX_LOADUNA_INT64(p0, enc_buf, 0);

					CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

					enc_in += fill_bytes;

					CVMX_MF_3DES_RESULT(enc_result);
                    enc_result = cvmx_htobe64(enc_result);

					/* copy into the space left */
					memcpy(enc_out, &enc_result, eout_space_left);	

					/* move to next output iovec */
					enc_out = outp_vec[idx].ptr;
					memcpy(enc_out, ((uint8_t *)&enc_result + eout_space_left), fill_bytes);
					enc_out += fill_bytes;
					eout_space_left = 0;
					enc_copied_bytes = 0;

					/* encrypt DES_BLOCK_SIZE chunks if any */
					while (enc_rem_num >= DES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);

						CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

						enc_in += DES_BLOCK_SIZE;
						enc_rem_num -= DES_BLOCK_SIZE;

						CVMX_MF_3DES_RESULT(p0);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

						enc_out += DES_BLOCK_SIZE;
					}
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;
				}
			} else {
				if (idx < inp_cnt) {
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					enc_rem_num = (enc_end - enc_in);
				}

				if (idx < outp_cnt)
					enc_out = outp_vec[idx].ptr;

				while (enc_rem_num >= DES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);

					CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

					enc_in += DES_BLOCK_SIZE;
					enc_rem_num -= DES_BLOCK_SIZE;

					CVMX_MF_3DES_RESULT(p0);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

					enc_out += DES_BLOCK_SIZE;
				}
				if (enc_in) {
					memcpy(enc_buf, enc_in, enc_rem_num);	
					enc_copied_bytes = enc_rem_num;
					enc_rem_num = 0;
				}
			}
			goto get_inner_hash;
		}

		/* more than fill bytes in next iovec */
		fill_bytes = SHA_CBLOCK - hash_copied_bytes;
		memcpy(hash_buf + hash_copied_bytes, hash_in, fill_bytes);
		hash_copied_bytes += fill_bytes;

		hash_in += fill_bytes;
		hash_rem_num -= fill_bytes;

		ptr = (uint64_t *)hash_buf;
		CVMX_LOADUNA_INT64(p0, ptr, 0);
		CVMX_LOADUNA_INT64(p1, ptr, 8);
		CVMX_LOADUNA_INT64(p2, ptr, 16);
		CVMX_LOADUNA_INT64(p3, ptr, 24);
		CVMX_LOADUNA_INT64(p4, ptr, 32);
		CVMX_LOADUNA_INT64(p5, ptr, 40);

		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
		CVMX_LOADUNA_INT64(p0, ptr, 48);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_LOADUNA_INT64(p1, ptr, 56);

		CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

		while (enc_rem_num >= DES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);

			CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

			enc_in += DES_BLOCK_SIZE;
			enc_rem_num -= DES_BLOCK_SIZE;

			CVMX_MF_3DES_RESULT(p0);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);

			enc_out += DES_BLOCK_SIZE;
		}

		/* if sha1 remain bytes are less than SHA_CBLOCK, finish inner hash */
		if (hash_rem_num < SHA_CBLOCK) {
			memcpy(hash_buf, hash_in, hash_rem_num);	
			hash_copied_bytes = hash_rem_num;
			hash_buf[hash_copied_bytes] = 0x80;

			memset(hash_buf + hash_copied_bytes + 1, 0, (SHA_CBLOCK - hash_copied_bytes - 1));

			SHA1_FINISH(hash_buf, hash_copied_bytes, total_bits);
			hash_rem_num = 0;
			hash_copied_bytes = 0;

			start_inner_hash = 1;
		}

		if (enc_rem_num) {
			uint64_t enc_result;

			memcpy(enc_buf, enc_in, enc_rem_num);
			enc_copied_bytes = enc_rem_num;
			eout_space_left = enc_rem_num;

			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;	
			enc_rem_num = (enc_end - enc_in);

			if (enc_rem_num <= (DES_BLOCK_SIZE - enc_copied_bytes)) {
				/* copy the remaining enc bytes */
				memcpy(enc_buf + enc_copied_bytes, enc_in, enc_rem_num);
				enc_copied_bytes += enc_rem_num;
				enc_rem_num = 0;
				if (idx < outp_cnt)
					next_output_iovec = 1;

				goto get_inner_hash;
			} else {
				fill_bytes = (DES_BLOCK_SIZE - enc_copied_bytes);

				/* chance that enc bytes are crossing hash bytes */
				if ((enc_rem_num - fill_bytes) < hash_rem_num) {
					ptr = (uint64_t *)hash_in;	
					CVMX_LOADUNA_INT64(p0, ptr, 0);
					CVMX_LOADUNA_INT64(p1, ptr, 8);
					CVMX_LOADUNA_INT64(p2, ptr, 16);
					CVMX_LOADUNA_INT64(p3, ptr, 24);
					CVMX_LOADUNA_INT64(p4, ptr, 32);
					CVMX_LOADUNA_INT64(p5, ptr, 40);

					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
					CVMX_LOADUNA_INT64(p0, ptr, 48);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
					CVMX_LOADUNA_INT64(p1, ptr, 56);

					CVMX_MT_HSH_STARTSHA(cvmx_htobe64(p1));

					hash_in += SHA_CBLOCK;
					hash_rem_num -= SHA_CBLOCK;
				}

				memcpy(enc_buf + enc_copied_bytes, enc_in, fill_bytes); 
				enc_copied_bytes += fill_bytes;

				CVMX_LOADUNA_INT64(p0, enc_buf, 0);

				CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

				enc_in += fill_bytes;
				enc_rem_num -= fill_bytes;

				CVMX_MF_3DES_RESULT(enc_result);
                enc_result = cvmx_htobe64(enc_result);

				memcpy(enc_out, &enc_result, eout_space_left);

				/* move to next output iovec */
				enc_out = outp_vec[idx].ptr;
				memcpy(enc_out, ((uint8_t *)&enc_result + eout_space_left), fill_bytes);
				enc_out += fill_bytes;

				eout_space_left = 0;
				enc_copied_bytes = 0;

				if (start_inner_hash) {
					/* encrypt DES_BLOCK_SIZE chunks if any */
					while (enc_rem_num >= DES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);

						CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

						enc_in += DES_BLOCK_SIZE;
						enc_rem_num -= DES_BLOCK_SIZE;

						CVMX_MF_3DES_RESULT(p0);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

						enc_out += DES_BLOCK_SIZE;
					}
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;

					goto get_inner_hash;
				}
			}
		} else {
			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
			enc_rem_num = (enc_end - enc_in);

			if (idx < outp_cnt)
				enc_out = outp_vec[idx].ptr;

			if (start_inner_hash) {
				while (enc_rem_num >= DES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);

					CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

					enc_in += DES_BLOCK_SIZE;
					enc_rem_num -= DES_BLOCK_SIZE;

					CVMX_MF_3DES_RESULT(p0);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	

					enc_out += DES_BLOCK_SIZE;
				}

				if (enc_in) {
					memcpy(enc_buf, enc_in, enc_rem_num);
					enc_copied_bytes = enc_rem_num;
					enc_rem_num = 0;
				}

				goto get_inner_hash;
			}
		}
	} /* for each iovec */

get_inner_hash:
	/* get inner hash */
	CVMX_MF_HSH_IV(tmpvar, 0);
	CVMX_MF_HSH_IV(tmpvar2, 1);
	CVMX_MF_HSH_IV(p0, 2);

    if(key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar),outp_vec[0].ptr,0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(tmpvar2),outp_vec[0].ptr,8);
        CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(p0)),outp_vec[0].ptr,16);
        return 1;
    } else if (key->ex_mode == EX_MODE_ENCRYPT) {
        return 1;
    }

#ifdef DEBUG
	cavium_dbgprint("hash_copied_bytes: %d, hash_rem_num: %d (%d)\n", hash_copied_bytes, hash_rem_num, __LINE__);
	_hexdump(hash_buf, hash_copied_bytes);
	cavium_dbgprint("enc_copied_bytes: %d, enc_rem_num: %d (%d)\n", enc_copied_bytes, enc_rem_num, __LINE__);
	_hexdump(enc_buf, enc_copied_bytes);;
	cavium_dbgprint("innerhash\n");
	_hexdump(&tmpvar, 8);
	_hexdump(&tmpvar2, 8);
	cavium_dbgprint("%llx\n", p0);
#endif

	/* start outer hash (mac secret ^ pad_2 + inner hash) */

	/* SHA init */
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);


	CVMX_LOADUNA_INT64(p1, mac_secret, 0);
	p1 = p1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 0);
	CVMX_LOADUNA_INT64(p1, mac_secret, 8);
	p1 = p1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);

	p1 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	p1 = p1 << 32;
	p1 = p1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(p1, 2);

	CVMX_LOADUNA_INT64(p1, CRY_HMAC_PAD2_64, 0);
	CVMX_MT_HSH_DAT(p1, 3);
	CVMX_MT_HSH_DAT(p1, 4);
	CVMX_MT_HSH_DAT(p1, 5);
	CVMX_MT_HSH_DAT(p1, 6);
	CVMX_MT_HSH_STARTSHA(p1);


	ptr = (uint64_t *)&enc_buf[enc_copied_bytes];


	CVMX_MT_HSH_DAT(tmpvar, 0 );
	CVMX_MT_HSH_DAT(tmpvar2, 1 );
	p0 = p0 & 0xFFFFFFFF00000000ULL;
	p0 = p0 | 0x0000000080000000ULL;

	CVMX_MT_HSH_DAT(p0, 2);
	CVMX_MT_HSH_DATZ(3);
	CVMX_MT_HSH_DATZ(4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	total_bits = ((SHA_CBLOCK + SHA_DIGEST_LENGTH) * 8);
	CVMX_MT_HSH_STARTSHA(total_bits);

	enc_copied_bytes += SHA_DIGEST_LENGTH;

	/* calculate pad length */
	padlen = DES_BLOCK_SIZE - ((enc_copied_bytes) % DES_BLOCK_SIZE);

	/* add pad bytes pattren (padlen - 1) */
	memset(enc_buf + enc_copied_bytes, padlen - 1, padlen);
	enc_copied_bytes += padlen;

	/* get outer hash */
	CVMX_MF_HSH_IV(p0, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p0), ptr, 0);
	CVMX_MF_HSH_IV(p1, 1);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p1), ptr, 8);
	CVMX_MF_HSH_IV(p0, 2);
	CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(p0)), ptr, 16);

    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr,ptr,SHA_DIGEST_LENGTH);
        return 1;
    }

#ifdef DEBUG
	cavium_dbgprint("Outer Hash\n");
	_hexdump(ptr, SHA_DIGEST_LENGTH);
#endif
	uint8_t *out_enc_ptr = final_buf;
	uint8_t *in_enc_ptr = enc_buf;
	enc_rem_num = (enc_copied_bytes - eout_space_left);

#ifdef DEBUG
	cavium_dbgprint("enc_copied_bytes: %d eout_space_left: %d (%d)\n", enc_copied_bytes, eout_space_left, __LINE__);
	_hexdump(enc_buf, enc_copied_bytes);
#endif
	while (enc_copied_bytes >= DES_BLOCK_SIZE) {
		CVMX_LOADUNA_INT64(p0, in_enc_ptr, 0);

		CVMX_MT_3DES_ENC_CBC(cvmx_htobe64(p0));

		in_enc_ptr += DES_BLOCK_SIZE;
		enc_copied_bytes -= DES_BLOCK_SIZE;	

		CVMX_MF_3DES_RESULT(p0);
#ifdef DEBUG
		cavium_dbgprint("out: %llx\n", p0);
#endif

		CVMX_STOREUNA_INT64(cvmx_htobe64(p0), out_enc_ptr, 0);

		out_enc_ptr += DES_BLOCK_SIZE;
	}
	if (enc_copied_bytes)
		cavium_dbgprint("ERROR: enc_copied_bytes: %d\n", enc_copied_bytes);

	memcpy(enc_out, final_buf, eout_space_left);

	if (next_output_iovec)
		enc_out = outp_vec[idx].ptr;			
	else
		enc_out += eout_space_left;

	if (enc_out)
		memcpy(enc_out, final_buf + eout_space_left, enc_rem_num);


	//update ssl output header
	*len = ROUNDUP8(DES_BLOCK_SIZE + inp_len + SHA_DIGEST_LENGTH + 1); 

	// External IV is included between Record HDR and Record Data, so increase iovec[0] size


	return 1;
}


/*******************************************************************************
 *                                                                             *
 *    Decrypts input record and generates TLS-1.1 / TLS - 1.2 SHA1 HMAC;       *
 *  authenticates the record and stores result in output iovec buffer.         *
 *  Initialisation vector to be used for decrytion is extracted from the       *
 *  intial bytes of input buffer passed to the function. Key and MAC secret to *
 *  be for cipher decryption and HMAC generation respectively are taken from   *
 *  the key structure passed. Sequence number and record header used in        *
 *  generating TLS-1.1 / TLS-1.2 HMAC are extracted from AAD stored in KEY     *
 *  structure. First 8bytes of  AAD holds sequence number and next 5 bytes     * 
 *  holds the TLS-1.1 / TLS-1.2 record header.                                 *
 *  Protocol : TLS-1.1  / TLS-1.2                                              *
 *  Cipher   : DES-CBC / DES3-CBC                                              *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing decryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to decrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store decryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_des_sha1_dec_tls11(EVP_DESOCT_HMAC_SHA1 *key,
		cav_iovec_t *inp_vec, uint16_t inp_cnt,
		cav_iovec_t *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t r0, r1, r2, r3;
	uint8_t padarray[2*16] = {0}, hash_buf[128] = {0}, hash_rem_buf[128] = {0}, buf[128] = {0};
	uint8_t des = 0, inner_hash_done = 0, output_overlap = 0, cipher_type = 0, mac_size = 0;
	int32_t enc_data_len = 0, raw_data_len = 0, pad_data_len = 0, hash_input_rem_len = 0;
	int32_t overlap_buf_size = 0, enc_rem_cur_chunk = 0, inner_hash_total_bytes = 0, outer_hash_total_bytes = 0, prev_iovec_rem_bytes = 0, ret_val = 1;
	int32_t i = 0;
	uint64_t seq_num = 0, inner_hash_total_bits = 0, outer_hash_total_bits = 0;
	uint8_t *hash_bal = NULL, *out_ptr = NULL, *in_chunk_cur = NULL, *out_chunk_cur = NULL, *enc_rem_ptr = NULL, *mac_secret = NULL;
    uint8_t *ext_iv_ptr = NULL;
	uint64_t *enc_key = NULL, *key_data = NULL;
	struct tls_rec_hdr *in_rec_hdr = NULL;
    uint16_t *hash_buf_ptr_16 = NULL;
    uint32_t *hash_buf_ptr_32 = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct iovec_cav hash_overlap_buf;
	hash_overlap_buf.size = 0;

#ifdef DEBUG
	uint32_t itr = 0;

	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
	for (itr = 0; itr < inp_cnt; itr++)
	{
		cavium_dbgprint("inp_vec[%d].ptr: %p, inp_vec[%d].size: %d\n", itr, inp_vec[itr].ptr, itr, inp_vec[itr].size);
		_hexdump(inp_vec[itr].ptr, inp_vec[itr].size);
	}

#endif

	in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];
	enc_data_len = cvmx_htobe16(in_rec_hdr->length);

#ifdef DEBUG
	cavium_dbgprint("enc_data_len: %d (%d)\n", enc_data_len, __LINE__);
#endif

	cipher_type = key->cipher_type;

	if ((enc_data_len % DES_BLOCK_SIZE) && (key->ex_mode == EX_MODE_DEFAULT || key->ex_mode == EX_MODE_ENCRYPT))
	{
		cavium_dbgprint(
			"%s:%d Not properly padded data is received for cipher %s\n",
			__FUNCTION__,__LINE__, cipher_strings[cipher_type]);
		ret_val = 0;
		return ret_val;
	}

	switch(cipher_type)
	{
		case OCT_DES_40:
		case OCT_DES_64:
			des = 1;
			break;

		case OCT_3DES:
			break;
		default:
			break;
	}

	mac_size = SHA_DIGEST_LENGTH;

	seq_num = cvmx_htobe64(key->aux.tls_aad_64[0]);
	enc_key = (uint64_t *)key->cipher_key;
	mac_secret = key->mac_secret;
	key_data = (uint64_t *)key->cipher_iv;


	/*Load DES key*/
	if(des == 1)
	{
		enc_key[1] = enc_key[0];
		enc_key[2] = enc_key[0];
	}
	CVMX_MT_3DES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enc_key[1]), 1);
	CVMX_MT_3DES_KEY(cvmx_htobe64(enc_key[2]), 2);

	/* Pad data can not be more than DES_BLOCK_SIZE
	 * As per DES CBC implementation, previous des block result is fed as IV for current des block
	 * So decrypt last 16 bytes to know PAD length*/


	// calculate pad length by decrypting last 16 bytes
	// if the last iovec buffer contains data >= 16
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    if(inp_vec[inp_cnt - 1].size >= 2*DES_BLOCK_SIZE)
	    {
	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 16, 0);

	    	CVMX_MT_3DES_IV(cvmx_htobe64(r0));

	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 8, 0);

	    	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r1));


	    	CVMX_MF_3DES_RESULT(r0);

	    	pad_data_len = (r0 & 0xff ) + 1;
	    }
	    else
	    {
	    	// X + Y = 16
	    	// X bytes from iovec[inp_cnt - 2]
	    	// Y bytes from iovec[inp_cnt - 1]
	    	memcpy(padarray, inp_vec[inp_cnt - 2].ptr + (inp_vec[inp_cnt - 2].size - (16 - inp_vec[inp_cnt - 1].size)), (16 - inp_vec[inp_cnt - 1].size));
	    	memcpy(padarray + (16 - inp_vec[inp_cnt - 1].size), inp_vec[inp_cnt - 1].ptr, inp_vec[inp_cnt - 1].size);
	    	CVMX_LOADUNA_INT64(r0, padarray, 0);
	    	CVMX_LOADUNA_INT64(r1, padarray, 8);

	    	CVMX_MT_3DES_IV(cvmx_htobe64(r0));

	    	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r1));

	    	CVMX_MF_3DES_RESULT(r0);

	    	pad_data_len = (r0 & 0xff ) + 1;
	    }
    }
#ifdef DEBUG
	cavium_dbgprint("pad_data_len: %d (%d)\n", pad_data_len, __LINE__);
#endif


	CVMX_MT_3DES_IV(cvmx_htobe64(key_data[0]));

    if(key->ex_mode == EX_MODE_DEFAULT)
    	raw_data_len = enc_data_len - mac_size - pad_data_len - DES_BLOCK_SIZE;
    else
        raw_data_len = enc_data_len;

	/* If message length < 0, then we are sure we received error packet,
 	   to avoid timing attack decrypt packet and calculate HMAC and throw error at end */
	if (raw_data_len < 0)
	{
		raw_data_len = enc_data_len - mac_size;
		pad_data_len = 0;
	}

#ifdef DEBUG
	cavium_dbgprint("raw_data_len: %d (%d)\n", raw_data_len, __LINE__);
#endif
	// first chunk - 16+48

	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_LOADUNA_INT64(r0, mac_secret, 0);
	    CVMX_LOADUNA_INT64(r1, mac_secret, 8);

	    r0 = r0 ^ (0x3636363636363636);
	    r1 = r1 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);

	    r0 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	    r0 = r0 << 32;
	    r0 = r0 ^ (0x3636363636363636);
	    CVMX_MT_HSH_DAT(r0, 2);

	    CVMX_LOADUNA_INT64(r1, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(r1, 3);
	    CVMX_MT_HSH_DAT(r1, 4);
	    CVMX_MT_HSH_DAT(r1, 5);
	    CVMX_MT_HSH_DAT(r1, 6);
	    CVMX_MT_HSH_STARTSHA(r1);

    }
	/*8 + 1 + 2 + 2 = 13 bytes are left over*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    CVMX_STOREUNA_INT64(cvmx_htobe64(seq_num), hash_buf, 0);
	    hash_buf[8] = in_rec_hdr->type;
        hash_buf_ptr_16 = (uint16_t *)&hash_buf[9];
	    hash_buf_ptr_16[0]  = in_rec_hdr->u.version;
	    hash_buf_ptr_16[1]  = cvmx_htobe16(raw_data_len);
    } else if (key->ex_mode & 0x02) {
        if(raw_data_len > 16) {
            memcpy(hash_buf, inp_vec[0].ptr, 16);
            inp_vec[0].ptr += 16;
            inp_vec[0].size -= 16;
        } else {
            memcpy(hash_buf, inp_vec[0].ptr, raw_data_len);
            inp_vec[0].ptr += raw_data_len;
            inp_vec[0].size -= raw_data_len;
        }
    }

	hash_bal = hash_buf;

#ifdef DEBUG
	cavium_dbgprint("hash bal : %d\n", __LINE__);
	_hexdump(hash_bal, 13);
#endif

	hash_input_rem_len = raw_data_len;

/* Remove External IV*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
        ext_iv_ptr = inp_vec[0].ptr;
	    CVMX_LOADUNA_INT64(r0, ext_iv_ptr, 0);
	    CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r0));
        inp_vec[0].ptr += DES_BLOCK_SIZE;
        inp_vec[0].size -= DES_BLOCK_SIZE;
        outp_vec[0].size -= DES_BLOCK_SIZE;
    }

	if (inp_cnt == 1) {
		in_chunk_cur = inp_vec[0].ptr;
        if(key->ex_mode & 0x02)
            out_chunk_cur = inp_vec[0].ptr;
        else
		    out_chunk_cur = outp_vec[0].ptr;

		while (hash_input_rem_len >= 64) {
			hash_bal = block_processing_des_sha1_dec_tls10(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);

#ifdef DEBUG
			cavium_dbgprint("hash_bal : %d\n", __LINE__);
			_hexdump(hash_bal, 13);
#endif
			in_chunk_cur += 64;
			out_chunk_cur += 64;

			hash_input_rem_len -= 64;
		}
		out_ptr = out_chunk_cur;
#ifdef DEBUG
		cavium_dbgprint("hash_bal : %d \n", __LINE__);
		_hexdump(hash_bal, 13);
#endif
		// remaining data is compulsory a multiple of 8
         if(key->ex_mode == EX_MODE_DEFAULT || 
                 (key->ex_mode == EX_MODE_ENCRYPT && hash_input_rem_len)) {
		    do {
		    	CVMX_LOADUNA_INT64(r0, in_chunk_cur, 0);

		    	CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r0));

		    	in_chunk_cur += 8;

		    	CVMX_MF_3DES_RESULT(r1);
#ifdef DEBUG
			cavium_dbgprint("r1 : %llx\n", r1);
#endif
		    	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), out_chunk_cur, 0);

		    	out_chunk_cur += 8;

		    } while (((inp_vec[0].ptr + inp_vec[0].size) - in_chunk_cur) > 0);
        }
		// Along with 13 bytes hash balance, hash_input_rem_len 
		// should be given to hash engine as input
		memset(hash_rem_buf, 0, 128);
         if(key->ex_mode & 0x02 && hash_input_rem_len) 
 		    memcpy(hash_rem_buf, hash_bal, (hash_input_rem_len>16)?16:hash_input_rem_len);
         else if(key->ex_mode == EX_MODE_DEFAULT)
		    memcpy(hash_rem_buf, hash_bal, 13);

		if (hash_input_rem_len)	{
            if(key->ex_mode == EX_MODE_DEFAULT)
    			memcpy(hash_rem_buf + 13, out_ptr, hash_input_rem_len);
            else if ((key->ex_mode & 0x02) && (hash_input_rem_len > 16))
    			memcpy(hash_rem_buf + 16, out_ptr, hash_input_rem_len - 16);
		}

        if(key->ex_mode & 0x02) 
        	hash_rem_buf[hash_input_rem_len] = 0x80;
        else if (key->ex_mode == EX_MODE_DEFAULT)
    		hash_rem_buf[13 + hash_input_rem_len] = 0x80;
		uint64_t * ptr = (uint64_t *)hash_rem_buf;
#ifdef DEBUG
		cavium_dbgprint("hash_rem_buf\n", __LINE__);
		_hexdump(hash_rem_buf, 128);
#endif
		if((((hash_input_rem_len + 13) >= 56) && (key->ex_mode == EX_MODE_DEFAULT)) ||
           ((key->ex_mode & 0x02) && (hash_input_rem_len >= 56))) {
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

			ptr += 8;
		}
#ifdef DEBUG
		cavium_dbgprint("raw_data_length : %d\n", raw_data_len);
#endif
        if(key->ex_mode == EX_MODE_DIGEST)
		    inner_hash_total_bytes = raw_data_len;
        if(key->ex_mode == EX_MODE_HMAC)
		    inner_hash_total_bytes = SHA_CBLOCK + raw_data_len;
        if(key->ex_mode == EX_MODE_DEFAULT)
		    inner_hash_total_bytes = SHA_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
		inner_hash_total_bits = inner_hash_total_bytes * 8;
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
		CVMX_MT_HSH_STARTSHA(inner_hash_total_bits);
	} else if(inp_cnt > 1) {
		hash_overlap_buf.ptr = buf;
		hash_overlap_buf.size = 0;
		iovec_record_process_t *iovec_stat = NULL;
		iovec_stat = malloc(inp_cnt * sizeof(iovec_record_process_t));
		memset(iovec_stat, 0, inp_cnt * sizeof(iovec_record_process_t));

		prev_iovec_rem_bytes = 0;
		// populate loop counters
		while(hash_input_rem_len && (i < inp_cnt))
		{
			if(hash_input_rem_len >= inp_vec[i].size)
			{
				iovec_stat[i].hash_size = inp_vec[i].size;
				hash_input_rem_len -= inp_vec[i].size;
			}
			else
			{
				iovec_stat[i].hash_size = hash_input_rem_len;
				hash_input_rem_len = 0;
			}

			if(prev_iovec_rem_bytes)
			{
				iovec_stat[i].hash_req_bytes = 64 - prev_iovec_rem_bytes;
				// in case of last chunk not holding enough bytes to fill 64 byte block
				if(iovec_stat[i].hash_size < iovec_stat[i].hash_req_bytes)
				{
					iovec_stat[i].hash_req_bytes = iovec_stat[i].hash_size;
				}
			}
			else
			{
				iovec_stat[i].hash_req_bytes = 0;
			}
			iovec_stat[i].crypto_itr = (iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) / DES_BLOCK_SIZE;
			iovec_stat[i].hash_itr = iovec_stat[i].crypto_itr / 8;

			iovec_stat[i].hash_rem_bytes = ((iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) - (iovec_stat[i].hash_itr * 64));
			prev_iovec_rem_bytes = iovec_stat[i].hash_rem_bytes;

			i++;
		}
		uint16_t num_hash_iovec = i;
		// need to update enc_digest_ptr
		if(iovec_stat[i-1].hash_size == inp_vec[i-1].size)
		{
			enc_rem_ptr = inp_vec[i].ptr;
		}
		else
		{
			enc_rem_ptr = inp_vec[i-1].ptr + iovec_stat[i-1].hash_size;
			enc_rem_cur_chunk = inp_vec[i-1].size - iovec_stat[i-1].hash_size;

		}

#ifdef DEBUG
		for(i = 0; i < inp_cnt; i++)
		{
			cavium_dbgprint("Chunk : %d\n", i+1);
			cavium_dbgprint("hash_req_bytes : %d\n",iovec_stat[i].hash_req_bytes);
			cavium_dbgprint("crypto_itr : %d\n", iovec_stat[i].crypto_itr);
			cavium_dbgprint("hash_itr : %d\n", iovec_stat[i].hash_itr);
			cavium_dbgprint("hash_rem_bytes : %d\n",iovec_stat[i].hash_rem_bytes);
		}
#endif

		for (i = 0; i < num_hash_iovec; i++) {
			in_chunk_cur = inp_vec[i].ptr;
			out_chunk_cur = outp_vec[i].ptr;

			if (iovec_stat[i].hash_req_bytes) {	
				memcpy(hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, in_chunk_cur, iovec_stat[i].hash_req_bytes);
				in_chunk_cur += iovec_stat[i].hash_req_bytes;

				if ((iovec_stat[i-1].hash_rem_bytes + iovec_stat[i].hash_req_bytes) < 64) {
					hash_overlap_buf.size += iovec_stat[i].hash_req_bytes;
					out_chunk_cur = outp_vec[i].ptr + iovec_stat[i-1].hash_req_bytes + (64 * iovec_stat[i-1].hash_itr);
					output_overlap = 1;
					continue;
				} else {
					hash_bal = block_processing_des_sha1_dec_tls10(hash_overlap_buf.ptr, hash_overlap_buf.ptr, hash_bal, key->ex_mode);
#ifdef DEBUG
					cavium_dbgprint("hash_bal : %d\n", __LINE__);
					_hexdump(hash_bal, 13);
#endif

					memcpy(outp_vec[i].ptr + (outp_vec[i].size - iovec_stat[i-1].hash_rem_bytes), hash_overlap_buf.ptr, iovec_stat[i-1].hash_rem_bytes);
					memcpy(out_chunk_cur, hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, iovec_stat[i].hash_req_bytes);

					out_chunk_cur += iovec_stat[i].hash_req_bytes;
					hash_overlap_buf.size = 0;
				}
			}

			register uint8_t hash_itr = iovec_stat[i].hash_itr;
			while (hash_itr) {
				hash_bal = block_processing_des_sha1_dec_tls10(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);
#ifdef DEBUG
				cavium_dbgprint("hash_bal : %d\n", __LINE__);
				_hexdump(hash_bal, 13);
#endif

				in_chunk_cur += 64;
				out_chunk_cur += 64;

				hash_itr--;
			}

			if (iovec_stat[i].hash_rem_bytes) {
				//copy rem bytes to hash overlap buf
				memcpy(hash_overlap_buf.ptr, in_chunk_cur, iovec_stat[i].hash_rem_bytes);
				in_chunk_cur += iovec_stat[i].hash_rem_bytes;
				hash_overlap_buf.size = iovec_stat[i].hash_rem_bytes;
			}
		}

#ifdef DEBUG
		cavium_dbgprint("hash_overlap_buf.size : %d \n", hash_overlap_buf.size);
#endif
		//obviously this case gets hit all the time
		if (inner_hash_done == 0) {
			//if(hash_overlap_buf.size)
			{
				uint8_t *ptr = NULL;
				overlap_buf_size = hash_overlap_buf.size;
				memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size, enc_rem_ptr, enc_rem_cur_chunk);
				overlap_buf_size += enc_rem_cur_chunk;
#ifdef DEBUG
				cavium_dbgprint("enc_rem_cur_chunk : %d\n", enc_rem_cur_chunk);
#endif
				if (enc_rem_cur_chunk < (SHA_DIGEST_LENGTH + pad_data_len)) {
					memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size + enc_rem_cur_chunk, 
							inp_vec[inp_cnt-1].ptr, inp_vec[inp_cnt-1].size);

					overlap_buf_size += inp_vec[inp_cnt-1].size;
#ifdef DEBUG
					cavium_dbgprint("overlap_buf_size : %d\n", overlap_buf_size);
#endif
				}
				ptr = hash_overlap_buf.ptr;
				uint16_t num_itr = 0;
				num_itr = overlap_buf_size / DES_BLOCK_SIZE;
				while (num_itr)	{
					CVMX_LOADUNA_INT64(r0, ptr, 0);

					CVMX_MT_3DES_DEC_CBC(cvmx_htobe64(r0));

					num_itr--;

					CVMX_MF_3DES_RESULT(r1);

#ifdef DEBUG
					cavium_dbgprint("r1 : %llx\n", r1);
#endif
					CVMX_STOREUNA_INT64(cvmx_htobe64(r1), ptr, 0);
					ptr += 8;

				}

				if (hash_overlap_buf.size) {
					if (output_overlap) {
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, iovec_stat[num_hash_iovec - 2].hash_rem_bytes);
						memcpy(outp_vec[outp_cnt - 1].ptr, hash_overlap_buf.ptr + iovec_stat[num_hash_iovec - 2].hash_rem_bytes, iovec_stat[num_hash_iovec - 1].hash_req_bytes);
					} else {
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, hash_overlap_buf.size);
					}
				}
			}

			memset(hash_rem_buf, 0, 128);
			memcpy(hash_rem_buf, hash_bal, 13);
			memcpy(hash_rem_buf + 13, hash_overlap_buf.ptr, hash_overlap_buf.size);
			hash_rem_buf[hash_overlap_buf.size + 13] = 0x80;
			uint64_t *ptr = (uint64_t *)hash_rem_buf;
#ifdef DEBUG
			cavium_dbgprint("hash rem buf\n");
			_hexdump(hash_rem_buf, 128);
#endif
			if((hash_overlap_buf.size + 13) >= 56)
			{
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
				CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

				ptr += 8;
			}
			inner_hash_total_bytes = SHA_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
			inner_hash_total_bits = inner_hash_total_bytes * 8;
			CVMX_MT_HSH_DAT(cvmx_htobe64(cvmx_htobe64(ptr[0])), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(cvmx_htobe64(ptr[1])), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(cvmx_htobe64(ptr[2])), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(cvmx_htobe64(ptr[3])), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(cvmx_htobe64(ptr[4])), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(cvmx_htobe64(ptr[5])), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(cvmx_htobe64(ptr[6])), 6);
			CVMX_MT_HSH_STARTSHA(inner_hash_total_bits);

			inner_hash_done = 1;
		}
		free(iovec_stat);
	}

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

#ifdef DEBUG
	cavium_dbgprint("Inner Hash \n");
	cavium_dbgprint("r0 : %llx\n", r0);
	cavium_dbgprint("r1 : %llx\n", r1);
	cavium_dbgprint("r2 : %llx\n", r2);
#endif

    if(key->ex_mode == EX_MODE_ENCRYPT)
        return 1;
    else if (key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(r0), outp_vec[0].ptr, 0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(r1), outp_vec[0].ptr, 8);
        CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(r2)), outp_vec[0].ptr, 16);
        return 1;
    }

	//Calculate Outer Hash
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

	outer_hash_total_bytes = SHA_CBLOCK + SHA_DIGEST_LENGTH;
	outer_hash_total_bits = outer_hash_total_bytes * 8;

	CVMX_LOADUNA_INT64(r3, mac_secret, 0);
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 0);
	CVMX_LOADUNA_INT64(r3, mac_secret, 8);
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 1);

	r3 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	r3 = r3 << 32;
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(r3, 2);

	CVMX_LOADUNA_INT64(r3, CRY_HMAC_PAD2_64, 0);
	CVMX_MT_HSH_DAT(r3, 3);
	CVMX_MT_HSH_DAT(r3, 4);
	CVMX_MT_HSH_DAT(r3, 5);
	CVMX_MT_HSH_DAT(r3, 6);
	CVMX_MT_HSH_STARTSHA(r3);

	inp_vec[0].ptr -= ( DES_BLOCK_SIZE) ;
	inp_vec[0].size += ( DES_BLOCK_SIZE);

	CVMX_MT_HSH_DAT(r0, 0);
	CVMX_MT_HSH_DAT(r1, 1);
	r2 = r2 & 0xFFFFFFFF00000000ULL;
	r2 = r2 | 0x0000000080000000ULL;

	CVMX_MT_HSH_DAT(r2, 2);
	CVMX_MT_HSH_DATZ(3);
	CVMX_MT_HSH_DATZ(4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	CVMX_MT_HSH_STARTSHA(outer_hash_total_bits);

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

	CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 8);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 16);

    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr, hash_buf, SHA_DIGEST_LENGTH);
        return 1;
    }
#ifdef DEBUG
	cavium_dbgprint("Outer Hash Calculated : %s\n", __func__);
	_hexdump(hash_buf, SHA_DIGEST_LENGTH);
#endif
#ifdef DEBUG

	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
	for (itr = 0; itr < inp_cnt; itr++)
	{
		cavium_dbgprint("inp_vec[%d].ptr: %p, inp_vec[%d].size: %d\n", itr, inp_vec[itr].ptr, itr, inp_vec[itr].size);
		_hexdump(inp_vec[itr].ptr, inp_vec[itr].size);
	}

#endif

	//Compare calculated HMAC with received HMAC
	if (ret_val == 1) {
		if (inp_cnt == 1) {
            hash_buf_ptr_64 = (uint64_t *)hash_buf;
            hash_buf_ptr_32 = (uint32_t *)(hash_buf+16);
			if ((hash_buf_ptr_64[0] != (((uint64_t *)(out_ptr + hash_input_rem_len))[0])) || 
					(hash_buf_ptr_64[1] != (((uint64_t *)(out_ptr + hash_input_rem_len))[1])) || 
					(hash_buf_ptr_32[0] != (((uint32_t *)(out_ptr + hash_input_rem_len + 16))[0]))) {
				ret_val = 0;
				cavium_dbgprint(
						"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
				cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
				cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(out_ptr + hash_input_rem_len))[0]));
				cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(out_ptr + hash_input_rem_len))[1]));
#endif
			}
		} else {
			if (memcmp(hash_buf, hash_overlap_buf.ptr + hash_overlap_buf.size, SHA_DIGEST_LENGTH)) {
				ret_val = 0;
				cavium_dbgprint(
						"%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
				cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
				cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[0]));
				cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[1]));
#endif
			}
		}
	}



// May be need to update outp_vec[*].size 
	*len = raw_data_len;
	return ret_val;
}


/*******************************************************************************
 *                                                                             *
 *    Decrypts input record and generates TLS-1.1 / TLS - 1.2 SHA1 HMAC;       *
 *  authenticates the record and stores result in output iovec buffer.         *
 *  Initialisation vector to be used for decrytion is extracted from the       *
 *  intial bytes of input buffer passed to the function. Key and MAC secret to *
 *  be for cipher decryption and HMAC generation respectively are taken from   *
 *  the key structure passed. Sequence number and record header used in        *
 *  generating TLS-1.1 / TLS-1.2 HMAC are extracted from AAD stored in KEY     *
 *  structure. First 8bytes of  AAD holds sequence number and next 5 bytes     * 
 *  holds the TLS-1.1 / TLS-1.2 record header.                                 *
 *  Protocol : TLS-1.1  / TLS-1.2                                              *
 *  Cipher   : AES128-CBC / AES256-CBC                                         *
 *  HMAC     : SHA1                                                            *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing decryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to decrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store decryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_aes_sha1_dec_tls11(EVP_AESOCT_HMAC_SHA1 *key,
		cav_iovec_t *inp_vec, uint16_t inp_cnt,
		cav_iovec_t *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t r0, r1, r2, r3;

	uint8_t padarray[32] = {0}, hash_buf[128] = {0}, hash_rem_buf[128] = {0}, buf[128] = {0};
	uint8_t keylength = 1, inner_hash_done = 0, output_overlap = 0, cipher_type = 0;
	int32_t enc_data_len = 0, raw_data_len = 0, pad_data_len = 0, hash_input_rem_len = 0;
	int32_t overlap_buf_size = 0, enc_rem_cur_chunk = 0, inner_hash_total_bytes = 0, outer_hash_total_bytes = 0, prev_iovec_rem_bytes = 0, ret_val = 1;
	int32_t i = 0;
	uint64_t seq_num = 0, inner_hash_total_bits = 0, outer_hash_total_bits = 0;
	uint8_t *hash_bal = NULL, *out_ptr = NULL, *mac_secret = NULL, *key_data = NULL ;
	uint8_t *in_chunk_cur = NULL, *out_chunk_cur = NULL, *enc_rem_ptr = NULL;
    uint8_t *ext_iv_ptr = NULL;
	struct tls_rec_hdr *in_rec_hdr = NULL;
	uint64_t *enc_key = NULL;
    uint16_t *hash_buf_ptr_16 = NULL;
    uint32_t *hash_buf_ptr_32 = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct iovec_cav hash_overlap_buf;
	hash_overlap_buf.size = 0;

#ifdef DEBUG
	uint32_t itr = 0;

	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
	for (itr = 0; itr < inp_cnt; itr++) {
		cavium_dbgprint("inp_vec[%d].ptr: %p, inp_vec[%d].size: %d\n", itr, inp_vec[itr].ptr, itr, inp_vec[itr].size);
	}

	for (itr = 0; itr < outp_cnt; itr++) {
		cavium_dbgprint("outp_vec[%d].ptr: %p, outp_vec[%d].size: %d\n", itr, outp_vec[itr].ptr, itr, outp_vec[itr].size);
	}
#endif

	in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];
	enc_data_len = cvmx_htobe16(in_rec_hdr->length);

#ifdef DEBUG
	cavium_dbgprint("enc_data_len: %d (%d)\n", enc_data_len, __LINE__);
#endif

	cipher_type = key->cipher_type;

	if ((enc_data_len % AES_BLOCK_SIZE) && (key->ex_mode == EX_MODE_DEFAULT || key->ex_mode == EX_MODE_ENCRYPT)){
		cavium_dbgprint(
			"%s:%d Not properly padded data is received for cipher %s\n",
			__FUNCTION__,__LINE__, cipher_strings[cipher_type]);
		ret_val = 0;
		return ret_val;
	}

	seq_num = cvmx_htobe64(key->aux.tls_aad_64[0]);
	enc_key = (uint64_t *)key->cipher_key;
	mac_secret = key->mac_secret;
	key_data = key->cipher_iv;

	/*Load AES key*/
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[1]), 1);
	if (cipher_type == OCT_AES_256) {
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[2]), 2);
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[3]), 3);
		keylength = 3;
	} else {
		CVMX_MT_AES_KEY((uint64_t)0, 2);
		CVMX_MT_AES_KEY((uint64_t)0, 3);
	}
	CVMX_MT_AES_KEYLENGTH(keylength);


	/* Pad data can not be more than AES_BLOCK_SIZE
	 * As per AES CBC implementation, previous aes block result is fed as IV for current aes block
	 * So decrypt last 32 bytes to know PAD length*/


	// calculate pad length by decrypting last 32 bytes
	// if the last iovec buffer contains data >= 32
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    if (inp_vec[inp_cnt - 1].size >= 2*AES_BLOCK_SIZE) {
	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 32, 0);
	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 24, 0);

	    	CVMX_MT_AES_IV(cvmx_htobe64(r0), 0);
	    	CVMX_MT_AES_IV(cvmx_htobe64(r1), 1);

	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 16, 0);
	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 8, 0);

	    	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
	    	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));


	    	CVMX_MF_AES_RESULT(r0, 0);
	    	CVMX_MF_AES_RESULT(r1, 1);

	    	pad_data_len = (r1 & 0xff ) + 1;
	    } else {
	    	// X + Y = 32
	    	// X bytes from iovec[inp_cnt - 2]
	    	// Y bytes from iovec[inp_cnt - 1]
	    	memcpy(padarray, inp_vec[inp_cnt - 2].ptr + (inp_vec[inp_cnt - 2].size - (32 - inp_vec[inp_cnt - 1].size)), (32 - inp_vec[inp_cnt - 1].size));
	    	memcpy(padarray + (32 - inp_vec[inp_cnt - 1].size), inp_vec[inp_cnt - 1].ptr, inp_vec[inp_cnt - 1].size);
	    	CVMX_LOADUNA_INT64(r0, padarray, 0);
	    	CVMX_LOADUNA_INT64(r1, padarray, 8);

	    	CVMX_MT_AES_IV(cvmx_htobe64(r0), 0);
	    	CVMX_MT_AES_IV(cvmx_htobe64(r1), 1);

	    	CVMX_LOADUNA_INT64(r0, padarray, 16);
	    	CVMX_LOADUNA_INT64(r1, padarray, 24);

	    	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
	    	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

	    	CVMX_MF_AES_RESULT(r0, 0);
	    	CVMX_MF_AES_RESULT(r1, 1);

	    	pad_data_len = (r1 & 0xff ) + 1;
	    }
    }

#ifdef DEBUG
	cavium_dbgprint("pad_data_len: %d (%d)\n", pad_data_len, __LINE__);
#endif

	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)key_data), 0);
	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)(&key_data[8])), 1);

    if(key->ex_mode == EX_MODE_DEFAULT)
        raw_data_len = enc_data_len - SHA_DIGEST_LENGTH - pad_data_len - AES_BLOCK_SIZE;
    else
        raw_data_len = enc_data_len;

	/* If message length < 0, then we are sure we received error packet,
 	   to avoid timing attack decrypt packet and calculate HMAC and throw error at end */
	if (raw_data_len < 0)
	{
		raw_data_len = enc_data_len - SHA_DIGEST_LENGTH;
		pad_data_len = 0;
	}

#ifdef DEBUG
	cavium_dbgprint("raw_data_len: %d (%d)\n", raw_data_len, __LINE__);
#endif
	// first chunk - 16 + 48

	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_LOADUNA_INT64(r0, mac_secret, 0);
	    CVMX_LOADUNA_INT64(r1, mac_secret, 8);

	    r0 = r0 ^ (0x3636363636363636);
	    r1 = r1 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);

	    r0 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	    r0 = r0 << 32;
	    r0 = r0 ^ (0x3636363636363636);

	    CVMX_MT_HSH_DAT(r0, 2);

	    CVMX_LOADUNA_INT64(r1, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(r1, 3);
	    CVMX_MT_HSH_DAT(r1, 4);
	    CVMX_MT_HSH_DAT(r1, 5);
	    CVMX_MT_HSH_DAT(r1, 6);
	    CVMX_MT_HSH_STARTSHA(r1);
    }
	/*8 + 1 + 2 + 2 = 13 bytes are left over*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    CVMX_STOREUNA_INT64(cvmx_htobe64(seq_num), hash_buf, 0);
	    hash_buf[8] = in_rec_hdr->type;
        hash_buf_ptr_16 = (uint16_t *)&hash_buf[9];
	    hash_buf_ptr_16[0]  = in_rec_hdr->u.version;
	    hash_buf_ptr_16[1]  = cvmx_htobe16(raw_data_len);
	    // hash bal is 13 bytes
    } else if (key->ex_mode & 0x02) {
        if(raw_data_len > 16) {
            memcpy(hash_buf, inp_vec[0].ptr, 16);
            inp_vec[0].ptr += 16;
            inp_vec[0].size -= 16;
        } else {
            memcpy(hash_buf, inp_vec[0].ptr, raw_data_len);
            inp_vec[0].ptr += raw_data_len;
            inp_vec[0].size -= raw_data_len;
        }
    }

	hash_bal = hash_buf;
	hash_input_rem_len = raw_data_len;

/* Remove External IV*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
        ext_iv_ptr = inp_vec[0].ptr;
	    CVMX_LOADUNA_INT64(r0, ext_iv_ptr, 0);
	    CVMX_LOADUNA_INT64(r1, ext_iv_ptr, 8);
	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));
        inp_vec[0].ptr += AES_BLOCK_SIZE;
        inp_vec[0].size -= AES_BLOCK_SIZE;
        outp_vec[0].size -= AES_BLOCK_SIZE;
    }
	if (inp_cnt == 1) {
		in_chunk_cur = inp_vec[0].ptr;
        if(key->ex_mode & 0x02)
            out_chunk_cur = inp_vec[0].ptr;
        else
		    out_chunk_cur = outp_vec[0].ptr;

		while (hash_input_rem_len >= SHA_CBLOCK) {
			hash_bal = block_processing_aes_sha1_dec_tls10(in_chunk_cur, out_chunk_cur, hash_bal,key->ex_mode);

#ifdef DEBUG
			cavium_dbgprint("hash_bal\n");
			_hexdump(hash_bal, 11);
#endif
			in_chunk_cur += SHA_CBLOCK;
			out_chunk_cur += SHA_CBLOCK;

			hash_input_rem_len -= SHA_CBLOCK;
		}
		out_ptr = out_chunk_cur;
#ifdef DEBUG
		cavium_dbgprint("hash_bal\n");
		_hexdump(hash_bal, 13);
#endif
        if(key->ex_mode == EX_MODE_DEFAULT || 
                (key->ex_mode == EX_MODE_ENCRYPT && hash_input_rem_len)) {
		// remaining data is compulsory a multiple of 16
		    do {
		    	CVMX_LOADUNA_INT64(r0, in_chunk_cur, 0);
		    	CVMX_LOADUNA_INT64(r1, in_chunk_cur, 8);

		    	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
		    	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

		    	in_chunk_cur += AES_BLOCK_SIZE;

		    	CVMX_MF_AES_RESULT(r2, 0);
		    	CVMX_MF_AES_RESULT(r3, 1);

#ifdef DEBUG
			cavium_dbgprint("r2 : %llx  r3 : %llx\n", r2, r3);
#endif
		    	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), out_chunk_cur, 0);
		    	CVMX_STOREUNA_INT64(cvmx_htobe64(r3), out_chunk_cur, 8);

		    	out_chunk_cur += AES_BLOCK_SIZE;

		    } while (((inp_vec[0].ptr + inp_vec[0].size) - in_chunk_cur) > 0);
        }
		// Along with 13 bytes hash balance, hash_input_rem_len 
		// should be given to hash engine as input
        if((key->ex_mode & 0x02) && hash_input_rem_len)
		    memcpy(hash_rem_buf, hash_bal, (hash_input_rem_len > 16)?16:hash_input_rem_len);
        else if(key->ex_mode == EX_MODE_DEFAULT)
		    memcpy(hash_rem_buf, hash_bal, 13);

		if (hash_input_rem_len)	{
                if(key->ex_mode & 0x02 && hash_input_rem_len > 16)
			        memcpy(hash_rem_buf + 16, out_ptr, hash_input_rem_len-16);
                else if(key->ex_mode == EX_MODE_DEFAULT)
			        memcpy(hash_rem_buf + 13, out_ptr, hash_input_rem_len);
		}

        if(key->ex_mode & 0x02)
            hash_rem_buf[hash_input_rem_len] = 0x80;
        else
		    hash_rem_buf[13 + hash_input_rem_len] = 0x80;
#ifdef DEBUG
		cavium_dbgprint("hash_rem_buf\n");
		_hexdump(hash_rem_buf, 64);
#endif
		uint64_t * ptr = (uint64_t *)hash_rem_buf;
		if((((hash_input_rem_len + 13) >= 56) && (key->ex_mode == EX_MODE_DEFAULT))||
           ((key->ex_mode & 0x02) && (hash_input_rem_len >= 56))) {
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

			ptr += 8;
		}
        if(key->ex_mode == EX_MODE_DIGEST)
		    inner_hash_total_bytes = raw_data_len;
        if(key->ex_mode == EX_MODE_HMAC)
		    inner_hash_total_bytes = SHA_CBLOCK + raw_data_len;
        if(key->ex_mode == EX_MODE_DEFAULT)
		    inner_hash_total_bytes = SHA_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
		inner_hash_total_bits = inner_hash_total_bytes * 8;
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
		CVMX_MT_HSH_STARTSHA(inner_hash_total_bits);
	} else if(inp_cnt > 1) {
		hash_overlap_buf.ptr = buf;
		hash_overlap_buf.size = 0;
		iovec_record_process_t *iovec_stat = NULL;
		iovec_stat = malloc(inp_cnt * sizeof(iovec_record_process_t));
		memset(iovec_stat, 0, inp_cnt * sizeof(iovec_record_process_t));

		prev_iovec_rem_bytes = 0;
		// populate loop counters
		while (hash_input_rem_len && (i < inp_cnt)) {
			if (hash_input_rem_len >= inp_vec[i].size) {
				iovec_stat[i].hash_size = inp_vec[i].size;
				hash_input_rem_len -= inp_vec[i].size;
			} else {
				iovec_stat[i].hash_size = hash_input_rem_len;
				hash_input_rem_len = 0;
			}

			if (prev_iovec_rem_bytes) {
				iovec_stat[i].hash_req_bytes = SHA_CBLOCK - prev_iovec_rem_bytes;
				// in case of last chunk not holding enough bytes to fill 64 byte block
				if (iovec_stat[i].hash_size < iovec_stat[i].hash_req_bytes) {
					iovec_stat[i].hash_req_bytes = iovec_stat[i].hash_size;
				}
			} else {
				iovec_stat[i].hash_req_bytes = 0;
			}
			iovec_stat[i].hash_itr = (iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) / SHA_CBLOCK;
			iovec_stat[i].crypto_itr = iovec_stat[i].hash_itr / AES_BLOCK_SIZE;
			//iovec_stat[i].crypto_itr = (iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) / AES_BLOCK_SIZE;
			//iovec_stat[i].hash_itr = iovec_stat[i].crypto_itr / 4;

			iovec_stat[i].hash_rem_bytes = ((iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) - (iovec_stat[i].hash_itr * SHA_CBLOCK));
			prev_iovec_rem_bytes = iovec_stat[i].hash_rem_bytes;

			i++;
		}
		uint16_t num_hash_iovec = i;
		// need to update enc_digest_ptr
		if (iovec_stat[i-1].hash_size == inp_vec[i-1].size) {
			enc_rem_ptr = inp_vec[i].ptr;
		} else {
			enc_rem_ptr = inp_vec[i-1].ptr + iovec_stat[i-1].hash_size;
			enc_rem_cur_chunk = inp_vec[i-1].size - iovec_stat[i-1].hash_size;
		}

#ifdef DEBUG
		for (i = 0; i < inp_cnt; i++) {
			cavium_dbgprint("Chunk : %d\n", i+1);
			cavium_dbgprint("hash_req_bytes : %d\n",iovec_stat[i].hash_req_bytes);
			cavium_dbgprint("crypto_itr : %d\n", iovec_stat[i].crypto_itr);
			cavium_dbgprint("hash_itr : %d\n", iovec_stat[i].hash_itr);
			cavium_dbgprint("hash_rem_bytes : %d\n",iovec_stat[i].hash_rem_bytes);
		}
#endif

		for (i = 0; i < num_hash_iovec; i++) {
			in_chunk_cur = inp_vec[i].ptr;
			out_chunk_cur = outp_vec[i].ptr;

			if (iovec_stat[i].hash_req_bytes) {	
				memcpy(hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, in_chunk_cur, iovec_stat[i].hash_req_bytes);
				in_chunk_cur += iovec_stat[i].hash_req_bytes;

				if ((iovec_stat[i-1].hash_rem_bytes + iovec_stat[i].hash_req_bytes) < SHA_CBLOCK) {
					hash_overlap_buf.size += iovec_stat[i].hash_req_bytes;
					out_chunk_cur = outp_vec[i].ptr + iovec_stat[i-1].hash_req_bytes + (SHA_CBLOCK * iovec_stat[i-1].hash_itr);
					output_overlap = 1;
					continue;
				} else {
					hash_bal = block_processing_aes_sha1_dec_tls10(hash_overlap_buf.ptr, hash_overlap_buf.ptr, hash_bal, key->ex_mode);
#ifdef DEBUG
					cavium_dbgprint("hash_bal : %d\n", __LINE__);
					_hexdump(hash_bal, 13);
#endif

					memcpy(outp_vec[i].ptr + (outp_vec[i].size - iovec_stat[i-1].hash_rem_bytes), hash_overlap_buf.ptr, iovec_stat[i-1].hash_rem_bytes);
					memcpy(out_chunk_cur, hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, iovec_stat[i].hash_req_bytes);

					out_chunk_cur += iovec_stat[i].hash_req_bytes;
					hash_overlap_buf.size = 0;
				}
			}

			register uint8_t hash_itr = iovec_stat[i].hash_itr;
			while (hash_itr) {
				hash_bal = block_processing_aes_sha1_dec_tls10(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);
#ifdef DEBUG
				cavium_dbgprint("hash_bal : %d\n", __LINE__);
				_hexdump(hash_bal, 13);
#endif

				in_chunk_cur += SHA_CBLOCK;
				out_chunk_cur += SHA_CBLOCK;

				hash_itr--;
			}

			if (iovec_stat[i].hash_rem_bytes) {
				//copy rem bytes to hash overlap buf
				memcpy(hash_overlap_buf.ptr, in_chunk_cur, iovec_stat[i].hash_rem_bytes);
				in_chunk_cur += iovec_stat[i].hash_rem_bytes;
				hash_overlap_buf.size = iovec_stat[i].hash_rem_bytes;
			}
		}

#ifdef DEBUG
		cavium_dbgprint("hash_overlap_buf.size : %d \n", hash_overlap_buf.size);
#endif
		//obviously this case gets hit all the time
		if (inner_hash_done == 0) {
			//if(hash_overlap_buf.size)
			{
				uint8_t *ptr = NULL;
				overlap_buf_size = hash_overlap_buf.size;
				memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size, enc_rem_ptr, enc_rem_cur_chunk);
				overlap_buf_size += enc_rem_cur_chunk;
#ifdef DEBUG
				cavium_dbgprint("enc_rem_cur_chunk : %d\n", enc_rem_cur_chunk);
#endif
				if (enc_rem_cur_chunk < (SHA_DIGEST_LENGTH + pad_data_len)) {
					memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size + enc_rem_cur_chunk, inp_vec[inp_cnt-1].ptr, inp_vec[inp_cnt-1].size);

					overlap_buf_size += inp_vec[inp_cnt-1].size;
#ifdef DEBUG
					cavium_dbgprint("overlap_buf_size : %d\n", overlap_buf_size);
#endif
				}
				ptr = hash_overlap_buf.ptr;
				uint16_t num_itr = 0;
				num_itr = overlap_buf_size / AES_BLOCK_SIZE;
				while (num_itr) {
					CVMX_LOADUNA_INT64(r0, ptr, 0);
					CVMX_LOADUNA_INT64(r1, ptr, 8);

					CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
					CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

					CVMX_MF_AES_RESULT(r2, 0);
					CVMX_MF_AES_RESULT(r3, 1);

#ifdef DEBUG
					cavium_dbgprint("r2 : %llx  r3 : %llx\n", r2, r3);
#endif
					CVMX_STOREUNA_INT64(cvmx_htobe64(r2), ptr, 0);
					CVMX_STOREUNA_INT64(cvmx_htobe64(r3), ptr, 8);

					ptr += AES_BLOCK_SIZE;
					num_itr--;
				}

				if (hash_overlap_buf.size) {
					if (output_overlap) {
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, iovec_stat[num_hash_iovec - 2].hash_rem_bytes);
						memcpy(outp_vec[outp_cnt -1].ptr, hash_overlap_buf.ptr + iovec_stat[num_hash_iovec - 2].hash_rem_bytes, iovec_stat[num_hash_iovec - 1].hash_req_bytes);
					} else {
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, hash_overlap_buf.size);
					}
				}
			}

			memset(hash_rem_buf, 0, 128);
			memcpy(hash_rem_buf, hash_bal, 13);
			memcpy(hash_rem_buf + 13, hash_overlap_buf.ptr, hash_overlap_buf.size);
			hash_rem_buf[hash_overlap_buf.size + 13] = 0x80;
			uint64_t *ptr = (uint64_t *)hash_rem_buf;
#ifdef DEBUG
			cavium_dbgprint("hash rem buf\n");
			_hexdump(hash_rem_buf, 128);
#endif
			if ((hash_overlap_buf.size + 13) >= 56) {
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
				CVMX_MT_HSH_STARTSHA(cvmx_htobe64(ptr[7]));

				ptr += 8;
			}
			inner_hash_total_bytes = SHA_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
			inner_hash_total_bits = inner_hash_total_bytes * 8;
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA(inner_hash_total_bits);

			inner_hash_done = 1;
		}
		free(iovec_stat);
	}

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

#ifdef DEBUG
	cavium_dbgprint("Inner Hash\n");
	cavium_dbgprint("r0 : %llx\n", r0);
	cavium_dbgprint("r1 : %llx\n", r1);
	cavium_dbgprint("r2 : %llx\n", r2);
#endif

    if(key->ex_mode == EX_MODE_ENCRYPT)
        return 1;
    else if (key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(r0), outp_vec[0].ptr, 0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(r1), outp_vec[0].ptr, 8);
        CVMX_STOREUNA_INT32(cvmx_htobe32(upper_32_bits(r2)), outp_vec[0].ptr, 16);
        return 1;
    }
	//calculate outer hash
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA1_INIT2, 2);

	outer_hash_total_bytes = SHA_CBLOCK + SHA_DIGEST_LENGTH;
	outer_hash_total_bits = outer_hash_total_bytes * 8;

	CVMX_LOADUNA_INT64(r3, mac_secret, 0);
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 0);
	CVMX_LOADUNA_INT64(r3, mac_secret, 8);
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 1);

	r3 = cvmx_htobe32(*(uint32_t *)(mac_secret + 16));
	r3 = r3 << 32;
	r3 = r3 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(r3, 2);

	CVMX_LOADUNA_INT64(r3, CRY_HMAC_PAD2_64, 0);
	CVMX_MT_HSH_DAT(r3, 3);
	CVMX_MT_HSH_DAT(r3, 4);
	CVMX_MT_HSH_DAT(r3, 5);
	CVMX_MT_HSH_DAT(r3, 6);
	CVMX_MT_HSH_STARTSHA(r3);


	CVMX_MT_HSH_DAT(r0, 0);
	CVMX_MT_HSH_DAT(r1, 1);
	r2 = r2 & 0xFFFFFFFF00000000ULL;
	r2 = r2 | 0x0000000080000000ULL;

	CVMX_MT_HSH_DAT(r2, 2);
	CVMX_MT_HSH_DATZ(3);
	CVMX_MT_HSH_DATZ(4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	CVMX_MT_HSH_STARTSHA(outer_hash_total_bits);

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);

	CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 8);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 16);

#ifdef DEBUG
	cavium_dbgprint("Outer Hash Calculated : %s\n", __func__);
	_hexdump(hash_buf, SHA_DIGEST_LENGTH);

#endif
    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr, hash_buf, SHA_DIGEST_LENGTH);
        return 1;
    }
	//Compare Calculated HMAC and Received HMAC
	if (ret_val == 1) {
		if (inp_cnt == 1) {
            hash_buf_ptr_64 = (uint64_t *)hash_buf;
            hash_buf_ptr_32 = (uint32_t *)&hash_buf[16];
			if ((hash_buf_ptr_64[0] != (((uint64_t *)(out_ptr + hash_input_rem_len))[0])) ||
					(hash_buf_ptr_64[1] != (((uint64_t *)(out_ptr + hash_input_rem_len))[1])) ||
					(hash_buf_ptr_32[0] != (((uint32_t *)(out_ptr + hash_input_rem_len + 16))[0]))) {
				ret_val = 0;
				cavium_dbgprint( "%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
				cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
				cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(out_ptr + hash_input_rem_len))[0]));
				cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(out_ptr + hash_input_rem_len))[1]));
				cavium_dbgprint("3. :%x  : %x\n", ((uint32_t *)(hash_buf + 16))[0], (((uint32_t *)(out_ptr + hash_input_rem_len + 16))[0]));
#endif
			}
		} else {
			if (memcmp(hash_buf, hash_overlap_buf.ptr + hash_overlap_buf.size, SHA_DIGEST_LENGTH)) {
				ret_val = 1;
				cavium_dbgprint( "%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
				cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
				cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[0]));
				cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[1]));
				cavium_dbgprint("3. :%x  : %x\n", ((uint32_t *)(hash_buf + 16))[0], (((uint32_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size + 16))[0]));
#endif
			}
		}
	}





// May be need to update outp_vec[*].size 
	return raw_data_len;
}

/*******************************************************************************
 *                                                                             *
 *    Decrypts input record and generates TLS - 1.2 SHA256 HMAC; authenticates *
 *  the record and stores result in output iovec buffer. Initialisation vector *
 *  to be used for decrytion is extracted from the intial bytes of input       *
 *  buffer passed to the function. Key and MAC secret to be for cipher         *
 *  decryption and HMAC generation respectively are taken from the key         *
 *  structure passed. Sequence number and record header used in generating     *
 *  TLS-1.2 HMAC are extracted from AAD stored in KEY structure. First 8bytes  *
 *  of  AAD holds sequence number and next 5 bytes holds the TLS-1.2 record    *
 *  header.                                                                    *
 *  Protocol : TLS-1.2                                                         *
 *  Cipher   : AES128-CBC / AES256-CBC                                         *
 *  HMAC     : SHA256                                                          *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing decryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to decrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store decryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/

int32_t oct_aes_sha256_dec_tls12(EVP_AESOCT_HMAC_SHA1 *key,
		cav_iovec_t *inp_vec, uint16_t inp_cnt,
		cav_iovec_t *outp_vec, uint16_t outp_cnt,
		uint16_t *len)
{
	register uint64_t r0, r1, r2, r3, r4;

	uint8_t padarray[32] = {0}, hash_buf[128] = {0}, hash_rem_buf[128] = {0}, buf[128] = {0};
	uint8_t keylength = 1, inner_hash_done = 0, output_overlap = 0, cipher_type = 0;
	int32_t enc_data_len = 0, raw_data_len = 0, pad_data_len = 0, hash_input_rem_len = 0;
	int32_t overlap_buf_size = 0, enc_rem_cur_chunk = 0, inner_hash_total_bytes = 0, outer_hash_total_bytes = 0, prev_iovec_rem_bytes = 0, ret_val = 1;
	int32_t i = 0;
	uint64_t seq_num = 0, inner_hash_total_bits = 0, outer_hash_total_bits = 0;
	uint8_t *hash_bal = NULL, *out_ptr = NULL, *mac_secret = NULL, *key_data = NULL;
	uint8_t *in_chunk_cur = NULL, *out_chunk_cur = NULL, *enc_rem_ptr = NULL;
    uint8_t *ext_iv_ptr = NULL; 
	struct tls_rec_hdr *in_rec_hdr = NULL;
	uint64_t *enc_key = NULL;
    uint16_t *hash_buf_ptr_16 = NULL;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct iovec_cav hash_overlap_buf;
	hash_overlap_buf.size = 0;

#ifdef DEBUG
	uint32_t itr = 0;

	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
	for (itr = 0; itr < inp_cnt; itr++) {
		cavium_dbgprint("inp_vec[%d].ptr: %p, inp_vec[%d].size: %d\n", itr, inp_vec[itr].ptr, itr, inp_vec[itr].size);
	}

	for (itr = 0; itr < outp_cnt; itr++) {
		cavium_dbgprint("outp_vec[%d].ptr: %p, outp_vec[%d].size: %d\n", itr, outp_vec[itr].ptr, itr, outp_vec[itr].size);
	}
#endif

	in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];
	enc_data_len = cvmx_htobe16(in_rec_hdr->length);

#ifdef DEBUG
	cavium_dbgprint("enc_data_len: %d (%d)\n", enc_data_len, __LINE__);
#endif

	cipher_type = key->cipher_type;

	if ((enc_data_len % AES_BLOCK_SIZE) && (key->ex_mode == EX_MODE_DEFAULT || key->ex_mode == EX_MODE_ENCRYPT)){
		cavium_dbgprint(
			"%s:%d Not properly padded data is received for cipher %s\n",
			__FUNCTION__,__LINE__, cipher_strings[cipher_type]);
		ret_val = 1;
		return ret_val;
	}

	seq_num = cvmx_htobe64(key->aux.tls_aad_64[0]);
	enc_key = (uint64_t *)key->cipher_key;
	mac_secret = key->mac_secret;
	key_data = key->cipher_iv;

	/*Load AES key*/
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[1]), 1);
	if (cipher_type == OCT_AES_256) {
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[2]), 2);
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[3]), 3);
		keylength = 3;
	} else {
		CVMX_MT_AES_KEY((uint64_t)0, 2);
		CVMX_MT_AES_KEY((uint64_t)0, 3);
	}
	CVMX_MT_AES_KEYLENGTH(keylength);


	/* Pad data can not be more than AES_BLOCK_SIZE
	 * As per AES CBC implementation, previous aes block result is fed as IV for current aes block
	 * So decrypt last 32 bytes to know PAD length*/


	// calculate pad length by decrypting last 32 bytes
	// if the last iovec buffer contains data >= 32
	if(key->ex_mode == EX_MODE_DEFAULT) {
        if (inp_vec[inp_cnt - 1].size >= 2*AES_BLOCK_SIZE) {
	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 32, 0);
	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 24, 0);

	    	CVMX_MT_AES_IV(cvmx_htobe64(r0), 0);
	    	CVMX_MT_AES_IV(cvmx_htobe64(r1), 1);

	    	CVMX_LOADUNA_INT64(r0, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 16, 0);
	    	CVMX_LOADUNA_INT64(r1, inp_vec[inp_cnt - 1].ptr + inp_vec[inp_cnt - 1].size - 8, 0);

	    	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
	    	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));


	    	CVMX_MF_AES_RESULT(r0, 0);
	    	CVMX_MF_AES_RESULT(r1, 1);

	    	pad_data_len = (r1 & 0xff ) + 1;
	    } else {
	    	// X + Y = 32
	    	// X bytes from iovec[inp_cnt - 2]
	    	// Y bytes from iovec[inp_cnt - 1]
	    	memcpy(padarray, inp_vec[inp_cnt - 2].ptr + (inp_vec[inp_cnt - 2].size - (32 - inp_vec[inp_cnt - 1].size)), (32 - inp_vec[inp_cnt - 1].size));
	    	memcpy(padarray + (32 - inp_vec[inp_cnt - 1].size), inp_vec[inp_cnt - 1].ptr, inp_vec[inp_cnt - 1].size);
	    	CVMX_LOADUNA_INT64(r0, padarray, 0);
	    	CVMX_LOADUNA_INT64(r1, padarray, 8);

	    	CVMX_MT_AES_IV(cvmx_htobe64(r0), 0);
	    	CVMX_MT_AES_IV(cvmx_htobe64(r1), 1);

	    	CVMX_LOADUNA_INT64(r0, padarray, 16);
	    	CVMX_LOADUNA_INT64(r1, padarray, 24);

	    	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
	    	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

	    	CVMX_MF_AES_RESULT(r0, 0);
	    	CVMX_MF_AES_RESULT(r1, 1);

	    	pad_data_len = (r1 & 0xff ) + 1;
	    }
    }

#ifdef DEBUG
	cavium_dbgprint("pad_data_len: %d (%d)\n", pad_data_len, __LINE__);
#endif

	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)key_data), 0);
	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)(&key_data[8])), 1);

    if(key->ex_mode == EX_MODE_DEFAULT)
    	raw_data_len = enc_data_len - SHA256_DIGEST_LENGTH - pad_data_len - AES_BLOCK_SIZE;
    else 
        raw_data_len = enc_data_len;

	/* If message length < 0, then we are sure we received error packet,
 	   to avoid timing attack decrypt packet and calculate HMAC and throw error at end */
	if (raw_data_len < 0)
	{
		raw_data_len = enc_data_len - SHA256_DIGEST_LENGTH;
		pad_data_len = 0;
	}

#ifdef DEBUG
	cavium_dbgprint("raw_data_len: %d (%d)\n", raw_data_len, __LINE__);
#endif

	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT2, 2);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT3, 3);

	CVMX_LOADUNA_INT64(r0, mac_secret, 0);
	CVMX_LOADUNA_INT64(r1, mac_secret, 8);
	CVMX_LOADUNA_INT64(r2, mac_secret, 16);
	CVMX_LOADUNA_INT64(r3, mac_secret, 24);

	r0 = r0 ^ (0x3636363636363636);
	r1 = r1 ^ (0x3636363636363636);
	r2 = r2 ^ (0x3636363636363636);
	r3 = r3 ^ (0x3636363636363636);

    if(key->ex_mode != EX_MODE_DIGEST) {
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r0), 0);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r2), 2);
	    CVMX_MT_HSH_DAT(cvmx_htobe64(r3), 3);

	    CVMX_LOADUNA_INT64(r0, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(r0, 4);
	    CVMX_MT_HSH_DAT(r0, 5);
	    CVMX_MT_HSH_DAT(r0, 6);
	    CVMX_MT_HSH_STARTSHA256(r0);
    }

    if(key->ex_mode == EX_MODE_DEFAULT) {
	    /*8 + 1 + 2 + 2 = 13 bytes are left over*/
	    CVMX_STOREUNA_INT64(cvmx_htobe64(seq_num), hash_buf, 0);
	    hash_buf[8] = in_rec_hdr->type;
        hash_buf_ptr_16 = (uint16_t *)&hash_buf[9];
	    hash_buf_ptr_16[0]  = in_rec_hdr->u.version;
	    hash_buf_ptr_16[1]  = cvmx_htobe16(raw_data_len);
	    hash_bal = hash_buf;
    } else if (key->ex_mode & 0x02) {
        if(raw_data_len > 16) {
            memcpy(hash_buf, inp_vec[0].ptr, 16);
            inp_vec[0].ptr += 16;
            inp_vec[0].size -=16;
            hash_bal = hash_buf;
        } else {
            memcpy(hash_buf, inp_vec[0].ptr, raw_data_len);
            inp_vec[0].ptr +=  raw_data_len;
            inp_vec[0].size -= raw_data_len;
            hash_bal = hash_buf;
        }
    } else
        hash_bal = padarray; /* We don't care, what hash_bal holds as long as it holds something */

	hash_input_rem_len = raw_data_len;

/* Remove External IV*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
        ext_iv_ptr = inp_vec[0].ptr;
	    CVMX_LOADUNA_INT64(r0, ext_iv_ptr, 0);
	    CVMX_LOADUNA_INT64(r1, ext_iv_ptr, 8);
	    CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
	    CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));
	    outp_vec[0].size -= AES_BLOCK_SIZE;
        inp_vec[0].ptr += AES_BLOCK_SIZE;
        inp_vec[0].size -= AES_BLOCK_SIZE;
    }
	if (inp_cnt == 1) {
		in_chunk_cur = inp_vec[0].ptr;
		out_chunk_cur = outp_vec[0].ptr;

		while (hash_input_rem_len >= SHA256_CBLOCK) {
			hash_bal = block_processing_aes_sha256_dec_tls12(in_chunk_cur, out_chunk_cur, hash_bal,key->ex_mode);

#ifdef DEBUG
			cavium_dbgprint("hash_bal\n");
			_hexdump(hash_bal, 11);
#endif
			in_chunk_cur += SHA256_CBLOCK;
			out_chunk_cur += SHA256_CBLOCK;

			hash_input_rem_len -= SHA256_CBLOCK;
		}
        if(key->ex_mode & 0x02)
            out_ptr = in_chunk_cur;
        else
		    out_ptr = out_chunk_cur;
#ifdef DEBUG
		cavium_dbgprint("hash_bal\n");
		_hexdump(hash_bal, 13);
#endif
		// remaining data is compulsory a multiple of 16
        if((key->ex_mode == EX_MODE_ENCRYPT && hash_input_rem_len) || 
                key->ex_mode == EX_MODE_DEFAULT) {
		    do {
		    	CVMX_LOADUNA_INT64(r0, in_chunk_cur, 0);
		    	CVMX_LOADUNA_INT64(r1, in_chunk_cur, 8);

		    	CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
		    	CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

		    	in_chunk_cur += AES_BLOCK_SIZE;

		    	CVMX_MF_AES_RESULT(r2, 0);
		    	CVMX_MF_AES_RESULT(r3, 1);

#ifdef DEBUG
			cavium_dbgprint("r2 : %llx  r3 : %llx\n", r2, r3);
#endif
	    		CVMX_STOREUNA_INT64(cvmx_htobe64(r2), out_chunk_cur, 0);
	    		CVMX_STOREUNA_INT64(cvmx_htobe64(r3), out_chunk_cur, 8);

	    		out_chunk_cur += AES_BLOCK_SIZE;

	    	} while (((inp_vec[0].ptr + inp_vec[0].size) - in_chunk_cur) > 0);
        }
		// Along with 13 bytes hash balance, hash_input_rem_len 
		// should be given to hash engine as input
        if(key->ex_mode == EX_MODE_DEFAULT) {
    		memcpy(hash_rem_buf, hash_bal, 13);
		    if (hash_input_rem_len)
    			memcpy(hash_rem_buf + 13, out_ptr, hash_input_rem_len);
		    hash_rem_buf[13 + hash_input_rem_len] = 0x80;
        } else if(key->ex_mode & 0x02){
            memcpy(hash_rem_buf, hash_bal, (hash_input_rem_len>16)?16:hash_input_rem_len);
		    if (hash_input_rem_len > 16) 
    			memcpy(hash_rem_buf + 16, out_ptr, hash_input_rem_len-16);
		    hash_rem_buf[hash_input_rem_len] = 0x80;
        }


#ifdef DEBUG
		cavium_dbgprint("hash_rem_buf , hash_input_rem_len : %d\n", hash_input_rem_len);
		_hexdump(hash_rem_buf, 64);
#endif
		uint64_t * ptr = (uint64_t *)hash_rem_buf;
		if (((key->ex_mode == EX_MODE_DEFAULT) && (hash_input_rem_len + 13) >= 56)||
             ((key->ex_mode & 0x02) && (hash_input_rem_len >= 56))) {
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA256(cvmx_htobe64(ptr[7]));

			ptr += 8;
		}
        if(key->ex_mode == EX_MODE_DEFAULT)
    		inner_hash_total_bytes = SHA256_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
        else if (key->ex_mode == EX_MODE_HMAC)
    		inner_hash_total_bytes = SHA256_CBLOCK + raw_data_len;
        else if (key->ex_mode == EX_MODE_DIGEST)
            inner_hash_total_bytes = raw_data_len;
		inner_hash_total_bits = inner_hash_total_bytes * 8;
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
		CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
		CVMX_MT_HSH_STARTSHA256(inner_hash_total_bits);
	} else if(inp_cnt > 1) {
		hash_overlap_buf.ptr = buf;
		hash_overlap_buf.size = 0;
		iovec_record_process_t *iovec_stat = NULL;
		iovec_stat = malloc(inp_cnt * sizeof(iovec_record_process_t));
		memset(iovec_stat, 0, inp_cnt * sizeof(iovec_record_process_t));

		prev_iovec_rem_bytes = 0;
		// populate loop counters
		while (hash_input_rem_len && (i < inp_cnt)) {
			if (hash_input_rem_len >= inp_vec[i].size) {
				iovec_stat[i].hash_size = inp_vec[i].size;
				hash_input_rem_len -= inp_vec[i].size;
			} else {
				iovec_stat[i].hash_size = hash_input_rem_len;
				hash_input_rem_len = 0;
			}

			if (prev_iovec_rem_bytes) {
				iovec_stat[i].hash_req_bytes = SHA256_CBLOCK - prev_iovec_rem_bytes;
				// in case of last chunk not holding enough bytes to fill 64 byte block
				if (iovec_stat[i].hash_size < iovec_stat[i].hash_req_bytes) {
					iovec_stat[i].hash_req_bytes = iovec_stat[i].hash_size;
				}
			} else {
				iovec_stat[i].hash_req_bytes = 0;
			}
			iovec_stat[i].hash_itr = (iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) / SHA256_CBLOCK;
			iovec_stat[i].crypto_itr = iovec_stat[i].hash_itr / AES_BLOCK_SIZE;
			//iovec_stat[i].crypto_itr = (iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) / AES_BLOCK_SIZE;
			//iovec_stat[i].hash_itr = iovec_stat[i].crypto_itr / 4;

			iovec_stat[i].hash_rem_bytes = ((iovec_stat[i].hash_size - iovec_stat[i].hash_req_bytes) - (iovec_stat[i].hash_itr * SHA256_CBLOCK));
			prev_iovec_rem_bytes = iovec_stat[i].hash_rem_bytes;

			i++;
		}
		uint16_t num_hash_iovec = i;
		// need to update enc_digest_ptr
		if (iovec_stat[i-1].hash_size == inp_vec[i-1].size) {
			enc_rem_ptr = inp_vec[i].ptr;
		} else {
			enc_rem_ptr = inp_vec[i-1].ptr + iovec_stat[i-1].hash_size;
			enc_rem_cur_chunk = inp_vec[i-1].size - iovec_stat[i-1].hash_size;
		}

#ifdef DEBUG
		for (i = 0; i < inp_cnt; i++) {
			cavium_dbgprint("Chunk : %d\n", i+1);
			cavium_dbgprint("hash_req_bytes : %d\n",iovec_stat[i].hash_req_bytes);
			cavium_dbgprint("crypto_itr : %d\n", iovec_stat[i].crypto_itr);
			cavium_dbgprint("hash_itr : %d\n", iovec_stat[i].hash_itr);
			cavium_dbgprint("hash_rem_bytes : %d\n",iovec_stat[i].hash_rem_bytes);
		}
#endif

		for (i = 0; i < num_hash_iovec; i++) {
			in_chunk_cur = inp_vec[i].ptr;
			out_chunk_cur = outp_vec[i].ptr;

			if (iovec_stat[i].hash_req_bytes) {	
				memcpy(hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, in_chunk_cur, iovec_stat[i].hash_req_bytes);
				in_chunk_cur += iovec_stat[i].hash_req_bytes;

				if ((iovec_stat[i-1].hash_rem_bytes + iovec_stat[i].hash_req_bytes) < SHA256_CBLOCK) {
					hash_overlap_buf.size += iovec_stat[i].hash_req_bytes;
					out_chunk_cur = outp_vec[i].ptr + iovec_stat[i-1].hash_req_bytes + (SHA256_CBLOCK * iovec_stat[i-1].hash_itr);
					output_overlap = 1;
					continue;
				} else {
					hash_bal = block_processing_aes_sha256_dec_tls12(hash_overlap_buf.ptr, hash_overlap_buf.ptr, hash_bal, key->ex_mode);
#ifdef DEBUG
					cavium_dbgprint("hash_bal : %d\n", __LINE__);
					_hexdump(hash_bal, 13);
#endif

					memcpy(outp_vec[i].ptr + (outp_vec[i].size - iovec_stat[i-1].hash_rem_bytes), hash_overlap_buf.ptr, iovec_stat[i-1].hash_rem_bytes);
					memcpy(out_chunk_cur, hash_overlap_buf.ptr + iovec_stat[i-1].hash_rem_bytes, iovec_stat[i].hash_req_bytes);

					out_chunk_cur += iovec_stat[i].hash_req_bytes;
					hash_overlap_buf.size = 0;
				}
			}

			register uint8_t hash_itr = iovec_stat[i].hash_itr;
			while (hash_itr) {
				hash_bal = block_processing_aes_sha256_dec_tls12(in_chunk_cur, out_chunk_cur, hash_bal, key->ex_mode);
#ifdef DEBUG
				cavium_dbgprint("hash_bal : %d\n", __LINE__);
				_hexdump(hash_bal, 13);
#endif

				in_chunk_cur += SHA256_CBLOCK;
				out_chunk_cur += SHA256_CBLOCK;

				hash_itr--;
			}

			if (iovec_stat[i].hash_rem_bytes) {
				//copy rem bytes to hash overlap buf
				memcpy(hash_overlap_buf.ptr, in_chunk_cur, iovec_stat[i].hash_rem_bytes);
				in_chunk_cur += iovec_stat[i].hash_rem_bytes;
				hash_overlap_buf.size = iovec_stat[i].hash_rem_bytes;
			}
		}

#ifdef DEBUG
		cavium_dbgprint("hash_overlap_buf.size : %d \n", hash_overlap_buf.size);
#endif
		//obviously this case gets hit all the time
		if (inner_hash_done == 0) {
			//if(hash_overlap_buf.size)
			{
				uint8_t *ptr = NULL;
				overlap_buf_size = hash_overlap_buf.size;
				memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size, enc_rem_ptr, enc_rem_cur_chunk);
				overlap_buf_size += enc_rem_cur_chunk;
#ifdef DEBUG
				cavium_dbgprint("enc_rem_cur_chunk : %d\n", enc_rem_cur_chunk);
#endif
				if (enc_rem_cur_chunk < (SHA256_DIGEST_LENGTH + pad_data_len)) {
					memcpy(hash_overlap_buf.ptr + hash_overlap_buf.size + enc_rem_cur_chunk, inp_vec[inp_cnt-1].ptr, inp_vec[inp_cnt-1].size);

					overlap_buf_size += inp_vec[inp_cnt-1].size;
#ifdef DEBUG
					cavium_dbgprint("overlap_buf_size : %d\n", overlap_buf_size);
#endif
				}
				ptr = hash_overlap_buf.ptr;
				uint16_t num_itr = 0;
				num_itr = overlap_buf_size / AES_BLOCK_SIZE;
				while (num_itr) {
					CVMX_LOADUNA_INT64(r0, ptr, 0);
					CVMX_LOADUNA_INT64(r1, ptr, 8);

					CVMX_MT_AES_DEC_CBC0(cvmx_htobe64(r0));
					CVMX_MT_AES_DEC_CBC1(cvmx_htobe64(r1));

					CVMX_MF_AES_RESULT(r2, 0);
					CVMX_MF_AES_RESULT(r3, 1);

#ifdef DEBUG
					cavium_dbgprint("r2 : %llx  r3 : %llx\n", r2, r3);
#endif
					CVMX_STOREUNA_INT64(cvmx_htobe64(r2), ptr, 0);
					CVMX_STOREUNA_INT64(cvmx_htobe64(r3), ptr, 8);

					ptr += AES_BLOCK_SIZE;
					num_itr--;
				}

				if (hash_overlap_buf.size) {
					if (output_overlap) {
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, iovec_stat[num_hash_iovec - 2].hash_rem_bytes);
						memcpy(outp_vec[outp_cnt -1].ptr, hash_overlap_buf.ptr + iovec_stat[num_hash_iovec - 2].hash_rem_bytes, iovec_stat[num_hash_iovec - 1].hash_req_bytes);
					} else {
						memcpy(out_chunk_cur, hash_overlap_buf.ptr, hash_overlap_buf.size);
					}
				}
			}

			memset(hash_rem_buf, 0, 128);
			memcpy(hash_rem_buf, hash_bal, 13);
			memcpy(hash_rem_buf + 13, hash_overlap_buf.ptr, hash_overlap_buf.size);
			hash_rem_buf[hash_overlap_buf.size + 13] = 0x80;
			uint64_t *ptr = (uint64_t *)hash_rem_buf;
#ifdef DEBUG
			cavium_dbgprint("hash rem buf\n");
			_hexdump(hash_rem_buf, 128);
#endif
			if ((hash_overlap_buf.size + 13) >= 56) {
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
				CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
				CVMX_MT_HSH_STARTSHA256(cvmx_htobe64(ptr[7]));

				ptr += 8;
			}
			inner_hash_total_bytes = SHA256_CBLOCK + 8 + 1 + 2 + 2 + raw_data_len;
			inner_hash_total_bits = inner_hash_total_bytes * 8;
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[0]), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[1]), 1);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[2]), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[3]), 3);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[4]), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[5]), 5);
			CVMX_MT_HSH_DAT(cvmx_htobe64(ptr[6]), 6);
			CVMX_MT_HSH_STARTSHA256(inner_hash_total_bits);

			inner_hash_done = 1;
		}
		free(iovec_stat);
	}

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);
	CVMX_MF_HSH_IV(r3, 3);

#ifdef DEBUG
	cavium_dbgprint("Inner Hash\n");
	cavium_dbgprint("r0 : %llx\n", r0);
	cavium_dbgprint("r1 : %llx\n", r1);
	cavium_dbgprint("r2 : %llx\n", r2);
	cavium_dbgprint("r3 : %llx\n", r3);
#endif

    if(key->ex_mode == EX_MODE_ENCRYPT)
        return 1;
    else if (key->ex_mode == EX_MODE_DIGEST) {
        CVMX_STOREUNA_INT64(cvmx_htobe64(r0),   outp_vec[0].ptr, 0);
        CVMX_STOREUNA_INT64(cvmx_htobe64(r1),   outp_vec[0].ptr, 8);
        CVMX_STOREUNA_INT64(cvmx_htobe64(r2),   outp_vec[0].ptr, 16);
        CVMX_STOREUNA_INT64(cvmx_htobe64(r3),   outp_vec[0].ptr, 24);
        return 1;
    }
	//calculate outer hash
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT2, 2);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT3, 3);

	outer_hash_total_bytes = SHA256_CBLOCK + SHA256_DIGEST_LENGTH;
	outer_hash_total_bits = outer_hash_total_bytes * 8;

	CVMX_LOADUNA_INT64(r4, mac_secret, 0);
	r4 = r4 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r4), 0);
	CVMX_LOADUNA_INT64(r4, mac_secret, 8);
	r4 = r4 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r4), 1);
	CVMX_LOADUNA_INT64(r4, mac_secret, 16);
	r4 = r4 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r4), 2);
	CVMX_LOADUNA_INT64(r4, mac_secret, 24);
	r4 = r4 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r4), 3);

	CVMX_LOADUNA_INT64(r4, CRY_HMAC_PAD2_64, 0);
	CVMX_MT_HSH_DAT(r4, 4);
	CVMX_MT_HSH_DAT(r4, 5);
	CVMX_MT_HSH_DAT(r4, 6);
	CVMX_MT_HSH_STARTSHA256(r4);

	inp_vec[0].ptr -=  (AES_BLOCK_SIZE);
	inp_vec[0].size += (AES_BLOCK_SIZE);

	CVMX_MT_HSH_DAT(r0, 0);
	CVMX_MT_HSH_DAT(r1, 1);
	CVMX_MT_HSH_DAT(r2, 2);
	CVMX_MT_HSH_DAT(r3, 3);
	CVMX_MT_HSH_DAT((uint64_t)CRY_HSH_BLK1, 4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	CVMX_MT_HSH_STARTSHA256(outer_hash_total_bits);

	CVMX_MF_HSH_IV(r0, 0);
	CVMX_MF_HSH_IV(r1, 1);
	CVMX_MF_HSH_IV(r2, 2);
	CVMX_MF_HSH_IV(r3, 3);

	CVMX_STOREUNA_INT64(cvmx_htobe64(r0), hash_buf, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r1), hash_buf, 8);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r2), hash_buf, 16);
	CVMX_STOREUNA_INT64(cvmx_htobe64(r3), hash_buf, 24);

#ifdef DEBUG
	cavium_dbgprint("Outer Hash Calculated : %s\n", __func__);
	_hexdump(hash_buf, SHA256_DIGEST_LENGTH);

#endif
    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr, hash_buf, SHA256_DIGEST_LENGTH);
        return 1;
    }

	//Compare Calculated HMAC and Received HMAC
	if (ret_val == 1) {
		if (inp_cnt == 1) {
            hash_buf_ptr_64 = (uint64_t *)hash_buf;
			if ((hash_buf_ptr_64[0] != (((uint64_t *)(out_ptr + hash_input_rem_len))[0])) ||
					(hash_buf_ptr_64[1] != (((uint64_t *)(out_ptr + hash_input_rem_len))[1])) ||
					(hash_buf_ptr_64[2] != (((uint64_t *)(out_ptr + hash_input_rem_len))[2])) ||
					(hash_buf_ptr_64[3] != (((uint64_t *)(out_ptr + hash_input_rem_len))[3]))) {
				ret_val = 0;
				cavium_dbgprint( "%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
				cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
				cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(out_ptr + hash_input_rem_len))[0]));
				cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(out_ptr + hash_input_rem_len))[1]));
				cavium_dbgprint("3. %llx : %llx\n", ((uint64_t *)hash_buf)[2], (((uint64_t *)(out_ptr + hash_input_rem_len))[2]));
				cavium_dbgprint("4. %llx : %llx\n", ((uint64_t *)hash_buf)[3], (((uint64_t *)(out_ptr + hash_input_rem_len))[3]));
#endif
			}
		} else {
			if (memcmp(hash_buf, hash_overlap_buf.ptr + hash_overlap_buf.size, SHA256_DIGEST_LENGTH)) {
				ret_val = 1;
				cavium_dbgprint( "%s : %d Invalid HMAC for Data Length : %d\n", __func__, __LINE__, raw_data_len);
#ifdef DEBUG
				cavium_dbgprint("Calculated HMAC vs Received HMAC\n");
				cavium_dbgprint("1. %llx : %llx\n", ((uint64_t *)hash_buf)[0], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[0]));
				cavium_dbgprint("2. %llx : %llx\n", ((uint64_t *)hash_buf)[1], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[1]));
				cavium_dbgprint("3. %llx : %llx\n", ((uint64_t *)hash_buf)[2], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[2]));
				cavium_dbgprint("4. %llx : %llx\n", ((uint64_t *)hash_buf)[3], (((uint64_t *)(hash_overlap_buf.ptr + hash_overlap_buf.size))[3]));
#endif
			}
		}
	}



// May be need to update outp_vec[*].size 
	*len = raw_data_len; 
	return ret_val;
}


/*******************************************************************************
 *                                                                             *
 *      Generates  TLS-1.2 SHA256 HMAC, adds padding bytes and encrypts        *
 *  payload data and stores the result in output iovec buffer. Initialisation  *
 *  vector  be used for encrytion is taken from the ext_iv_data pointer passed *
 *  to the function. Key and MAC secret to be used for cipher encryption and   *
 *  HMAC generation respectively are taken from the key structure passed.      *
 *  Sequence number and record header to be used for used for HMAC generation  *
 *  are extracted from AAD stored in KEY structure. First 8bytes of AAD holds  *
 *  sequence number and next 5 bytes holds the  TLS-1.2 record header.         *
 *  Protocol : TLS-1.2                                                         *
 *  Cipher   : AES128-CBC / AES256-CBC                                         *
 *  HMAC     : SHA256                                                          *
 *                                                                             *
 *  Input :-                                                                   *
 *  key  : Cipher data structure containing encryption key, IV, HMAC secret,   *
 *         AAD.                                                                *
 *  inp_vec : Input data to encrypt as IOVEC structure.                        *
 *  inp_cnt : Number of input IOVEC buffers.                                   *
 *  outp_vec: Destination IOVEC structure to store encryption result.          *
 *  outp_cnt : Number of outputput IOVEC buffers.                              *
 *  ext_iv_data : Pointer to explicit IV to be used for encryption.            *
 *  len     : Length of input payload data.                                    *
 *                                                                             *
 *******************************************************************************/
int32_t oct_aes_sha256_enc_tls12(EVP_AESOCT_HMAC_SHA1 *key,
		cav_iovec_t *inp_vec, uint16_t inp_cnt,
		cav_iovec_t *outp_vec, uint16_t outp_cnt,
		uint8_t *ext_iv_ptr,uint16_t *len)
{
	register uint64_t p0, p1, p2, p3, p4, p5, r0, r1;
	uint8_t sha1_rem[128] = {0};
	uint8_t aes_rem[128] = {0};
	uint8_t aes_out[128] = {0};
	uint8_t idx = 0;
	uint16_t inp_len = 0, cipher_type = 0;
	uint32_t aes_rem_num = 0, sha1_rem_num = 0;
	int32_t aes_out_len = 0, keylength = 1, cal_enc_len = 0, eout_space_left = 0, aes_copied_bytes = 0, sha1_copied_bytes = 0;
	int32_t fill_bytes = 0, next_output_iovec = 0, start_inner_hash = 0 ;
	uint64_t tmpvar = 0, total_bits = 0, seq = cvmx_htobe64(key->aux.tls_aad_64[0]);
      	uint8_t *mac_secret = NULL, *key_data = NULL, *hash_in = NULL, *hash_end = NULL, *enc_in = NULL, *enc_end = NULL, *enc_out = NULL;
	uint64_t *ptr = NULL, *enc_key = (uint64_t *)key->cipher_key;
    uint64_t *hash_buf_ptr_64 = NULL;
	struct tls_rec_hdr *in_rec_hdr = (struct tls_rec_hdr *)&key->aux.tls_aad_8[8];

#ifdef DEBUG
	cavium_dbgprint("rec length: %d, seq: %llx\n", in_rec_hdr->length, seq);
	cavium_dbgprint("inp_cnt: %d, outp_cnt: %d\n", inp_cnt, outp_cnt);
#endif
	mac_secret = key->mac_secret;

	cipher_type = key->cipher_type;

/*Generate External IV*/

	key_data =key->cipher_iv;


	/* assume AES keylen 128 bit */
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[0]), 0);
	CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[1]), 1);
	if (cipher_type == OCT_AES_256) {
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[2]), 2);
		CVMX_MT_AES_KEY(cvmx_htobe64(enc_key[3]), 3);
		keylength = 3;
	} else {
		CVMX_MT_AES_KEY((uint64_t)0, 2);
		CVMX_MT_AES_KEY((uint64_t)0, 3);
	}
	CVMX_MT_AES_KEYLENGTH(keylength);
	
	/* Init AES iv */
	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)key_data), 0);
	CVMX_MT_AES_IV(cvmx_htobe64(*(uint64_t *)(&key_data[8])), 1);

	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT2, 2);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT3, 3);

	CVMX_LOADUNA_INT64(p0, mac_secret, 0);
	CVMX_LOADUNA_INT64(p1, mac_secret, 8);
	CVMX_LOADUNA_INT64(p2, mac_secret, 16);
	CVMX_LOADUNA_INT64(p3, mac_secret, 24);

	p0 = p0 ^ (0x3636363636363636);
	p1 = p1 ^ (0x3636363636363636);
	p2 = p2 ^ (0x3636363636363636);
	p3 = p3 ^ (0x3636363636363636);

    if(key->ex_mode != EX_MODE_DIGEST) {
        CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
        CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
        CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
        CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
	    CVMX_LOADUNA_INT64(r0, CRY_HMAC_PAD1_64, 0);
	    CVMX_MT_HSH_DAT(r0, 4);
	    CVMX_MT_HSH_DAT(r0, 5);
	    CVMX_MT_HSH_DAT(r0, 6);
	    CVMX_MT_HSH_STARTSHA256(r0);
    }

	CVMX_PREFETCH0(inp_vec[idx].ptr);

	inp_len = cvmx_htobe16(in_rec_hdr->length);
    if(key->ex_mode == EX_MODE_DIGEST)
        total_bits = inp_len * 8;
    else if (key->ex_mode == EX_MODE_HMAC)
        total_bits = (SHA256_CBLOCK + inp_len)*8;
    else
    	total_bits = ((SHA256_CBLOCK + 8 + 1 + 2 + 2 + inp_len) * 8);

    if(key->ex_mode == EX_MODE_DEFAULT) {
	/*store 8 bytes seq_num , 1 byte record type, 2 bytes record version, 2 bytes record length*/
        hash_buf_ptr_64 = (uint64_t *)sha1_rem;
        hash_buf_ptr_64[0] = cvmx_htobe64(seq);
	    CVMX_STOREUNA_INT32(key->aux.tls_aad_32[2], sha1_rem, 8); //12
	    sha1_rem[12] = key->aux.tls_aad_8[12]; //13
	    sha1_copied_bytes = 13;
        /* seq num (8 bytes) + record type (1 byte) + record version (2 bytes) + record length (2 bytes) */
    } else {
        if(key->ex_mode & 0x02) {
            if(inp_len >= 16) {
                memcpy(sha1_rem,inp_vec[0].ptr,16);
                inp_vec[0].ptr += 16;
                inp_vec[0].size -= 16;
                sha1_copied_bytes = 16;
            } else {
                memcpy(sha1_rem,inp_vec[0].ptr,inp_len);
                inp_vec[0].ptr += inp_len;
                inp_vec[0].size -= inp_len;
                sha1_copied_bytes = inp_len;
            }
        }
    }
	aes_copied_bytes = 0;


/*Encrpyt External IV*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    enc_in = ext_iv_ptr;
	    CVMX_LOADUNA_INT64(p0, enc_in, 0);
	    CVMX_LOADUNA_INT64(p1, enc_in, 8);

	    CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
	    CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

	    enc_out = outp_vec[idx].ptr ;

	    CVMX_MF_AES_RESULT(p0, 0);	
	    CVMX_MF_AES_RESULT(p1, 1);	

	    CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

    }
	hash_in = inp_vec[idx].ptr ;
	hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
	sha1_rem_num = (hash_end - hash_in) + sha1_copied_bytes;

	enc_in = inp_vec[idx].ptr ;
	enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
	aes_rem_num = (enc_end - enc_in);

    if(key->ex_mode == EX_MODE_DEFAULT)
    	enc_out = outp_vec[idx].ptr  + AES_BLOCK_SIZE;
    else
        enc_out = outp_vec[idx].ptr ;

	if (sha1_rem_num <= SHA256_CBLOCK || key->ex_mode == EX_MODE_ENCRYPT) {
		ptr = (uint64_t *)(sha1_rem + sha1_copied_bytes);	
/*memset 64 bytes to 0*/
		CVMX_STOREUNA_INT64(0, ptr, 0);
		CVMX_STOREUNA_INT64(0, ptr, 8);
		CVMX_STOREUNA_INT64(0, ptr, 16);
		CVMX_STOREUNA_INT64(0, ptr, 24);
		CVMX_STOREUNA_INT64(0, ptr, 32);
		CVMX_STOREUNA_INT64(0, ptr, 40);
		CVMX_STOREUNA_INT64(0, ptr, 48);
		CVMX_STOREUNA_INT64(0, ptr, 56);

		memcpy(sha1_rem + sha1_copied_bytes, hash_in, inp_vec[idx].size );	
		sha1_copied_bytes += (inp_vec[idx].size );
		sha1_rem[sha1_copied_bytes] = 0x80;

		
        if(key->ex_mode != EX_MODE_ENCRYPT)
            SHA256_FINISH(sha1_rem, sha1_copied_bytes, total_bits);
		sha1_copied_bytes = 0;
		sha1_rem_num = 0;

		while (aes_rem_num >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	
			CVMX_LOADUNA_INT64(p1, enc_in, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
			
			enc_in += AES_BLOCK_SIZE;
			aes_rem_num -= AES_BLOCK_SIZE;

			CVMX_MF_AES_RESULT(p0, 0);	
			CVMX_MF_AES_RESULT(p1, 1);	

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

			enc_out += AES_BLOCK_SIZE;
		}

		memcpy(aes_rem, enc_in, aes_rem_num);
		aes_copied_bytes = aes_rem_num;

		goto get_inner_hash;
	}

	/* inp_vec[0] has payload(inp_vec[0].size - sizeof(struct tls_rec_hdr)) > 51 bytes */
	/*sha1 copied bytes = 13, so pushing 3 more bytes from hash_in to sha1_rem*/
    if(key->ex_mode == EX_MODE_DEFAULT) {
	    sha1_rem[sha1_copied_bytes] = hash_in[0];
	    sha1_rem[sha1_copied_bytes + 1] = hash_in[1];
	    sha1_rem[sha1_copied_bytes + 2] = hash_in[2];
	    ptr = (uint64_t *)(hash_in + 3);
    } else {
        ptr = (uint64_t *)hash_in;
    }

	CVMX_LOADUNA_INT64(p0, ptr, 0);
	CVMX_LOADUNA_INT64(p1, ptr, 8);
	CVMX_LOADUNA_INT64(p2, ptr, 16);
	CVMX_LOADUNA_INT64(p3, ptr, 24);
	CVMX_LOADUNA_INT64(p4, ptr, 32);
	
    hash_buf_ptr_64 = (uint64_t *)sha1_rem;
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[0]), 0);
	CVMX_MT_HSH_DAT(cvmx_htobe64(hash_buf_ptr_64[1]), 1);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 4);
    CVMX_LOADUNA_INT64(p0, ptr, 40);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 5);
	CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 6);

	CVMX_MT_HSH_STARTSHA256(cvmx_htobe64(p0));

    if(key->ex_mode & 0x02)
        hash_in += 48;
    else
	    hash_in += 51;
	sha1_rem_num = (hash_end - hash_in);
	sha1_copied_bytes = 0;
	
	/* SHA256_CBLOCK chunks in 1st iovec */

	while ((hash_in + SHA256_CBLOCK) <= hash_end) {
		/* load 1st aes bytes */
		CVMX_LOADUNA_INT64(p0, enc_in, 0);	
		CVMX_LOADUNA_INT64(p1, enc_in, 8);
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 0);
		CVMX_LOADUNA_INT64(p1, hash_in, 8);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		
		/* get 2nd aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 16);
		CVMX_LOADUNA_INT64(p1, enc_in, 24);

		/* 1st aes result */
		CVMX_MF_AES_RESULT(p2, 0);	
		CVMX_MF_AES_RESULT(p3, 1);	

		/* load 2nd aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 16);
		CVMX_LOADUNA_INT64(p1, hash_in, 24);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

		/* store 1st aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 0);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

		/* get 3rd aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 32);
		CVMX_LOADUNA_INT64(p1, enc_in, 40);
		
		/* 2nd aes result */
		CVMX_MF_AES_RESULT(p2, 0);	
		CVMX_MF_AES_RESULT(p3, 1);	

		/* load 3rd aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 32);
		CVMX_LOADUNA_INT64(p1, hash_in, 40);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

		/* store 2nd aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 16);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

		/* get 4th aes input */
		CVMX_LOADUNA_INT64(p0, enc_in, 48);
		CVMX_LOADUNA_INT64(p1, enc_in, 56);
		
		/* get 3rd aes result */
		CVMX_MF_AES_RESULT(p2, 0);
		CVMX_MF_AES_RESULT(p3, 1);

		/* load 4th aes bytes */
		CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
		CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

		/* hash bytes */
		CVMX_LOADUNA_INT64(p0, hash_in, 48);
		CVMX_LOADUNA_INT64(p1, hash_in, 56);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_MT_HSH_STARTSHA256(cvmx_htobe64(p1));

		hash_in += SHA256_CBLOCK;
		enc_in += SHA256_CBLOCK;
		sha1_rem_num -= SHA256_CBLOCK;
		aes_rem_num -= SHA256_CBLOCK;

		/* store 3rd aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 32);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

		/* get 4th aes result */
		CVMX_MF_AES_RESULT(p2, 0);
		CVMX_MF_AES_RESULT(p3, 1);
			
		CVMX_PREFETCH(enc_in, SHA256_CBLOCK);

		/* store 4th aes result */
		CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 48);
		CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);
		
		enc_out += SHA256_CBLOCK;
	}

	/* for each iovec */
	while (idx++ < inp_cnt) {
		/* for each SHA256_CBLOCK in iovec */
		while ((hash_in + SHA256_CBLOCK) <= hash_end) {
			/* load 1st aes bytes */
			CVMX_LOADUNA_INT64(p0, enc_in, 0);	
			CVMX_LOADUNA_INT64(p1, enc_in, 8);
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 0);
			CVMX_LOADUNA_INT64(p1, hash_in, 8);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		
			/* get 2nd aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 16);
			CVMX_LOADUNA_INT64(p1, enc_in, 24);

			/* 1st aes result */
			CVMX_MF_AES_RESULT(p2, 0);	
			CVMX_MF_AES_RESULT(p3, 1);	

			/* load 2nd aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 16);
			CVMX_LOADUNA_INT64(p1, hash_in, 24);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 2);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 3);

			/* store 1st aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 8);

			/* get 3rd aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 32);
			CVMX_LOADUNA_INT64(p1, enc_in, 40);
		
			/* 2nd aes result */
			CVMX_MF_AES_RESULT(p2, 0);	
			CVMX_MF_AES_RESULT(p3, 1);	

			/* load 3rd aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 32);
			CVMX_LOADUNA_INT64(p1, hash_in, 40);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 4);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 5);

			/* store 2nd aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 16);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 24);

			/* get 4th aes input */
			CVMX_LOADUNA_INT64(p0, enc_in, 48);
			CVMX_LOADUNA_INT64(p1, enc_in, 56);
		
			/* get 3rd aes result */
			CVMX_MF_AES_RESULT(p2, 0);
			CVMX_MF_AES_RESULT(p3, 1);

			/* load 4th aes bytes */
			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			/* hash bytes */
			CVMX_LOADUNA_INT64(p0, hash_in, 48);
			CVMX_LOADUNA_INT64(p1, hash_in, 56);
			CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
			CVMX_MT_HSH_STARTSHA256(cvmx_htobe64(p1));

			hash_in += SHA256_CBLOCK;
			enc_in += SHA256_CBLOCK;
			sha1_rem_num -= SHA256_CBLOCK;
			aes_rem_num -= SHA256_CBLOCK;

			/* store 3rd aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 32);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 40);

			/* get 4th aes result */
			CVMX_MF_AES_RESULT(p2, 0);
			CVMX_MF_AES_RESULT(p3, 1);
			
			CVMX_PREFETCH(enc_in, SHA256_CBLOCK);

			/* store 4th aes result */
			CVMX_STOREUNA_INT64(cvmx_htobe64(p2), enc_out, 48);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p3), enc_out, 56);
		
			enc_out += SHA256_CBLOCK;
		} /* for each SHA256_CBLOCK in iovec */

		sha1_rem_num = (hash_end - hash_in);
		aes_rem_num = (hash_end - enc_in);

#ifdef DEBUG
		cavium_dbgprint("iovec_idx : %d sha1_rem_num: %d, aes_rem_num: %d (%d)\n", idx-1, sha1_rem_num, aes_rem_num, __LINE__);
#endif
		/* copy the remaining sha1 bytes */
		memcpy(sha1_rem, hash_in, sha1_rem_num);
		sha1_copied_bytes = sha1_rem_num;
		sha1_rem_num = 0;

		if (idx < inp_cnt) {
			/* try fill the sha1 block, get bytes from next iovec */
			hash_in = inp_vec[idx].ptr;
			hash_end = inp_vec[idx].ptr + inp_vec[idx].size;
			sha1_rem_num = (hash_end - hash_in);
		}
#ifdef DEBUG
		cavium_dbgprint("iovec_idx : %d sha1_rem_num: %d, aes_rem_num: %d (%d)\n", idx, sha1_rem_num, aes_rem_num, __LINE__);
#endif

		/* less bytes than needed, try to finish inner hash */
		if (sha1_rem_num < (SHA256_CBLOCK - sha1_copied_bytes)) {
			memcpy(sha1_rem + sha1_copied_bytes, hash_in, sha1_rem_num);
			sha1_copied_bytes += sha1_rem_num;
			sha1_rem[sha1_copied_bytes] = 0x80;
			
			memset(sha1_rem + sha1_copied_bytes + 1, 0, (SHA256_CBLOCK - sha1_copied_bytes - 1));

			SHA256_FINISH(sha1_rem, sha1_copied_bytes, total_bits);
			sha1_rem_num = 0;
			sha1_copied_bytes = 0;

			/* complete aes bytes in previous iovec */
			while (aes_rem_num >= AES_BLOCK_SIZE) {
				CVMX_LOADUNA_INT64(p0, enc_in, 0);
				CVMX_LOADUNA_INT64(p1, enc_in, 8);

				CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
				CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

				enc_in += AES_BLOCK_SIZE;
				aes_rem_num -= AES_BLOCK_SIZE;

				CVMX_MF_AES_RESULT(p0, 0);
				CVMX_MF_AES_RESULT(p1, 1);

				CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
				CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

				enc_out += AES_BLOCK_SIZE;
			}

#ifdef DEBUG
		cavium_dbgprint("iovec_idx : %d sha1_rem_num: %d, aes_rem_num: %d (%d)\n", idx, sha1_rem_num, aes_rem_num, __LINE__);
#endif
			if (aes_rem_num) {
				/* copy the remaining aes bytes from previous iovec */
				memcpy(aes_rem, enc_in, aes_rem_num);
				aes_copied_bytes = aes_rem_num;
				eout_space_left = aes_rem_num;
				aes_rem_num = 0;

				if (idx < inp_cnt) {
					/* move to next iovec */
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					aes_rem_num = (enc_end - enc_in);
				}

				/* fill bytes from next iovec */
				if (aes_rem_num <= (AES_BLOCK_SIZE - aes_copied_bytes)) {
					memcpy(aes_rem + aes_copied_bytes, enc_in, aes_rem_num);
					aes_copied_bytes += aes_rem_num;
					aes_rem_num = 0;
					if (idx < outp_cnt)
						next_output_iovec = 1;
				} else {
					fill_bytes = (AES_BLOCK_SIZE - aes_copied_bytes);
					uint64_t enc_result[2];
					memcpy(aes_rem + aes_copied_bytes, enc_in, fill_bytes);
					aes_copied_bytes = AES_BLOCK_SIZE;
					aes_rem_num -= fill_bytes;

					CVMX_LOADUNA_INT64(p0, aes_rem, 0);
					CVMX_LOADUNA_INT64(p1, aes_rem, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

					enc_in += fill_bytes;

					CVMX_MF_AES_RESULT(enc_result[0], 0);
					CVMX_MF_AES_RESULT(enc_result[1], 1);
                    
                    enc_result[0] = cvmx_htobe64(enc_result[0]);
                    enc_result[1] = cvmx_htobe64(enc_result[1]);

					/* copy into the space left */
					memcpy(enc_out, enc_result, eout_space_left);	

					/* move to next output iovec */
					enc_out = outp_vec[idx].ptr;
					memcpy(enc_out, ((uint8_t *)enc_result + eout_space_left), fill_bytes);
					enc_out += fill_bytes;
					eout_space_left = 0;
					aes_copied_bytes = 0;

					/* encrypt AES_BLOCK_SIZE chunks if any */
					while (aes_rem_num >= AES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);
						CVMX_LOADUNA_INT64(p1, enc_in, 8);

						CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
						CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
						enc_in += AES_BLOCK_SIZE;
						aes_rem_num -= AES_BLOCK_SIZE;

						CVMX_MF_AES_RESULT(p0, 0);
						CVMX_MF_AES_RESULT(p1, 1);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
						CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
						enc_out += AES_BLOCK_SIZE;
					}
					memcpy(aes_rem, enc_in, aes_rem_num);
					aes_copied_bytes = aes_rem_num;
				}
			} else {
				if (idx < inp_cnt) {
					enc_in = inp_vec[idx].ptr;
					enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
					aes_rem_num = (enc_end - enc_in);
				}
			
				if (idx < outp_cnt)
					enc_out = outp_vec[idx].ptr;

				while (aes_rem_num >= AES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);
					CVMX_LOADUNA_INT64(p1, enc_in, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
					enc_in += AES_BLOCK_SIZE;
					aes_rem_num -= AES_BLOCK_SIZE;

					CVMX_MF_AES_RESULT(p0, 0);
					CVMX_MF_AES_RESULT(p1, 1);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
					CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
					enc_out += AES_BLOCK_SIZE;
				}
				if (enc_in) {
					memcpy(aes_rem, enc_in, aes_rem_num);	
					aes_copied_bytes = aes_rem_num;
					aes_rem_num = 0;
				}
			}
			goto get_inner_hash;
		}

		/* more than fill bytes in next iovec */
		fill_bytes = SHA256_CBLOCK - sha1_copied_bytes;
		memcpy(sha1_rem + sha1_copied_bytes, hash_in, fill_bytes);
		sha1_copied_bytes += fill_bytes;

		hash_in += fill_bytes;
		sha1_rem_num -= fill_bytes;

		ptr = (uint64_t *)sha1_rem;
		CVMX_LOADUNA_INT64(p0, ptr, 0);
		CVMX_LOADUNA_INT64(p1, ptr, 8);
		CVMX_LOADUNA_INT64(p2, ptr, 16);
		CVMX_LOADUNA_INT64(p3, ptr, 24);
		CVMX_LOADUNA_INT64(p4, ptr, 32);
		CVMX_LOADUNA_INT64(p5, ptr, 40);
	
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
		CVMX_LOADUNA_INT64(p0, ptr, 48);
		CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
		CVMX_LOADUNA_INT64(p1, ptr, 56);
		
		CVMX_MT_HSH_STARTSHA256(cvmx_htobe64(p1));
		
		while (aes_rem_num >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, enc_in, 0);
			CVMX_LOADUNA_INT64(p1, enc_in, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			enc_in += AES_BLOCK_SIZE;
			aes_rem_num -= AES_BLOCK_SIZE;

			CVMX_MF_AES_RESULT(p0, 0);
			CVMX_MF_AES_RESULT(p1, 1);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);

			enc_out += AES_BLOCK_SIZE;
		}

		/* if sha256 remain bytes are less than SHA256_CBLOCK, finish inner hash */
		if (sha1_rem_num < SHA256_CBLOCK) {
			memcpy(sha1_rem, hash_in, sha1_rem_num);	
			sha1_copied_bytes = sha1_rem_num;
			sha1_rem[sha1_copied_bytes] = 0x80;

			memset(sha1_rem + sha1_copied_bytes + 1, 0, (SHA256_CBLOCK - sha1_copied_bytes - 1));

			SHA256_FINISH(sha1_rem, sha1_copied_bytes, total_bits);
			sha1_rem_num = 0;
			sha1_copied_bytes = 0;

			start_inner_hash = 1;
		}

		if (aes_rem_num) {
			uint64_t enc_result[2];

			memcpy(aes_rem, enc_in, aes_rem_num);
			aes_copied_bytes = aes_rem_num;
			eout_space_left = aes_rem_num;

			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;	
			aes_rem_num = (enc_end - enc_in);

			if (aes_rem_num <= (AES_BLOCK_SIZE - aes_copied_bytes)) {
				/* copy the remaining aes bytes */
				memcpy(aes_rem + aes_copied_bytes, enc_in, aes_rem_num);
				aes_copied_bytes += aes_rem_num;
				aes_rem_num = 0;
				if (idx < outp_cnt)
					next_output_iovec = 1;

				goto get_inner_hash;
			} else {
				fill_bytes = (AES_BLOCK_SIZE - aes_copied_bytes);

				/* chance that aes bytes are crossing sha1 bytes */
				if ((aes_rem_num - fill_bytes) < sha1_rem_num) {
					ptr = (uint64_t *)hash_in;	
					CVMX_LOADUNA_INT64(p0, ptr, 0);
					CVMX_LOADUNA_INT64(p1, ptr, 8);
					CVMX_LOADUNA_INT64(p2, ptr, 16);
					CVMX_LOADUNA_INT64(p3, ptr, 24);
					CVMX_LOADUNA_INT64(p4, ptr, 32);
					CVMX_LOADUNA_INT64(p5, ptr, 40);
	
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 0);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p1), 1);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p2), 2);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p3), 3);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p4), 4);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p5), 5);
					CVMX_LOADUNA_INT64(p0, ptr, 48);
					CVMX_MT_HSH_DAT(cvmx_htobe64(p0), 6);
					CVMX_LOADUNA_INT64(p1, ptr, 56);
		
					CVMX_MT_HSH_STARTSHA256(cvmx_htobe64(p1));

					hash_in += SHA256_CBLOCK;
					sha1_rem_num -= SHA256_CBLOCK;
				}

				memcpy(aes_rem + aes_copied_bytes, enc_in, fill_bytes); 
				aes_copied_bytes += fill_bytes;

				CVMX_LOADUNA_INT64(p0, aes_rem, 0);
				CVMX_LOADUNA_INT64(p1, aes_rem, 8);

				CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
				CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

				enc_in += fill_bytes;
				aes_rem_num -= fill_bytes;

				CVMX_MF_AES_RESULT(enc_result[0], 0);
				CVMX_MF_AES_RESULT(enc_result[1], 1);

                enc_result[0] = cvmx_htobe64(enc_result[0]);
                enc_result[1] = cvmx_htobe64(enc_result[1]);
			
				memcpy(enc_out, enc_result, eout_space_left);

				/* move to next output iovec */
				enc_out = outp_vec[idx].ptr;
				memcpy(enc_out, ((uint8_t *)enc_result + eout_space_left), fill_bytes);
				enc_out += fill_bytes;

				eout_space_left = 0;
				aes_copied_bytes = 0;

				if (start_inner_hash) {
					/* encrypt AES_BLOCK_SIZE chunks if any */
					while (aes_rem_num >= AES_BLOCK_SIZE) {
						CVMX_LOADUNA_INT64(p0, enc_in, 0);
						CVMX_LOADUNA_INT64(p1, enc_in, 8);

						CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
						CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
						enc_in += AES_BLOCK_SIZE;
						aes_rem_num -= AES_BLOCK_SIZE;

						CVMX_MF_AES_RESULT(p0, 0);
						CVMX_MF_AES_RESULT(p1, 1);

						CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
						CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
						enc_out += AES_BLOCK_SIZE;
					}
					memcpy(aes_rem, enc_in, aes_rem_num);
					aes_copied_bytes = aes_rem_num;

					goto get_inner_hash;
				}
			}
		} else {
			enc_in = inp_vec[idx].ptr;
			enc_end = inp_vec[idx].ptr + inp_vec[idx].size;
			aes_rem_num = (enc_end - enc_in);

			if (idx < outp_cnt)
				enc_out = outp_vec[idx].ptr;

			if (start_inner_hash) {
				while (aes_rem_num >= AES_BLOCK_SIZE) {
					CVMX_LOADUNA_INT64(p0, enc_in, 0);
					CVMX_LOADUNA_INT64(p1, enc_in, 8);

					CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
					CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));
						
					enc_in += AES_BLOCK_SIZE;
					aes_rem_num -= AES_BLOCK_SIZE;

					CVMX_MF_AES_RESULT(p0, 0);
					CVMX_MF_AES_RESULT(p1, 1);

					CVMX_STOREUNA_INT64(cvmx_htobe64(p0), enc_out, 0);	
					CVMX_STOREUNA_INT64(cvmx_htobe64(p1), enc_out, 8);	
						
					enc_out += AES_BLOCK_SIZE;
				}

				if (enc_in) {
					memcpy(aes_rem, enc_in, aes_rem_num);
					aes_copied_bytes = aes_rem_num;
					aes_rem_num = 0;
				}

				goto get_inner_hash;
			}
		}
	} /* for each iovec */
    if(key->ex_mode == EX_MODE_ENCRYPT) {
        return 1;
    }

get_inner_hash:
	/* get inner hash */
	//CVMX_MF_HSH_IV(tmpvar, 0);
	//CVMX_MF_HSH_IV(tmpvar2, 1);
	//CVMX_MF_HSH_IV(r0, 2);
	CVMX_MF_HSH_IV(p0, 0);
	CVMX_MF_HSH_IV(p1, 1);
	CVMX_MF_HSH_IV(p2, 2);
	CVMX_MF_HSH_IV(p3, 3);
    if(key->ex_mode == EX_MODE_DIGEST) {
        ptr = (uint64_t *)outp_vec[0].ptr;
        CVMX_STOREUNA_INT64(cvmx_htobe64(p0),ptr,0);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(p1), ptr, 8);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(p2), ptr, 16);
	    CVMX_STOREUNA_INT64(cvmx_htobe64(p3), ptr, 24);
        return 1;
    }

#ifdef DEBUG
	cavium_dbgprint("sha1_copied_bytes: %d, sha1_rem_num: %d (%d)\n", sha1_copied_bytes, sha1_rem_num, __LINE__);
	_hexdump(sha1_rem, sha1_copied_bytes);
	cavium_dbgprint("aes_copied_bytes: %d, aes_rem_num: %d (%d)\n", aes_copied_bytes, aes_rem_num, __LINE__);
	_hexdump(aes_rem, aes_copied_bytes);;
	cavium_dbgprint("Inner Hash\n");
	cavium_dbgprint("%llx\n", p0);
	cavium_dbgprint("%llx\n", p1);
	cavium_dbgprint("%llx\n", p2);
	cavium_dbgprint("%llx\n", p3);
#endif

	/* start outer hash ((mac secret ^ pad_2) + inner hash) */


	/* SHA init */
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT0, 0);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT1, 1);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT2, 2);
	CVMX_MT_HSH_IV((uint64_t)CRY_SHA256_INIT3, 3);

	CVMX_LOADUNA_INT64(r1, mac_secret, 0);
	r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 0);
	CVMX_LOADUNA_INT64(r1, mac_secret, 8);
	r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 1);
	CVMX_LOADUNA_INT64(r1, mac_secret, 16);
	r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 2);
	CVMX_LOADUNA_INT64(r1, mac_secret, 24);
	r1 = r1 ^ (0x5c5c5c5c5c5c5c5c);
	CVMX_MT_HSH_DAT(cvmx_htobe64(r1), 3);
	
	CVMX_LOADUNA_INT64(r1, CRY_HMAC_PAD2_64, 0);
	CVMX_MT_HSH_DAT(r1, 4);
	CVMX_MT_HSH_DAT(r1, 5);
	CVMX_MT_HSH_DAT(r1, 6);
	CVMX_MT_HSH_STARTSHA256(r1);

	ptr = (uint64_t *)&aes_rem[aes_copied_bytes];

	CVMX_MT_HSH_DAT(p0, 0);
	CVMX_MT_HSH_DAT(p1, 1);
	CVMX_MT_HSH_DAT(p2, 2);
	CVMX_MT_HSH_DAT(p3, 3);

	CVMX_MT_HSH_DAT((uint64_t)CRY_HSH_BLK1, 4);
	CVMX_MT_HSH_DATZ(5);
	CVMX_MT_HSH_DATZ(6);

	total_bits = ((SHA256_CBLOCK + SHA256_DIGEST_LENGTH) * 8);
	CVMX_MT_HSH_STARTSHA256(total_bits);
      
#ifdef DEBUG
	cavium_dbgprint("Aes copied bytes to aes rem buf : %d\n", aes_copied_bytes);
#endif
	aes_copied_bytes += SHA256_DIGEST_LENGTH;

	/* calculate pad length */
      	tmpvar = AES_BLOCK_SIZE - ((aes_copied_bytes) % AES_BLOCK_SIZE);
	
	/* add pad bytes pattren (padlen - 1) */
      	memset(aes_rem + aes_copied_bytes, tmpvar - 1, tmpvar);
	aes_copied_bytes += tmpvar;

	/* get outer hash */
	CVMX_MF_HSH_IV(p0, 0);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p0), ptr, 0);
	CVMX_MF_HSH_IV(p1, 1);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p1), ptr, 8);
	CVMX_MF_HSH_IV(p2, 2);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p2), ptr, 16);
	CVMX_MF_HSH_IV(p3, 3);
	CVMX_STOREUNA_INT64(cvmx_htobe64(p3), ptr, 24);

    if(key->ex_mode == EX_MODE_HMAC) {
        memcpy(outp_vec[0].ptr,ptr,SHA256_DIGEST_LENGTH);
        return 1;
    }
#ifdef DEBUG
	cavium_dbgprint("Outer Hash\n");
	cavium_dbgprint("%llx\n", p0);
	cavium_dbgprint("%llx\n", p1);
	cavium_dbgprint("%llx\n", p2);
	cavium_dbgprint("%llx\n", p3);
#endif
#ifdef DEBUG
	cavium_dbgprint("Aes copied bytes to aes rem buf : %d\n", aes_copied_bytes);
#endif
	if (aes_copied_bytes) {
		uint8_t *out_aes_ptr = aes_out;
		uint8_t *in_aes_ptr = aes_rem;
		aes_out_len = aes_copied_bytes;

		while (aes_copied_bytes >= AES_BLOCK_SIZE) {
			CVMX_LOADUNA_INT64(p0, in_aes_ptr, 0);
			CVMX_LOADUNA_INT64(p1, in_aes_ptr, 8);

			CVMX_MT_AES_ENC_CBC0(cvmx_htobe64(p0));
			CVMX_MT_AES_ENC_CBC1(cvmx_htobe64(p1));

			in_aes_ptr += AES_BLOCK_SIZE;
			aes_copied_bytes -= AES_BLOCK_SIZE;	

			CVMX_MF_AES_RESULT(p0, 0);
			CVMX_MF_AES_RESULT(p1, 1);

			CVMX_STOREUNA_INT64(cvmx_htobe64(p0), out_aes_ptr, 0);
			CVMX_STOREUNA_INT64(cvmx_htobe64(p1), out_aes_ptr, 8);
			
			out_aes_ptr += AES_BLOCK_SIZE;
		}
		if (aes_copied_bytes)
			cavium_dbgprint("ERROR: aes_copied_bytes: %d\n", aes_copied_bytes);
	}	

#ifdef DEBUG
	cavium_dbgprint("eout space left in previous output iovec : %d\n", eout_space_left);
#endif
	memcpy(enc_out, aes_out, eout_space_left);

	if (next_output_iovec)
		enc_out = outp_vec[idx].ptr;
	else
		enc_out += eout_space_left;

	aes_rem_num = aes_out_len - eout_space_left;

	if (enc_out)
		memcpy(enc_out, aes_out + eout_space_left, aes_rem_num);


	//update ssl output header

        cal_enc_len = ROUNDUP16(AES_BLOCK_SIZE + inp_len + SHA256_DIGEST_LENGTH + 1);
	    *len= cal_enc_len;

	// External IV is included between Record HDR and Record Data, so increase iovec[0] size
	
#ifdef DEBUG	
        _hexdump(outp_vec[0].ptr, cal_enc_len);
#endif

	return 1;
}
#endif
